//
// globals.c - This is the file containing global variables
//             for Mr. Ham's ECMA-55 Minimal BASIC compiler.
//
// Copyright (C) 2013,2014  John Gatewood Ham
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// version 2 of the Free Software Foundation License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
#define _POSIX_C_SOURCE 200809L
#include <features.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>
#include "globals.h"
const char version[]="1.7";// version number for this compiler
const uint32_t EXRAD_NARROW_OUTPUT=UINT32_C(2);                 // width of exponent field for display with narrow output mode
const uint32_t EXRAD_WIDE_OUTPUT=UINT32_C(3);                   // width of exponent field for display with wide output mode
// NOTE: In the code, normally 1 is subtracted from the SIGNIFICANCE value since David M. Gay's g_fmt() expects the number of digits after 
// the decimal point for normalized output.
const uint32_t SIGNIFICANCE_WIDTH_NARROW_OUTPUT=UINT32_C(7);    // number of significant digits to output for display with narrow output mode
const uint32_t SIGNIFICANCE_WIDTH_WIDE_OUTPUT=UINT32_C(18);     // number of significant digits to output for display with wide output mode
bool use_SSE4_1  = false;  // if true, SSE4.1 instructions can be used
bool use_AVX     = false;  // if true, AVX instructions can be used
bool verbose     = false;  // this flag indicates whether verbose diagnostics should be displayed
bool base_is_one = false;  // if true, array indexes begin at 1, otherwise they begin at 0
bool need_RND    = false;  // if true, runtime library code for the RND() function must be emitted
bool need_SIN    = false;  // if true, runtime library code for the SIN() function must be emitted
bool need_COS    = false;  // if true, runtime library code for the COS() function must be emitted
bool need_TAN    = false;  // if true, runtime library code for the TAN() function must be emitted
bool need_ATN    = false;  // if true, runtime library code for the ATN() function must be emitted
bool need_LOG    = false;  // if true, runtime library code for the LOG() function must be emitted
bool need_EXP    = false;  // if true, runtime library code for the EXP() function must be emitted
bool need_POW    = false;  // if true, runtime library code for the POW() function must be emitted
bool need_SQR    = false;  // if true, runtime library code for the SQR() function must be emitted
bool need_strcpy = false;  // if true, runtime library code for the strcpy() function must be emitted
bool need_strcmp = false;  // if true, runtime library code for the strcmp() function must be emitted
bool need_INPUT  = false;  // if true, runtime library code for the INPUT statement must be emitted
bool need_PRINT  = false;  // if true, runtime library code for the PRINT statement must be emitted
bool need_READ   = false;  // if true, runtime library code for the READ statement must be emitted
bool need_BSS_numeric = false; // if true, we must output a .bss section for floating point variables
bool need_BSS_string = false;  // if true, we must output a .bss section for string variables
bool use_double  = true;   // if true, use 64 bit floating point math, otherwise use 32 bit floating point math
bool use_double_output = false; // if true, output in full precision, otherwise use lower precision output
int optimization_level=0; // optimization level, 0 indicates no optimization (not used at this time)
const uint32_t MAXLINENO=UINT32_C(9999); // maximum permitted Minimal BASIC line number
const uint32_t MAXCOLUMN=UINT32_C(73);   // maximum permitted column width for input Minimal BASIC source code
static FILE *diagnostic_stream=NULL;  // high-level file handle for debug message output file
static int diagnostic_stream_fileno=-1; // low-level file handle for debug message output file

//
// This procedure will intialize the file handles for the debug message output file
//
void init_debug_printf(FILE *f) {
  diagnostic_stream=f;
  diagnostic_stream_fileno=fileno(f);
  return;
}

//
// This procedure will output a message to the debug message output file.  THe init_debug_printf()
// funtion must be called before ever using this function.  This function mimics the printf() 
// function and has the same API, but sends all of its otuput to the debug message output file.
//
void debug_printf(const char *format, ...) {
  va_list arg;

  if (verbose) {
    if (NULL==diagnostic_stream) {
       fprintf(stderr,"You failed to call init_debug_printf before debug_printf\n");
       abort();
    }
    va_start(arg, format);
    vfprintf(diagnostic_stream, format, arg);
    va_end(arg);
    if (diagnostic_stream!=stderr) {
       fflush(diagnostic_stream);
       fsync(diagnostic_stream_fileno);
    }
  }
  return;
}

//
// This function returns the appropriate suffix to use for a number in error
// messages.  It is best explained by examples:
//
// 1 - st, for 1st
// 2 - nd, for 2nd
// 3 - rd, for 3rd
// 4 - th, for 4th
// 11 - th, for 11th
// 12 - th, for 12th
// 13 - th, for 13th
// 21 - st, for 21st
// 22 - nd, for 22nd
// 23 - rd, for 23rd
//
const char *suffix(const uint32_t n) {
  static const char *nth[]={"th","st","nd","rd","th","th","th","th","th","th"};
  uint32_t it;

  if ( ((n%UINT32_C(100))>=UINT32_C(11)) && ((n%UINT32_C(100))<=UINT32_C(13)) )
    it=UINT32_C(0);
  else
    it=n%UINT32_C(10);
  return nth[it];
}

