//
// lineno.c - This is the file containing the line number module
//            for Mr. Ham's ECMA-55 Minimal BASIC compiler.
//
// Copyright (C) 2013,2014  John Gatewood Ham
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// version 2 of the Free Software Foundation License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include "lineno.h"
#include "globals.h"

#define ERROR_BUFFER_LEN 80  // length of error message buffers used in this module

//
// The same structure is used for both line number tables to allow coding only
// one sort and one search routine.
//
struct lineinfo {
  uint32_t fromlineno;   // fromlineno only used for targets
  uint32_t lineno;       // line number in BASIC program
  char scope[MAX_SCOPE_BYTES+1];  // REXX-like scope indicator
}; 
// contains one entry for every Minimal BASIC line number of the source program
static struct lineinfo *linedata=NULL;
// contains one entry for every Minimal BASIC line number of the source program
// that was used as a jump target (destination) by GOTO, GOSUB, ON GOSUB, etc.
static struct lineinfo *targetdata=NULL;

static void domerge(struct lineinfo *thelist,
           uint32_t list1_start,
           uint32_t list1_stop,
           uint32_t list2_stop);
static void mergesort(struct lineinfo *thelist,
           const uint32_t element_start,
               uint32_t element_stop);
static uint32_t binary_search(struct lineinfo*thelist,
           const uint32_t element_count, const uint32_t lno);

// module variables
static uint32_t linedata_size=UINT32_C(0);
static uint32_t linedata_lastline=UINT32_C(0);
static uint32_t linedata_next=UINT32_C(0);
static uint32_t targetdata_size=UINT32_C(0);
static uint32_t targetdata_lastline=UINT32_C(0);
static uint32_t targetdata_next=UINT32_C(0);
static bool sorted=false;	// flag indicating whether linedata array data is sorted

//
// This procedure deallocates the line number tracking arrays
//
void clear_lineno_data(void) {
  free(linedata); linedata=NULL;
  free(targetdata); targetdata=NULL;
  return;
}

//
// This function returns true if it can convert the non-empty, non-NULL 
// ASCIIZ string passed in parameter s to a valid Minimal BASIC line number,
// otherwise it returns false.  If the conversion is successful, the result
// is stored in the parameter lineno, otherwise an ASCIIZ string containing
// the reason for failure is stored in the parameter why.
//
bool convert_line_number(const char *s,uint32_t *lineno, char *why) {
  char *endptr=NULL;
  uint64_t incoming;
  uint32_t tlineno;

  errno=0; // clear any existing error
  if ('-'==s[0]) {  // if the first byte is a minus then fail with an error
    snprintf(why,ERROR_BUFFER_LEN,"ERROR: Invalid line number %s; line numbers cannot be negative\n",s);
    why[ERROR_BUFFER_LEN-1]=0; // terminate error buffer for safety
    return false;
  }
  incoming=strtoul(s,&endptr,10);  // attempt the conversion
  switch (errno) {
    case 0:     // OK
       break;
    case ERANGE:  // range error
       snprintf(why,ERROR_BUFFER_LEN,"ERROR: Invalid line number %s; must be between 1 and %" PRIu32 "\n",s,MAXLINENO);
       why[ERROR_BUFFER_LEN-1]=0; // terminate error buffer for safety
       return false;
    default:  // some other problem with the conversion
       snprintf(why,ERROR_BUFFER_LEN,"ERROR: Invalid line number %s; cannot convert to unsigned long\n",s);
       why[ERROR_BUFFER_LEN-1]=0; // terminate error buffer for safety
       return false;
  }
  if (endptr!=s+strlen(s)) { // s parameter had trailing junk
    snprintf(why,ERROR_BUFFER_LEN,"ERROR: Invalid line number %s; trailing garbage is '%s'(scanner failure?)\n",s,endptr);
    why[ERROR_BUFFER_LEN-1]=0; // terminate error buffer for safety
    return false;
  }
  if ((incoming>UINT_MAX)||(incoming>MAXLINENO)) { // line number was too big
    snprintf(why,ERROR_BUFFER_LEN,"ERROR: Invalid line number %s; must be between 1 and %" PRIu32 "\n",s,MAXLINENO);
    why[ERROR_BUFFER_LEN-1]=0; // terminate error buffer for safety
    return false;
  }
  tlineno=(uint32_t)incoming; // convert type to 32 bit unsigned integer
  *lineno=tlineno;            // store line number in outgoing parameter
  return true;
}

//
// This function is used to add a new line number specifed by the lno
// parameter to the line number data.  If it succeeds, true is returned, 
// otherwise false is returned and and ASCIIZ string containing the 
// reason for failure is stored in the parameter why.  The scope parameter
// is also passed in to be stored.
//
bool add_new_line_number(const uint32_t lno,char *why,const char *scope) {
  debug_printf("add_new_line_number('%" PRIu32 "')\n",lno);
  if (linedata_lastline>UINT32_C(0)) { // if we already have some lines
     if (binary_search(linedata,linedata_lastline,lno)!=UINT_MAX) { // ensure this is not a duplicate
       snprintf(why,80,"ERROR: Duplicate line number %" PRIu32 "\n",lno);
       return false;
     }
     // linedata must be sorted as a side-effect of the search
     if (linedata[linedata_lastline-UINT32_C(1)].lineno>lno) { // if the line number lno is not strictly greater than the last line stored in the linedata array
       // the input Minimal BASIC source file was not sorted in ascending line number order 
       // and we return an error in that case
       snprintf(why,80,"ERROR: Line number %" PRIu32 " is less than previous line number %" PRIu32 "\n",
          lno,linedata[linedata_lastline-UINT32_C(1)].lineno);
       return false;
     }
  }
  if (UINT32_C(0)==linedata_size) { // if no storage has ever been allocated for linedata
     // allocate storage for linedata
     linedata_size=UINT32_C(1000); // choose 1000 elements as the default initial size
     linedata=(struct lineinfo *)malloc(sizeof(struct lineinfo)*linedata_size);
     if (NULL==linedata) {
       fprintf(stderr,"Error: out of memory in add_new_line_number\n");
       abort();
     }
  } else { // otherwise storage exists
    if (linedata_lastline>=linedata_size) { // if the allocated storage for linedata is exhausted
       struct lineinfo *temp=NULL;
       // re-allocate the storage with a size double the previous size
       linedata_size<<=1;
       temp=(struct lineinfo *)realloc(linedata,sizeof(struct lineinfo)*linedata_size);
       if (NULL==temp) {
         fprintf(stderr,"Error: out of memory #2 in add_new_line_number\n");
         abort();
       }
       linedata=temp;
    }
  }
  // at this point there is definitely room in the linedata array and lno is NOT
  // already in it (not a duplicate), so add lno and its accompanying scope.
  linedata[linedata_lastline].fromlineno=UINT32_C(0); // store dummy 0 value for origin (caller) line number
  linedata[linedata_lastline].lineno=lno;  // store target (destination) line number
  strncpy(linedata[linedata_lastline].scope,scope,MAX_SCOPE_BYTES);  // store scope indicator
  linedata[linedata_lastline++].scope[MAX_SCOPE_BYTES]=(char)0; // force termination for safety
  sorted=false; // mark this table as unsorted
  return true;
}

//
// This function is used to add a new jump target to the tracking data
// for jump targets in this module.  The jump target line number is specified
// by the to_lno parameter, the place where the jump came from is specified
// by the from_lno parameter, and the scope indicator is passed in the scope
// paramter.  On succes, the function returns true.  Otherwise it calls abort().
//
bool add_new_jump_target(const uint32_t to_lno, const uint32_t from_lno, const char *scope) {
  debug_printf("add_new_jump_target(from %" PRIu32 " to %" PRIu32 ")\n",from_lno,to_lno);
  if (UINT32_C(0)==targetdata_size) { // if no storage has ever been allocated for targetdata
     // allocate storage for targetdata
     targetdata_size=UINT32_C(1000); // choose 1000 elements as the default initial size
     targetdata=(struct lineinfo *)malloc(sizeof(struct lineinfo)*targetdata_size);
     if (NULL==targetdata) {
       fprintf(stderr,"Error: out of memory in add_new_jump_target\n");
       abort();
     }
  } else { // otherwise storage exists
    if (targetdata_lastline>=targetdata_size) { // if the allocated storage for targetdata is exhausted
       struct lineinfo *temp=NULL;
       // re-allocate the storage with a size double the previous size
       targetdata_size<<=1;
       temp=(struct lineinfo *)realloc(targetdata,sizeof(struct lineinfo)*targetdata_size);
       if (NULL==temp) {
         fprintf(stderr,"Error: out of memory #2 in add_new_jump_target\n");
         abort();
       }
       targetdata=temp;
    }
  }
  // at this point there is definitely room in the targetdata array
  // so add lno with its corresponding from_lno and scope.
  targetdata[targetdata_lastline].fromlineno=from_lno;	// store origin (caller) line number
  targetdata[targetdata_lastline].lineno=to_lno;	// store target (destination) line number
  strncpy(targetdata[targetdata_lastline].scope,scope,MAX_SCOPE_BYTES);  // store scope indicator
  targetdata[targetdata_lastline++].scope[MAX_SCOPE_BYTES]=(char)0; // force termination for safety
  sorted=false; // mark this table as unsorted
  return true;
}

//
// This is a helper function for mergesort() that merges
// two sorted runs, both of which are stored in the array specified
// by the thelist parameter.  The first run starts with the index
// specified by the list1_start parameter and ends with the index
// specified by the list1_stop parameter.  The second run starts
// with the index specified by the list2_start parameter and ends
// with the index specified by the list2_stop parameter.  This function
// will dynamically allocate a second array called scratch to be used
// as a temporary destination of merges, and the last step is to
// copy that back to the thelist array.
//
static void domerge(struct lineinfo *thelist,  // input array
           uint32_t list1_start,               // index of start of first run
           uint32_t list1_stop,                // index of end of first run
           uint32_t list2_stop) {              // index of end of second run
  uint32_t list2_start=list1_stop+UINT32_C(1); // index of start of second run
  uint32_t leftspot=list1_start;  // index in first list of element to consider during compares
  uint32_t rightspot=list2_start; // index in second list of element to consider during compares
  uint32_t outspot=UINT32_C(0);   // index of output position in destination array
  struct lineinfo *scratch;       // scratch (temporary) array

  // allocate scratch array
  if ((scratch=(struct lineinfo *)malloc(sizeof(struct lineinfo)*linedata_lastline))==NULL) {
    fprintf(stderr,"Error: out of memory in domerge\n");
    abort();
  }
  // while both runs have data
  while ((leftspot<=list1_stop) && (rightspot<=list2_stop)) {
    if (thelist[leftspot].lineno<thelist[rightspot].lineno) // if the element in the left list is smaller
       scratch[outspot++]=thelist[leftspot++];              // copy it to scratch
    else                                                    // else
       scratch[outspot++]=thelist[rightspot++];             // copy the element in the right list to scratch
  }
  while (leftspot<=list1_stop) {        // if the left list still has data append the left tail
    scratch[outspot++]=thelist[leftspot++];
  }
  while (rightspot<=list2_stop) {       // if the right list still has data append the right tail
    scratch[outspot++]=thelist[rightspot++];
  }
  // copy data back from scratch array back into mylist array overwriting old contents
  for (uint32_t i=UINT32_C(0);i<outspot;i++)
     thelist[list1_start+i]=scratch[i];
  free(scratch); // de-allocate scratch array
  scratch=NULL;
  return;
}

//
// This function will sort a lineinfo array using a recursive 
// implementation of a standard mergesort.  This function does
// the partitioning, and the domerge() help function will merge
// the sorted runs.  The sort is done in ascending order by
// the lineno field in the lineinfo record of the thelist array.
// The input array is specified by the thelist parameter, and the 
// array indixes for the start and stop are specified by the
// element_start and element_stop parameters respectively.
//
static void mergesort(struct lineinfo *thelist,
               const uint32_t element_start,
               uint32_t element_stop) {
  uint32_t tlen_minus_one=element_stop-element_start; // length of the run minus one
  uint32_t delta=tlen_minus_one/UINT32_C(2);          // start of second sub-run
  switch(tlen_minus_one) {
    case UINT32_C(0): // one element, by definition it is sorted, nothing to do
      break;
    case UINT32_C(1): // two elements, compare them and swap if necessary
      if (thelist[element_stop].lineno<thelist[element_start].lineno) {
         struct lineinfo temp=thelist[element_start];
         thelist[element_start]=thelist[element_stop];
         thelist[element_stop]=temp;
      }
      break;
    default: // more than two elements, partition into two sub lists each approximately
             // half the size of the original, and send each of those to runs recursively
             // for sorting by the mergesort() function, and once they are both sorted,
             // send them to the domerge() function to merge the two sorted sub runs into
             // one longer sorted run, resulting in sorting the run passed into this
             // invocation of mergesort()
      mergesort(thelist,element_start,element_start+delta-1);
      mergesort(thelist,element_start+delta,element_stop);
      domerge(thelist,element_start,element_start+delta-1,element_stop);
      break;
  }
  sorted=true;
  return;
}

//
// This function will use a binary search to look for the line number
// specifed by the lno parameter in the list specified by thelist
// parameter.  If an add to the list has occurred and it is potentially
// not sorted, a call to the mergesort function is used to sort the
// data in the thelist array.  If lno is found, the index of the
// record in the array is returned, otherwise UNIT_MAX is returned
// to indicate failure.  The input list must not contain duplicate 
// line number (the key) values.
//
static uint32_t binary_search(struct lineinfo *thelist, // list to search
         const uint32_t element_count,                  // number of elements in list to search
         const uint32_t lno) {                          // value we are searching for
  uint32_t low=UINT32_C(0);                 // index of start of partition
  uint32_t high=element_count-UINT32_C(1);  // index of end of partition
  uint32_t middle=UINT32_C(0);              // index of midway point in partition

  if (!sorted) mergesort(thelist,UINT32_C(0),high); // if thelist is not sorted, sort it since
                                                    // binary search algorithm requires the array be sorted

  // while the partition still contains elements keep looking
  while (low<=high) {
    middle=low+(high-low)/UINT32_C(2);  // compute the midway point
    debug_printf("TRACING: low=%" PRIu32 ", middle=%" PRIu32 ", high=%" PRIu32 ", checking if %" PRIu32 "==%" PRIu32 "\n",
        low,middle,high,thelist[middle].lineno,lno);
    if (thelist[middle].lineno==lno) // if we found lno
       return middle;                // return the index of the record that contains lno
    debug_printf("TRACING: nope\n");
    if (thelist[middle].lineno<lno) {  // if the value in the midway point is too small
       debug_printf("TRACING: too low\n");
       low=middle+UINT32_C(1);         // look in the partition to the right of the midway point
    } else {                           // otherwise the value at the midway point is too big
       if (UINT32_C(0)==middle) break; // if the midpoint was the first element, the value is not in the list
       high=middle-UINT32_C(1);        // look in the partition to the left of the midway point
    }
  }
  debug_printf("TRACING: gave up with low=%" PRIu32 ", middle=%" PRIu32 ", high=%" PRIu32 ", and %" PRIu32 "==%" PRIu32 "\n",
        low,middle,high,thelist[middle].lineno,lno);
  return UINT_MAX;  // lno was not found, so return the error indicator value
}

//
// This function returns true if lineno was found in the linedata table,
// and false otherwise.  If the search was successful, the index in the
// table of the record containing lineno is stored in the ndex parameter.
//
bool line_exists(const uint32_t lineno,uint32_t *ndex) {
  *ndex=binary_search(linedata,linedata_lastline,lineno);
  debug_printf("line_exists(%" PRIu32 ") got %" PRIu32 " from binary_search()\n",lineno,*ndex);
  return (*ndex!=UINT_MAX);
}

//
// This function returns true if lineno was found in the targetdata table,
// and false otherwise.
//
bool jump_target_exists(const uint32_t lineno) {
  uint32_t temp;

  temp=binary_search(targetdata,targetdata_lastline,lineno);
  debug_printf("jump_target_exists(%" PRIu32 ") got %" PRIu32 " from binary_search()\n",lineno,temp);
  return (temp!=UINT_MAX);
}

//
// This function is for debugging, and will dump the data stored in the linedata
// array to STDOUT.
//
void dump_line_numbers(void) {
  uint32_t i;
  printf("line numbers\n");
  for (i=UINT32_C(0);i<linedata_lastline;i++) {
    printf("source line %" PRIu32 " at scope 'global%s'\n",linedata[i].lineno,linedata[i].scope);
  }
  printf("\n");
  return;
}

//
// This function is for debugging, and will dump the data stored in the targetdata
// array to STDOUT.
//
void dump_jump_targets(void) {
  uint32_t i;
  printf("jump targets\n");
  for (i=UINT32_C(0);i<targetdata_lastline;i++) {
    printf("jump to target line %" PRIu32 " from source %" PRIu32 " at scope 'global%s'\n",
      targetdata[i].lineno,targetdata[i].fromlineno,linedata[i].scope);
  }
  printf("\n");
  return;
}

//
// This function is used to verify that every line number in the Minimal BASIC
// source program specified as a jump target (destination) actually exists.  If
// all jump targets do exist in the program, this function returns true.  Otherwise
// it returns false.
//
bool verify_all_jump_targets_exist(void) {
  uint32_t i,j,fromscopelen,toscopelen;
  bool retval=true;

  if (UINT32_C(0)==targetdata_lastline) return true;
  mergesort(targetdata,UINT32_C(0),targetdata_lastline-UINT32_C(1));

  if (verbose) { // if the verbose flag is set, show the contents of the module's line number lists
    dump_line_numbers();
    dump_jump_targets();
  }
  for (i=UINT32_C(0);i<targetdata_lastline;i++) { // for every jump target line number
    if (verbose)
      fprintf(stderr,"checking jump target %" PRIu32 "\n",targetdata[i].lineno);
    if (!line_exists(targetdata[i].lineno,&j)) { // if the jump target line number does not exist fail with an error
      fprintf(stderr,"Jump target %" PRIu32 " specified in program on line %" PRIu32 ", but no line %" PRIu32 " exists in the program.\n",
        targetdata[i].lineno,targetdata[i].fromlineno,targetdata[i].lineno);
      retval=false;
      continue;
    } else { // line does exists
      if (verbose)
        fprintf(stderr,"Jump target %" PRIu32 " specified in program, and line %" PRIu32 " exists in the program.\n",
          targetdata[i].lineno,targetdata[i].lineno);
    }
    if (verbose)
      fprintf(stderr,"jump from %" PRIu32 " at scope 'global%s' to %" PRIu32 " at scope 'global%s'\n",
        targetdata[i].fromlineno,targetdata[i].scope,linedata[j].lineno,linedata[j].scope);
    // line number exists, but is it in a permitted scope?
    fromscopelen=strlen(targetdata[i].scope);	// get scope of caller
    toscopelen=strlen(linedata[j].scope);       // get scope of destination
    if (fromscopelen<toscopelen) {              // if the jump is to an outer scope it is not OK, note the error
      fprintf(stderr,"Jumping from %" PRIu32 " into loop at %" PRIu32 " not permitted\n",
        targetdata[i].fromlineno,linedata[j].lineno);
      retval=false;
      continue;
    }
    // if the jump is to a scope of the same level but still a different scope, that is not OK, note the error
    if ((fromscopelen==toscopelen) && (strcmp(targetdata[i].scope,linedata[j].scope)!=0)) {
      fprintf(stderr,"Jumping from %" PRIu32 " into loop at %" PRIu32 " not permitted\n",
        targetdata[i].fromlineno,linedata[j].lineno);
      retval=false;
      continue;
    }
  }
  // if no errors were seen, return true, else return false
  return retval;
}

//
// This function is an iterator over the values of the linedata array.
// If the reset parameter is false, the next Minimal BASIC program line number is returned.
// If the reset parameter is true, the iterator is reset to the beginning of the list
// and the first line number is returned.  If the end of the list is reached,
// return UINT_MAX to indicate that that condition.
//
uint32_t get_next_line(const bool reset) {
   if (reset) linedata_next=UINT32_C(0); // if reset was true, reposition the iterator
   debug_printf("get_next_line(%s), linedata_next=%" PRIu32 ",linedata_lastline=%" PRIu32 "\n",
         (reset?"true":"false"),linedata_next,linedata_lastline);
   if (linedata_next==linedata_lastline) {
      debug_printf("END OF LIST OF LINE NUMBERS\n");
      return UINT_MAX;
   }
   return linedata[linedata_next++].lineno;
}

//
// This function is an iterator over the values of the targetdata array.
// If the reset parameter is false, the next target linenumber is returned.
// If the reset parameter is true, the iterator is reset to the beginning of the list
// and the first line number is returned.  If the end of the list is reached,
// return UINT_MAX to indicate that that condition.
//
uint32_t get_next_jump_target(const bool reset) {
   if (reset) targetdata_next=UINT32_C(0); // if reset was true, reposition the iterator
   debug_printf("get_next_jump_target(%s), targetdata_next=%" PRIu32 ",targetdata_lastline=%" PRIu32 "\n",
         (reset?"true":"false"),targetdata_next,targetdata_lastline);
   if (targetdata_next==targetdata_lastline) {
      debug_printf("END OF LIST OF JUMP TARGETS\n");
      return UINT_MAX;
   }
   return targetdata[targetdata_next++].lineno;
}

//
// This is code to test the module and is not normally used.
// If you compile with UNIT_TEST defined, then this module will
// be a stand-alone program with its own main() that runs some
// self-tests.
//
#ifdef UNIT_TEST
#include <string.h>

static const uint32_t lnos[]={
  UINT32_C(10),
  UINT32_C(20),
  UINT32_C(30),
  UINT32_C(40),
  UINT32_C(50),
  UINT32_C(60),
  UINT32_C(70),
  UINT32_C(80),
  UINT32_C(90),
  UINT32_C(100),
};
static const uint32_t lnos_count=sizeof(lnos)/sizeof(uint32_t);
static const uint32_t tgts[]={
  UINT32_C(50),
  UINT32_C(70),
  UINT32_C(100),
};
static const uint32_t tgts_count=sizeof(tgts)/sizeof(uint32_t);

int main(int argc, char *argv[]) {
  char why[ERROR_BUFFER_LEN+1];
  uint32_t temp;
  bool quit_on_fail=false,ok=true;
  uint32_t i,j;

  verbose=false;
  if ((2==argc)&&(strcmp(argv[1],"-V")==0)) {
     verbose=true;
     quit_on_fail=true;
  } else {
     quit_on_fail=false;
     if ((2==argc)&&(strcmp(argv[1],"-v")==0))
       verbose=true;
     else
       verbose=false;
  }
  init_debug_printf(stderr);
  // positive testing
  printf("test get_next_line() when no lines exist: ");
  temp=get_next_line(false);
  if (temp!=UINT_MAX) {
     fprintf(stderr,"get_next_line() returns %" PRIu32 " instead of UINT_MAX when no lines exist\n",temp);
     ok=false;
     printf("FAIL\n");
     if (quit_on_fail) goto done;
  } else {
     printf("OK\n");
  }
  printf("test get_next_jump_target() when no targets exist: ");
  temp=get_next_jump_target(false);
  if (temp!=UINT_MAX) {
     fprintf(stderr,"get_next_jump_target() returns %" PRIu32 " instead of UINT_MAX when no lines exist\n",temp);
     ok=false;
     printf("FAIL\n");
     if (quit_on_fail) goto done;
  } else {
     printf("OK\n");
  }
  printf("verify_all_jump_targets_exist() when no targets or lines exist: ");
  if (!verify_all_jump_targets_exist()) {
     fprintf(stderr,"verify_all_jump_targets_exist() fails when no lines exist\n");
     ok=false;
     printf("FAIL\n");
     if (quit_on_fail) goto done;
  } else {
     printf("OK\n");
  }
  for (i=UINT32_C(0);i<lnos_count;i++) {
    printf("verify add_new_line_number(%4" PRIu32 ") succeeds: ",lnos[i]);
    if (!add_new_line_number(lnos[i],why,"")) {
      ok=false;
      printf("FAIL\n");
      fputs(why,stderr);
      if (quit_on_fail) goto done;
    } else {
      printf("OK\n");
    }
  }
  if (verbose) dump_line_numbers();
  for (i=UINT32_C(0);i<lnos_count;i++) {
    printf("verify line_exists(%4" PRIu32 ") succeeds: ",lnos[i]);
    if (!line_exists(lnos[i],&j)) {
      ok=false;
      printf("FAIL\n");
      if (quit_on_fail) goto done;
    } else {
      printf("OK\n");
    }
  }
  for (i=UINT32_C(0);i<tgts_count;i++) {
    printf("verify add_new_jump_target(%4" PRIu32 ") succeeds: ",tgts[i]);
    if (!add_new_jump_target(tgts[i],UINT32_C(10),"")) {
      ok=false;
      printf("FAIL\n");
      if (quit_on_fail) goto done;
    } else {
      printf("OK\n");
    }
  }
  if (verbose) dump_jump_targets();
  for (i=UINT32_C(0);i<tgts_count;i++) {
    printf("verify jump_target_exists(%4" PRIu32 ") succeeds: ",tgts[i]);
    if (!jump_target_exists(tgts[i])) {
      ok=false;
      printf("FAIL\n");
      if (quit_on_fail) goto done;
    } else {
      printf("OK\n");
    }
  }
  printf("verify_all_jump_targets_exist() succeeds when it should: ");
  if (!verify_all_jump_targets_exist()) {
     ok=false;
     printf("FAIL\n");
     if (quit_on_fail) goto done;
  } else {
     printf("OK\n");
  }
  // advance to non-first position
  printf("verify get_next_line(true) resets and gives us the first line %" PRIu32 ": ",lnos[UINT32_C(0)]);
  if (get_next_line(true)!=lnos[UINT32_C(0)]) {
    ok=false;
    printf("FAIL\n");
    if (quit_on_fail) goto done;
  } else {
    printf("OK\n");
  }
  printf("verify get_next_line(false) gives us the second line %" PRIu32 ": ",lnos[UINT32_C(1)]);
  if (get_next_line(false)!=lnos[UINT32_C(1)]) {
    ok=false;
    printf("FAIL\n");
    if (quit_on_fail) goto done;
  } else {
    printf("OK\n");
  }
  printf("verify get_next_line(true) resets and gives us the first line %" PRIu32 ": ",lnos[UINT32_C(0)]);
  if (get_next_line(true)!=lnos[UINT32_C(0)]) {
    ok=false;
    printf("FAIL\n");
    if (quit_on_fail) goto done;
  } else {
    printf("OK\n");
  }
  for (i=UINT32_C(1);i<lnos_count;i++) {
    printf("verify get_next_line(false) gives us the next line %" PRIu32 ": ",lnos[i]);
    if (get_next_line(false)!=lnos[i]) {
      ok=false;
      printf("FAIL\n");
      if (quit_on_fail) goto done;
    } else {
      printf("OK\n");
    }
  }
  // advance to non-first position
  printf("verify get_next_jump_target(true) resets and gives us the first jump target %" PRIu32 ": ",tgts[UINT32_C(0)]);
  if (get_next_jump_target(true)!=tgts[UINT32_C(0)]) {
    ok=false;
    printf("FAIL\n");
    if (quit_on_fail) goto done;
  } else {
    printf("OK\n");
  }
  printf("verify get_next_jump_target(false) gives us the second jump target %" PRIu32 ": ",tgts[UINT32_C(1)]);
  if (get_next_jump_target(false)!=tgts[UINT32_C(1)]) {
    ok=false;
    printf("FAIL\n");
    if (quit_on_fail) goto done;
  } else {
    printf("OK\n");
  }
  printf("verify get_next_jump_target(true) resets and gives us the first jump target %" PRIu32 ": ",tgts[UINT32_C(0)]);
  if (get_next_jump_target(true)!=tgts[UINT32_C(0)]) {
    ok=false;
    printf("FAIL\n");
    if (quit_on_fail) goto done;
  } else {
    printf("OK\n");
  }
  for (i=UINT32_C(1);i<tgts_count;i++) {
    printf("verify get_next_jump_target(false) gives us the next jump target %" PRIu32 ": ",tgts[i]);
    if (get_next_jump_target(false)!=tgts[i]) {
      ok=false;
      printf("FAIL\n");
      if (quit_on_fail) goto done;
    } else {
      printf("OK\n");
    }
  }
  // negative testing
  printf("verify line_exists fails for non-existent line: ");
  if (line_exists(UINT32_C(999),&j)) {
    ok=false;
    printf("FAIL\n");
    if (quit_on_fail) goto done;
  } else {
    printf("OK\n");
  }
  printf("verify jump_target_exists fails for non-existent target: ");
  if (jump_target_exists(UINT32_C(999))) {
    ok=false;
    printf("FAIL\n");
    if (quit_on_fail) goto done;
  } else {
    printf("OK\n");
  }
done:
  clear_lineno_data();
  printf("Summary: ");
  if (!ok) {
    printf("FAIL\n");
    return EXIT_FAILURE;
  }
  printf("OK\n");
  return EXIT_SUCCESS;
}
#endif
