//
// main.c - This is the file containing main() and options processing
//          for Mr. Ham's ECMA-55 Minimal BASIC compiler.
//
static const char * const license_msg = "\
Copyright (C) 2013,2014  John Gatewood Ham\n\
\n\
This program is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
version 2 of the Free Software Foundation License.\n\
\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
GNU General Public License for more details.\n\
\n\
You should have received a copy of the GNU General Public License along\n\
with this program; if not, write to the Free Software Foundation, Inc.,\n\
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.\n\
";

#define _POSIX_C_SOURCE 200809L
#include <features.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include "globals.h"
#include "scanner2.h"
#include "parser2.h"
#include "lineno.h"
#include "symbol_table.h"
#include "codegen.h"

static char errbuffer[4096]={0};	// error message buffer

void show_GPLv2(void);

//
// This procedure will print the version of the program.  The program
// parameter is the program name obtained from main's first magic argument.
//
static void show_version(const char *progname) {
  printf("ecma55 version %s\n",version);
  printf("Copyright (C) 2013,2014 John Gatewood Ham\n");
  printf("Compiler License is GNU GPL version 2 ONLY\n");
  printf("Executable was invoked with name '%s'\n",progname);
  return;
}

//
// This procedure will print a simple usage message explaing the
// parmeters, which are explained in more detail in the man page.
//
static void usage(const char *progname) {
  printf("Usage: %s [-v] [-s] [-w] [-4] [-A] [-O opt] [-o outputfilename ] inputfilename\n", progname);
  printf("              or\n");
  printf("       %s -h\n", progname);
  printf("              or\n");
  printf("       %s -l\n", progname);
  printf("              or\n");
  printf("       %s -L\n", progname);
  printf("              or\n");
  printf("       %s -V\n", progname);
  printf("    -O sets optimization level, and must be followed by an integer between\n");
  printf("       0 and 3 inclusive, which higher numbers requesting more optimization\n");
  printf("    -L shows brief license information\n");
  printf("    -L shows full GPLv2 license\n");
  printf("    -v generates verbose diagnostics\n");
  printf("    -s uses single precision float math for arithmetic expressions\n");
  printf("    -w uses double precision output, 132 columns instead of default 80 columns\n");
  printf("    -4 allows code generator to emit SSE4.1 instructions\n");
  printf("    -A allows code generator to emit AVX instructions\n");
  printf("    -h Show this usage information\n");
  printf("    -V shows the version number\n");
  return;
}

//
// This fuction will load the file specified in the filename parameter into
// a dynamically allocated buffer and return a pointer to that buffer.  The
// caller is responsible for deallocating that buffer when it is no longer
// needed.  If the file cannot be loaded, NULL is returned to indicate failure.
// The file must exist and be readable, and it must be a UNIX format text file.
// Much of the code used in this function is UNIX/Linux-specific.
//
static char *load_file_into_buffer(const char *filename) {
  char *buffer=NULL;
  FILE *f=NULL;
  int64_t offset,howmany,totalsofar;
  struct stat sb;

  if (stat(filename, &sb) == -1) {
    fprintf(stderr,"Unable to access file '%s'\n",filename);
    return NULL;
  }
  if (!((S_ISREG(sb.st_mode) || S_ISLNK(sb.st_mode)))) {
    fprintf(stderr,"File '%s' is not a regular file\n",filename);
    return NULL;
  }
  if ((f=fopen(filename,"r"))==NULL) {
    sprintf(errbuffer,"Unable to open file '%s' for input",filename);
    perror(errbuffer);
    return NULL;
  }
  if (fseek(f,0,SEEK_END)<0) {
    sprintf(errbuffer,"Unable to seek to end of input file '%s'",filename);
    perror(errbuffer);
    fclose(f);
    f=NULL;
    return NULL;
  }
  offset=ftell(f);
  if (offset<INT64_C(0)) {
    sprintf(errbuffer,"Unable to determine length of input file '%s'",filename);
    perror(errbuffer);
    fclose(f);
    f=NULL;
    return NULL;
  }
  if (offset<INT64_C(6)) {
    fprintf(stderr,"Error: Input file '%s' is too short to be a real ECMA-55 Minimal BASIC program\n",filename);
    fclose(f);
    f=NULL;
    return NULL;
  }
  if ((buffer=(char *)malloc(sizeof(char)*((uint64_t)offset+UINT64_C(1))))==NULL) {
    fclose(f);
    f=NULL;
    fprintf(stderr,"Unable to allocate %" PRIu64 " bytes in load_file_into_buffer\n",(uint64_t)offset+UINT32_C(1));
    abort();
  }
  rewind(f);
  buffer[0]=0;
  totalsofar=INT64_C(0);
  do {
    howmany=fread(buffer,1,offset-totalsofar,f);
    debug_printf("read %" PRIi64 " bytes from input file '%s'\n",howmany,filename);
    if (feof(f)) {
      free(buffer);buffer=NULL;
      sprintf(errbuffer,"fread() EOF for input file '%s'",filename);
      perror(errbuffer);
      fclose(f);
      return NULL;
    }
    if (ferror(f)) {
      free(buffer);buffer=NULL;
      sprintf(errbuffer,"fread() Error for input file '%s'",filename);
      perror(errbuffer);
      fclose(f);
      return NULL;
    }
    totalsofar+=howmany;
  } while (totalsofar<offset);
  fclose(f);
  f=NULL;
  buffer[offset]=0;     // guarantee termination
  return buffer;
}

//
// This main function is the starting point for running this compiler.
// Different methods for invocation are possible.  These are briefly
// shown in the usage() function, and explained in detail in the man
// page for this compiler.
//
int main(int argc, char **argv) {
  char *outname=NULL; // name of output assembly language file this compiler will generate
  char *inname=NULL;  // name of input Minimal BASIC source program this compiler will use as input
  char *inbuf=NULL;   // pointer to buffer that contains the Minimal BASIC source program loaded from the inname file
  int32_t opt; // variable used by getopt() function
  uint32_t tlen; // variable used to hold length of strings
  int32_t retval=EXIT_FAILURE; // return value of main(), pessimistically assumes falure
  FILE *outfile=NULL; // file handle for output aseembly file
  FILE *xtraoutfile=NULL; // file handle for file containing user-defined functions which
                          // will be appended to the output if any user-defined functions
                          // are defined in the Minimal BASIC input program

  init_debug_printf(stderr); // initialize the debug output system
  optimization_level = INT32_C(0); // set the optimization level to the default value of zero
  // process command-line arguments to the compiler
  while ((opt = getopt(argc, argv, "hsvw4AVlLo:O:")) != -1) {
    char *endptr=NULL;
    switch (opt) {
      case 'A': // use AVX instructions (which means SSE4.1 too)
        use_SSE4_1=true;
        use_AVX=true;
        break;
      case '4': // use SSE4.1 instructions
        use_SSE4_1=true;
        break;
      case 'w': // use full-precision output for displaying floating point values
        use_double_output=true;
        break;
      case 's': // use 32bit floating point math
        use_double=false;
        break;
      case 'v': // show verbose diagnostic messages during compilation
        verbose=true;
        break;
      case 'V': // show the version number of this compiler and then exit.
        show_version(argv[0]);
        if (outname!=NULL) { free(outname);outname=NULL; }
        return EXIT_SUCCESS;
      case 'l': // show the brief license summary text and then exit.
        fputs(license_msg,stdout);
        if (outname!=NULL) { free(outname);outname=NULL; }
        return EXIT_SUCCESS;
      case 'L': // show the full GPLv2 license text and then exit
        show_GPLv2();
        if (outname!=NULL) { free(outname);outname=NULL; }
        return EXIT_SUCCESS;
      case 'o': // specify the name of the output file to contain the assembly code this compiler generates
        if (outname!=NULL) {
          fprintf(stderr,"-o option can only be used once\n");
          free(outname);outname=NULL;
          return EXIT_FAILURE;
        }
        tlen=strlen(optarg);
        if ((outname=(char *)malloc(sizeof(char)*(UINT32_C(1)+tlen)))==NULL) {
          fprintf(stderr,"Unable to allocate even %" PRIu32 " bytes of memory in main.\n",UINT32_C(1)+tlen);
          abort();
        }
        strncpy(outname,optarg,tlen);
        outname[tlen]=0;
        break;
      case 'O': // specify the optimization level (not currently used)
        errno = 0;
        optimization_level = (int32_t)strtol(optarg, &endptr, 10);
        if  ((errno != 0) && (INT32_C(0) == optimization_level)) {
           fprintf(stderr,"Bogus argument to -O; you must use an integer\n");
           if (outname!=NULL) { free(outname);outname=NULL; }
           return EXIT_FAILURE;
        }
        if (endptr == optarg) {
           fprintf(stderr,"Bogus or missing argument to -O; you must use an integer\n");
           if (outname!=NULL) { free(outname);outname=NULL; }
           return EXIT_FAILURE;
        }
        if (*endptr != '\0') {
           fprintf(stderr,"trailing garbage in argument to -O; you must use an integer\n");
           if (outname!=NULL) { free(outname);outname=NULL; }
           return EXIT_FAILURE;
        }
        break;
      case 'h': // show help (usage) information and exit
        usage(argv[0]);
        if (outname!=NULL) { free(outname);outname=NULL; }
        return EXIT_SUCCESS;
      default: // unknown option encountered, show help (usage) information and exit with an indication of failure 
        usage(argv[0]);
        if (outname!=NULL) { free(outname);outname=NULL; }
        return EXIT_FAILURE;
    }
  }
  // every remaining option should be a file name
  if (optind >= argc) { // if no file to compile was specified, report the error and exit with an indication of failure
    if (outname!=NULL) { free(outname);outname=NULL; }
    fprintf(stderr, "Expected a filename after options\n");
    usage(argv[0]);
    return EXIT_FAILURE;
  }
  tlen=strlen(argv[optind]);
  // allocation buffer for input Minimal BASIC source file name
  if ((inname=(char *)malloc(sizeof(char)*(UINT32_C(1)+tlen)))==NULL) { // if OOM exit with failure
    if (outname!=NULL) { free(outname);outname=NULL; }
    fprintf(stderr,"Unable to allocate even %" PRIu32 " bytes of memory #2 in main.\n",UINT32_C(1)+tlen);
    abort();
  }
  strncpy(inname,argv[optind],tlen);
  inname[tlen]=0;
  if (NULL==outname) { // if no output file name was specified
    // allocate buffer for output assembly language file name
    if ((outname=(char *)malloc(sizeof(char)*(UINT32_C(3)+tlen)))==NULL) { // if OOM exit with failure
      fprintf(stderr,"Unable to allocate even %" PRIu32 " bytes of memory #3 in main.\n",UINT32_C(3)+tlen);
      abort();
    }
    // generate default output file name which is the input name with '.s' appended
    strncpy(outname,inname,tlen);
    outname[tlen]='.';
    outname[tlen+1]='s';
    outname[tlen+2]=0;
  }
  // if somehow the input name and output name are the same, report the error and exit with an indication of failure
  if (strcmp(outname,inname)==0) {
    fprintf(stderr,"Error: input and output must be different\n");
    free(inname);
    free(outname);
    inname=outname=NULL;
    usage(argv[0]);
    return EXIT_FAILURE;
  }
  debug_printf("verbose output = %s\noptimization level = %" PRIi32 "\ninput filename = '%s'\noutput filename is '%s'\n",
    (verbose?"true":"false"), optimization_level, inname, outname);
  if ((inbuf=load_file_into_buffer(inname))==NULL) { // if load of input Minimal BASIC source file fails
    retval=EXIT_FAILURE;                             // exit with indication of failure
    goto xit;
  }
  retval=(scan_buffer(inbuf)?EXIT_SUCCESS:EXIT_FAILURE); // run the scanner on the buffer containing the
                                                         // Minimal BASIC source code and generate the token stream
  free(inbuf);inbuf=NULL; // deallocate the buffer containing the Minimal BASIC source code
  if (retval!=EXIT_SUCCESS) { // if the scan failed, report the error and exit with an indication of failure
    fprintf(stderr,"scan_buffer() failed\n");
    retval=EXIT_FAILURE;
    goto xit;
  }
  if ((outfile=fopen(outname,"w"))==NULL) { // if the output assembly file cannot be opened
                                            // report the error and exit with an indication of failure
    fprintf(stderr,"Unable to open file '%s for output.\n",outname);
    retval=EXIT_FAILURE;
    goto xit;
  }
  if ((xtraoutfile=tmpfile())==NULL) { // if the output assembly file for user-defined functions cannot be opened
                                       // report the error and exit with an indication of failure
    perror("Unable to open temporary file");
    abort();
  }
  // if you get here, the input was successfully scanned and both output files are ready to use
  write_prologue(outfile,inname); // emit the assembly language prologue

  // run the parser
  if (!parseit(outfile,xtraoutfile)) { // if the parse failed report the error and exit with an indication of failure
    fflush(stdout);
    fprintf(stderr,"parseit() failed\n");
    if (verbose) {
      debug_printf("Remaining unparsed tokens:\n");
      dump_token_stream();
      dump_symbol_table(true);
    }
    retval=EXIT_FAILURE;
    goto xit;
  }
  // if we get here, the parse was successful
  if (!verify_all_jump_targets_exist()) { // if a jump was specified to a line number that does not exist or
                                          // the jump was not permitted because of scope issues, report the error
                                          // and exit with an indication of failure
    retval=EXIT_FAILURE;
    goto xit;
  }
  if (verbose) {
    dump_symbol_table(true);
    printf("Syntax OK\n");
  }
  write_epilogue(outfile); // write the assembly language epilogue, which includes the bulk of the runtime library

  rewind(xtraoutfile); // reset the file pointer on the file contianing user-defined function definitions
  {
    char *line = NULL;
    size_t len = 0;

    // append the lines in the xtraoutput file to the assembly language output file outfile and
    // deallocate the memory used by them
    while (getline(&line, &len, xtraoutfile) != -1) {
      fputs(line,outfile);
    }
    free(line);
    fflush(outfile);
  }

  // attempt to write the constant and variable defintions to the assembly language output file outfile
  if (!generate_symbol_table(outfile)) { // if there were problems, report the error and exit with an
                                         // indication of failure
    fprintf(stderr,"Unable to write data segment to file '%s'.\n",outname);
    retval=EXIT_FAILURE;
    goto xit;
  }
  gen_stack(outfile); // emit code for the runtime stack definition
  // if we get here, everything was OK
  retval=EXIT_SUCCESS; // change return value to indicate the compile was successful

xit:
  fflush(stdout);
  clear_lineno_data(); // deallocate lineno module dynamic memory
  clear_symbol_table(); // deallocateion symbol table dynamic memory
  clear_token_stream(); // deallocation scanner dynamic memory (the token stream)
  if (outfile!=NULL) {
    fclose(outfile);outfile=NULL; // if an output file exists, close it
  }
  free(inname);inname=NULL; // deallocate the input file name buffer
  if (xtraoutfile!=NULL) {
    fclose(xtraoutfile);xtraoutfile=NULL; // if an xtraoutput file exists, close it
  }
  if (retval==EXIT_FAILURE) { // if the compile failed
    if (!verbose) { // and verbose diagnostics were not requested
      if (outname!=NULL) { // and the output filename is not NULL
        remove(outname); // remove the output assembly language file
      }
    } else { // else verbose diagnostics were requested, but the compile failed
      // so leave the partial output file for inspection and let the user know it was
      // not removed.
      fprintf(stderr,"Leaving partial assembly file '%s' for analysis.\n",outname);
    }
  }
  free(outname);outname=NULL; // deallocate the output filename buffer
  return retval;
}

static const char * const GPLv2[]={"\
                    GNU GENERAL PUBLIC LICENSE\n\
                       Version 2, June 1991\n\
\n\
 Copyright (C) 1989, 1991 Free Software Foundation, Inc.,\n\
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA\n\
 Everyone is permitted to copy and distribute verbatim copies\n\
 of this license document, but changing it is not allowed.\n\
\n\
                            Preamble\n\
\n\
  The licenses for most software are designed to take away your\n\
freedom to share and change it.  By contrast, the GNU General Public\n\
License is intended to guarantee your freedom to share and change free\n\
software--to make sure the software is free for all its users.  This\n\
General Public License applies to most of the Free Software\n\
Foundation's software and to any other program whose authors commit to\n\
using it.  (Some other Free Software Foundation software is covered by\n\
the GNU Lesser General Public License instead.)  You can apply it to\n\
your programs, too.\n\
\n\
  When we speak of free software, we are referring to freedom, not\n\
price.  Our General Public Licenses are designed to make sure that you\n\
have the freedom to distribute copies of free software (and charge for\n\
this service if you wish), that you receive source code or can get it\n\
if you want it, that you can change the software or use pieces of it\n\
in new free programs; and that you know you can do these things.\n\
\n\
  To protect your rights, we need to make restrictions that forbid\n\
anyone to deny you these rights or to ask you to surrender the rights.\n\
These restrictions translate to certain responsibilities for you if you\n\
distribute copies of the software, or if you modify it.\n\
\n\
  For example, if you distribute copies of such a program, whether\n\
gratis or for a fee, you must give the recipients all the rights that\n\
you have.  You must make sure that they, too, receive or can get the\n\
source code.  And you must show them these terms so they know their\n\
rights.\n\
\n\
  We protect your rights with two steps: (1) copyright the software, and\n\
(2) offer you this license which gives you legal permission to copy,\n\
distribute and/or modify the software.\n\
\n\
  Also, for each author's protection and ours, we want to make certain\n\
that everyone understands that there is no warranty for this free\n\
software.  If the software is modified by someone else and passed on, we\n\
want its recipients to know that what they have is not the original, so\n\
that any problems introduced by others will not reflect on the original\n\
authors' reputations.\n\
\n\
  Finally, any free program is threatened constantly by software\n\
patents.  We wish to avoid the danger that redistributors of a free\n\
program will individually obtain patent licenses, in effect making the\n\
program proprietary.  To prevent this, we have made it clear that any\n\
patent must be licensed for everyone's free use or not licensed at all.\n\
\n\
  The precise terms and conditions for copying, distribution and\n\
modification follow.\n\
\n",
"\
                    GNU GENERAL PUBLIC LICENSE\n\
   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION\n\
\n\
  0. This License applies to any program or other work which contains\n\
a notice placed by the copyright holder saying it may be distributed\n\
under the terms of this General Public License.  The \"Program\", below,\n\
refers to any such program or work, and a \"work based on the Program\"\n\
means either the Program or any derivative work under copyright law:\n\
that is to say, a work containing the Program or a portion of it,\n\
either verbatim or with modifications and/or translated into another\n\
language.  (Hereinafter, translation is included without limitation in\n\
the term \"modification\".)  Each licensee is addressed as \"you\".\n\
\n\
Activities other than copying, distribution and modification are not\n\
covered by this License; they are outside its scope.  The act of\n\
running the Program is not restricted, and the output from the Program\n\
is covered only if its contents constitute a work based on the\n\
Program (independent of having been made by running the Program).\n\
Whether that is true depends on what the Program does.\n\
\n\
  1. You may copy and distribute verbatim copies of the Program's\n\
source code as you receive it, in any medium, provided that you\n\
conspicuously and appropriately publish on each copy an appropriate\n\
copyright notice and disclaimer of warranty; keep intact all the\n\
notices that refer to this License and to the absence of any warranty;\n\
and give any other recipients of the Program a copy of this License\n\
along with the Program.\n\
\n",
"\
You may charge a fee for the physical act of transferring a copy, and\n\
you may at your option offer warranty protection in exchange for a fee.\n\
\n\
  2. You may modify your copy or copies of the Program or any portion\n\
of it, thus forming a work based on the Program, and copy and\n\
distribute such modifications or work under the terms of Section 1\n\
above, provided that you also meet all of these conditions:\n\
\n\
    a) You must cause the modified files to carry prominent notices\n\
    stating that you changed the files and the date of any change.\n\
\n\
    b) You must cause any work that you distribute or publish, that in\n\
    whole or in part contains or is derived from the Program or any\n\
    part thereof, to be licensed as a whole at no charge to all third\n\
    parties under the terms of this License.\n\
\n\
    c) If the modified program normally reads commands interactively\n\
    when run, you must cause it, when started running for such\n\
    interactive use in the most ordinary way, to print or display an\n\
    announcement including an appropriate copyright notice and a\n\
    notice that there is no warranty (or else, saying that you provide\n\
    a warranty) and that users may redistribute the program under\n\
    these conditions, and telling the user how to view a copy of this\n\
    License.  (Exception: if the Program itself is interactive but\n\
    does not normally print such an announcement, your work based on\n\
    the Program is not required to print an announcement.)\n\
\n\
These requirements apply to the modified work as a whole.  If\n\
identifiable sections of that work are not derived from the Program,\n\
and can be reasonably considered independent and separate works in\n\
themselves, then this License, and its terms, do not apply to those\n\
sections when you distribute them as separate works.  But when you\n\
distribute the same sections as part of a whole which is a work based\n\
on the Program, the distribution of the whole must be on the terms of\n\
this License, whose permissions for other licensees extend to the\n\
entire whole, and thus to each and every part regardless of who wrote it.\n\
\n\
Thus, it is not the intent of this section to claim rights or contest\n\
your rights to work written entirely by you; rather, the intent is to\n\
exercise the right to control the distribution of derivative or\n\
collective works based on the Program.\n\
\n\
In addition, mere aggregation of another work not based on the Program\n\
with the Program (or with a work based on the Program) on a volume of\n\
a storage or distribution medium does not bring the other work under\n\
the scope of this License.\n\
\n",
"\
  3. You may copy and distribute the Program (or a work based on it,\n\
under Section 2) in object code or executable form under the terms of\n\
Sections 1 and 2 above provided that you also do one of the following:\n\
\n\
    a) Accompany it with the complete corresponding machine-readable\n\
    source code, which must be distributed under the terms of Sections\n\
    1 and 2 above on a medium customarily used for software interchange; or,\n\
\n\
    b) Accompany it with a written offer, valid for at least three\n\
    years, to give any third party, for a charge no more than your\n\
    cost of physically performing source distribution, a complete\n\
    machine-readable copy of the corresponding source code, to be\n\
    distributed under the terms of Sections 1 and 2 above on a medium\n\
    customarily used for software interchange; or,\n\
\n\
    c) Accompany it with the information you received as to the offer\n\
    to distribute corresponding source code.  (This alternative is\n\
    allowed only for noncommercial distribution and only if you\n\
    received the program in object code or executable form with such\n\
    an offer, in accord with Subsection b above.)\n\
\n\
The source code for a work means the preferred form of the work for\n\
making modifications to it.  For an executable work, complete source\n\
code means all the source code for all modules it contains, plus any\n\
associated interface definition files, plus the scripts used to\n\
control compilation and installation of the executable.  However, as a\n\
special exception, the source code distributed need not include\n\
anything that is normally distributed (in either source or binary\n\
form) with the major components (compiler, kernel, and so on) of the\n\
operating system on which the executable runs, unless that component\n\
itself accompanies the executable.\n\
\n\
If distribution of executable or object code is made by offering\n\
access to copy from a designated place, then offering equivalent\n\
access to copy the source code from the same place counts as\n\
distribution of the source code, even though third parties are not\n\
compelled to copy the source along with the object code.\n\
\n\
  4. You may not copy, modify, sublicense, or distribute the Program\n\
except as expressly provided under this License.  Any attempt\n\
otherwise to copy, modify, sublicense or distribute the Program is\n\
void, and will automatically terminate your rights under this License.\n\
However, parties who have received copies, or rights, from you under\n\
this License will not have their licenses terminated so long as such\n\
parties remain in full compliance.\n\
\n\
  5. You are not required to accept this License, since you have not\n\
signed it.  However, nothing else grants you permission to modify or\n\
distribute the Program or its derivative works.  These actions are\n\
prohibited by law if you do not accept this License.  Therefore, by\n\
modifying or distributing the Program (or any work based on the\n\
Program), you indicate your acceptance of this License to do so, and\n\
all its terms and conditions for copying, distributing or modifying\n\
the Program or works based on it.\n\
\n\
  6. Each time you redistribute the Program (or any work based on the\n\
Program), the recipient automatically receives a license from the\n\
original licensor to copy, distribute or modify the Program subject to\n\
these terms and conditions.  You may not impose any further\n\
restrictions on the recipients' exercise of the rights granted herein.\n\
You are not responsible for enforcing compliance by third parties to\n\
this License.\n\
\n",
"\
  7. If, as a consequence of a court judgment or allegation of patent\n\
infringement or for any other reason (not limited to patent issues),\n\
conditions are imposed on you (whether by court order, agreement or\n\
otherwise) that contradict the conditions of this License, they do not\n\
excuse you from the conditions of this License.  If you cannot\n\
distribute so as to satisfy simultaneously your obligations under this\n\
License and any other pertinent obligations, then as a consequence you\n\
may not distribute the Program at all.  For example, if a patent\n\
license would not permit royalty-free redistribution of the Program by\n\
all those who receive copies directly or indirectly through you, then\n\
the only way you could satisfy both it and this License would be to\n\
refrain entirely from distribution of the Program.\n\
\n\
If any portion of this section is held invalid or unenforceable under\n\
any particular circumstance, the balance of the section is intended to\n\
apply and the section as a whole is intended to apply in other\n\
circumstances.\n\
\n\
It is not the purpose of this section to induce you to infringe any\n\
patents or other property right claims or to contest validity of any\n\
such claims; this section has the sole purpose of protecting the\n\
integrity of the free software distribution system, which is\n\
implemented by public license practices.  Many people have made\n\
generous contributions to the wide range of software distributed\n\
through that system in reliance on consistent application of that\n\
system; it is up to the author/donor to decide if he or she is willing\n\
to distribute software through any other system and a licensee cannot\n\
impose that choice.\n\
\n\
This section is intended to make thoroughly clear what is believed to\n\
be a consequence of the rest of this License.\n\
\n",
"\
  8. If the distribution and/or use of the Program is restricted in\n\
certain countries either by patents or by copyrighted interfaces, the\n\
original copyright holder who places the Program under this License\n\
may add an explicit geographical distribution limitation excluding\n\
those countries, so that distribution is permitted only in or among\n\
countries not thus excluded.  In such case, this License incorporates\n\
the limitation as if written in the body of this License.\n\
\n\
  9. The Free Software Foundation may publish revised and/or new versions\n\
of the General Public License from time to time.  Such new versions will\n\
be similar in spirit to the present version, but may differ in detail to\n\
address new problems or concerns.\n\
\n\
Each version is given a distinguishing version number.  If the Program\n\
specifies a version number of this License which applies to it and \"any\n\
later version\", you have the option of following the terms and conditions\n\
either of that version or of any later version published by the Free\n\
Software Foundation.  If the Program does not specify a version number of\n\
this License, you may choose any version ever published by the Free Software\n\
Foundation.\n\
\n\
  10. If you wish to incorporate parts of the Program into other free\n\
programs whose distribution conditions are different, write to the author\n\
to ask for permission.  For software which is copyrighted by the Free\n\
Software Foundation, write to the Free Software Foundation; we sometimes\n\
make exceptions for this.  Our decision will be guided by the two goals\n\
of preserving the free status of all derivatives of our free software and\n\
of promoting the sharing and reuse of software generally.\n\
\n\
                            NO WARRANTY\n\
\n\
  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY\n\
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN\n\
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES\n\
PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED\n\
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n\
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS\n\
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE\n\
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,\n\
REPAIR OR CORRECTION.\n\
\n\
  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING\n\
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR\n\
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,\n\
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING\n\
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED\n\
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY\n\
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER\n\
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE\n\
POSSIBILITY OF SUCH DAMAGES.\n\
\n",
"\
                     END OF TERMS AND CONDITIONS\n\
\n\
            How to Apply These Terms to Your New Programs\n\
\n\
  If you develop a new program, and you want it to be of the greatest\n\
possible use to the public, the best way to achieve this is to make it\n\
free software which everyone can redistribute and change under these terms.\n\
\n\
  To do so, attach the following notices to the program.  It is safest\n\
to attach them to the start of each source file to most effectively\n\
convey the exclusion of warranty; and each file should have at least\n\
the \"copyright\" line and a pointer to where the full notice is found.\n\
\n\
    <one line to give the program's name and a brief idea of what it does.>\n\
    Copyright (C) <year>  <name of author>\n\
\n\
    This program is free software; you can redistribute it and/or modify\n\
    it under the terms of the GNU General Public License as published by\n\
    the Free Software Foundation; either version 2 of the License, or\n\
    (at your option) any later version.\n\
\n\
    This program is distributed in the hope that it will be useful,\n\
    but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
    GNU General Public License for more details.\n\
\n\
    You should have received a copy of the GNU General Public License along\n\
    with this program; if not, write to the Free Software Foundation, Inc.,\n\
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.\n\
\n\
Also add information on how to contact you by electronic and paper mail.\n\
\n\
If the program is interactive, make it output a short notice like this\n\
when it starts in an interactive mode:\n\
\n\
    Gnomovision version 69, Copyright (C) year name of author\n\
    Gnomovision comes with ABSOLUTELY NO WARRANTY; for details type `show w'.\n\
    This is free software, and you are welcome to redistribute it\n\
    under certain conditions; type `show c' for details.\n\
\n\
The hypothetical commands `show w' and `show c' should show the appropriate\n\
parts of the General Public License.  Of course, the commands you use may\n\
be called something other than `show w' and `show c'; they could even be\n\
mouse-clicks or menu items--whatever suits your program.\n\
\n\
You should also get your employer (if you work as a programmer) or your\n\
school, if any, to sign a \"copyright disclaimer\" for the program, if\n\
necessary.  Here is a sample; alter the names:\n\
\n\
  Yoyodyne, Inc., hereby disclaims all copyright interest in the program\n\
  `Gnomovision' (which makes passes at compilers) written by James Hacker.\n\
\n\
  <signature of Ty Coon>, 1 April 1989\n\
  Ty Coon, President of Vice\n\
\n\
This General Public License does not permit incorporating your program into\n\
proprietary programs.  If your program is a subroutine library, you may\n\
consider it more useful to permit linking proprietary applications with the\n\
library.  If this is what you want to do, use the GNU Lesser General\n\
Public License instead of this License.\n\
"};
static uint32_t GPLv2_linecount=sizeof(GPLv2)/sizeof(char *);

//
// This procedure is used to display the full GPLv2 license on STDOUT
//
void show_GPLv2(void) {
  for (uint32_t i=UINT32_C(0);i<GPLv2_linecount;++i)
    fputs(GPLv2[i],stdout);
  return;
}
