//
// parseinput.c - This is C version of the INPUT support
//                for Mr. Ham's ECMA-55 Minimal BASIC compiler.
//
// Copyright (C) 2014  John Gatewood Ham
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// version 2 of the Free Software Foundation License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#if defined(UNIT_TEST) || defined(TROUBLE)
#include <stdio.h>
#endif

#define T_COMMA 4              // token for ','
#define T_EOL 12               // token for '\n'
#define T_REAL 43              // token for a real number string 
#define T_UQSTRING 62          // token for an unquoted string
#define T_QSTRING 40           // token for a quoted string
#define MAX_STRING_BYTES 18U   // maximum number of bytes that a quoted or unquoted
                               // string can contain
#define MAX_INPUT_BYTES 80U    // maximum number of bytes that can be input at once in
                               // response to an INPUT query
#define MAX_INPUT_VARS 30U     // maximum number of variables permitted on one INPUT line (and in runtime INPUT stack)
#define STDIN 0                // file number for STDIN
#define STDOUT 1               // file number for STDOUT

static struct rt_token {                // type of one element in temporary input stack
  unsigned char tokeid;                 // numeric token id
  char toketext[2*MAX_STRING_BYTES+1];  // 2*MAX_STRING_BYTES + 1 terminator
} rt_nput_stack[MAX_INPUT_VARS];                       // runtime INPUT temporary stack
static unsigned long int rt_nput_stack_top=0UL;        // runtime INPUT stack pointer, 0UL means empty
static char informat[MAX_INPUT_VARS+1]={0};            // format statement (+1 for terminator) for actual types found by scan

// At runtime, scanning values is required for the INPUT statment.
// This is done using a DFSM (deterministic finite state machine).
// This machine uses character classes and states.
//
// The character_class is a conversion map to convert between
// an input 7 bit ASCII value and a character class number for use
// with the DFSM used by the INPUT statement scanner.  
// character_class is a vector in which there are 127 elements.
// The ASCII code of the byte in question is used as an
// index into the character_class vector to get the
// character class number which is stored in that element.
//
//  0  is A..D,F..Z upper case
//  1  is E upper case
//  2  is DIGIT
//  3  is ADD
//  4  is MINUS
//  5  is COMMA
//  6  is SPACE
//  7  is NEWLINE
//  8  is DOT
//  9  is QUOTE
// 10  is anything else
static const unsigned char character_class[128]={
10,10,10,10,10,10,10,10,10,10, 7,10,10,10,10,10,10,10,10,10, 
10,10,10,10,10,10,10,10,10,10,10,10, 6,10, 9,10,10,10,10,10,
10,10,10, 3, 5, 4, 8,10, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,10,10,
10,10,10,10,10, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,10,10,10,10,10,10,10,10,10,
10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,
10,10,10,10,10,10,10,10
};

// states is a matrix containing the DFSM used to scan
// data from an INPUT statement.  The rows correspond to the
// states, and the columns correspond to character classes.
// To find the next action, the character class value for the
// current byte is used as the column index, and the current
// state is used as the row index in the states matrix.
// Actions can be accept (see list below), moving to a new
// state, or reject.
//  13 accept T_COMMA, 14 accept T_EOL, 15 accept T_INTEGER, 16 accept T_REAL,
//  17 accept T_UQSTRING, 18 accept T_QSTRING, 19 reject
//
//state      A..D     E    digit    +      -      ,   space   eol      .      "   other
//           F..Z
static const unsigned char states[][11] = {
/*   0 */   { 8UL,   8UL,   2UL,   1UL,   1UL,  13UL,   0UL,  14UL,   3UL,   9UL,  19UL, },
/*   1 */   { 8UL,   8UL,   2UL,   8UL,   8UL,  17UL,   8UL,  17UL,   3UL,  19UL,  19UL, },
/*   2 */   { 8UL,   5UL,   2UL,   8UL,   8UL,  15UL,  11UL,  15UL,   3UL,  19UL,  19UL, },
/*   3 */   { 8UL,   5UL,   4UL,   8UL,   8UL,  16UL,  12UL,  16UL,   8UL,  19UL,  19UL, },
/*   4 */   { 8UL,   5UL,   4UL,   8UL,   8UL,  16UL,  12UL,  16UL,   8UL,  19UL,  19UL, },
/*   5 */   { 8UL,   8UL,   7UL,   6UL,   6UL,  17UL,   8UL,  17UL,   8UL,  19UL,  19UL, },
/*   6 */   { 8UL,   8UL,   7UL,   8UL,   8UL,  17UL,   8UL,  17UL,   8UL,  19UL,  19UL, },
/*   7 */   { 8UL,   8UL,   7UL,   8UL,   8UL,  16UL,  12UL,  16UL,   8UL,  19UL,  19UL, },
/*   8 */   { 8UL,   8UL,   8UL,   8UL,   8UL,  17UL,  10UL,  17UL,   8UL,  19UL,  19UL, },
/*   9 */   { 9UL,   9UL,   9UL,   9UL,   9UL,   9UL,   9UL,  19UL,   9UL,  18UL,   9UL, },
/*  10 */   { 8UL,   8UL,   8UL,   8UL,   8UL,  17UL,  10UL,  17UL,   8UL,  19UL,  19UL, },
/*  11 */   { 8UL,   8UL,   8UL,   8UL,   8UL,  15UL,  11UL,  15UL,   8UL,  19UL,  19UL, },
/*  12 */   { 8UL,   8UL,   8UL,   8UL,   8UL,  16UL,  12UL,  16UL,   8UL,  19UL,  19UL, },
};

void mystrcpy(char *dest, const char *source);
unsigned long int mystrlen(const char *s);
int mystrcmp(const char *left,const char *right);
void mytrim(char *s);
bool parseinput(const char buffer[],const unsigned long int buflen);
void printnewline(void);
void printstring2(const char *s);
void doinput(char *infmt);
signed long int INPUT(char *buffer,unsigned long int buflen);

// ASCIIZ string copy routine
// Copy the ASCIIZ string passed in source buffer to the
// dest buffer.  The dest buffer must be large enough, and
// it cannot be NULL.
void mystrcpy(char *dest, const char *source) {
  while ((*dest++=*source++));
}

// ASCIIZ string length routine
// this will return the number of bytes in the ASCIIZ string passed
// in the s parameter.  zero is returned for two cases:
// s is NULL, or the first byte of S is zero.
unsigned long int mystrlen(const char *s) {
  unsigned long int n=0UL;                           // number of bytes in ASCIIZ string in s, initialize to zero

  while (*s++)                                       // while the next byte in s is not the end of string
    ++n;                                             // add one to the counter n and move to the next byte
  return n;                                          // return number of bytes in ASCIIZ string in s
}

// ASCIIZ string compare routine
// this will compare two ASCIIZ strings stored in the left and right
// buffers and return a value indicating whether the left is less than,
// equal to, or greater than the right.  This will return a number less
// than zero if ASCIIZ string stored in the left buffer is less than 
// the ASCIIZ string stored in the right buffer, zero if they are equal, 
// or a positive number if the ASCIIZ string stored in the left buffer
// is greater than the ASCIIZ string stored in the right buffer.
// left is a buffer containing an ASCIIZ string
// right is a buffer containing an ASCIIZ string
int mystrcmp(const char *left,const char *right) {
  while ((*left!=0) && (*right!=0) && (*left==*right)) // while the left string still has a non-zero byte and
                                                       // the right string still has a non-zero byte and
                                                       // the correspnding bytes are equal
     *left++,*right++;                                 // move to the next byte in both strings
  return (*left-*right);                               // return the result
}

// ASCIIZ string trim (remove leading and trailing spaces
//                     inplace)
// s is an input ASCIIZ string that we modify in place (destructive)
// s cannot be NULL
void mytrim(char *s) {
  unsigned long int sindex; // index from front of string
  unsigned long int eindex; // index from end of string
  unsigned long int i;
  unsigned long int l;      // length of string after trim

  if (s[0]==0) return;      // don't do anything if the string is empty
                            // first we go forward
  sindex=0UL;               // start with first byte
  while ((s[sindex]!=0) &&  // while it is not the end of the string
         ((s[sindex]==' ')||(s[sindex]=='\t')))  // but the current byte is whitespace
    ++sindex;               // move to the next byte
  eindex=sindex;            // assume first non-blank is final non-blank
  while (s[eindex]!=0)      // while it is not the end of the string
    ++eindex;               // move to the next byte
  --eindex;                 // ok, that last move was past the end, so undo it
                            // now we back up
  while ((eindex>sindex) && // while we are still in the part that is not leading whitespace
         ((s[eindex]==' ')||(s[eindex]=='\t'))) // and we are still in trailing whitespace
    --eindex;               // move to the previous byte
  l=eindex-sindex+1UL;      // calculate length of string after trim
  for (i=0;i<l;++i)         // for each final destination byte
    s[i]=s[i+sindex];       // move source byte into place
  s[l]=0;                   // terminate string
}

// This boolean function will parse the input line buffer passed in 
// buffer with the length passed inbuflen.  It will push the values 
// obtained onto the rt_nput_stack and generate a new format string 
// in informat.  returns true if the input is valid, false otherwise
// buffer is the buffer that was fetched from the user with a line input
// buflen is the length of the buffer in bytes
bool parseinput(const char buffer[],const unsigned long int buflen) {
  bool done=false;
  unsigned long int curstate=0UL;     // current state in DFSM
  unsigned long int bufpos=0UL;       // current position in buffer[]
  unsigned long int tokepos=0UL;      // current position in tokestring[]
  unsigned long int itemcount=0UL;    // total number of tokens encountered so far
  char tokestring[MAX_INPUT_BYTES+1]; // text of current token we are building (+1 for terminator)

  rt_nput_stack_top=0UL;              // reset stack to empty
  informat[0]=0;                      // reset format string to empty
  tokestring[tokepos]=0;              // reset current token string to empty
  while (!done) {                     // while bytes remain in buffer[] to process
#ifdef TROUBLE
    unsigned long int oldstate=curstate; // save current state (becomes previous state)
#endif
    // move to next state
    curstate=states[curstate][character_class[(unsigned int)buffer[bufpos]]];
#ifdef TROUBLE
    fprintf(stderr,"transition from %lu to %lu on %u\n",oldstate,curstate,buffer[bufpos]);
#endif
    switch (curstate) {                                     // check for final states
      case 13UL:                                            // accept comma token
          ++itemcount;                                      // add 1 to the total number of items accepted so far
          bufpos++;
          if (!((itemcount & 1UL)==0UL)) {                  // if an odd number of previous tokens have been processed
            printstring2("Error: comma encountered when data expected\n"); // output error message
            done=true;                                      // end processing buffer[]
            curstate=19UL;                                  // and reject
            break;
          }
          tokepos=0UL;                                      // reset token string to empty
          tokestring[0]=0;                                  // make tokestring[] an empty string
          curstate=0UL;                                     // reset  DFSM state to start state to search for next token
          break;
      case 14UL:                                            // accept end-of-line token
          ++itemcount;                                      // add 1 to the total number of items accepted so far
          if (!((itemcount & 1UL)==0UL)) {                  // if an odd number of previous tokens have been processed
            printstring2("Error: newline encountered when data expected\n"); // output error message
            curstate=19UL;                                  // reject
          }
          informat[rt_nput_stack_top]=0;                    // terminate token string
          done=true;                                        // and end processing buffer[]
          break;
      case 15UL:                                            // accept integer
      case 16UL:                                            // accept real number
          ++itemcount;                                      // add 1 to the total number of items accepted so far
          if (!((itemcount & 1UL)==1UL)) {                  // if an even number of previous tokens have been processed
            printstring2("Error: number encountered when comma or newline expected\n"); // output error message
            done=true;                                      // end processing buffer[]
            curstate=19UL;                                  // and reject
            break;
          }
          tokestring[tokepos]=0;                            // terminate token string
          mytrim(tokestring);                               // remove any leading and trailing whitespace
          if (mystrlen(tokestring)>2*MAX_STRING_BYTES) {
            //printf("Error: numeric constant string maximum size is %u bytes\n",2*MAX_STRING_BYTES);
            printstring2("Error: numeric constant string maximum size is 36 bytes\n"); // output error message
            done=true;                                      // end processing buffer[]
            curstate=19UL;                                  // and reject
            break;
          }
          if (rt_nput_stack_top>=MAX_INPUT_VARS) {          // if runtime INPUT stack is full
            printstring2("Error: Runtime failure: too many input variables\n"); // output error message
            done=true;                                      // end processing buffer[]
            curstate=19UL;                                  // and reject
            break;
          }
          informat[rt_nput_stack_top]='N';                  // mark this item as numeric in format string
          rt_nput_stack[rt_nput_stack_top].tokeid=T_REAL;   // token ID
          mystrcpy(rt_nput_stack[rt_nput_stack_top++].toketext,tokestring); // token text
          tokestring[0]=0;                                  // make tokestring[] an empty string
          tokepos=0UL;                                      // reset index into token string to the beginning
          curstate=0UL;                                     // reset current state to start state for next token
          break;
      case 17UL:                                            // accept unquoted string
          ++itemcount;                                      // add 1 to the total number of items accepted so far
          if (!((itemcount & 1UL)==1UL)) {                  // if an even number of previous tokens have been processed
            printstring2("Error: unquoted string encountered when comma or newline expected\n"); // output error message
            done=true;                                      // end processing buffer[]
            curstate=19UL;                                  // and reject
            break;
          }
          tokestring[tokepos]=0;                            // terminate token string
          mytrim(tokestring);                               // remove any leading or trailing whitespace
          if (mystrlen(tokestring)>MAX_STRING_BYTES) {      // if input token text is too long
            //printf("Error: unquoted string maximum size is %u bytes\n",MAX_STRING_BYTES);
            printstring2("Error: unquoted string maximum size is 18 bytes\n"); // output error message
            done=true;                                      // end processing buffer[]
            curstate=19UL;                                  // and reject
            break;
          }
#ifdef TROUBLE
          fprintf(stderr,"Length of tokestring '%s' is %u\n",tokestring,mystrlen(tokestring));
#endif
          if (rt_nput_stack_top>=MAX_INPUT_VARS) {          // if runtime INPUT stack is full
            printstring2("Error: Runtime failure: too many input variables\n"); // output error message
            done=true;                                      // end processing buffer[]
            curstate=19UL;                                  // and reject
            break;
          }
          informat[rt_nput_stack_top]='U';                  // mark this item as an unquoted string in format string
          // push token record onto runtime INPUT stack
          rt_nput_stack[rt_nput_stack_top].tokeid=T_QSTRING;// token ID
          mystrcpy(rt_nput_stack[rt_nput_stack_top++].toketext,tokestring); // token text
          tokestring[0]=0;                                  // make tokestring[] an empty string
          tokepos=0UL;                                      // reset index into token string to the beginning
          curstate=0UL;                                     // reset current state to start state for next token
          break;
      case 18UL:                                            // accept quoted string
          ++itemcount;                                      // add 1 to the total number of items accepted so far
          // tricky - we must consume the quote for mytrim(), then remove it after
          tokestring[tokepos++]=buffer[bufpos];             // DO NOT INCREMENT bufpos HERE
          if (!((itemcount & 1UL)==1UL)) { // error
            printstring2("Error: quoted string encountered when comma or newline expected\n"); // output error message
            done=true;                                      // end processing buffer[]
            curstate=19UL;                                  // and reject
            break;
          }
          tokestring[tokepos]=0;                            // terminate token string
          mytrim(tokestring);                               // remove any leading and trailing whitespace
          if (mystrlen(tokestring)>MAX_STRING_BYTES+2U) {   // add 2 for the quotes that get removed
            //printf("Error: quoted string maximum size is %u bytes\n",MAX_STRING_BYTES);
            printstring2("Error: quoted string maximum size is 18 bytes\n"); // output error message
            done=true;                                      // end processing buffer[]
            curstate=19UL;                                  // and reject
            break;
          }
#ifdef TROUBLE
          fprintf(stderr,"Length of tokestring '%s' is %u\n",tokestring,mystrlen(tokestring));
#endif
          if (tokestring[0]!=0)                             // if tokestring is not empty
             tokestring[mystrlen(tokestring)-1U]=0;         // remove trailing quote now
          bufpos++;                                         // and move to next byte
          if (rt_nput_stack_top>=MAX_INPUT_VARS) {          // if runtime INPUT stack is full
            printstring2("Error: Runtime failure: too many input variables\n"); // output error message
            done=true;                                      // end processing buffer[]
            curstate=19UL;                                  // and reject
            break;
          }
          informat[rt_nput_stack_top]='S';                  // record the fact that this token was a string in the 'found' format string
          // push token record onto runtime INPUT stack
          rt_nput_stack[rt_nput_stack_top].tokeid=T_EOL;    // token ID
          mystrcpy(rt_nput_stack[rt_nput_stack_top++].toketext,tokestring+1);  // token text, +1 to skip leading quote
          tokestring[0]=0;                                  // make tokestring[] an empty string
          tokepos=0UL;                                      // reset index into tokestring[] to beginning
          curstate=0UL;                                     // reset current state to start state for next token
          break;
      case 19UL:                                            // reject
          done=true;                                        // so end processing buffer[]
          break;
      default:                                              // this is not a final state or error state so
          tokestring[tokepos++]=buffer[bufpos++];           // add current byte to tokestring[]
          if ((tokepos>MAX_INPUT_BYTES)||(bufpos>buflen)) { // but if tokestring[] is too long
            curstate=19UL;                                  // reject
            done=true;                                      // and end processing buffer[]
          }
          break;
    }
  }
#ifdef TROUBLE
  printstring2(informat);printnewline();
#endif
  return (curstate==14UL);                                  // return true if everything is OK and last token was an end-of-line
                                                            // otherwise return false
}

// Line input routine (get one line of input from STDIN)
// This had to be one one byte at a time in order to work with redirection in the
// shell.  We MUST stop when a newline is encountered.
// buffer is a pointer to the INPUT line buffer
// buflen is the length of buffer
// returns number of bytes in this line of input on succes, -1 on error
signed long int INPUT(char *buffer,unsigned long int buflen) {
  char *tbuffer=buffer;                                     // pointer to the start of the actual input buffer
  char *bufmax=buffer+buflen;                               // pointer to end of input buffer

  // read one byte at a time so we don't read too far
  do {                                                      // repeat
    if (read(STDIN,tbuffer,1)<1) {                          // read one byte
      printstring2("EOF ON INPUT - PROGRAM TERMINATED IN AN ERROR STATE\n"); // output error message
      printnewline();                                       // output a newline
      return -1UL;                                          // return an error indicator
    }
    if (*(tbuffer++)=='\n') break;                          // if this byte was a newline, break out of loop
  } while (tbuffer<bufmax);                                 // while more bytes of input exist
  *tbuffer=0;                                               // terminate the input buffer
  return (unsigned long int)(tbuffer-buffer);               // return the number of bytes in this line of input
}

// This procedure will output a newline to STDOUT
// and if it cannot, the program will abort().
void printnewline(void) {
  if (write(STDOUT,"\n",1)<1L) abort();
}

// This procedure will display the ASCIIZ string passed in the 
// parameter s to STDOUT.  If there is an error, display a 
// diagnostic message and abort() the program.
void printstring2(const char *s) {
  unsigned int slen=mystrlen(s);                            // length of string passed in s parameter
  unsigned int offset=0U;                                   // offset of first byte in s that has not yet been printed
  signed long int byteswritten=0L;                          // total number of bytes written so far
  signed long int so;                                       // if greater than zero then this is the number of bytes written by write()
                                                            // otherwise there was an error

  if (slen>0U) {                                            // if the string passed in parameter s was not empty
    do {                                                    // repeat
      so=write(STDOUT,s+offset,slen-byteswritten);          // try to write the remaining bytes to STDOUT, with so containg the 
                                                            // actual number output in case we had a partial write()
      if (so<1L) {                                          // if the write() failed
        fprintf(stderr,"so=%ld,offset=%lu,byteswritten=%ld,slen=%lu\n",   // output an error message with low-level details
            so,offset,byteswritten,slen);
        abort();                                            // and abort() the program
      }
      byteswritten+=so;                                     // otherwise add the number of bytes written this time to the total number
                                                            // written so far
    } while (byteswritten<slen);                            // until all bytes have been output
  }
}

// When this returns you definitely have the right stuff
// in the input queue.  fmt is the format string containg
// 1 type byte per item.  Type bytes can be
// U: unquoted string
// S: quoted string
// N: number (floating point or integer, stored as floating point)
void doinput(char *fmt) {
  char buffer[MAX_INPUT_BYTES+1];                          // INPUT line buffer (+1 for terminator)
  bool rez=true;                                           // flag indicating succes (true) or failure (false)
  unsigned int i;                                          // index used to iterate through the runtime input stack
  unsigned int needcount;                                  // number of items user needs to INPUT
  unsigned int tries=0U;                                   // number of times the user has input a line of items
  unsigned int maxtries=5U;                                // maximum number of tries the user can have
  signed long int bytesread;                               // number of bytes returned in buffer after the call to INPUT()

  do {                                                     // repeat
    buffer[0]=0;                                           // initialize buffer
    buffer[1]=0;
    printstring2("? ");                                    // print prompt
    bytesread=INPUT(buffer,MAX_INPUT_BYTES);               // call INPUT() function to get a line of input and store it in buffer
    if (bytesread<1) {                                     // if there was no input available
      printstring2("EOF ON INPUT");                        // output error message
      printnewline();                                      // print a newline
      rez=false;                                           // indicate failure
      break;
    }
    rez=parseinput(buffer,bytesread);                      // scan buffer into tokens using parseinput() function
    if (rez && (mystrcmp(fmt,informat)!=0)) {              // if parse was OK and types do not match desired types or number of items is wrong
      rez=false;                                           // assume we will fail
      unsigned int lhave=mystrlen(informat);               // get number of items in format string (1 type byte/item)
      if (rt_nput_stack_top==lhave) {                      // if the number of items is correct then
        for (i=0U;i<lhave;i++) {                           // for each item
          if ((informat[i]=='N') && (fmt[i]=='S')) {       // if the type that was found numeric and WE NEED A STRING
             if (mystrlen(rt_nput_stack[i].toketext)>MAX_STRING_BYTES) { // if the token was too long then
                printstring2("INVALID INPUT: string exceeds 18 bytes");  // output error message
                printnewline();                            // and print newline
                rez=false;                                 // and set rez flag to indicate failure
                continue;                                  // and keep going, since we need all types in informat to compare against fmt
             }                                             // and we want to test later entries that might be OK
             informat[i]='S';                              // OK, change found type to string
             rt_nput_stack[i].tokeid=T_QSTRING;            // and change type of item on stack to string
          }
        }
        rez=(mystrcmp(fmt,informat)==0);                   // after all checks, if the input format and output format are compatible
                                                           // indicate success, else indicate failure
      }
    }
    if (!rez) {                                            // if there is a failure
      printstring2("INVALID INPUT: expected ");            // begin display of error message
      needcount=mystrlen(fmt);                             // record number of INPUT items that were required
      for (i=0U;i<needcount;++i) {                         // for each required INPUT item
        if ((i>0U) && (i<needcount)) printstring2(",");    // if it is not the first item, output a comma separator
        switch(fmt[i]) {                                   // and display the type that was required which can be
            case 'S': printstring2("QUOTED STRING"); break;       // quoted string
            case 'N': printstring2("NUMBER"); break;              // or numeric
            case 'U': printstring2("UNQUOTED STRING"); break;     // or unquoted string
            default: printstring2("UNKNOWN FORMAT!"); break;      // and if known (ICE) let user know that too
        }
      }
      printnewline();                                      // print a newline after the complete error message text is output
    }
    tries++;                                               // increment the number of tries
  } while ((tries<maxtries) && (!rez));                    // until number of tries exceeds maxtries and we still have bad input
  return;
}

// Self-test code
#ifdef UNIT_TEST
int main(void) {
  unsigned long int i;
  char buffer[MAX_INPUT_VARS+1U]={0};

  fputs("Enter the format string of N, S and U:",stdout);
  fflush(stdout);
  INPUT(buffer,MAX_INPUT_VARS);
  if (buffer[0]!=0)
     buffer[mystrlen(buffer)-1U]=0; // chop trailing newline
  doinput(buffer);
  fprintf(stderr,"item count=%lu\n",rt_nput_stack_top);
  for (i=0UL;i<rt_nput_stack_top;++i) {
    fprintf(stderr,"Token #%lu: %u, \"%s\"\n",i,
           rt_nput_stack[i].tokeid,rt_nput_stack[i].toketext);
  }
  return 0;
}
#endif
