//
// parser2.c - This is the file containing the parser
//             for Mr. Ham's ECMA-55 Minimal BASIC compiler.
//
// Copyright (C) 2013,2014  John Gatewood Ham
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// version 2 of the Free Software Foundation License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
#define _POSIX_C_SOURCE 200809L
#include <features.h>
#include <stdlib.h>
#include <limits.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "scanner2.h"
#include "globals.h"
#include "parser2.h"
#include "lineno.h"
#include "symbol_table.h"
#include "g_fmt_BASIC_normal.h"
#include "codegen.h"

#define ERROR_BUFFER_MAX 81   // specify size for error message ASCIIZ buffers (80 bytes + terminator)

#ifdef UNIT_TEST
// map used to convert relop values to strings for display
const char * const relop_names[]={"<","<=","=",">=",">","<>"};
#endif


static bool have_DATA=false; // if true, program contains a DATA statement
static uint32_t READ_lineno=UINT32_C(0);  // if need_READ is true, then this will be updated during
                             // the parse to contain the line number of the first READ statement in
                             // the program for use in the case when there is a READ but no DATA
                             // statement and an error message must be issued.

// dedicated compile-time FOR loop stack to ensure every NEXT matches
// the preceeding FOR, and every FOR has a corrsponding NEXT.
static struct forstack_entry {
  uint32_t lineno;
  char tol_label[MAX_FOR_LABEL_BYTES+1]; // label for start of loop test in assembly output
  char bol_label[MAX_FOR_LABEL_BYTES+1]; // label for exiting the loop in assembly output
  char ndex_var[MAX_FOR_VAR_BYTES+1];    // the ASCIIZ name of the loop index variable
  char inc_var[MAX_FOR_VAR_BYTES+1];     // the ASCIIZ name of the hidden variable that holds the loop increment
  char limit_var[MAX_FOR_VAR_BYTES+1];   // the ASCIIZ name of the hidden variable that holds the limit used in the loop test
  char scope[MAX_SCOPE_BYTES+1];         // the scope indicator for this loop (.X.Y.Z would be FOR Z inside FOR Y inside 
                                         // FOR X which is at global scope level
                                         // The '.' is used as a delimiter, and a the global scope has no index variable name.

} forstack[FORSTACK_DEPTH];
static uint32_t forstack_ptr=UINT32_C(0); // stack pointer in forstack
// FOR loop scope indicator; empty string is GLOBAL scope
// This begins with a . and has the index variable name.  For example, .A0.B.Z9 would mean
// FOR Z9 inside of FOR B inside of FOR A0 which is at global level
// The '.' is used as a delimiter, and a the global scope has no index variable name.
static char current_scope[MAX_SCOPE_BYTES+1]="";

// set to false when you see an array reference or a DIM statement
// OPTION BASE is only permitted if this is true
static bool can_option_base=true;

// output file handles (DO NOT CLOSE LOCALLY, THEY ARE CLOSED IN MAIN)
static FILE *outfile=NULL;  // main assembly language output file
static FILE *xtraoutfile=NULL;  // extra assembly language output file for UDF definitions
// set to true during user-defined function definition (from just after DEF to the next newline)
static bool in_udf_definition=false;
// current UDF name,pattern is "FN?" with ? is a letter 'A'..'Z'
static char curfunc_name[4]={0};
// current UDF argument name, pattern is ".LFN?_?" with first ? is letter 'A'..'Z', 2nd ? is arg name
static char curfunc_argname[10]={0};

// numeric expression
static bool g_primary(void);
static bool g_factor(void);
static bool g_term(void);
static bool g_string_expression(void);
static bool g_numeric_expression(void);
static bool g_expression(void);
static bool g_relational_expression(enum relop *op,bool *stringexpression);
// miscellaneous
static bool g_lineno(uint32_t *lno,char *why,const bool jumptarget);
// normal statements
static bool g_defstmt(void);
static bool g_ifstmt(void);
static bool g_on_gotostmt(void);
static bool g_target_variable(void);
static bool g_inputstmt(void);
static bool g_readstmt(void);
static bool g_datum(void);
static bool g_datastmt(void);
static bool g_dimstmt(void);
static bool g_nextstmt(void);
static bool g_forstmt(void);
static bool g_print_item(void);
static bool g_print_item_tail(bool *neednewline);
static bool g_printlist(bool *neednewline);
static bool g_printstmt(void);
static bool g_letstmt(void);
static bool g_randomizestmt(void);
static bool g_restorestmt(void);
static bool g_optionbasestmt(void);
static bool g_gotostmt(void);
static bool g_go_stmt(void);
static bool g_gosubstmt(void);
static bool g_returnstmt(void);
static bool g_remstmt(void);
static bool g_stopstmt(void);
static bool g_statement(void);
static bool g_lines(void);
static bool g_program(void);
// built-in functions
static bool g_abs(void);
static bool g_atn(void);
static bool g_cos(void);
static bool g_exp(void);
static bool g_int(void);
static bool g_log(void);
static bool g_sgn(void);
static bool g_sin(void);
static bool g_sqr(void);
static bool g_tan(void);
static bool g_rnd(void);
static bool g_integer(void);
static bool g_real(void);
static bool g_udf(void);
static bool g_nvar(void);
static bool g_navar(void);
// helper functions
static bool convert_unsigned_integer(const char *s,uint32_t *lineno, char *why);
static bool match(enum token_ids tid,const char *from);

//
// This function is used to convert an ASCIIZ string to an unsigned integer.
// The incoming string is specified by the s parameter, a pointer to the outgoing 
// line number is specified by lineno, and the why parameter is a pointer to a
// buffer to contain an error message if the conversion was not successful.
// The error buffer must have room for an 80 byte ASCIIZ error message.
// This function returns true on success, false otherwise.
//
static bool convert_unsigned_integer(const char *s,uint32_t *lineno, char *why) {
  char *endptr=NULL;
  static uint32_t tlineno;

  errno=0;
  if ('-'==s[0]) {
    snprintf(why,80,"ERROR: Invalid number %s; unsigned integers cannot be negative",s);
    return false;
  }
  tlineno=strtoul(s,&endptr,10);
  switch (errno) {
    case 0:     // OK
       break;
    case ERANGE:
       snprintf(why,80,"ERROR: Invalid number %s; must be between 1 and %" PRIu32 " inclusive\n",s,UINT_MAX);
       return false;
    default:
       snprintf(why,80,"ERROR: Invalid number %s; cannot convert to uint32_t\n",s);
       return false;
  }
  if (endptr!=s+strnlen(s,20)) {
    snprintf(why,80,"ERROR: Invalid number %s; trailing garbage is '%s'(scanner failure?)\n",s,endptr);
    return false;
  }
  *lineno=tlineno;
  return true;
}

//
// This fuction is used to ensure the next token is the desired token.
// The paraemter tid is the token identifier of the desired token.
// The parameter from is the ASCIIZ string is used for error messages
// and is design to carry the name of the calling function.  Returns
// true on success, false otherwise.
//
static bool match(enum token_ids tid,const char *from) {
  struct token curtoken=get_next_token();
  bool retval;

  if (verbose) {
     if ((T_INTEGER==tid)||(T_REAL==tid)||(T_NVAR==tid)||(T_SVAR==tid)||(T_FNID==tid))
       debug_printf("match(%s,%s) with value '%s'\n",token_names[tid].namestring,from,curtoken.toketext);
     else
       debug_printf("match(%s,%s)\n",token_names[tid].namestring,from);
  }
  if (curtoken.tid==tid) {
    retval=true;
  } else {
    // +2 is to skip the 'T_' prefix
    fprintf(stderr,"ERROR on line %" PRIu32 ": Expected %s, but got %s with value '%s' instead\n",
        curtoken.lno, token_names[tid].namestring+2, token_names[curtoken.tid].namestring+2,
        curtoken.toketext);
    retval=false;
  }
  free(curtoken.toketext);
  curtoken.toketext=NULL;
  return retval;
}

//
// This function will emit code to call the RND function within an arithmetic expression.
// Returns true on success, false otherwise.
//
// g_rnd -> T_RND
//
static bool g_rnd(void) {
  // this is a wierd one - it is a psuedo-variable, no argument function
  debug_printf("g_rnd -> T_RND\n");
  if (!match(T_RND,__func__)) return false;
  if (!dornd(outfile)) return false;
  return true;
}

//
// This function will emit code to call the ABS function within an arithmetic expression.
// Returns true on success, false otherwise.
//
// g_abs -> T_ABS T_LPAREN g_numeric_expression T_RPAREN
//
static bool g_abs(void) {
  debug_printf("g_abs -> T_ABS T_LPAREN g_numeric_expression T_RPAREN\n");
  if (!match(T_ABS,__func__)) return false;
  if (!match(T_LPAREN,__func__)) return false;
  if (!g_numeric_expression()) return false;
  if (!match(T_RPAREN,__func__)) return false;
  if (!doabs(outfile)) return false;
  return true;
}

//
// This function will emit code to call the ATN function within an arithmetic expression.
// Returns true on success, false otherwise.
//
// g_atn -> T_ATN T_LPAREN g_numeric_expression T_RPAREN
//
static bool g_atn(void) {
  debug_printf("g_atn -> T_ATN T_LPAREN g_numeric_expression T_RPAREN\n");
  if (!match(T_ATN,__func__)) return false;
  if (!match(T_LPAREN,__func__)) return false;
  if (!g_numeric_expression()) return false;
  if (!match(T_RPAREN,__func__)) return false;
  if (!doatan(outfile)) return false;
  return true;
}

//
// This function will emit code to call the COS function within an arithmetic expression.
// Returns true on success, false otherwise.
//
// g_cos -> T_COS T_LPAREN g_numeric_expression T_RPAREN
//
static bool g_cos(void) {
  debug_printf("g_cos -> T_COS T_LPAREN g_numeric_expression T_RPAREN\n");
  if (!match(T_COS,__func__)) return false;
  if (!match(T_LPAREN,__func__)) return false;
  if (!g_numeric_expression()) return false;
  if (!match(T_RPAREN,__func__)) return false;
  if (!docos(outfile)) return false;
  return true;
}

//
// This function will emit code to call the EXP function within an arithmetic expression.
// Returns true on success, false otherwise.
//
// g_exp -> T_EXP T_LPAREN g_numeric_expression T_RPAREN
//
static bool g_exp(void) {
  debug_printf("g_exp -> T_EXP T_LPAREN g_numeric_expression T_RPAREN\n");
  if (!match(T_EXP,__func__)) return false;
  if (!match(T_LPAREN,__func__)) return false;
  if (!g_numeric_expression()) return false;
  if (!match(T_RPAREN,__func__)) return false;
  if (!doexp(outfile)) return false;
  return true;
}

//
// This function will emit code to call the INT function within an arithmetic expression.
// Returns true on success, false otherwise.
//
// g_int -> T_INT T_LPAREN g_numeric_expression T_RPAREN
//
static bool g_int(void) {
  debug_printf("g_int -> T_INT T_LPAREN g_numeric_expression T_RPAREN\n");
  if (!match(T_INT,__func__)) return false;
  if (!match(T_LPAREN,__func__)) return false;
  if (!g_numeric_expression()) return false;
  if (!match(T_RPAREN,__func__)) return false;
  if (!doint(outfile)) return false;
  return true;
}

//
// This function will emit code to call the LOG function within an arithmetic expression.
// Returns true on success, false otherwise.
//
// g_log -> T_LOG T_LPAREN g_numeric_expression T_RPAREN
//
static bool g_log(void) {
  debug_printf("g_log -> T_LOG T_LPAREN g_numeric_expression T_RPAREN\n");
  if (!match(T_LOG,__func__)) return false;
  if (!match(T_LPAREN,__func__)) return false;
  if (!g_numeric_expression()) return false;
  if (!match(T_RPAREN,__func__)) return false;
  if (!dolog(outfile)) return false;
  return true;
}

//
// This function will emit code to call the SGN function within an arithmetic expression.
// Returns true on success, false otherwise.
//
// g_sgn -> T_SGN T_LPAREN g_numeric_expression T_RPAREN
//
static bool g_sgn(void) {
  debug_printf("g_sgn -> T_SGN T_LPAREN g_numeric_expression T_RPAREN\n");
  if (!match(T_SGN,__func__)) return false;
  if (!match(T_LPAREN,__func__)) return false;
  if (!g_numeric_expression()) return false;
  if (!match(T_RPAREN,__func__)) return false;
  if (!dosgn(outfile)) return false;
  return true;
}

//
// This function will emit code to call the SIN function within an arithmetic expression.
// Returns true on success, false otherwise.
//
// g_sin -> T_SIN T_LPAREN g_numeric_expression T_RPAREN
//
static bool g_sin(void) {
  debug_printf("g_sin -> T_SIN T_LPAREN g_numeric_expression T_RPAREN\n");
  if (!match(T_SIN,__func__)) return false;
  if (!match(T_LPAREN,__func__)) return false;
  if (!g_numeric_expression()) return false;
  if (!match(T_RPAREN,__func__)) return false;
  if (!dosin(outfile)) return false;
  return true;
}

//
// This function will emit code to call the SQR function within an arithmetic expression.
// Returns true on success, false otherwise.
//
// g_sqr -> T_SQR T_LPAREN g_numeric_expression T_RPAREN
//
static bool g_sqr(void) {
  debug_printf("g_sqr -> T_SQR T_LPAREN g_numeric_expression T_RPAREN\n");
  if (!match(T_SQR,__func__)) return false;
  if (!match(T_LPAREN,__func__)) return false;
  if (!g_numeric_expression()) return false;
  if (!match(T_RPAREN,__func__)) return false;
  if (!dosqr(outfile)) return false;
  return true;
}

//
// This function will emit code to call the TAN function within an arithmetic expression.
// Returns true on success, false otherwise.
//
// g_tan -> T_TAN T_LPAREN g_numeric_expression T_RPAREN
//
static bool g_tan(void) {
  debug_printf("g_tan -> T_TAN T_LPAREN g_numeric_expression T_RPAREN\n");
  if (!match(T_TAN,__func__)) return false;
  if (!match(T_LPAREN,__func__)) return false;
  if (!g_numeric_expression()) return false;
  if (!match(T_RPAREN,__func__)) return false;
  if (!dotan(outfile)) return false;
  return true;
}

//
// This function will convert the T_INTEGER token value to an ASCIIZ string
// using David M. Gay's g_fmt(), add a symbol table entry for the float literal,
// and emit code to push that value onto the dedicated floating point stack at runtime.
// Returns true on success, false otherwise.  The g_fmt() is used to ensure the
// compile-time conversion will behave in exactly the same way as a run-time conversion
// using the assmebly version of g_fmt used in the runtime library.
//
// g_integer -> T_INTEGER
//
static bool g_integer(void) {
  struct token *curtoken=clone_next_token(); // copy of next token
  float floatval; // float value used for 32 bit math
  double doubleval; // double value use for 64 bit math
  static char tbuf[ERROR_BUFFER_MAX]; // buffer used to hold the result of g_fmt() conversion
  static char symname[21]; // buffer to hold a copy of the generated float literal name
  char *s=NULL; // pointer for use returning the generated float literal name from the code generator
  bool retval=true; // function return value

  debug_printf("g_integer -> T_INTEGER <<<value is '%s'>>>\n",curtoken->toketext);
  if (!match(T_INTEGER,__func__)) {
     retval=false;
     goto g_integer_xit;
  }
  if (use_double) {
    // reformat 64 bit float to David M. Gay's format
    sscanf(curtoken->toketext,"%lf",&doubleval);
    g_fmt(tbuf,doubleval,SIGNIFICANCE_WIDTH_WIDE_OUTPUT-UINT32_C(1),EXRAD_WIDE_OUTPUT);
  } else {
    // reformat 32 bit float to David M. Gay's format
    sscanf(curtoken->toketext,"%f",&floatval);
    g_fmt(tbuf,floatval,SIGNIFICANCE_WIDTH_NARROW_OUTPUT-UINT32_C(1),EXRAD_NARROW_OUTPUT);
  }
  if (!add_float_literal(tbuf,(char **)&s)) {
    retval=false;
    goto g_integer_xit;
  }
  strncpy(symname,s,20);
  symname[20]=0;
  s=NULL;
  if (!push_float_literal(outfile,symname,tbuf)) {
    retval=false;
    // fall through to g_integer_xit
  }
g_integer_xit:
  free(curtoken->toketext);curtoken->toketext=NULL;
  free(curtoken);curtoken=NULL;
  return retval;
}

//
// This function will convert the T_REAL token value to an ASCIIZ string
// using David M. Gay's g_fmt(), add a symbol table entry for the float literal,
// and emit code to push that value onto the dedicated floating point stack at runtime.
// Returns true on success, false otherwise.  The g_fmt() is used to ensure the
// compile-time conversion will behave in exactly the same way as a run-time conversion
// using the assmebly version of g_fmt used in the runtime library.
//
// Returns true on success, false otherwise.
//
// g_real -> T_REAL
//
static bool g_real(void) {
  struct token *curtoken=clone_next_token(); // copy of next token
  float floatval; // float value used for 32 bit math
  double doubleval; // double value used for 64 bit math
  static char tbuf[ERROR_BUFFER_MAX]; // buffer used to hold the result of g_fmt() conversion
  static char symname[21]; // buffer to hold a copy of the generated float literal name
  char *s=NULL; // pointer for returning the generated float literal name from the code generator
  bool retval=true; // function return value

  debug_printf("g_real -> T_REAL <<<value is '%s'>>>\n",curtoken->toketext);
  if (!match(T_REAL,__func__)) { 
     retval=false;
     goto g_real_xit;
  }
  if (use_double) {
    // reformat 64 bit float to David M. Gay's format
    sscanf(curtoken->toketext,"%lf",&doubleval);
    g_fmt(tbuf,doubleval,SIGNIFICANCE_WIDTH_WIDE_OUTPUT-UINT32_C(1),EXRAD_WIDE_OUTPUT);
  } else {
    // reformat 32 bit float to David M. Gay's format
    sscanf(curtoken->toketext,"%f",&floatval);
    g_fmt(tbuf,floatval,SIGNIFICANCE_WIDTH_NARROW_OUTPUT-UINT32_C(1),EXRAD_NARROW_OUTPUT);
  }
  if (!add_float_literal(tbuf,(char **)&s)) {
    retval=false;
    goto g_real_xit;
  }
  strncpy(symname,s,20);
  symname[20]=0;
  s=NULL;
  if (!push_float_literal(outfile,symname,tbuf)) {
    retval=false;
    // fall through to g_real_xit
  }
g_real_xit:
  free(curtoken->toketext);curtoken->toketext=NULL;
  free(curtoken);curtoken=NULL;
  return retval;
}

//
// This function will emit code to make a call to a UDF within an arithmetic expression.
// Semantic checks include verifying the function is defined and the correct number
// of arguments are passed, using information from the symbol table module.  Minimal BASIC
// explicitely disallows recursive UDFs, and a semantic check is made to enforce that rule.
// The in_udf_definition flag will be true if this production is called as part of the
// arithmetic expression in a DEF statement, and false otherwise.
// Returns true on success, false otherwise.
//
// g_udf -> T_FNID g_function_tail
// g_function_tail -> T_LPAREN g_numeric_expression T_RPAREN
// g_function_tail -> epsilon
//
static bool g_udf(void) {
  struct token *curtoken=clone_next_token(); // copy of next token
  bool retval=true; // function return value
  char argname[2]; // buffer to hold one-byte UDF parameter variable (argument) name

  debug_printf("g_udf -> T_FNID g_function_tail\n");
  if (!match(T_FNID,__func__)) {
    retval=false;
    goto g_udf_xit;
  }
  if (in_udf_definition) { // if we are in arithmetic expression in a DEF statement
    // recursion is not permitted - report the error and set error return code to incidate error
    if (strcmp(curtoken->toketext,curfunc_name)==0) {
      fprintf(stderr,"ERROR: User-defined function '%s' cannot reference itself on line %" PRIu32 "\n",curfunc_name,curtoken->lno);
      retval=false;
      goto g_udf_xit;
    }
  }
  if (!udf_exists(curtoken->toketext,argname)) { // calling an undefined UDF is an error
    retval=false;
    goto g_udf_xit;
  }
  // verify number of arguments
  if (peek_next_token().tid==T_LPAREN) { // function has an argument
    if (0==argname[0]) {
      fprintf(stderr,"ERROR: function '%s' called with an argument but defined without one.\n",
        curtoken->toketext);
      retval=false;
      goto g_udf_xit;
    }
    debug_printf("g_function_tail -> T_LPAREN g_numeric_expression T_RPAREN\n");
    if (!match(T_LPAREN,__func__)) {
      retval=false;
      goto g_udf_xit;
    }
    if (!g_numeric_expression()) {
      retval=false;
      goto g_udf_xit;
    }
    if (!match(T_RPAREN,__func__)) {
      retval=false;
      goto g_udf_xit;
    }
    if (!doudf(outfile,curtoken->toketext,true)) {
      retval=false;
      goto g_udf_xit;
    }
  } else {
    debug_printf("g_function_tail -> epsilon\n");
    if (argname[0]!=0) {
      fprintf(stderr,"ERROR: function '%s' called without an argument but defined with one.\n",
        curtoken->toketext);
      retval=false;
      goto g_udf_xit;
    }
    if (!doudf(outfile,curtoken->toketext,false)) {
      retval=false;
      goto g_udf_xit;
    }
  }
g_udf_xit:
  free(curtoken->toketext);curtoken->toketext=NULL;
  free(curtoken);curtoken=NULL;
  return retval;
}

//
// This will emit code to read a numeric variable in an arithmetic expression
// and leave the value on top of the dedicated floating point stack.
// The variable loop in the symbol table has two paths:
// 1) normal, just look up a scalar numeric variable in the symbol table.
//         This path is taken if the flag in_udf_definition is false.
//         If the variable exists, emit code to push its value.
// 2) UDF, if we are in an arithmetic expression as part of a DEF statement,
//         first check if the variable is the UDF parameter variable name (if there is one).
//         If the parameter exists, emit code to push its value.
//         If that fails look up the scalar numeric variable in the symbol table.
//         If the variable exists, emit code to push its value.
//         This UDF path is only taken if the flag in_udf_definition is true.
//         
// Returns true on success, false otherwise.
//
// g_nvar -> T_NVAR
//
static bool g_nvar(void) {
  struct token *curtoken=clone_next_token(); // copy of next token
  bool retval=true; // function return value
  char varname[10]={0}; // buffer to hold UDF parameter (argument) name

  debug_printf("g_nvar -> T_NVAR\n");
  if (!match(T_NVAR,__func__)) {
    retval=false;
    goto g_nvar_xit;
  }
  if (in_udf_definition) {
    if (curtoken->toketext[0]==curfunc_argname[6]) {
      strcpy(varname,curfunc_argname);   // REDIRECT FOR FUNCTION PARAMETER SCALAR NUMERIC VARIABLE
      debug_printf("using '%s' instead of '%s' because '%s' is UDF parameter\n",
        varname,curtoken->toketext,curtoken->toketext);
    } else {
      debug_printf("using unchanged '%s' in UDF since it is not a parameter\n",
        curtoken->toketext);
      strcpy(varname,curtoken->toketext);
    }
  } else {
    debug_printf("using unchanged '%s' since we are not in a UDF\n",
      curtoken->toketext);
    strcpy(varname,curtoken->toketext);
  }
  if ((!in_udf_definition) && (!nvar_exists(varname))) {
     if (!assign_nvar(varname,curtoken->lno)) {
       if (navar_exists(varname)) {
          fprintf(stderr,"ERROR: on line %" PRIu32 " attempt to use array variable %s as a scalar\n",
               curtoken->lno,varname);
       } else {
          fprintf(stderr,"ERROR: on line %" PRIu32 " attempt to use scalar variable %s for read failed and I do not know why\n",
               curtoken->lno,varname);
       }
       retval=false;
       goto g_nvar_xit;
     }
  }
  if (in_udf_definition && ('.'==varname[0])) {
    if (!push_nparm(outfile,varname,curfunc_name)) {
      retval=false;
      // fall through to g_nvar_xit
    }
  } else {
    if (!push_nvar(outfile, varname)) {
      retval=false;
      // fall through to g_nvar_xit
    }
  }
g_nvar_xit:
  free(curtoken->toketext);curtoken->toketext=NULL;
  free(curtoken);curtoken=NULL;
  return retval;
}

//
// This will emit code to read a numeric array variable in an arithmetic expression
// and leave the value on top of the dedicated floating point stack.  As part of
// semantic checking, the number of indexes is checked to ensure it is valid.
// Actual bounds checking for the expression used as indexes is performed at run time,
// not at compile time, since the compiler does not evaluate constant expressions
// at compile time.
// Returns true on success, false otherwise.
//
// g_navar -> T_NAVAR T_LPAREN g_numeric_expression g_array_tail2 T_RPAREN
// g_array_tail2 -> T_COMMA g_numeric_expression
// g_array_tail2 -> epsilon
//
static bool g_navar(void) {
  struct token *curtoken=clone_next_token(); // copy of next token
  uint32_t lno=curtoken->lno;                // line number of current token
  uint32_t dims;                             // number of array dimensions
  bool retval=true;                          // return value for this function
  struct arraydata *ad=NULL;                 // pointer to a row of array information in the symbol table module

  debug_printf("g_navar -> T_NAVAR T_LPAREN g_numeric_expression g_array_tail2 T_RPAREN\n");
  if (!match(T_NAVAR,__func__)) {
     retval=false;
     goto g_navar_xit;
  }
  debug_printf("T_NAVAR <<<value is '%s'>>>\n",curtoken->toketext);
  if (!match(T_LPAREN,__func__)) {
     retval=false;
     goto g_navar_xit;
  }
  if (!g_numeric_expression()) {
     retval=false;
     goto g_navar_xit;
  }
  if (peek_next_token().tid==T_COMMA) { // 2-d matrix
      debug_printf("g_array_tail2 -> T_COMMA g_numeric_expression\n");
      if (!match(T_COMMA,__func__)) {
         retval=false;
         goto g_navar_xit;
      }
      if (!g_numeric_expression()) {
         retval=false;
         goto g_navar_xit;
      }
      dims=UINT32_C(2);
  } else { // vector
     debug_printf("g_array_tail2 -> epsilon\n");
     dims=UINT32_C(1);
  }
  if (!match(T_RPAREN,__func__)) {
     retval=false;
     goto g_navar_xit;
  }
  if (!navar_exists(curtoken->toketext)) {
    if (!assign_navar(curtoken->toketext,curtoken->lno)) {
      debug_printf("assign_navar(%s) failed on line %" PRIu32 "\n",
         curtoken->toketext,lno);
      if (nvar_exists(curtoken->toketext)) {
        fprintf(stderr,"ERROR: on line %" PRIu32 " attempt to use scalar %s as an array\n",
          lno,curtoken->toketext);
      } else {
        fprintf(stderr,"ERROR: on line %" PRIu32 " attempt to use array %s for read failed and I do not know why\n",
          lno,curtoken->toketext);
      }
      retval=false;
      goto g_navar_xit;
    }
    ad=get_array_data(curtoken->toketext);
    if (NULL==ad) {
      debug_printf("No array data for '%s' found\n",curtoken->toketext);
      if (!put_array_data(curtoken->toketext,dims,UINT32_C(10),(UINT32_C(1)==dims?UINT32_C(0):UINT32_C(10)))) {
        debug_printf("put_array_data(%s,%" PRIu32 ",%" PRIu32 ",%" PRIu32 ") failed\n",
          curtoken->toketext,dims,UINT32_C(10),(UINT32_C(1)==dims?UINT32_C(0):UINT32_C(10)));
        retval=false;
        goto g_navar_xit;
      } else {
        debug_printf("put_array_data(%s,%" PRIu32 ",%" PRIu32 ",%" PRIu32 ") succeeded\n",
          curtoken->toketext,dims,UINT32_C(10),(UINT32_C(1)==dims?UINT32_C(0):UINT32_C(10)));
      }
      ad=get_array_data(curtoken->toketext);
    } else {
      debug_printf("array data for '%s' found\n",curtoken->toketext);
      debug_printf("dims=%" PRIu32 ",dim1=%" PRIu32 ",dim2=%" PRIu32 "\n",ad->dims,ad->dim1,ad->dim2);
      if (UINT32_C(0)==ad->dims) {
        debug_printf("update necessary\n");
        if (!put_array_data(curtoken->toketext,dims,UINT32_C(10),(UINT32_C(1)==dims?UINT32_C(0):UINT32_C(10)))) {
          debug_printf("put_array_data(%s,%" PRIu32 ",%" PRIu32 ",%" PRIu32 ") failed\n",
            curtoken->toketext,dims,UINT32_C(10),(UINT32_C(1)==dims?UINT32_C(0):UINT32_C(10)));
          retval=false;
          goto g_navar_xit;
        } else {
          debug_printf("put_array_data(%s,%" PRIu32 ",%" PRIu32 ",%" PRIu32 ") succeeded\n",
            curtoken->toketext,dims,UINT32_C(10),(UINT32_C(1)==dims?UINT32_C(0):UINT32_C(10)));
        }
      } else {
        if (ad->dims!=dims) {
          fprintf(stderr,"ERROR: on line %" PRIu32 " read of array %s is using %" PRIu32 " dimensions, but it actually has %" PRIu32 " dimensions\n",
            lno,curtoken->toketext,dims,ad->dims);
          retval=false;
          goto g_navar_xit;
        }
        debug_printf("no update necessary\n");
      }
    }
  } else {
    ad=get_array_data(curtoken->toketext);
    if (NULL==ad) {
      debug_printf("No array data for '%s' found\n",curtoken->toketext);
      retval=false;
      goto g_navar_xit;
    }
    debug_printf("array data for '%s' found\n",curtoken->toketext);
    debug_printf("dims=%" PRIu32 ",dim1=%" PRIu32 ",dim2=%" PRIu32 "\n",ad->dims,ad->dim1,ad->dim2);
    if (ad->dims!=dims) {
      fprintf(stderr,"ERROR: on line %" PRIu32 " read of array %s is using %" PRIu32 " dimensions, but it actually has %" PRIu32 " dimensions\n",
        lno,curtoken->toketext,dims,ad->dims);
      retval=false;
      goto g_navar_xit;
    }
  }
  switch (dims) {
    case UINT32_C(1):
      if (!push_1D_navar(outfile,curtoken->toketext,ad->dim1)) {
        retval=false;
        goto g_navar_xit;
      }
      break;
    case UINT32_C(2):
      if (!push_2D_navar(outfile,curtoken->toketext,ad->dim1,ad->dim2)) {
        retval=false;
        goto g_navar_xit;
      }
      break;
    default:
      fprintf(stderr,"ICE: on line %" PRIu32 " read of array %s is using %" PRIu32 " dimensions is not supported.\n",
        lno,curtoken->toketext,dims);
      retval=false;
      goto g_navar_xit;
  }
  can_option_base=false;
g_navar_xit:
  free(curtoken->toketext);curtoken->toketext=NULL;
  free(curtoken);curtoken=NULL;
  return retval;
}

//
// This is part of the processing of arithmetic expressions.
// It will emit code for calling the built-in arithmetic functions and
// user-defined functions.  It will emit code to push scalar numeric,
// array numeric, or numeric constant constant values to the dedicated
// run time floating point stack.  It will call g_numeric_expression for
// parenthesized sub-expressions.
// Returns true on success, false otherwise.
//
//  g_primary -> g_abs
//  g_primary -> g_atn
//  g_primary -> g_cos
//  g_primary -> g_exp
//  g_primary -> g_int
//  g_primary -> g_log
//  g_primary -> g_sgn
//  g_primary -> g_sin
//  g_primary -> g_sqr
//  g_primary -> g_tan
//  g_primary -> g_rnd
//  g_primary -> g_integer
//  g_primary -> g_real
//  g_primary -> g_udf
//  g_primary -> g_nvar
//  g_primary -> g_navar
//  g_primary -> T_LPAREN g_numeric_expression T_RPAREN
//
static bool g_primary(void) {
  struct token curtoken=peek_next_token(); // copy of next token
  bool retval=true; // function return value

  switch(curtoken.tid) {
    case T_ABS:
       debug_printf("g_primary -> g_abs\n");
       retval=g_abs();
       break;
    case T_ATN:
       debug_printf("g_primary -> g_atn\n");
       retval=g_atn();
       break;
    case T_COS:
       debug_printf("g_primary -> g_cos\n");
       retval=g_cos();
       break;
    case T_EXP:
       debug_printf("g_primary -> g_exp\n");
       retval=g_exp();
       break;
    case T_INT:
       debug_printf("g_primary -> g_int\n");
       retval=g_int();
       break;
    case T_LOG:
       debug_printf("g_primary -> g_log\n");
       retval=g_log();
       break;
    case T_SGN:
       debug_printf("g_primary -> g_sgn\n");
       retval=g_sgn();
       break;
    case T_SIN:
       debug_printf("g_primary -> g_sin\n");
       retval=g_sin();
       break;
    case T_SQR:
       debug_printf("g_primary -> g_sqr\n");
       retval=g_sqr();
       break;
    case T_TAN:
       debug_printf("g_primary -> g_tan\n");
       retval=g_tan();
       break;
    case T_RND:
       debug_printf("g_primary -> g_rnd\n");
       retval=g_rnd();
       break;
    case T_INTEGER:
       debug_printf("g_primary -> g_integer\n");
       retval=g_integer();
       break;
    case T_REAL:
       debug_printf("g_primary -> g_real\n");
       retval=g_real();
       break;
    case T_FNID:        // user defined function call
       debug_printf("g_primary -> g_udf\n");
       retval=g_udf();
       break;
    case T_NVAR:        // numeric variable
       debug_printf("g_primary -> g_nvar\n");
       retval=g_nvar();
       break;
    case T_NAVAR:       // numeric array variable
       debug_printf("g_primary -> g_navar\n");
       retval=g_navar();
       break;
    case T_LPAREN:
       debug_printf("g_primary -> T_LPAREN g_numeric_expression T_RPAREN\n");
       if (!match(T_LPAREN,__func__)) {
         retval=false;
       } else {
         retval=g_numeric_expression();
         if (retval)
           retval=match(T_RPAREN,__func__);
       }
       break;
    default:
       fprintf(stderr,"ERROR on line %4" PRIu32 ": Expected an arithmetic function, a numeric scalar\nor array variable, a numeric constant, or a left parenthesis\n",curtoken.lno);
       retval=false;
  }
  return retval;
}

//
// This is part of the processing of arithmetic expressions.
// This will emit code for exponentiation.
// Returns true on success, false otherwise.
//
// g_factor -> g_primary g_factor_tail
// g_factor_tail -> T_POW g_primary
// g_factor_tail -> epsilon
//
static bool g_factor(void) {
  enum token_ids curtid; // current token ID number

  debug_printf("g_factor -> g_primary g_factor_tail\n");
  if (!g_primary()) return false;
  curtid=peek_next_token().tid;
  while (T_POW==curtid) {
    debug_printf("g_factor_tail -> T_POW g_primary\n");
    if (!match(T_POW,__func__)) return false;
    if (!g_primary()) return false;
    dopow(outfile);
    curtid=peek_next_token().tid;
  }
  debug_printf("g_factor_tail -> epsilon\n");
  return true;
}

//
// This is part of the processing of arithmetic expressions.
// This will emit code for binary multiplication and division.
// Returns true on success, false otherwise.
//
// g_term -> g_factor g_term_tail
// g_term_tail -> T_MUL g_factor g_term_tail
// g_term_tail -> T_DIV g_factor g_term_tail
// g_term_tail -> epsilon
//
static bool g_term(void) {
  enum token_ids curtid; // current token ID number

  debug_printf("g_term -> g_factor g_term_tail\n");
  if (!g_factor()) return false;
  curtid=peek_next_token().tid;
  while ((T_MUL==curtid)||(T_DIV==curtid)) {
    debug_printf("g_term_tail -> %s g_factor g_term_tail\n",token_names[curtid].namestring);
    if (!match(curtid,__func__)) return false;
    if (!g_factor()) return false;
    if (T_MUL==curtid)
      domultiply(outfile);
    else
      dodivide(outfile);
    curtid=peek_next_token().tid;
  }
  debug_printf("g_term_tail -> epsilon\n");
  return true;
}

//
// This will process a string expression.  It is quite simple since Minimal BASIC
// has no string operators.  This will emit code for pushing the address of the
// string variable or quoted string literal to the dedicated runtime string address
// stack.
// Returns true on success, false otherwise.
//
// g_string_expression -> T_SVAR
// g_string_expression -> T_QSTRING
//
static bool g_string_expression(void) {
  bool retval=true; // function return value
  char slit_name[21]; // buffer to hold generated name of string literal
  char *s=NULL; // pointer to generated name of string literal
  struct token *curtoken=clone_next_token(); // copy of next token

  switch (curtoken->tid) {
    case T_SVAR:
       debug_printf("g_string_expression -> T_SVAR\n");
       if (!match(T_SVAR,__func__)) {
         retval=false;
         break;
       }
       if (!svar_exists(curtoken->toketext)) {
         if (!assign_svar(curtoken->toketext)) {
           retval=false;
           break;
         }
       }
       if (!push_svar(outfile,curtoken->toketext)) {
         retval=false;
         break;
       }
       break;
    case T_QSTRING:
       debug_printf("g_string_expression -> T_QSTRING\n");
       if (!match(T_QSTRING,__func__)) {
         fprintf(stderr,"Expected quoted string on line %" PRIu32 "\n",curtoken->lno);
         retval=false;
         break;
       }
       if (!add_string_literal(curtoken->toketext,&s)) {
         fprintf(stderr,"add_string_literal() failure on line %" PRIu32 "\n",curtoken->lno);
         retval=false;
         break;
       }
       strncpy(slit_name,s,20);
       slit_name[20]=0;
       debug_printf("g_string_expression added string literal '%s' and got symbol '%s'\n",
           curtoken->toketext,slit_name);
       retval=push_string_literal(outfile,slit_name,curtoken->toketext);
       break;
    default:
       retval=false;
  }
  free(curtoken->toketext);curtoken->toketext=NULL;
  free(curtoken);curtoken=NULL;
  return retval;
}

//
// This is the production that processes an arithmetic expression.
// It will emit code for a leading unary minus, and for binary
// addition and subtraction.
// Returns true on success, false otherwise.
//
// g_numeric_expression -> g_sign g_term g_numeric_expression_tail
// g_numeric_expression_tail -> %s g_term g_numeric_expression_tail
// g_numeric_expression_tail -> epsilon
//
static bool g_numeric_expression(void) {
  struct token curtoken=peek_next_token(); // copy of next token
  bool neednegate=false; // flag to indicate g_sign was a unary minus
  enum token_ids curtid=curtoken.tid; // current token ID number

  debug_printf("g_numeric_expression -> g_sign g_term g_numeric_expression_tail\n");
  if ((T_ADD==curtid)||(T_SUBTRACT==curtid)) {
    debug_printf("g_sign -> %s\n",token_names[curtid].namestring);
    if (T_SUBTRACT==curtid)
      neednegate=true;
    if (!match(curtid,__func__)) return false;
  } else {
    debug_printf("g_sign ->\n");
  }
  if (!g_term()) return false;
  if (neednegate)
    donegate(outfile);
  curtoken=peek_next_token();
  curtid=curtoken.tid;
  while ((T_ADD==curtid)||(T_SUBTRACT==curtid)) {
    debug_printf("g_numeric_expression_tail -> %s g_term g_numeric_expression_tail\n",token_names[curtid].namestring);
    if (!match(curtid,__func__)) return false;
    if (!g_term()) return false;
    if (T_ADD==curtid)
      doadd(outfile);
    else
      dosubtract(outfile);
    curtoken=peek_next_token();
    curtid=curtoken.tid;
  }
  debug_printf("g_numeric_expression_tail -> epsilon\n");
  return true;
}

//
// This is used to represent expressions in general, either
// string or arithmetic expressions.  The real work is done
// in g_string_expression or g_numeric_expression.
// Returns true on success, false otherwise.
//
// g_expression -> g_string_expression
// g_expression -> g_numeric_expression
//
static bool g_expression(void) {
  switch (peek_next_token().tid) {
    case T_SVAR:
    case T_QSTRING:
      debug_printf("g_expression -> g_string_expression\n");
      return g_string_expression();
    default:
      ; // fall through
  }
  debug_printf("g_expression -> g_numeric_expression\n");
  return g_numeric_expression();
}

//
// This will process relational expressions.  It does not emit code.  Instead
// parameters are passed back to the caller and the caller is responsible for
// generating code.  The op parameter is used to pass back the specific
// relational operator used.  The stringexpression parameter is used to pass
// back whether the relational expression was between strings (if true) or
// numeric values (if false).  Returns true on success, false otherwise.
//
// g_relational_expression -> g_string_expression g_relop2 g_string_expression
// g_relop2 -> T_EQ
// g_relop2 -> T_NE
// g_relational_expression -> g_numeric_expression g_relop g_numeric_expression
// g_relop -> T_EQ
// g_relop -> T_NE
// g_relop -> T_GT
// g_relop -> T_GE
// g_relop -> T_LT
// g_relop -> T_LE
//
static bool g_relational_expression(enum relop *op,bool *stringexpression) {
  struct token curtoken=peek_next_token(); // copy of next token

  *stringexpression=false; // assume a comparison between arithmetic expressions
  switch (curtoken.tid) {
    case T_QSTRING:
    case T_SVAR:
       *stringexpression=true; // signal this was a comparson between string expressions
    default:
       ; // fall through
  }
  if (verbose) {
     if (*stringexpression) {
       debug_printf("g_relational_expression -> g_string_expression g_relop2 g_string_expression\n");
     } else {
       debug_printf("g_relational_expression -> g_numeric_expression g_relop g_numeric_expression\n");
     }
  }
  if (*stringexpression) {
    if (!g_string_expression()) {
      fprintf(stderr,"ERROR on line %" PRIu32 ": expected string expression on left hand side of relational expression\n",curtoken.lno);
      return false;
    }
  } else {
    if (!g_numeric_expression()) {
      fprintf(stderr,"ERROR on line %" PRIu32 ": expected numeric expression on left hand side of relational expression\n",curtoken.lno);
      return false;
    }
  }
  curtoken=peek_next_token();
  switch(curtoken.tid) {
    case T_LE:
      if (*stringexpression) {
        fprintf(stderr,"ERROR on line %" PRIu32 ": string expressions do not permit <= operator\n",curtoken.lno);
        return false;
      }
      if (!match(T_LE,__func__)) return false;
      *op=LE;
      break;
    case T_LT:
      if (*stringexpression) {
        fprintf(stderr,"ERROR on line %" PRIu32 ": string expressions do not permit < operator\n",curtoken.lno);
        return false;
      }
      if (!match(T_LT,__func__)) return false;
      *op=LT;
      break;
    case T_GE:
      if (*stringexpression) {
        fprintf(stderr,"ERROR on line %" PRIu32 ": string expressions do not permit >= operator\n",curtoken.lno);
        return false;
      }
      if (!match(T_GE,__func__)) return false;
      *op=GE;
      break;
    case T_GT:
      if (*stringexpression) {
        fprintf(stderr,"ERROR on line %" PRIu32 ": string expressions do not permit > operator\n",curtoken.lno);
        return false;
      }
      if (!match(T_GT,__func__)) return false;
      *op=GT;
      break;
    case T_EQ:
      if (!match(T_EQ,__func__)) return false;
      *op=EQ;
      break;
    case T_NE:
      if (!match(T_NE,__func__)) return false;
      *op=NE;
      break;
    default:
      // bogus relop
      return false;
  }
  if (verbose) {
    if (*stringexpression) {
      debug_printf("g_relop2 -> %s\n",relop_names[*op]);
    } else {
      debug_printf("g_relop -> %s\n",relop_names[*op]);
    }
  }
  if (*stringexpression) {
    if (!g_string_expression()) {
      fprintf(stderr,"ERROR on line %" PRIu32 ": expected string expression on right hand side of relational expression\n",curtoken.lno);
      return false;
    }
  } else {
    if (!g_numeric_expression()) {
      fprintf(stderr,"ERROR on line %" PRIu32 ": expected numeric expression on right hand side of relational expression\n",curtoken.lno);
      return false;
    }
  }
  return true;
}

//
// This is used to process a DEF statement for a UDF (user-defined function).
// Before calling g_numeric_expression,
// this function will set in_udf_definition to true so that arithmetic expression
// evaluation will prefix the normal symbol table lookup for scalar numeric variables
// by a test to see if the scalar variable name is actually the name of the parameter
// for the function, if one exists.  After g_numeric_expression returns, the
// in_udf_definition flag is set to false to resume normal processing of arithmetic
// expressions.
// During processing of the DEF, curfunc_name is set the the UDF function name and
// curfunc_argname is set the the UDF function parameter (argument) name, if one exists.
// Both of these are cleared after the DEF processing is complete.
// Returns true on success, false otherwise.
//
// g_defstmt -> T_DEF T_FNID g_def_arglist T_EQ g_numeric_expression
// g_def_arglist -> TLPAREN T_NVAR TRPAREN
// g_def_arglist -> epsilon
//
static bool g_defstmt(void) {
  struct token *funcname_token=NULL; // pointer to name of UDF
  struct token *funcarg_token=NULL;  // pointer to name of UDF parameter (if there is one)
  bool retval=true; // return value of this function
  FILE *thandle=NULL; // file handle of normal assembly output, used so we can switch
                      // to alternate UDF assembly output file xtraoutput and then 
                      // switch back after the UDF processing is done.

  debug_printf("g_defstmt -> T_DEF T_FNID g_def_arglist T_EQ g_numeric_expression\n");
  if (!match(T_DEF,__func__)) {
    retval=false;
    goto g_defstmt_xit;
  }
  funcname_token=clone_next_token();
  if (!match(T_FNID,__func__)) {
    retval=false;
    goto g_defstmt_xit;
  }
  debug_printf("T_FNID is <<<%s>>>\n",funcname_token->toketext);
  if (peek_next_token().tid==T_LPAREN) {
     debug_printf("g_def_arglist -> TLPAREN T_NVAR TRPAREN\n");
     if (!match(T_LPAREN,__func__)) {
       retval=false;
       goto g_defstmt_xit;
     }
     funcarg_token=clone_next_token();
     if (!match(T_NVAR,__func__)) {
       retval=false;
       goto g_defstmt_xit;
     }
     debug_printf("T_NVAR is <<<%s>>>\n",funcarg_token->toketext);
     if (!match(T_RPAREN,__func__)) {
       retval=false;
       goto g_defstmt_xit;
     }
  } else {
     debug_printf("g_def_arglist -> epsilon\n");
  }
  if (!match(T_EQ,__func__)) {
    retval=false;
    goto g_defstmt_xit;
  }
  if (NULL==funcarg_token) {
    if (!add_udf(funcname_token->toketext,NULL,funcname_token->lno)) {
      retval=false;
      goto g_defstmt_xit;
    }
    curfunc_argname[0]=0;
    if (!doudf_header(xtraoutfile,funcname_token->toketext,NULL)) {
      retval=false;
      goto g_defstmt_xit;
    }
  } else {
    if (!add_udf(funcname_token->toketext,funcarg_token->toketext,funcname_token->lno)) {
      retval=false;
      goto g_defstmt_xit;
    }
    strcpy(curfunc_name,funcname_token->toketext); // set current function name
    sprintf(curfunc_argname,".LFN%c_%c",           // set current function argument name
       funcname_token->toketext[2],
       funcarg_token->toketext[0]);
    debug_printf(">>>>>>>>>>>>> UDF ARG XFORM FROM '%s' TO curfunc_argname='%s'\n",
        funcarg_token->toketext,curfunc_argname);
    if (!doudf_header(xtraoutfile,funcname_token->toketext,funcarg_token->toketext)) {
      retval=false;
      goto g_defstmt_xit;
    }
  }
  in_udf_definition=true; // signal we are inside a DEF statement
  thandle=outfile; // save handle of normal assembly output file
  outfile=xtraoutfile; // set output to UDF assembly output file
  retval=g_numeric_expression(); // process body of UDF
  in_udf_definition=false; // signal we are NOT inside a DEF statement
  outfile=thandle; // restore handle of normal assembly output file
  thandle=NULL; // clear thandle for safety
  if (!doudf_footer(xtraoutfile,funcname_token->toketext)) {
    retval=false;
    goto g_defstmt_xit;
  }
  curfunc_name[0]=0; // clear curent function name for safety
  debug_printf(">>>>>>>>>>>>> curfunc_name is now empty\n");
  curfunc_argname[0]=0; // clear curent function argument name for safety
  debug_printf(">>>>>>>>>>>>> curfunc_argname is now empty\n");
g_defstmt_xit:
  if (funcarg_token!=NULL) {
    free(funcarg_token->toketext);funcarg_token->toketext=NULL;
    free(funcarg_token);funcarg_token=NULL;
  }
  if (funcname_token!=NULL) {
    free(funcname_token->toketext);funcname_token->toketext=NULL;
    free(funcname_token);funcname_token=NULL;
  }
  return retval;
}

//
// This is used when parsing a Minimal BASIC line number as part of another production.
// If the jumptarget parameter is true, the line number must start in the first column,
// and thus be a valid jump target.
// The why parameter, used for error messages, must have room for an 80 byte string.
// The lno is the actual Minimal BASIC line number T_INTEGER contains.
// Returns true on success, false otherwise.
//
// g_lineno -> T_INTEGER
//
static bool g_lineno(uint32_t *lno,char *why,const bool jumptarget) {
  struct token *curtoken=clone_next_token();
  uint32_t tlineno; // line number we matched
  bool retval=true; // function return value

  debug_printf("g_lineno -> T_INTEGER\n");
  if (!match(T_INTEGER,__func__)) {
     retval=false;
     goto g_lineno_xit;
  }
  
  debug_printf("LINE %s on line %" PRIu32 " starts in column %" PRIu32 "\n",
     curtoken->toketext,curtoken->lno,curtoken->cno);
  if (jumptarget && (curtoken->cno!=1)) { // jump target line numbers must begin in the first column
     snprintf(why,80,"ERROR: line number '%s' may not contain leading whitespace\n",curtoken->toketext);
     retval=false;
     goto g_lineno_xit;
  }
  if (!convert_line_number(curtoken->toketext,&tlineno,why)) {
     retval=false;
     goto g_lineno_xit;
  }
  // range checks - lines must be between 1 and MAXLINENO inclusive
  if (tlineno<UINT32_C(1)) {
     snprintf(why,80,"ERROR: Invalid line number %" PRIu32 "; minimum line number is 1\n",tlineno);
     retval=false;
     goto g_lineno_xit;
  }
  if (tlineno>MAXLINENO) {
     snprintf(why,80,"ERROR: Invalid line number %" PRIu32 "; maximum line number is %" PRIu32 "\n",tlineno,MAXLINENO);
     retval=false;
     goto g_lineno_xit;
  }
  why[0]=0; // no error, clear error message buffer
  *lno=tlineno; // copy line number value to return parameter
g_lineno_xit:
  free(curtoken->toketext);curtoken->toketext=NULL;
  free(curtoken);curtoken=NULL;
  return retval;
}

//
// This is the start of processing for an IF statement.
// It will emit code for the conditional jumps to g_lineno based on
// logical comparisons done by g_relational_expression.
// Returns true on success, false otherwise.
//
// g_ifstmt -> T_IF g_relational_expression T_THEN g_lineno
//
static bool g_ifstmt(void) {
  uint32_t tlineno; // line number to branch to if g_relational_expression is true
  uint32_t mylineno; // line number of the IF statement
  static char msg[ERROR_BUFFER_MAX]; // buffer for error messages
  enum relop op; // relational operator returned by g_relational_expression and
                 // passed to dostring_comparsion and donumeric_comparison
  bool isstring; // flag to indicate whether a string (true) or numeric (false)
                 // comparison is needed

  debug_printf("g_ifstmt -> T_IF g_relational_expression T_THEN g_lineno\n");
  mylineno=peek_next_token().lno;
  if (!match(T_IF,__func__)) return false;
  if (!g_relational_expression(&op,&isstring)) return false;
  if (!match(T_THEN,__func__)) return false;
  if (!g_lineno(&tlineno,msg,false)) {
    fputs(msg,stderr);
    return false;
  }
  if (!add_new_jump_target(tlineno,mylineno,current_scope)) return false;
  if (isstring) {
    need_strcmp=true;
    if (!dostring_comparison(outfile,op,tlineno)) return false;
  } else {
    if (!donumeric_comparison(outfile,op,tlineno)) return false;
  }
  return true;
}

//
// This will emit code for a jump table for an ON..GOTO statement.
// g_on_gotostmt -> T_ON g_numeric_expression g_on_gotostmt_tail
// g_on_gotostmt_tail -> T_COMMA g_lineno g_on_gotostmt_tail
// g_on_gotostmt_tail -> epsilon
//
static bool g_on_gotostmt(void) {
  uint32_t tlineno; // variable to holde a jump target line number
  uint32_t mylineno; // line number of the ON..GOTO statement
  uint32_t i; // loop index variable used to iterate over lineno[] when emiting the jump table
  static char msg[ERROR_BUFFER_MAX]; // buffer for error messages
  uint32_t lineno[50]; // vector of jump target line numbers
  uint32_t curlinenoslot=UINT32_C(0); // index of current slot in lineno[]

  debug_printf("g_on_gotostmt -> T_ON g_numeric_expression g_on_gotostmt_tail\n");
  if (!match(T_ON,__func__)) return false;
  mylineno=peek_next_token().lno;
  if (!g_numeric_expression()) return false;
  switch (peek_next_token().tid) {
    case T_GOTO:
       debug_printf("g_on_gotostmt_tail -> T_GOTO g_lineno g_on_gotostmt_tail\n");
       if (!match(T_GOTO,__func__)) return false;
       break;
    case T_GO:
       debug_printf("g_on_gotostmt_tail -> T_GO T_TO g_lineno g_on_gotostmt_tail\n");
       if (!match(T_GO,__func__)) return false;
       if (!match(T_TO,__func__)) return false;
       break;
    default:
       return false;
  }
  do {
    if (!g_lineno(&tlineno,msg,false)) {
      fputs(msg,stderr);
      return false;
    }
    if (!add_new_jump_target(tlineno,mylineno,current_scope)) return false;
    lineno[curlinenoslot++]=tlineno;
    debug_printf("g_on_gotostmt_tail -> T_COMMA g_lineno g_on_gotostmt_tail\n");
    if (peek_next_token().tid!=T_COMMA) break;
    if (!match(T_COMMA,__func__)) return false;
  } while (true);
  if (!doongotostart(outfile,curlinenoslot)) return false;
  for (i=UINT32_C(0);i<curlinenoslot;++i)
    if (!doongotochoice(outfile,lineno[i])) return false;
  debug_printf("g_on_gotostmt_tail -> epsilon\n");
  if (!doongotoend(outfile)) return false;
  return true;
}

//
// This is used as part of the PRINT statement processing.  It will process a 
// destiation variable expression.
// Returns true on success, false otherwise.
//
// g_target_variable -> T_NAVAR T_LPAREN g_numeric_expression g_target_arraytail2 T_RPAREN
// g_target_arraytail2 -> T_COMMA g_numeric_expression
// g_target_arraytail2 -> epsilon
// g_target_variable -> T_NVAR
// g_target_variable -> T_SVAR
//
static bool g_target_variable(void) {
  struct arraydata *ad=NULL; // pointer to a row of array information in the symbol table module
  struct token *curtoken=clone_next_token(); // copy of next token
  uint32_t dims; // number of array dimensions
  bool retval=true; // return value of this function

  switch(curtoken->tid) {
    case T_NAVAR:       // numeric array variable
      debug_printf("g_target_variable -> T_NAVAR T_LPAREN g_numeric_expression g_target_arraytail2 T_RPAREN\n");
      if (!match(T_NAVAR,__func__)) {
        retval=false;
        goto g_target_variable_xit;
      }
      debug_printf("T_NAVAR <<<value is '%s'>>>\n",curtoken->toketext);
      if (!match(T_LPAREN,__func__)) {
        retval=false;
        goto g_target_variable_xit;
      }
      if (!g_numeric_expression()) {
        retval=false;
        goto g_target_variable_xit;
      }
      if (peek_next_token().tid==T_COMMA) { // 2-d matrix
        debug_printf("g_target_arraytail2 -> T_COMMA g_numeric_expression\n");
        if (!match(T_COMMA,__func__)) {
          retval=false;
          goto g_target_variable_xit;
        }
        if (!g_numeric_expression()) {
          retval=false;
          goto g_target_variable_xit;
        }
        dims=UINT32_C(2);
      } else { // vector
        debug_printf("g_target_arraytail2 -> epsilon\n");
        dims=UINT32_C(1);
      }
      if (!match(T_RPAREN,__func__)) {
        retval=false;
        goto g_target_variable_xit;
      }
      if (!assign_navar(curtoken->toketext,curtoken->lno)) {
        retval=false;
        goto g_target_variable_xit;
      }
      ad=get_array_data(curtoken->toketext);
      if (NULL==ad) {
        debug_printf("No array data for '%s' found\n",curtoken->toketext);
        if (!put_array_data(curtoken->toketext,dims,UINT32_C(10),(UINT32_C(1)==dims?UINT32_C(0):UINT32_C(10)))) {
          debug_printf("put_array_data(%s,%" PRIu32 ",%" PRIu32 ",%" PRIu32 ") failed\n",
            curtoken->toketext,dims,UINT32_C(10),(UINT32_C(1)==dims?UINT32_C(0):UINT32_C(10)));
          retval=false;
          goto g_target_variable_xit;
        } else {
          debug_printf("put_array_data(%s,%" PRIu32 ",%" PRIu32 ",%" PRIu32 ") succeeded\n",
            curtoken->toketext,dims,UINT32_C(10),(UINT32_C(1)==dims?UINT32_C(0):UINT32_C(10)));
        }
      } else {
        debug_printf("array data for '%s' found\n",curtoken->toketext);
        debug_printf("dims=%" PRIu32 ",dim1=%" PRIu32 ",dim2=%" PRIu32 "\n",ad->dims,ad->dim1,ad->dim2);
        if (UINT32_C(0)==ad->dims) {
          debug_printf("update necessary\n");
          if (!put_array_data(curtoken->toketext,dims,UINT32_C(10),(UINT32_C(1)==dims?UINT32_C(0):UINT32_C(10)))) {
            debug_printf("put_array_data(%s,%" PRIu32 ",%" PRIu32 ",%" PRIu32 ") failed\n",
              curtoken->toketext,dims,UINT32_C(10),(UINT32_C(1)==dims?UINT32_C(0):UINT32_C(10)));
            retval=false;
            goto g_target_variable_xit;
          } else {
            debug_printf("put_array_data(%s,%" PRIu32 ",%" PRIu32 ",%" PRIu32 ") succeeded\n",
              curtoken->toketext,dims,UINT32_C(10),(UINT32_C(1)==dims?UINT32_C(0):UINT32_C(10)));
          }
        } else {
          if (ad->dims!=dims) {
             fprintf(stderr,"ERROR: on line %" PRIu32 " assignment to array %s is using %" PRIu32 " dimensions, but it actually has %" PRIu32 " dimensions\n",
                  curtoken->lno,curtoken->toketext,dims,ad->dims);
             retval=false;
             goto g_target_variable_xit;
          }
          debug_printf("no update necessary\n");
        }
      }
      can_option_base=false;
      break;
    case T_NVAR:        // numeric scalar variable
      debug_printf("g_target_variable -> T_NVAR\n");
      if (!match(T_NVAR,__func__)) {
        retval=false;
        goto g_target_variable_xit;
      }
      if (!assign_nvar(curtoken->toketext,curtoken->lno)) {
        retval=false;
        goto g_target_variable_xit;
      }
      debug_printf("T_NVAR <<<value is '%s'>>>\n",curtoken->toketext);
      break;
    case T_SVAR:        // string scalar variable
      debug_printf("g_target_variable -> T_SVAR\n");
      if (!match(T_SVAR,__func__)) {
        retval=false;
        goto g_target_variable_xit;
      }
      if (!assign_svar(curtoken->toketext)) {
        retval=false;
        goto g_target_variable_xit;
      }
      debug_printf("T_SVAR <<<value is '%s'>>>\n",curtoken->toketext);
      break;
    case T_COMMA:  // error production
      fprintf(stderr,"ERROR: Missing destination variable on line %" PRIu32 "\n",
        curtoken->lno);
      // intentionally fall through to default error case
    default:
      retval=false;
      goto g_target_variable_xit;
  }
g_target_variable_xit:
  free(curtoken->toketext);curtoken->toketext=NULL;
  free(curtoken);curtoken=NULL;
  return retval;
}

//
// This will process an INPUT statement.  It will emit code for the
// input from STDIN into the g_target_variables in several phases, using
// temporary output files and temporary redirection of output to those files.
// It must build a format string, add that to the symbol table which will,
// after the parse, generate the required definition for it in the assembly language
// output.
// Returns true on success, false otherwise.
//
// g_inputstmt -> T_INPUT g_target_variable g_inputstmt_tail
// g_inputstmt_tail -> g_target_variable g_inputstmt_tail
// g_inputstmt_tail -> epsilon
//
// The semantic rules for the INPUT are non-trivial. Essentially, at runtime the
// program must prompt the user for input and then wait for an entire line of
// input. After the line input is complete, the program must then scan it into
// tokens. The tokens that can be valid are:
// 
// 1. number (integer or real)
// 2. quoted string
// 3. unquoted string
// 4. comma
// 5. end-of-line (newline)
//
// The last token must be an end-of-line.
// The comma serves as the delimiter between input items, except when it
// is within a quoted string. The number of items must match the number of
// variables specified in the INPUT statement. Then the type of each item is
// checked to ensure it can actually be stored in the corresponding variable
// specified in the INPUT statement. A number can always be stored in a scalar
// string variable or a numeric scalar or array variable. Quoted and unquoted
// strings can be stored in scalar string variables.
//
// But no input item can be stored in its corresponding variable until all type
// checking for all input items is complete and no type errors were found. Only
// after all type checking is successful can the input items be copied to the
// corresponding variables specified in the INPUT statement. If any error in the
// number of items or type of items occurs, the program must flush the input
// buffer and re-prompt the user for the input.
//
// Generating code to actually implement those semantics requires two temporary
// output files for assembly code. It also requires runtime temporary holding
// hidden variables for each of the items that can be input, since the data cannot
// be stored in the variables specified in the INPUT statement until all type
// checking is successful. The generation of code in the parser to implement an
// INPUT statement is done like this:
//
// 1. Initialize the format string to the empty string (informat).
// 2. For each variable in the INPUT statement do:
//    (a) For numeric variables, emit to the first temporary file code that converts 
//        the ASCIIZ holding area string for this numeric item to a floating point
//        representation and stores that back to the temporary holding area, and
//        if an overflow occurs during the conversion, then the code must jump
//        back to the start of the statement. Append an 'N' to the format string.
//        Emit to the second temporary file code that copies the floating point
//        value from the temporary holding area to the actual variable.
//    (b) For string variables, append an 'S' to the format string. Emit to the
//        second temporary file code that copies ASCIIZ string value from the
//        temporary holding area to the actual variable.
// 3. Send the format string to the symbol table so it will be emitted after the 
//    parser is done (add_input_format()).
// 4. Emit code to the main output file to call a function to print a prompt the
//    user for input, do the line input to a buffer, and process that buffer,
//    performing the scan into tokens left to right.  As each number or string token
//    is encountered, if the type is compatible, the data is moved to a hidden
//    temporary storage area. Any time an error occurs, the code must jump back to
//    the start of the statment. The code in this phase verifies the type of each
//    item matches the required type from the format string created earlier. The
//    assembly language routine that does all of this is called doinput, and it will
//    reset the input stack pointer itself before processing input.
//    Verify that the number of items matches the required number of variables, if
//    it does not, then the code must jump back to the start of the statement. This
//    is possible by using the length of the ASCIIZ format string, since there is one
//    byte for each variable (do_INPUT()).
// 5. rewind both temporary files.
// 6. append contents of the first temporary file to the main output file
// 7. append contents of the second temporary file to the main output file
// 8. close and remove both temporary files
//
// The name of the structure for the temporary holding areas is .Lrt_nput_stack,
// the pointer used to access elements in this stack is called
// .Lrt_nput_stack_top, and doinput uses that area like a stack. However, the
// other code emitted will access the temporary holding areas directly by treating
// the stack like a vector instead of popping the items.
//
static bool g_inputstmt(void) {
  bool retval=true; // return value of this function
  struct token *curtoken=NULL; // pointer to current token
  struct arraydata *ad=NULL; // pointer to a row of array information in the symbol table module
  char informat[31]={0}; // input format string, dynamically built while processing the
                         // INPUT statement and used for code generation.  In each byte of 
                         // this ASCIIZ format string an 'N' or 'S' will be stored.  An 'N'
                         // signals that a numeric (floating point) value is needed.  An 'S'
                         // signals that a 7-bit ASCIIZ string value is needed.  Up to 30
                         // values can be accepted in one INPUT statement.
  char *nformatname=NULL; // this is a pointer to the generated name of the input format created by the symbol table
  uint32_t informat_next=UINT32_C(0); // this is a pointer to the next available slot in the informat[] format string.
  FILE *temp_outfile1=NULL; // temp_outfile1 contains generated code that converts numbers and checks for overflow
  FILE *temp_outfile2=NULL; // temp_outfile2 contains generated code that moves data from temp locations to real locations
  FILE *real_outfile=NULL; // handle for normal assembly language output file, used to
                           // allow switching output to alternate files and then restoring
                           // it before returning from this function.

  debug_printf("g_inputstmt -> T_INPUT g_target_variable g_inputstmt_tail\n");
  need_strcmp=true;
  need_strcpy=true;
  if (!match(T_INPUT,__func__)) {
    retval=false;
    goto g_inputstmt_xit;
  }
  if ((temp_outfile1=tmpfile())==NULL) {
    perror("Unable to open temporary file 1");
    abort();
  }
  if ((temp_outfile2=tmpfile())==NULL) {
    perror("Unable to open temporary file 2");
    abort();
  }
  do {
    if (curtoken!=NULL) {
      free(curtoken->toketext);curtoken->toketext=NULL;
      free(curtoken);curtoken=NULL;
    }
    curtoken=clone_next_token();
    // redirect outfile to emit code for the actual assignment of the value to temp_outfile2
    real_outfile=outfile;
    outfile=temp_outfile2;
    if (!g_target_variable()) {
      retval=false;
      goto g_inputstmt_xit;
    }
    fflush(outfile);
    outfile=real_outfile;
    // undeo redirection 
    switch (curtoken->tid) {
      case T_SVAR:
        // note that code for input of the scalar string variable is emitted to temp_outfile1
        // the input will go into a temporary holding vector until all input has been scanned
        // and verified to be of the correct types without any exceptions
        if (!doinput_svar(temp_outfile1,curtoken->toketext,informat_next)) {
          retval=false;
          goto g_inputstmt_xit;
        }
        // note that code for copying the scalar string variable value from the temporary holding
        // vector to the real destination variable is emitted to temp_outfile2.
        if (!doinput_svar_phase2(temp_outfile2,curtoken->toketext,informat_next)) {
          retval=false;
          goto g_inputstmt_xit;
        }
        informat[informat_next++]='S';
        break;
      case T_NVAR:
        // note that code for input of the scalar numeric variable is emitted to temp_outfile1
        // the input will go into a temporary holding vector until all input has been scanned
        // and verified to be of the correct types without any exceptions
        if (!doinput_nvar(temp_outfile1,curtoken->toketext,informat_next,curtoken->lno)) {
          retval=false;
          goto g_inputstmt_xit;
        }
        // note that code for copying the scalar numeric variable value from the temporary holding
        // vector to the real destination variable is emitted to temp_outfile2.
        if (!doinput_nvar_phase2(temp_outfile2,curtoken->toketext,informat_next,curtoken->lno)) {
          retval=false;
          goto g_inputstmt_xit;
        }
        informat[informat_next++]='N';
        break;
      case T_NAVAR:
        ad=get_array_data(curtoken->toketext);
        if (NULL==ad) {
          debug_printf("No array data for '%s' found\n",curtoken->toketext);
          retval=false;
          goto g_inputstmt_xit;
        }
        // note that code for input of the array numeric variable is emitted to temp_outfile1
        // the input will go into a temporary holding vector until all input has been scanned
        // and verified to be of the correct types without any exceptions
        if (!doinput_navar(temp_outfile1,curtoken->toketext,informat_next,curtoken->lno)) {
          retval=false;
          goto g_inputstmt_xit;
        }
        // note that code for copying the array numeric variable value from the temporary holding
        // vector to the real destination variable is emitted to temp_outfile2.
        if (!doinput_navar_phase2(temp_outfile2,curtoken->toketext,ad->dim1,ad->dim2,informat_next,curtoken->lno)) {
          retval=false;
          goto g_inputstmt_xit;
        }
        informat[informat_next++]='N';
        break;
      default:
        fprintf(stderr,"ICE: Unkown variable type for %s in INPUT statement on line %" PRIu32 "\n",
          curtoken->toketext,curtoken->lno);
        abort();
    }
    debug_printf("g_inputstmt_tail -> g_target_variable g_inputstmt_tail\n");
    if (peek_next_token().tid!=T_COMMA) break;
    if (!match(T_COMMA,__func__)) {
      retval=false;
      goto g_inputstmt_xit;
    }
  } while (true);
  debug_printf("g_inputstmt_tail -> epsilon\n");
  informat[informat_next]=0; // end of PRINT list, terminate format string
  debug_printf("INPUT FORMAT IS '%s'\n",informat);
  // add format string to symbol table
  if (!add_input_format(informat,&nformatname,curtoken->lno)) {
     retval=false;
     goto g_inputstmt_xit;
  }
  // emit code to read the raw input from STDIN to a line buffer
  if (!do_INPUT(outfile,nformatname)) {
     retval=false;
     goto g_inputstmt_xit;
  }
  // prepare our two temporary files for input
  fflush(temp_outfile1);
  fflush(temp_outfile2);
  rewind(temp_outfile1);
  rewind(temp_outfile2);
  {
    char *line = NULL; // pointer to line buffer; we let getline() allocated it but we must free it
    size_t len = 0;    // length of data in line buffer

    // append temp_outfile1 contents to normal assembly language output file
    // this code will scan the input line buffer verying data is of the correct type
    // and copy the data to temporary locations, or display errors and force re-entry
    // if required
    while (getline(&line, &len, temp_outfile1) != -1) {
      fputs(line,outfile);
    }
    free(line);line=NULL;
    // append temp_outfile2 contents to normal assembly language output file
    // this will copy values from the temporary holding areas to the actual
    // destination variables after all data has been entered and all validation
    // completed.
    while (getline(&line, &len, temp_outfile2) != -1) {
      fputs(line,outfile);
    }
    free(line);line=NULL;
  }
  fflush(outfile);
g_inputstmt_xit:
  if (temp_outfile1!=NULL) {
    fclose(temp_outfile1); // file is also removed automatically
    temp_outfile1=NULL;
  }
  if (temp_outfile2!=NULL) {
    fclose(temp_outfile2); // file is also removed automatically
    temp_outfile2=NULL;
  }
  if (curtoken!=NULL) {
    free(curtoken->toketext);curtoken->toketext=NULL;
    free(curtoken);curtoken=NULL;
  }
  return retval;
}

//
// This will emit code for a READ statement.
// Returns true on success, false otherwise.
//
// g_readstmt -> T_READ g_target_variable g_readstmt_tail
// g_readstmt_tail -> g_target_variable g_readstmt_tail
// g_readstmt_tail -> epsilon
//
static bool g_readstmt(void) {
  bool retval=true; // return value from this function
  struct token *curtoken=NULL; // pointer to current token
  struct arraydata *ad=NULL; // pointer to a row of array information in the symbol table module

  debug_printf("g_readstmt -> T_READ g_target_variable g_readstmt_tail\n");
  if (READ_lineno==UINT32_C(0))         // if this is the very first READ statement in the program
     READ_lineno=peek_next_token().lno; // save line number of this READ for error messages
  if (!match(T_READ,__func__)) {
    retval=false;
    goto g_readstmt_xit;
  }
  do {
    if (curtoken!=NULL) {
      free(curtoken->toketext);curtoken->toketext=NULL;
      free(curtoken);curtoken=NULL;
    }
    curtoken=clone_next_token();
    if (!g_target_variable()) {
      retval=false;
      goto g_readstmt_xit;
    }
    switch (curtoken->tid) {
      case T_SVAR:
        if (!doread_svar(outfile,curtoken->toketext)) {
          retval=false;
          goto g_readstmt_xit;
        }
        break;
      case T_NVAR:
        if (!doread_nvar(outfile,curtoken->toketext)) {
          retval=false;
          goto g_readstmt_xit;
        }
        break;
      case T_NAVAR:
        ad=get_array_data(curtoken->toketext);
        if (NULL==ad) {
          debug_printf("No array data for '%s' found\n",curtoken->toketext);
          retval=false;
          goto g_readstmt_xit;
        }
        if (!doread_navar(outfile,curtoken->toketext,ad->dim1,ad->dim2)) {
          retval=false;
          goto g_readstmt_xit;
        }
        break;
      default:
        fprintf(stderr,"ICE: Unkown variable type for %s in READ statement on line %" PRIu32 "\n",
          curtoken->toketext,curtoken->lno);
        abort();
    }
    debug_printf("g_readstmt_tail -> g_target_variable g_readstmt_tail\n");
    if (peek_next_token().tid!=T_COMMA) break;
    if (!match(T_COMMA,__func__)) {
      retval=false;
      goto g_readstmt_xit;
    }
  } while (true);
  debug_printf("g_readstmt_tail -> epsilon\n");
g_readstmt_xit:
  if (curtoken!=NULL) {
    free(curtoken->toketext);curtoken->toketext=NULL;
    free(curtoken);curtoken=NULL;
  }
  return retval;
}

//
// This will update the symbol table with information about the constants
// stored in a DATA statement.  After a successful parse, the symbol table
// module will be called to actually emit the definitions to support the
// READ/DATA statements.
// Returns true on success, false otherwise.
//
// g_datum_value -> T_UQSTRING
// g_datum_value -> T_QSTRING
// g_datum_value -> T_INTEGER
// g_datum_value -> T_REAL
//
static bool g_datum(void) {
  struct token *curtoken=clone_next_token(); // copy of next token
  bool retval=true; // return value from this function
  uint32_t len; // variable to hold string length

  debug_printf("g_datum -> g_datam_value\n");
  switch(curtoken->tid) {
    case T_UQSTRING:
      debug_printf("g_datum_value -> T_UQSTRING\n");
      if (!match(T_UQSTRING,__func__)) {
        retval=false;
        goto g_datum_xit;
      }
      len=strlen(curtoken->toketext);
      if (len>MAX_STRING_BYTES) {
        fprintf(stderr,"ERROR: unquoted string datum on line %" PRIu32 " is too long.\nMaximum permitted length of an unquoted string datum is %" PRIu32 " bytes\n",
          curtoken->lno,MAX_STRING_BYTES);
        retval=false;
        goto g_datum_xit;
      }
      if (!add_data_literal(curtoken->toketext,UQSTR)) {
        retval=false;
        goto g_datum_xit;
      }
      break;
    case T_QSTRING:
      debug_printf("g_datum_value -> T_QSTRING\n");
      if (!match(T_QSTRING,__func__)) {
        retval=false;
        goto g_datum_xit;
      }
      len=strlen(curtoken->toketext);
      if (len>MAX_STRING_BYTES) {
        fprintf(stderr,"ERROR: quoted string datum on line %" PRIu32 " is too long.\nMaximum permitted length of a quoted string datum is %" PRIu32 " bytes\n",
          curtoken->lno,MAX_STRING_BYTES);
        retval=false;
        goto g_datum_xit;
      }
      if (!add_data_literal(curtoken->toketext,QSTR)) {
        retval=false;
        goto g_datum_xit;
      }
      break;
    case T_INTEGER:
      debug_printf("g_datum_value -> T_INTEGER\n");
      if (!match(T_INTEGER,__func__)) {
        retval=false;
        goto g_datum_xit;
      }
      if (!add_data_literal(curtoken->toketext,NUM)) {
        return false;
      }
      break;
    case T_REAL:
      debug_printf("g_datum_value -> T_REAL\n");
      if (!match(T_REAL,__func__)) {
        retval=false;
        goto g_datum_xit;
      }
      if (!add_data_literal(curtoken->toketext,NUM)) {
        retval=false;
        goto g_datum_xit;
      }
      break;
    case T_COMMA:       // error production
      fprintf(stderr,"ERROR: Empty datum in %s statement on line %" PRIu32 "\n",
        token_names[T_DATA].namestring+2,curtoken->lno);
      // fall through into default failure
    default:
      retval=false;
      goto g_datum_xit;
  }
g_datum_xit:
  free(curtoken->toketext);curtoken->toketext=NULL;
  free(curtoken);curtoken=NULL;
  return retval;
}

//
// This will begin the processing of a DATA statement.
// Returns true on success, false otherwise.
//
// g_datastmt -> T_DATA g_datum g_datastmt_tail
// g_datastmt_tail -> T_COMMA g_datum g_datastmt_tail
// g_datastmt_tail -> epsilon
//
static bool g_datastmt(void) {
  debug_printf("g_datastmt -> T_DATA g_datum g_datastmt_tail\n");
  if (!match(T_DATA,__func__)) return false;
  if (!g_datum()) return false;
  while (peek_next_token().tid==T_COMMA) {
    debug_printf("g_datastmt_tail -> T_COMMA g_datum g_datastmt_tail\n");
    if (!match(T_COMMA,__func__)) return false;
    if (!g_datum()) return false;
  }
  debug_printf("g_datastmt_tail -> epsilon\n");
  have_DATA=true; // we now have a DATA statement with items, so READ is OK
  return true;
}

//
// This will process the DIM statement.  This is used to update the symbol
// table with information about the number of dimensions and the maximum
// value for indices for each dimension for numeric arrays.
// Returns true on success, false otherwise.
//
// g_dimstmt -> T_DIM T_NAVAR T_LPAREN T_INTEGER g_matrix T_RPAREN g_dimtail
// g_matrix -> T_COMMA T_INTEGER
// g_matrix -> espilon
// g_dimtail -> T_COMMA T_NAVAR T_LPAREN T_INTEGER g_matrix T_RPAREN g_dimtail
// g_dimtail -> epsilon
//
static bool g_dimstmt(void) {
  struct token *curtoken=NULL; // pointer to current token
  uint32_t dims; // number of dimensions (0 or 1)
  uint32_t dim1; // maximum permitted subscript for first dimension
  uint32_t dim2; // maximum permitted subscript for second dimension (if there is one)
  uint32_t d1; // tentative value for dim1 after conversion from ASCIIZ to uint_32_t but before range check
  uint32_t d2; // tentative value for dim2 after conversion from ASCIIZ to uint_32_t but before range check
  char why[ERROR_BUFFER_MAX]; // buffer used for error messages
  char varname[2]; // buffer to hold name of array
  struct arraydata *ad=NULL; // pointer to a row of array information in the symbol table module
  bool retval=true; // return value for this function

  debug_printf("g_dimstmt -> T_DIM T_NAVAR T_LPAREN T_INTEGER g_matrix T_RPAREN g_dimtail\n");
  if (!match(T_DIM,__func__)) {
    retval=false;
    goto g_dimstmt_xit;
  }
  do {
    curtoken=clone_next_token();
    strncpy(varname,curtoken->toketext,2);
    varname[1]=0;       // terminate
    if (!match(T_NAVAR,__func__)) {
      retval=false;
      goto g_dimstmt_xit;
    }
    if (!assign_navar(varname,curtoken->lno)) {
      fprintf(stderr,"Unable to create array of name '%s' on line %" PRIu32 "\n",varname,curtoken->lno);
      retval=false;
      goto g_dimstmt_xit;
    }
    if (!match(T_LPAREN,__func__)) {
      retval=false;
      goto g_dimstmt_xit;
    }
    if (curtoken!=NULL) {
      free(curtoken->toketext);curtoken->toketext=NULL;
      free(curtoken);curtoken=NULL;
    }
    curtoken=clone_next_token();
    if (!match(T_INTEGER,__func__)) {
      retval=false;
      goto g_dimstmt_xit;
    }
    if (!convert_unsigned_integer(curtoken->toketext,&d1,why)) {
      fprintf(stderr,"%s on line %" PRIu32 "\n",why,curtoken->lno);
      retval=false;
      goto g_dimstmt_xit;
    }
    if (d1>MAX_SUBSCRIPT_VALUE) {
      fprintf(stderr,"%s on line %" PRIu32 " specified %" PRIu32 " rows; maximum permitted is %" PRIu32 "\n",
         token_names[T_DIM].namestring+2,curtoken->lno,d1,MAX_SUBSCRIPT_VALUE);
      retval=false;
      goto g_dimstmt_xit;
    }
    dim1=(uint32_t)d1;
    if ((base_is_one) && (UINT32_C(0)==d1)) {
      fprintf(stderr,"ERROR: %s specified an upper bound of zero but the minimum has been set\nto 1 by the %s %s statement on line %" PRIu32 "\n",
        token_names[T_DIM].namestring+2, token_names[T_OPTION].namestring+2,
        token_names[T_BASE].namestring+2, curtoken->lno);
      retval=false;
      goto g_dimstmt_xit;
    }
    if (curtoken!=NULL) {
      free(curtoken->toketext);curtoken->toketext=NULL;
      free(curtoken);curtoken=NULL;
    }
    curtoken=clone_next_token();
    if (T_COMMA==curtoken->tid) { // matrix
      debug_printf("g_matrix -> T_COMMA T_INTEGER\n");
      if (!match(T_COMMA,__func__)) {
        retval=false;
        goto g_dimstmt_xit;
      }
      if (curtoken!=NULL) {
        free(curtoken->toketext);curtoken->toketext=NULL;
        free(curtoken);curtoken=NULL;
      }
      curtoken=clone_next_token();
      if (!match(T_INTEGER,__func__)) {
        retval=false;
        goto g_dimstmt_xit;
      }
      if (!convert_unsigned_integer(curtoken->toketext,&d2,why)) {
        fprintf(stderr,"ERROR: %s specified an upper bound of zero but the minimum has been set\nto 1 by the %s %s statement on line %" PRIu32 "\n",
          token_names[T_DIM].namestring+2, token_names[T_OPTION].namestring+2,
          token_names[T_BASE].namestring+2, curtoken->lno);
        retval=false;
        goto g_dimstmt_xit;
      }
      if (d2>MAX_SUBSCRIPT_VALUE) {
        fprintf(stderr,"%s on line %" PRIu32 " specified %" PRIu32 " columns; maximum permitted is %" PRIu32 "\n",
           token_names[T_DIM].namestring+2,curtoken->lno,d2,MAX_SUBSCRIPT_VALUE);
        retval=false;
        goto g_dimstmt_xit;
      }
      dim2=(uint32_t)d2;
      dims=UINT32_C(2);
    } else { // vector
      debug_printf("g_matrix -> espilon\n");
      dims=UINT32_C(1);
      dim2=UINT32_C(0);
    }
    if (!match(T_RPAREN,__func__)) {
      retval=false;
      goto g_dimstmt_xit;
    }
    if (!put_array_data(varname,dims,dim1,dim2)) {
      ad=get_array_data(varname);
      if (NULL==ad) {
        fprintf(stderr,"Catestrophic symbol table failure while processing %s of variable %s on line %" PRIu32 "\n",
          token_names[T_DIM].namestring+2,varname,curtoken->lno);
        abort();
      } 
      if (ad->dims!=UINT32_C(0)) {
        fprintf(stderr,"ERROR: %s on line %" PRIu32 " for variable %s failed since it already has been dimensioned\n",
           token_names[T_DIM].namestring+2,curtoken->lno,varname);
        fprintf(stderr,"       with %" PRIu32 " dimensions, dim1=%" PRIu32,ad->dims,ad->dim1);
        if (UINT32_C(2)==ad->dims)
          fprintf(stderr,", dim2=%" PRIu32,ad->dim2);
        fprintf(stderr,"\n");
      } else {
        fprintf(stderr,"ERROR: %s on line %" PRIu32 " for variable %s failed but I do not know why\n",
           token_names[T_DIM].namestring+2,curtoken->lno,varname);
      }
      retval=false;
      goto g_dimstmt_xit;
    }
    if (peek_next_token().tid!=T_COMMA) break;
    debug_printf("g_dimtail -> T_COMMA T_NAVAR T_LPAREN T_INTEGER g_matrix T_RPAREN g_dimtail\n");
    if (!match(T_COMMA,__func__)) {
      retval=false;
      goto g_dimstmt_xit;
    }
    if (curtoken!=NULL) {
      free(curtoken->toketext);curtoken->toketext=NULL;
      free(curtoken);curtoken=NULL;
    }
  } while (true);
  debug_printf("g_dimtail -> epsilon\n");
  can_option_base=false;
g_dimstmt_xit:
  if (curtoken!=NULL) {
    free(curtoken->toketext);curtoken->toketext=NULL;
    free(curtoken);curtoken=NULL;
  }
  return retval;
}

//
// This will process the NEXT statement.  This is the end of a FOR loop, and it will
// verify that the entry on top of the the compile-time FOR stack called forstack contains
// the correct loop index variable and scope indicator.  It will use the information for the loop
// test label, the loop exit lable, and the hidden variable contataining the increment value.
// If the scope is wrong or the index variable is wrong, an error is reported.
// Returns true on success, false otherwise.
//
// g_nextstmt -> T_NEXT T_NVAR
//
static bool g_nextstmt(void) {
  struct token *curtoken=NULL; // pointer to current token
  bool retval=true; // return value of function
  static struct forstack_entry fse; // record to hold one entry from the compile-time FOR stack forstack
  char *dotspot=NULL; // pointer to scope delimiter in current scope string

  debug_printf("g_nextstmt -> T_NEXT T_NVAR\n");
  if (!match(T_NEXT,__func__)) {
    retval=false;
    goto g_nextstmt_xit;
  }
  curtoken=clone_next_token();
  if (!match(T_NVAR,__func__)) {
    retval=false;
    goto g_nextstmt_xit;
  }
  debug_printf("T_NVAR <<<value is '%s'>>>\n",curtoken->toketext);
  if (0==forstack_ptr) {
    fprintf(stderr,"NEXT %s without corresponding FOR on line %" PRIu32 "\n",curtoken->toketext,curtoken->lno);
    retval=false;
    goto g_nextstmt_xit;
  }
  fse=forstack[--forstack_ptr];
  // check index variable name
  if (strcmp(curtoken->toketext,fse.ndex_var)!=0) {
    fprintf(stderr,"NEXT %s but expecting NEXT %s on line %" PRIu32 "\n",curtoken->toketext,fse.ndex_var,curtoken->lno);
    retval=false;
    goto g_nextstmt_xit;
  }
  if (strcmp(fse.scope,current_scope)!=0) {
    fprintf(stderr,"NEXT %s at scope %s but expecting scope %s on line %" PRIu32 "\n",
       curtoken->toketext,current_scope,fse.scope,curtoken->lno);
    retval=false;
    goto g_nextstmt_xit;
  }
  // remove the innermost scope level (the rightmost component of current_scope)
  dotspot=strrchr(current_scope,'.');
  if (NULL==dotspot) {
    fprintf(stderr,"NEXT %s but scope is global on line %" PRIu32 "\n",curtoken->toketext,curtoken->lno);
    retval=false;
    goto g_nextstmt_xit;
  }
  *dotspot=(char)0; // the actual scope string chop
  if (verbose)
    fprintf(stderr,"After NEXT %s current scope is now '%s' on line %" PRIu32 "\n",
      curtoken->toketext,(0==current_scope[0]?"global":current_scope),curtoken->lno);
  // actually emit some code for the NEXT
  if (!doforloopend(outfile,fse.tol_label,fse.bol_label,fse.ndex_var,fse.inc_var))
    retval=false;
g_nextstmt_xit:
  if (curtoken!=NULL) {
    free(curtoken->toketext);curtoken->toketext=NULL;
    free(curtoken);curtoken=NULL;
  }
  return retval;
}

//
// This will process the FOR statement.  It will emit code for the start of the FOR
// loop, and add an entry to the compile-time FOR stack called forstack containing
// the loop index variable, the scope indicator, and the information for the loop
// test label, the loop exit lable, the hidden variable contataining the increment value,
// and the hidden variable containing the loop termination value.  It will also record
// the line where the FOR loop occurred for use in error messages.
// Returns true on success, false otherwise.
//
// g_forstmt -> T_FOR T_NVAR T_EQ g_numeric_expression T_TO g_numeric_expresion g_for_tail
// g_for_tail -> T_STEP g_numeric_expression
// g_for_tail -> epsilon
//
static bool g_forstmt(void) {
  struct token *curtoken=NULL; // pointer to current token
  bool retval=true; // return value from this function
  static struct forstack_entry fse; // record to hold one entry from the compile-time FOR stack forstack
  uint32_t i=forstack_ptr; // index into forstack, intialized to the top of stack

  debug_printf("g_forstmt -> T_FOR T_NVAR T_EQ g_numeric_expression T_TO g_numeric_expresion g_for_tail\n");
  if (!match(T_FOR,__func__)) {
    retval=false;
    goto g_forstmt_xit;
  }
  curtoken=clone_next_token();
  if (!match(T_NVAR,__func__)) {
    retval=false;
    goto g_forstmt_xit;
  }
  debug_printf("T_NVAR <<<value is '%s'>>>\n",curtoken->toketext);
  // create new numeric scalar if possible, or verify existing variable is a scalar and not an array
  if (!assign_nvar(curtoken->toketext,curtoken->lno)) {
    fprintf(stderr,"ERROR: Unable to use %s as a loop index variable in FOR on line %" PRIu32 "\n",curtoken->toketext,curtoken->lno);
    retval=false;
    goto g_forstmt_xit;
  }
  // ensure the same index variable is not in use in an outer scope
  while (i-->UINT32_C(0)) {
    if (strcmp(curtoken->toketext,forstack[i].ndex_var)==0) {
      fprintf(stderr,"ERROR: Variable %s in FOR on line %" PRIu32 " is a control variable in an outer scope FOR statement on line %" PRIu32 "\n",
         curtoken->toketext,curtoken->lno,forstack[i].lineno);
      retval=false;
      goto g_forstmt_xit;
    }
  }
  // save index variable name
  strncpy(fse.ndex_var,curtoken->toketext,MAX_FOR_VAR_BYTES);
  fse.ndex_var[MAX_FOR_VAR_BYTES]=0;  // force termination
  if (!match(T_EQ,__func__)) {
    retval=false;
    goto g_forstmt_xit;
  }
  emit_comment(outfile,"compute expression and push custom FOR loop with index %s start value",fse.ndex_var);
  // initial value for loop index
  if (!g_numeric_expression()) {
    retval=false;
    goto g_forstmt_xit;
  }
  if (!match(T_TO,__func__)) {
    retval=false;
    goto g_forstmt_xit;
  }
  emit_comment(outfile,"compute expression and push custom FOR loop with index %s limit value",fse.ndex_var);
  // termination value for loop index
  if (!g_numeric_expression()) {
    retval=false;
    goto g_forstmt_xit;
  }
  if (curtoken!=NULL) {
    free(curtoken->toketext);curtoken->toketext=NULL;
    free(curtoken);curtoken=NULL;
  }
  curtoken=clone_next_token();
  if (T_STEP==curtoken->tid) { // if they specify an increment
    emit_comment(outfile,"compute expression and push custom FOR loop with index %s increment value",fse.ndex_var);
    debug_printf("g_for_tail -> T_STEP g_numeric_expression\n");
    if (!match(T_STEP,__func__)) {
      retval=false;
      goto g_forstmt_xit;
    }
    // increment value for loop index
    if (!g_numeric_expression()) {
      retval=false;
      goto g_forstmt_xit;
    }
  } else { // use default increment of one
    debug_printf("g_for_tail -> epsilon\n");
    emit_comment(outfile,"push default FOR loop with index %s increment value of 1",fse.ndex_var);
    if (!doforpush1(outfile)) {
      retval=false;
      goto g_forstmt_xit;
    }
  }
  // update scope (append this scope level information)
  strcat(current_scope,".");
  strcat(current_scope,fse.ndex_var);
  if (verbose)
    fprintf(stderr,"After FOR %s current scope is now 'global%s' on line %" PRIu32 "\n",
      fse.ndex_var,current_scope,curtoken->lno);
  strcpy(fse.scope,current_scope);
  // actually emit code for start of FOR loop
  if (!doforloopstart(outfile,fse.ndex_var,fse.tol_label,fse.bol_label,fse.inc_var,fse.limit_var))
    retval=false;
  // push fse onto forstack (must be after code emission since we need data from the emitter for the forstack entry)
  if (forstack_ptr+1>=FORSTACK_DEPTH) {
    fprintf(stderr,"FOR stack depth of %" PRIu32 " exceeded.\n",FORSTACK_DEPTH);
    retval=false;
    goto g_forstmt_xit;
  }
  fse.lineno=curtoken->lno;
  forstack[forstack_ptr++]=fse;
  // inform symbol table about this loop's increment hidden variable
  if (!assign_magicvar(fse.inc_var)) {
    fprintf(stderr,"Unable to use %s as a loop increment variable\n",fse.inc_var);
    retval=false;
    goto g_forstmt_xit;
  }
  // inform symbol table about this loop's increment loop limit variable
  if (!assign_magicvar(fse.limit_var)) {
    fprintf(stderr,"Unable to use %s as a loop limit variable\n",fse.limit_var);
    retval=false;
    goto g_forstmt_xit;
  }
g_forstmt_xit:
  if (curtoken!=NULL) {
    free(curtoken->toketext);curtoken->toketext=NULL;
    free(curtoken);curtoken=NULL;
  }
  return retval;
}

//
// This is part of the PRINT statement processing.
// Returns true on success, false otherwise.
//
// g_print_item -> g_string_expression
// g_print_item -> g_numeric_expression
// g_print_item -> T_TAB T_LPAREN g_numeric_expression T_RPAREN
// g_print_item -> epsilon
//
static bool g_print_item(void) {
  struct token curtoken=peek_next_token(); // copy of next token
  
  switch (curtoken.tid) {
      return true;
    case T_QSTRING:
    case T_SVAR:
      debug_printf("g_print_item -> g_string_expression\n");
      if (!g_string_expression()) return false;
      if (!doprint_string(outfile)) return false;
      return true;
    case T_FNID:
    case T_REAL:
    case T_INTEGER:
    case T_ADD:
    case T_SUBTRACT:
    case T_LPAREN:
    case T_NVAR:
    case T_NAVAR:
    case T_ABS:
    case T_ATN:
    case T_COS:
    case T_EXP:
    case T_INT:
    case T_LOG:
    case T_SGN:
    case T_SIN:
    case T_SQR:
    case T_TAN:
    case T_RND:
      debug_printf("g_print_item -> g_numeric_expression\n");
      if (!g_numeric_expression()) return false;
      if (!doprint_number(outfile)) return false;
      return true;
    case T_TAB:
      debug_printf("g_print_item -> T_TAB T_LPAREN g_numeric_expression T_RPAREN\n");
      if (!match(T_TAB,__func__)) return false;
      if (!match(T_LPAREN,__func__)) return false;
      if (!g_numeric_expression()) return false;
      if (!match(T_RPAREN,__func__)) return false;
      if (!dotab(outfile)) return false;
      return true;
    case T_COMMA:
    case T_SEMI:
    case T_EOL:
      debug_printf("g_print_item -> epsilon\n");
      return true;
    default:
      fprintf(stderr,"ERROR on line %" PRIu32 ": I cannot print that\n",curtoken.lno);
      ; // fall through
  }
  return false;
}

//
// This is part of the PRINT statement processing.
// Returns true on success, false otherwise.
//
// g_print_item_tail -> T_COMMA g_print_item g_print_item_tail
// g_print_item_tail -> T_SEMI g_print_item g_print_item_tail
// g_print_item_tail -> epsilon
//
static bool g_print_item_tail(bool *neednewline) {
  struct token curtoken=peek_next_token(); // copy of next token
  
  while ((T_COMMA==curtoken.tid)||(T_SEMI==curtoken.tid)) {
    switch (peek_next_token().tid) {
      case T_COMMA:
        debug_printf("g_print_item_tail -> T_COMMA g_print_item g_print_item_tail\n");
        if (!match(T_COMMA,__func__)) return false;
        if (!doprint_nextfield(outfile)) return false;
        if (peek_next_token().tid==T_EOL) {
           // trailing comma case
           debug_printf("g_print_item_tail() trailing comma case\n");
           *neednewline=false;
           return true;
        }
        if (!g_print_item()) return false;
        curtoken=peek_next_token();
        break;
      case T_SEMI:
        debug_printf("g_print_item_tail -> T_SEMI g_print_item g_print_item_tail\n");
        if (!match(T_SEMI,__func__)) return false;
        if (peek_next_token().tid==T_EOL) {
           // trailing semicolon case
           debug_printf("g_print_item_tail() trailing semicolon case\n");
           *neednewline=false;
           return true;
        }
        if (!g_print_item()) return false;
        curtoken=peek_next_token();
        break;
      default:
        ; // fall through, got crud?
    }
  }
  debug_printf("g_print_item_tail -> epsilon\n");
  return true;
}

//
// This is part of the PRINT statement processing.
// Returns true on success, false otherwise.
//
// g_printlist -> g_print_item g_print_item_tail
// g_printlist -> epsilon
//
static bool g_printlist(bool *neednewline) {
  bool retval; // return value of function

  debug_printf("g_printlist() start neednewline is %s\n",(*neednewline?"true":"false"));
  switch (peek_next_token().tid) {
    case T_FNID:
    case T_REAL:
    case T_INTEGER:
    case T_ADD:
    case T_SUBTRACT:
    case T_LPAREN:
    case T_NVAR:
    case T_NAVAR:
    case T_SVAR:
    case T_COMMA:
    case T_SEMI:
    case T_QSTRING:
    case T_ABS:
    case T_ATN:
    case T_COS:
    case T_EXP:
    case T_INT:
    case T_LOG:
    case T_SGN:
    case T_SIN:
    case T_SQR:
    case T_TAN:
    case T_RND:
    case T_TAB:
      debug_printf("g_printlist -> g_print_item g_print_item_tail\n");
      if (!g_print_item()) {
        retval=false;
      } else {
        retval=g_print_item_tail(neednewline);
      }
      break;
    default:
      // only for bare prints
      *neednewline=true;
      debug_printf("g_printlist -> epsilon\n");
      retval=true;
      break;
  }
  debug_printf("g_printlist() end neednewline is now %s\n",
    (*neednewline?"true":"false"));
  return retval;
}

//
// This will process a PRINT statement.  If the PRINT statement
// does not end with T_SEMI or T_COMMA, code to flush the print
// line buffer will be emitted.
// Returns true on success, false otherwise.
//
// g_printstmt -> T_PRINT g_printlist
//
static bool g_printstmt(void) {
  bool neednewline=true; // flag passed back from the g_printlist to indicate whether
                         // to flush the output buffer to STDOUT or not

  debug_printf("g_printstmt -> T_PRINT g_printlist\n");
  if (!match(T_PRINT,__func__)) return false;
  if (!g_printlist(&neednewline)) return false;
  if (neednewline)
     if (!doprint_newline(outfile)) 
        return false;
  return true;
}

//
// This will emit code for the LET statement.
// Returns true on success, false otherwise.
//
// g_letstmt -> T_LET g_target_variable T_EQ g_expression
// g_expression -> g_string_expression
// g_expression -> g_numeric_expression
//
static bool g_letstmt(void) {
  struct token *curtoken=NULL;               // pointer to current token
  bool retval=true;                          // return value of this function

  debug_printf("g_letstmt -> T_LET g_target_variable T_EQ g_expression\n");
  if (!match(T_LET,__func__)) {
    retval=false;
    goto g_letstmt_xit;
  }
  curtoken=clone_next_token();
  if (T_SVAR==curtoken->tid) {
    if (!g_target_variable()) {
      retval=false;
      goto g_letstmt_xit;
    }
    if (!match(T_EQ,__func__)) {
      retval=false;
      goto g_letstmt_xit;
    }
    debug_printf("g_expression -> g_string_expression\n");
    if (!g_string_expression()) {
      fprintf(stderr,"ERROR: No string expression after = for assignment to %s on line %" PRIu32 "\n",
        curtoken->toketext,curtoken->lno);
      retval=false;
      goto g_letstmt_xit;
    }
    need_strcpy=true;
    retval=dostore_svar(outfile,curtoken->toketext);
    goto g_letstmt_xit;
  }
  // if we are here, it is not a string assignment
  if (!g_target_variable()) {
    retval=false;
    goto g_letstmt_xit;
  }
  if ((T_NVAR==curtoken->tid) && (peek_next_token().tid==T_LPAREN)) {
    fprintf(stderr,"ERROR: Numeric variable '%s' is a scalar, not an array on line %" PRIu32 "\n",
      curtoken->toketext,curtoken->lno);
    retval=false;
    goto g_letstmt_xit;
  }
  if (!match(T_EQ,__func__)) {
    retval=false;
    goto g_letstmt_xit;
  }
  debug_printf("g_expression -> g_numeric_expression\n");
  if (!g_numeric_expression()) {
    fprintf(stderr,"ERROR: No valid numeric expression after = for assignment to %s on line %" PRIu32 "\n",
       curtoken->toketext,curtoken->lno);
    retval=false;
    goto g_letstmt_xit;
  }
  if (T_NVAR==curtoken->tid) {
     retval=dostore_nvar(outfile,curtoken->toketext);
  } else {
     struct arraydata *ad=NULL; // pointer to a row of array information in the symbol table module

     ad=get_array_data(curtoken->toketext);
     if (NULL==ad) {
       debug_printf("ERROR: No array data for '%s' found\n",curtoken->toketext);
       retval=false;
       goto g_letstmt_xit;
     }
     switch(ad->dims) {
       case UINT32_C(1):
         retval=dostore_1D_navar(outfile,curtoken->toketext,ad->dim1);
         break;
       case UINT32_C(2):
         retval=dostore_2D_navar(outfile,curtoken->toketext,ad->dim1,ad->dim2);
         break;
       default:
         fprintf(stderr,"ICE: on line %" PRIu32 " write to array %s using %" PRIu32 " dimensions is not supported.\n",
           curtoken->lno,curtoken->toketext,ad->dims);
         retval=false;
         break;
     }
  }
g_letstmt_xit:
  if (curtoken!=NULL) {
    free(curtoken->toketext);curtoken->toketext=NULL;
    free(curtoken);curtoken=NULL;
  }
  return retval;
}

//
// This will emit code for the RANDOMIZE statement.
// Returns true on success, false otherwise.
//
// g_randomizestmt -> T_RANDOMIZE
//
static bool g_randomizestmt(void) {
  debug_printf("g_randomizestmt -> T_RANDOMIZE\n");
  if (!match(T_RANDOMIZE,__func__)) return false;
  if (!dorandomize(outfile)) return false;
  return true;
}

//
// This will emit code for the RESTORE statement.
// Returns true on success, false otherwise.
//
// g_restorestmt -> T_RESTORE
//
static bool g_restorestmt(void) {
  debug_printf("g_restorestmt -> T_RESTORE\n");
  if (!match(T_RESTORE,__func__)) return false;
  if (!dorestore(outfile)) return false;
  return true;
}

//
// This processes the OPTION BASE statement which does not
// emit code, but does set two flags.  If successful, before returning true
// this function will set the can_option_base to false, to prohibit multiple
// OPTION BASE statements. It also will update the base_is_one depending to the
// value of T_INTEGER, which must be 0 or 1.
//
// If can_option_base is false, then the program has already
// had an OPTION BASE, a DIM, or an array reference.  In that
// case it is not permitted to change the array base setting
// and an error message will be displayed.
// On error, false is returned.
//
// g_optionbasestmt -> T_OPTION T_BASE T_INTEGER
//
static bool g_optionbasestmt(void) {
  struct token *curtoken=NULL; // pointer to current token
  uint64_t value; // variable to hold argument of OPTION BASE (should be 0 or 1)
  bool retval=true; // return value of this function

  debug_printf("g_optionbasestmt -> T_OPTION T_BASE T_INTEGER\n");
  if (!match(T_OPTION,__func__)) {
    retval=false;
    goto g_optionbasestmt_xit;
  }
  if (!match(T_BASE,__func__)) {
    retval=false;
    goto g_optionbasestmt_xit;
  }
  curtoken=clone_next_token();
  debug_printf("T_INTEGER <<<value is %s>>>\n",curtoken->toketext);
  if (!match(T_INTEGER,__func__)) {
    printf("ERROR on line %" PRIu32 ": OPTION BASE requires an argument of 0 or 1\n",curtoken->lno);
    retval=false;
    goto g_optionbasestmt_xit;
  }
  if ( (strncmp(curtoken->toketext,"0",2)!=0) && (strncmp(curtoken->toketext,"1",2)!=0)) {
    printf("ERROR on line %" PRIu32 ": Invalid value for option base; must be 0 or 1\n",curtoken->lno);
    retval=false;
    goto g_optionbasestmt_xit;
  }
  value=(uint64_t)(curtoken->toketext[0]-48); // convert the digit from ASCIIZ to unsigned integer
  if ((value!=UINT64_C(0))&&(value!=UINT64_C(1))) { // verify it is either 0 or 1
    printf("ERROR on line %" PRIu32 ": Invalid value for option base; must be 0 or 1\n",curtoken->lno);
    retval=false;
    goto g_optionbasestmt_xit;
  }
  if (!can_option_base) { // ensure we are still permitted to set the base
    printf("ERROR on line %" PRIu32 ": Invalid use of OPTION BASE statement\n",curtoken->lno);
    retval=false;
    goto g_optionbasestmt_xit;
  }
  // finally, set the array minimum subscript (base) value and the flag prohibiting changing it
  can_option_base=false;
  base_is_one=((UINT64_C(1)==value)?true:false);
g_optionbasestmt_xit:
  if (curtoken!=NULL) {
    free(curtoken->toketext);curtoken->toketext=NULL;
    free(curtoken);curtoken=NULL;
  }
  return retval;
}

//
// This will emit code for either GOTO or GOSUB, depending on
// whether the token stream had 'SUB' or 'TO' after the 'GO'.
// Returns true on success, false otherwise.
//
// g_go_stmt -> T_GO g_gotail
// g_gotail -> T_TO glineno
// g_gotail -> T_SUB glineno
//
static bool g_go_stmt(void) {
  uint32_t tlineno; // target line number of the branch
  uint32_t mylineno; // line number of this GO statement
  static char msg[ERROR_BUFFER_MAX]; // buffer for error message
  bool isgoto=false; // flag to indicate whether this is a GOTO (true) or GOSUB (false)
   
  debug_printf("g_go_stmt -> T_GO g_gotail\n");
  if (!match(T_GO,__func__)) return false;
  mylineno=peek_next_token().lno;
  switch (peek_next_token().tid) {
    case T_TO:
      debug_printf("g_gotail -> T_TO glineno\n");
      if (!match(T_TO,__func__)) return false;
      isgoto=true;
      break;
    case T_SUB:
      debug_printf("g_gotail -> T_SUB glineno\n");
      if (!match(T_SUB,__func__)) return false;
      break;
    default:
      return false;
  }
  if (!g_lineno(&tlineno,msg,false)) {
    fputs(msg,stderr);
    return false;
  }
  // update lineno module information
  if (!add_new_jump_target(tlineno,mylineno,current_scope)) return false;
  // emit the code
  if (isgoto) { // for GOTO
    if (!dogoto(outfile,tlineno)) return false;
  } else { // FOR GOSUB
    if (!dogosub(outfile,tlineno)) return false;
  }
  return true;
}
  
//
// This will emit code for the GOTO statement.
// Returns true on success, false otherwise.
//
// g_gotostmt -> T_GOTO g_lineno
//
static bool g_gotostmt(void) {
  uint32_t tlineno; // target line number of the branch
  uint32_t mylineno; // line number of this GOTO statement
  static char msg[ERROR_BUFFER_MAX]; // error message buffer

  debug_printf("g_gotostmt -> T_GOTO g_lineno\n");
  if (!match(T_GOTO,__func__)) return false;
  mylineno=peek_next_token().lno;
  if (!g_lineno(&tlineno,msg,false)) {
    fputs(msg,stderr);
    return false;
  }
  // update lineno module information
  if (!add_new_jump_target(tlineno,mylineno,current_scope)) return false;
  // emit the code
  if (!dogoto(outfile,tlineno)) return false;
  return true;
}

//
// This will emit code for the GOSUB statement.
// Returns true on success, false otherwise.
//
// g_gosubstmt -> T_GOSUB g_lineno
//
static bool g_gosubstmt(void) {
  uint32_t tlineno; // target line number of the branch
  uint32_t mylineno; // line number of this GOSUB statement
  static char msg[ERROR_BUFFER_MAX]; // error message buffer

  debug_printf("g_gosubstmt -> T_GOSUB g_lineno\n");
  if (!match(T_GOSUB,__func__)) return false;
  mylineno=peek_next_token().lno;
  if (!g_lineno(&tlineno,msg,false)) {
    fputs(msg,stderr);
    return false;
  }
  // update lineno module information
  if (!add_new_jump_target(tlineno,mylineno,current_scope)) return false;
  // emit the code
  if (!dogosub(outfile,tlineno)) return false;
  return true;
}

//
// This will emit code for the RETURN statement.
// Returns true on success, false otherwise.
//
// g_returnstmt -> T_RETURN
//
static bool g_returnstmt(void) {
  debug_printf("g_returnstmt -> T_RETURN\n");
  if (!match(T_RETURN,__func__)) return false;
  // emit the code
  if (!doreturn(outfile)) return false;
  return true;
}

//
// This will not emit any code, since REM statements are comments and are discarded.
// Returns true on success, false otherwise.
//
// g_remstmt -> T_REM
//
static bool g_remstmt(void) {
  debug_printf("g_remstmt -> T_REM\n");
  return match(T_REM,__func__);
}

//
// This will emit code for the STOP statement.
// Returns true on success, false otherwise.
//
// g_stopstmt -> T_STOP
//
static bool g_stopstmt(void) {
  debug_printf("g_stopstmt -> T_STOP\n");
  if (!match(T_STOP,__func__)) return false;
  // emit the code
  if (!dostop(outfile)) return false;
  return true;
}

//
// Returns true on success, false otherwise.
//
// g_statement -> g_defstmt
// g_statement -> g_ifstmt
// g_statement -> g_on_gotostmt
// g_statement -> g_inputstmt
// g_statement -> g_readstmt
// g_statement -> g_datastmt
// g_statement -> g_dimstmt
// g_statement -> g_nextstmt
// g_statement -> g_forstmt
// g_statement -> g_printstmt
// g_statement -> g_letstmt
// g_statement -> g_randomizestmt
// g_statement -> g_restorestmt
// g_statement -> g_optionbasestmt
// g_statement -> g_remstmt
// g_statement -> g_stopstmt
// g_statement -> g_returnstmt
// g_statement -> g_go_stmt
// g_statement -> g_gotostmt
// g_statement -> g_gosubstmt
//
static bool g_statement(void) {
  struct token curtoken=peek_next_token();
  switch(curtoken.tid) {
    case T_DEF:
       debug_printf("g_statement -> g_defstmt\n");
       return g_defstmt();
    case T_IF:
       debug_printf("g_statement -> g_ifstmt\n");
       return g_ifstmt();
    case T_ON:
       debug_printf("g_statement -> g_on_gotostmt\n");
       return g_on_gotostmt();
    case T_INPUT:
       debug_printf("g_statement -> g_inputstmt\n");
       return g_inputstmt();
    case T_READ:
       debug_printf("g_statement -> g_readstmt\n");
       return g_readstmt();
    case T_DATA:
       debug_printf("g_statement -> g_datastmt\n");
       return g_datastmt();
    case T_DIM:
       debug_printf("g_statement -> g_dimstmt\n");
       return g_dimstmt();
    case T_NEXT:
       debug_printf("g_statement -> g_nextstmt\n");
       return g_nextstmt();
    case T_FOR:
       debug_printf("g_statement -> g_forstmt\n");
       return g_forstmt();
    case T_PRINT:
       debug_printf("g_statement -> g_printstmt\n");
       return g_printstmt();
    case T_LET:
       debug_printf("g_statement -> g_letstmt\n");
       return g_letstmt();
    case T_RANDOMIZE:
       debug_printf("g_statement -> g_randomizestmt\n");
       return g_randomizestmt();
    case T_RESTORE:
       debug_printf("g_statement -> g_restorestmt\n");
       return g_restorestmt();
    case T_OPTION:
       debug_printf("g_statement -> g_optionbasestmt\n");
       return g_optionbasestmt();
    case T_REM:
       debug_printf("g_statement -> g_remstmt\n");
       return g_remstmt();
    case T_STOP:
       debug_printf("g_statement -> g_stopstmt\n");
       return g_stopstmt();
    case T_RETURN:
       debug_printf("g_statement -> g_returnstmt\n");
       return g_returnstmt();
    case T_GO:
       debug_printf("g_statement -> g_go_stmt\n");
       return g_go_stmt();
    case T_GOTO:
       debug_printf("g_statement -> g_gotostmt\n");
       return g_gotostmt();
    case T_GOSUB:
       debug_printf("g_statement -> g_gosubstmt\n");
       return g_gosubstmt();
    case T_NVAR: // error production to catch the common forgotten LET in the assignment statement
    case T_SVAR:
    case T_NAVAR:
       fprintf(stderr,"ERROR on line %" PRIu32 ": missing %s keyword in assignment\n",
               curtoken.lno,token_names[T_LET].namestring+2);
       break;
    default:
       fprintf(stderr,"ERROR on line %" PRIu32 ": Expected a keyword after the line number but did not find one.\n",
               curtoken.lno);
       break;
  }
  return false;
}

//
// Returns true on success, false otherwise.
//
// glines -> g_lineno g_statement T_EOL g_linestail
// g_lines -> epsilon
// g_linestail -> g_lineno g_statement T_EOL g_linestail
// g_linestail -> epsilon
//
static bool g_lines(void) {
  uint32_t tlineno; // line number of current line of Minimal BASIC program being compiled
  static char msg[ERROR_BUFFER_MAX]; // error message buffer

  debug_printf("glines -> g_lineno g_statement T_EOL g_linestail\n");
  do {
    if (!g_lineno(&tlineno,msg,true)) {
       if (msg[0]!=0)
         fputs(msg,stderr);
       else
         fprintf(stderr,"ERROR: Missing line number\n");
       return false;
    }
    debug_printf("Processing line number %" PRIu32 "\n",tlineno);
    // emit statement label (for use when this line is a jump target)
    if (!doloc(outfile,(uint32_t)tlineno)) return false;
    // update symbol table
    if (!add_new_line_number(tlineno,msg,current_scope)) {
      fputs(msg,stderr);
      return false;
    }
    if (peek_next_token().tid==T_END) {
      debug_printf("g_lines -> epsilon\n");
      return true;
    }
    if (!g_statement()) return false;
    if (!match(T_EOL,__func__)) return false;
    if (peek_next_token().tid!=T_EOF)
       debug_printf("g_linestail -> g_lineno g_statement T_EOL g_linestail\n");
  } while (peek_next_token().tid!=T_EOF);
  debug_printf("g_linestail -> epsilon\n");
  return true;
}

//
// This is the start production for the grammar.  It returns true if the input
// token stream generates a program that is in the language described by the
// grammar, false otherwise.  To enforce the rule that the END statement must
// be the final statement, the rest of the parse is done by g_lines(), with
// this final production guaranteeing that the last line is an END statement.
// Returns true on success, false otherwise.
//
// g_program -> g_lines T_UINTEGER T_END T_EOL T_EOF
//
static bool g_program(void) {
  struct token curtoken; // current token

  debug_printf("g_program -> g_lines T_UINTEGER T_END T_EOL T_EOF\n");
  // final line number
  if (!g_lines()) return false;
  curtoken=peek_next_token();
  // end statement
  if (!match(T_END,__func__)) {
    fprintf(stderr,"ERROR: missing final %s statement.\n",token_names[T_END].namestring+2);
    return false;
  }
  if (!match(T_EOL,__func__)) {
    fprintf(stderr,"ERROR: final END statement on line %" PRIu32 " missing line terminator.\n",curtoken.lno);
    return false;
  }
  // final synthetic EOF token to signal end of token stream
  if (!match(T_EOF,__func__)) {
    fprintf(stderr,"ERROR: END statement on line %" PRIu32 " must be the last statement in the program.\n",curtoken.lno);
    return false;
  }
  return true;
}

//
// This is the external entry point for the parser.  This parser is a recursive-descent
// predictive parser with an LL(1) grammar.  Semantic checking is
// done as part of the parser.  Code generation is a side effect of the parse.
// Tokens are obtained using the scanner2 module API.  In a few places extra error
// checking for bad matches occurs which is semantically like using error productions.
// For example, when looking for statements some people used to other BASIC dialects
// forget the T_LET token, so when looking for statements if the first token is
// T_NVAR, T_NAVAR, or T_SVAR the parser will guess that an assignment with T_LET
// was intended, and it will report an error about missing T_LET token.
//
// It is the responsibility of the parser to free tokens after they are no longer
// needed.
//
// The f parameter is the file handle for the assembly output file, and the
// file must already be open for output.  The xf parameter is the file handle
// for the extra assembly output file that holds UDFs, and the file must
// already be open for output.  Returns true if the parser was successful,
// false otherwise.
//
bool parseit(FILE *f,FILE *xf) {
  bool retval; // function return value

  outfile=f;
  xtraoutfile=xf;
  retval=g_program(); // call the recursive descent parser's start symbol
  if (retval) { // if the parse was OK
     uint32_t i=forstack_ptr;	// index into FOR loop dedicated compile-time stack

     if (i!=UINT32_C(0)) { // if there were unclosed FOR loops during compilation
        while (i-->UINT32_C(0)) // for each unclosed FOR loop, report the error
          fprintf(stderr,"ERROR: FOR %s statement without NEXT on line %" PRIu32 "\n",forstack[i].ndex_var,forstack[i].lineno);
        retval=false; // and ensure we return an error
     }
     if (need_READ && (!have_DATA)) { // if there was a READ statement but no DATA statement
        fprintf(stderr,"Program contains a READ statement on line %" PRIu32 " but no DATA statements.\n",
          READ_lineno);
        retval=false; // and ensure we return an error
     }
  }
  return retval;
}

//
// This is code to test the module and is not normally used.
// If you compile with UNIT_TEST defined, then this module will
// be a stand-alone program with its own main() that runs some
// self-tests.
//
#ifdef UNIT_TEST
#include <string.h>

// BEGIN BLOCK OF STUBS for codegen module
bool dornd(FILE *f) { return true; }
bool doabs(FILE *f) { return true; }
bool dosgn(FILE *f) { return true; }
bool doatan(FILE *f) { return true; }
bool docos(FILE *f) { return true; }
bool doexp(FILE *f) { return true; }
bool doint(FILE *f) { return true; }
bool dolog(FILE *f) { return true; }
bool dosin(FILE *f) { return true; }
bool dosqr(FILE *f) { return true; }
bool dotan(FILE *f) { return true; }
bool dopow(FILE *f) { return true; }
bool push_float_literal(FILE *f,const char *flitname,const char *flitval) { return true; }
bool push_string_literal(FILE *f,const char *slitname,const char *txt) { return true; }
bool push_nvar(FILE *f, const char *varname) { return true; }
bool push_nparm(FILE *f, const char *varname, const char *fname) { return true; }
bool push_svar(FILE *f, const char *varname) { return true; }
bool push_1D_navar(FILE *f, const char *varname, const uint32_t maxindex) { return true; }
bool push_2D_navar(FILE *f, const char *varname, const uint32_t rows, const uint32_t columns) { return true; }
bool domultiply(FILE *f) { return true; }
bool dodivide(FILE *f) { return true; }
bool donegate(FILE *f) { return true; }
bool doadd(FILE *f) { return true; }
bool dosubtract(FILE *f) { return true; }
bool doprint_string(FILE *f) { return true; }
bool doprint_number(FILE *f) { return true; }
bool doprint_nextfield(FILE *f) { return true; }
bool doprint_newline(FILE *f) { return true; }
bool dostore_nvar(FILE *f, const char *varname) { return true; }
bool dostore_svar(FILE *f, const char *varname);
bool dostore_1D_navar(FILE *f, const char *varname, const uint32_t maxrows) { return true; }
bool dostore_2D_navar(FILE *f, const char *varname, const uint32_t maxrows, const uint32_t maxcolumns) { return true; }
bool dorandomize(FILE *f) { return true; }
bool doloc(FILE *f, const uint32_t lineno) { return true; }
bool dostop(FILE *f) { return true; }
bool dogoto(FILE *f, const uint32_t target_lineno) { return true; }
bool donumeric_comparison(FILE *f, const enum relop op, const uint32_t target_lineno) { return true; }
bool dogosub(FILE *f, const uint32_t target_lineno) { return true; }
bool doreturn(FILE *f) { return true; }
bool doongotostart(FILE *f, uint32_t maxchoice) { return true; }
bool doongotochoice(FILE *f, uint32_t target_lineno) { return true; }
bool doongotoend(FILE *f) { return true; }
bool doforpush1(FILE *f) { return true; }
bool doforloopstart(FILE *f, const char *ndex_var, char *tol_label, char *bol_label, char *inc_var, char *limit_var) { return true; }
bool doforloopend(FILE *f,const char *tol_label, const char *bol_label, const char *ndex_var, const char *inc_var) { return true; }
bool dostring_comparison(FILE *f, const enum relop op, const uint32_t target_lineno) { return true; }
bool dostore_svar(FILE *f, const char *varname) { return true; }
bool dotab(FILE *f) { return true; }
bool doinput_nvar(FILE *f, const char *varname,const uint32_t slotno,const uint32_t lno) { return true; }
bool doinput_svar(FILE *f, const char *varname,const uint32_t slotno) { return true; }
bool doinput_navar(FILE *f, const char *varname, const uint32_t slotno,const uint32_t lno) { return true; }
bool doinput_nvar_phase2(FILE *f, const char *varname,const uint32_t slotno,const uint32_t lno) { return true; }
bool doinput_svar_phase2(FILE *f, const char *varname,const uint32_t slotno) { return true; }
bool doinput_navar_phase2(FILE *f, const char *varname, const uint32_t rows, const uint32_t columns,const uint32_t slotno,const uint32_t lno) { return true; }
bool doread_nvar(FILE *f, const char *varname) { return true; }
bool doread_svar(FILE *f, const char *varname) { return true; }
bool doread_navar(FILE *f, const char *varname, const uint32_t rows, const uint32_t columns) { return true; }
bool dorestore(FILE *f) { return true; }
bool doudf(FILE *f,const char *fname,bool hasarg) { return true; }
bool doudf_header(FILE *f,const char *fname,const char *argname) { return true; }
bool doudf_footer(FILE *f,const char *fname) { return true; }
char *g_fmt(char * b, double x, signed long int digits_after_decimal, char digits_in_exponent) { return 0; }
bool do_INPUT(FILE *f,const char *nfmt) { return true; }
bool emit_comment(FILE *f,const char *fmt, ...) { return 0; }

// END BLOCK OF STUBS for codegen module

static const char sample_code[]="\
10 OPTION BASE 1\n\
11 REM This is a test\n\
12 RANDOMIZE\n\
13 LET A=INT(360/0.31415926535E-1+2*4/(1+(-2))-1)+SIN(B2)/TAN(LOG(0^0))\n\
14 LET A=INT(360/0.31415926535E-1+2*3/(4+(-5))-6)+SIN(B2)/TAN(LOG(7^8))\n\
15 FOR I=1 TO 10\n\
16   FOR J=9 TO 0 STEP -1\n\
17     PRINT I*J\n\
18   NEXT J\n\
19 NEXT I\n\
20 GOSUB 100\n\
21 GO SUB 100\n\
25 RESTORE\n\
26 LET X=RND\n\
29 GO TO 190\n\
30 GOTO 190\n\
31 PRINT ABS(-1)\n\
100 REM start of subroutine\n\
110 RETURN\n\
190 STOP\n\
200 DIM Z(20),B(5,5)\n\
210 LET X=RND/6\n\
220 LET B(X,X)=X\n\
230 LET Z(5)=4\n\
240 LET Z(Z(5))=5\n\
245 PRINT Q(3,3)\n\
250 PRINT Z(Q(3,2))\n\
254 ON RND/6 GO TO 10,20,30,100,200,256\n\
255 ON RND/6 GOTO 10,20,30,100,200,256\n\
256 IF Z(1)<(B(3,2)*SIN(X)) THEN 11\n\
260 READ Z(1),Z(2+X),Q(Z(1),RND),X\n\
300 DATA 1,0.1,.1,-1,-.1,-0.001E+3\n\
310 LET A$ = \"hello\"\n\
315 PRINT TAB(10);A1,A$,Z(1),Z(X)\n\
320 INPUT A1,A$,Z(1),Z(X)\n\
321 PRINT 3*X - Y^2, -X/Y, Z(1)+Z(2)+Z(3), SQR(X^2+Y^2),2^(-X)\n\
322 PRINT X,A$,V(3),W(X,X+Y/2),S$,C$\n\
330 READ A1,A$,Z(1),Z(X)\n\
340 DATA 1,\"dog\",99,\"big fat doggie\"\n\
341 DATA 1,500,-21.,.255,1E10,5E-1,.4E+1,\"XYZ\",\"X - 3B2\",\"1E10\"\n\
350 IF A$ = \"dog\" THEN 360\n\
360 IF \"dog\" <> A$ THEN 400\n\
361 DEF FNF(X) = X^4 - 1\n\
362 DEF FNA(X) = A*X + X\n\
363 DEF FNP = 3.14159\n\
364 LET P = 3.14159\n\
365 LET Q(X,3) = SIN(X)*Y + 1\n\
366 LET A$ = \"ABC\"\n\
367 LET A$ = B$\n\
400 END\n\
";

int main(int argc, char *argv[]) {
  struct token t;
  bool quit_on_fail=false,retval=EXIT_SUCCESS;
  FILE *ofile=NULL,*xofile=NULL;

  verbose=false;
  if ((2==argc)&&(strncmp(argv[1],"-V",3)==0)) {
     verbose=true;
     quit_on_fail=true;
  } else {
     quit_on_fail=false;
     if ((2==argc)&&(strncmp(argv[1],"-v",3)==0))
       verbose=true;
     else
       verbose=false;
  }
  init_debug_printf(stderr);
  if ((ofile=fopen("/dev/null","w"))==NULL) {
    fprintf(stderr,"Error: unable to open file '/dev/null' for output.\n");
    abort();
  }
  // convert text to token stream
  if (!scan_buffer(sample_code)) {
    fflush(stdout);
    fprintf(stderr,"scan_buffer failed\n");
    retval=EXIT_FAILURE;
    goto self_test_xit;
  }
  fprintf(stderr,"scan_buffer() OK\n");
  if (verbose)
     dump_token_stream();
  // ensure we can peek the first token
  t=peek_next_token();
  // and it is an unsigned integer
  if (t.tid!=T_INTEGER) {
    fclose(ofile);
    ofile=NULL;
    fprintf(stderr,"Missing line number\n");
    retval=EXIT_FAILURE;
    goto self_test_xit;
  }
  // with value '10'
  if (strncmp(t.toketext,"10",3)!=0) {
    fprintf(stderr,"Expected line '10', got line '%s'\n",t.toketext);
    if (quit_on_fail) {
      retval=EXIT_FAILURE;
      goto self_test_xit;
    }
  }
  if (parseit(ofile,xofile)) {
    fflush(stdout);
    if (!verify_all_jump_targets_exist()) {
      fprintf(stderr,"Some jump targets do not exist\n");
      fprintf(stderr,"Self-test failed\n");
      retval=EXIT_FAILURE;
      goto self_test_xit;
    }
    if (verbose) {
      fprintf(stderr,"\n********* SYMBOL TABLE ********\n\n");
      dump_symbol_table(true);
    }
    printf("Self-test passed\n");
    retval=EXIT_SUCCESS;
    goto self_test_xit;
  }
  fprintf(stderr,"parsit() failed\n");
  fprintf(stderr,"un-consumed tokens:\n");
  dump_token_stream();
  fflush(stdout);
  fprintf(stderr,"\n********* SYMBOL TABLE ********\n\n");
  dump_symbol_table(false);
  fflush(stdout);
  fprintf(stderr,"Self-test failed\n");
  retval=EXIT_FAILURE;

self_test_xit:
  clear_lineno_data();
  clear_symbol_table();
  clear_token_stream();
  fflush(ofile);
  fclose(ofile);
  ofile=NULL;
  fflush(stdout);
  return retval;
}
#endif
