//
// peephole.c - This is the file containing the peephole optimizer
//              for Mr. Ham's ECMA-55 Minimal BASIC compiler.
//
// Copyright (C) 2014  John Gatewood Ham
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// version 2 of the Free Software Foundation License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
#if __STDC_VERSION__ < 199901L
#error "You need a C99 compiler"
#endif
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <stdbool.h>

#define MAXLINEWIDTH 201	// number of coluns in input buffer
				// for a max of 200 bytes & trailing NULL
#define WINDOWSIZE 2		// number of lines in input buffer
static char BUFFER[WINDOWSIZE][MAXLINEWIDTH];	// input buffer matrix

bool match(const char *haystack,const char *needle);

//
// return true if string in needle parameter matches
// the string in the haystack parameter after the haystack
// parameter string has had whitespace removed, ignoring any
// trailing comment, otherwise return false
// Assumptions (which will be true for compiler generated assembly): 
// 1) neither haystack nor needle contain a newline
// 2) neither haystack nor needle contain any non-trailing NULL
// 3) both haystack and needle are non-empty ASCIIZ strings
//
bool match(const char *haystack,const char *needle) {
  size_t len=strlen(needle);
  while (*haystack!=0) {
    while ((*haystack==' ')||(*haystack=='\t')) haystack++; // eat leading whitespace
    if (strncmp(haystack,needle,len)!=0) return false;
    haystack+=len;
    while ((*haystack==' ')||(*haystack=='\t')) haystack++; // eat whitespace
    if ((*haystack=='#')||(*haystack=='\n')) return true;
    return false;
  }
  return false;
}

//
// This main function implments a simple peephole optimizer for GNU gas
// AMD64 code using the normal AT&T assmebly dialect which is output by
// the compiler.  This function will verify its arguments, open the files, 
// then // scan the input file with a 2 line logical window looking for 
// redundant push/pop sequences that can be safely removed.
// This function requires exactly two filenames to passed as arguments.
// The first is the assembly language input file (GNU GAS AMD64 code).
// The second is the assembly language output file (GNU GAS AMD64 code).
// Returns EXIT_SUCCESS if there are no errors, EXIT_FAILURE otherwise.
//
int main(int argc, char **argv) {
  int in;		// input file handle
  int out;		// output file handle
  int howmany;		// number of bytes read by the read() system call
  uint32_t curline;	// current line of the input file
  uint32_t curpos;	// current column position in read buffer
  uint32_t curlineno;	// current line in the input buffer
  uint32_t inlines;	// number of lines read in the input file
  uint32_t outlines;	// number of files written to the output file
  unsigned char c;	// input byte buffer used by the read() system call

  if (argc!=3) {
    printf("Usage:  %s INFILENAME OUTFILENAME\n",argv[0]);
    printf("had %d arguments, need 2 arguments\n",argc);
    return EXIT_FAILURE;
  }
  curline=curpos=UINT32_C(0);
  curlineno=inlines=outlines=UINT32_C(0);
  if ((in=open(argv[1],O_EXCL|O_RDONLY))<0) {
    printf("Unable to open '%s' for input\n",argv[1]);
    return EXIT_FAILURE;
  }
  if ((out=open(argv[2],O_CREAT|O_EXCL|O_WRONLY,S_IWUSR|S_IRGRP|S_IROTH|S_IRUSR))<0) {
    fprintf(stderr,"Unable to open '%s' for write\n",argv[2]);
    return EXIT_FAILURE;
  }
  // while true attempt to read 1 byte from input file
  while ((howmany=read(in,&c,1))) {
    if (howmany<1) break;	// if input file exhausted exit while
    if (UINT32_C(0)==curpos) ++curlineno;	// if start of new line
						// move to next line of buffer
    BUFFER[curline][curpos++]=c;	// put charater into buffer
    if ('\n'==c) {		// if input character is a newline then
      ++inlines;
      BUFFER[curline][curpos]=0; // terminate current line in buffer
      if (UINT32_C(1)==curline) {	// if we have two lines buffered
					// then look for patterns to match
        if ( match(BUFFER[0],"pushxmm 0") && match(BUFFER[1],"popxmm 0") ) {
          curline=curpos=UINT32_C(0);	// discard buffer because the pattern matched
        } else {
          if ( match(BUFFER[0],"pushsaddr") && match(BUFFER[1],"popsaddr %rdi") ) {
            curline=curpos=UINT32_C(0);	// discard buffer because the pattern matched
          } else {
            if (write(out,BUFFER[0],strlen(BUFFER[0]))<0) {	// write first line of buffer to output file
              fprintf(stderr,"Unable to write to file '%s'\n",argv[2]);
              return EXIT_FAILURE;
            }
            ++outlines;
            strcpy(BUFFER[0],BUFFER[1]);	// move second line of buffer to first line of buffer
            curpos=UINT32_C(0);			// and reset buffer to begin filling second line
          }
        }
      } else {	// saw end of first line of buffer, so switch to filling
		// second line of buffer
        curline=UINT32_C(1);
        curpos=UINT32_C(0);
      }
    } else {	// character is not a newline
      if (curpos>=MAXLINEWIDTH) {	// if the line was too long report error and exit while
        fprintf(stderr,"Line %" PRIu32 " of the assembly input is too long.\n",curlineno);
        fprintf(stderr,"Maximum supported line width is %" PRIu32 " bytes.\n",MAXLINEWIDTH);
        break;
      }
      // everything OK, loop back up to read next byte
    }
  }
  fflush(stdout);
  if (UINT32_C(1)==curline) {	// if we were filling the second line of the buffer
    if (write(out,BUFFER[0],strlen(BUFFER[0]))<0) {	// write the first line in the buffer to the output file
       fprintf(stderr,"Unable to write to file '%s'\n",argv[2]);
       return EXIT_FAILURE;
    }
    ++outlines;
    if (curpos>UINT32_C(0)) {	// if the second line of the buffer is non-empty
       if (write(out,BUFFER[1],strlen(BUFFER[0]))<0) {	// write the second line in the buffer to the output file
         fprintf(stderr,"Unable to write to file '%s'\n",argv[2]);
         return EXIT_FAILURE;
       }
       ++outlines;
    }
  } else {	// we were filling first line of the buffer
    if (curpos>UINT32_C(0)) {	// if the first line of the buffer is non-empty
       if (write(out,BUFFER[0],strlen(BUFFER[0]))<0) {	// write the first line in the buffer to the output file
         fprintf(stderr,"Unable to write to file '%s'\n",argv[2]);
         return EXIT_FAILURE;
       }
       ++outlines;
    }
  }
  close(out);	// close the output file
  close(in);	// close the input file
#ifdef VERBOSE
  // report statistics if compiled with VERBOSE flag
  if (inlines!=outlines)
    printf("Removed %" PRIu32 " lines\n",inlines-outlines);
  else
    printf("Removed no lines\n");
#endif
  return EXIT_SUCCESS;
}
