//
// scanner2.c - This is the file containing the scanner
//              for Mr. Ham's ECMA-55 Minimal BASIC compiler.
//
// Copyright (C) 2013,2014  John Gatewood Ham
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// version 2 of the Free Software Foundation License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
#include "scanner2.h"
#include "globals.h"
#include "lineno.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <ctype.h>

#define TBUFLEN 81

// map to convert token identifiers to text string for display
// and also can be used to determine if a token is a Minimal BASIC keyword or not
const struct tndata token_names[] = {
 { "T_ABS", false },
 { "T_ADD", false },
 { "T_ATN", false },
 { "T_BASE", true },
 { "T_COMMA", false },
 { "T_COS", false },
 { "T_DATA", true },
 { "T_DEF", true },
 { "T_DIM", true },
 { "T_DIV", false },
 { "T_END", true },
 { "T_EOF", false },
 { "T_EOL", false },
 { "T_EQ", false },
 { "T_EXP", false },
 { "T_FNID", false },
 { "T_FOR", true },
 { "T_GE", false },
 { "T_GO", true },
 { "T_GOSUB", true },
 { "T_GOTO", true },
 { "T_GT", false },
 { "T_IF", true },
 { "T_INPUT", true },
 { "T_INT", false },
 { "T_INTEGER", false },
 { "T_LE", false },
 { "T_LET", true },
 { "T_LOG", false },
 { "T_LPAREN", false },
 { "T_LT", false },
 { "T_MUL", false },
 { "T_NAVAR", false },
 { "T_NE", false },
 { "T_NEXT", true },
 { "T_NVAR", false },
 { "T_ON", true },
 { "T_OPTION", true },
 { "T_POW", false },
 { "T_PRINT", true },
 { "T_QSTRING", false },
 { "T_RANDOMIZE", true },
 { "T_READ", true },
 { "T_REAL", false },
 { "T_REM", true },
 { "T_RESTORE", true },
 { "T_RETURN", true },
 { "T_RND", false },
 { "T_RPAREN", false },
 { "T_SEMI", false },
 { "T_SGN", false },
 { "T_SIN", false },
 { "T_SQR", false },
 { "T_STEP", true },
 { "T_STOP", true },
 { "T_SUB", true },
 { "T_SUBTRACT", false },
 { "T_SVAR", false },
 { "T_TAB", false },
 { "T_TAN", false },
 { "T_THEN", true },
 { "T_TO", true },
 { "T_UQSTRING", false },
 { "T_UNKNOWN", false },        // not a real token
};
// numtokens does not include T_UNKNOWN
// but otherwise is the number of tokens in the token_names map
#define numtokens (sizeof(token_names)/sizeof(struct tndata)-UINT32_C(1))

// Character class definition for DFSM
//
// 0..25 is A..Z upper case
// 26  is DOLLAR
// 27  is DIGIT
// 28  is ADD
// 29  is MINUS
// 30  is MULTIPLY
// 31  is DIVIDE
// 32  is POWER
// 33  is LEFT PARENTHESIS
// 34  is RIGHT PARENTHESIS
// 35  is LESSTHAN
// 36  is GREATERTHAN
// 37  is EQUALS
// 38  is COMMA
// 39  is SEMICOLON
// 40  is SPACE
// 41  is NEWLINE
// 42  is DOT
// 43  is QUOTE
// 44  is anything else
//
static const unsigned char character_class[]={
44,44,44,44,44,44,44,44,44,44, 
41,44,44,44,44,44,44,44,44,44, 
44,44,44,44,44,44,44,44,44,44,
44,44,40,44,43,44,26,44,44,44,
33,34,30,28,38,29,42,31,27,27,
27,27,27,27,27,27,27,27,44,39,
35,37,36,44,44, 0, 1, 2, 3, 4,
 5, 6, 7, 8, 9,10,11,12,13,14,
15,16,17,18,19,20,21,22,23,24,
25,44,44,44,32,44,44,44,44,44,
44,44,44,44,44,44,44,44,44,44,
44,44,44,44,44,44,44,44,44,44,
44,44,44,44,44,44,44,44,
};
// number of classes in the character class table
#define numclasses (sizeof(character_class)/sizeof(unsigned char))

// Actions that can be taken in the DFSM
//
// REJ is reject
// ACC is accept and consume the current byte
// ANC is accept but do not consume the current byte
// MOV is move to new state consuming the current byte
enum action {REJ=0, ACC, ANC, MOV};
// map to convert action codes to strings for text display
static const char *action_names[]={"REJ","ACC","ANC","MOV"};

// vector used to determine whether a given state
// is a state a final state?  If so what token is accepted?
static const struct fstruct {
  const bool can_accept;        // if true, this state can accept, false otherwise
  const int32_t accept_state;   // if can_accept is true, identifier of token found
} is_final[] = {
/*  0 */ { false, -1},
/*  1 */ { true, T_NVAR},
/*  2 */ { false, -1},
/*  3 */ { false, -1},
/*  4 */ { true, T_NVAR},
/*  5 */ { false, -1},
/*  6 */ { false, -1},
/*  7 */ { true, T_REM},
/*  8 */ { false, -1},
/*  9 */ { false, -1},
/* 10 */ { true, T_NVAR},
/* 11 */ { false, -1},
/* 12 */ { true, T_NVAR},
/* 13 */ { false, -1},
/* 14 */ { false, -1},
/* 15 */ { true, T_NVAR},
/* 16 */ { false, -1},
/* 17 */ { false, -1},
/* 18 */ { false, -1},
/* 19 */ { false, -1},
/* 20 */ { true, T_NVAR},
/* 21 */ { false, -1},
/* 22 */ { false, -1},
/* 23 */ { false, -1},
/* 24 */ { true, T_NVAR},
/* 25 */ { false, -1},
/* 26 */ { false, -1},
/* 27 */ { true, T_NVAR},
/* 28 */ { false, -1},
/* 29 */ { false, -1},
/* 30 */ { true, T_NVAR},
/* 31 */ { false, -1},
/* 32 */ { false, -1},
/* 33 */ { false, -1},
/* 34 */ { false, -1},
/* 35 */ { true, T_NVAR},
/* 36 */ { false, -1},
/* 37 */ { false, -1},
/* 38 */ { false, -1},
/* 39 */ { true, T_NVAR},
/* 40 */ { false, -1},
/* 41 */ { false, -1},
/* 42 */ { false, -1},
/* 43 */ { false, -1},
/* 44 */ { false, -1},
/* 45 */ { false, -1},
/* 46 */ { false, -1},
/* 47 */ { false, -1},
/* 48 */ { false, -1},
/* 49 */ { false, -1},
/* 50 */ { false, -1},
/* 51 */ { false, -1},
/* 52 */ { false, -1},
/* 53 */ { false, -1},
/* 54 */ { false, -1},
/* 55 */ { false, -1},
/* 56 */ { false, -1},
/* 57 */ { true, T_NVAR},
/* 58 */ { false, -1},
/* 59 */ { false, -1},
/* 60 */ { false, -1},
/* 61 */ { false, -1},
/* 62 */ { false, -1},
/* 63 */ { false, -1},
/* 64 */ { false, -1},
/* 65 */ { true, T_NVAR},
/* 66 */ { false, -1},
/* 67 */ { false, -1},
/* 68 */ { false, -1},
/* 69 */ { true, T_NVAR},
/* 70 */ { false, -1},
/* 71 */ { false, -1},
/* 72 */ { false, -1},
/* 73 */ { false, -1},
/* 74 */ { false, -1},
/* 75 */ { false, -1},
/* 76 */ { true, T_ADD},
/* 77 */ { true, T_INTEGER},
/* 78 */ { true, T_REAL},
/* 89 */ { false, -1},
/* 80 */ { false, -1},
/* 81 */ { true, T_REAL},
/* 82 */ { true, T_SUB},
/* 83 */ { true, T_NVAR},
/* 84 */ { false, -1},
/* 85 */ { true, T_UQSTRING},
/* 86 */ { true, T_INTEGER},
/* 87 */ { true, T_REAL},
/* 88 */ { true, T_REAL},
/* 89 */ { false, -1},
/* 90 */ { false, -1},
/* 91 */ { true, T_REAL},
/* 92 */ { true, T_UQSTRING},
/* 93 */ { false, -1},
/* 94 */ { true, T_UQSTRING},
/* 95 */ { true, T_LT},
/* 96 */ { true, T_GT},
/* 97 */ { true, T_INTEGER},
/* 98 */ { true, T_REAL},
};

// This record is used to hold one element of the DFSM
// represented by FSM_TABLE
struct fsm_element {
  const int32_t next_state; // state number
                            // main scanner start state is 0
                            // DATA scanner start state is 84
  const enum action what2do; // action for this state
};

// DFSM table
// The rows are the states, and the columns are the character classes
// So if you are in state X and the next byte is of character class Y,
// The action to take is determined by FSM_TABLE[X][Y].  If the action
// is ACC or ANC, then the next_state will contain the token ID that
// was accepted.
//
// This DFSM table contains two distinct scanners.  The main scanner
// for Minimal BASIC stats with state zero.  A secondary scanner used
// for DATA statements start with state 84.  A copy of this smaller
// scanner subset is in the generated assembly code for use by INPUT
// so that it will process constants the same way that compiler does.
static const struct fsm_element FSM_TABLE[][numclasses] = {
//
//state                 A                B                C                D                E                F                G                H                I                J                K                L                M                N                O                P                Q                R                S                T                U                V                W                X                Y                Z                $            digit               +                 -                *                /                ^                (                )                <                >                =                ,                ;             space           newline              .                "        anything else
//
/*   0 */   { {      69, MOV}, {       1, MOV}, {      72, MOV}, {       4, MOV}, {      10, MOV}, {      12, MOV}, {      15, MOV}, {      83, MOV}, {      20, MOV}, {      83, MOV}, {      83, MOV}, {      24, MOV}, {      83, MOV}, {      27, MOV}, {      30, MOV}, {      35, MOV}, {      83, MOV}, {      39, MOV}, {      57, MOV}, {      65, MOV}, {      83, MOV}, {      83, MOV}, {      83, MOV}, {      83, MOV}, {      83, MOV}, {      83, MOV}, {  T_SVAR, ACC}, {      77, MOV}, {      76, MOV},{       82, MOV}, {   T_MUL, ACC}, {   T_DIV, ACC}, {   T_POW, ACC}, {T_LPAREN, ACC}, {T_RPAREN, ACC}, {      95, MOV}, {      96, MOV}, {    T_EQ, ACC}, { T_COMMA, ACC}, {  T_SEMI, ACC}, {       0, MOV}, {   T_EOL, ACC}, {      78, MOV}, {      75, MOV}, {       0, REJ} },
/*   1 */   { {       2, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*   2 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       3, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*   3 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_BASE, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*   4 */   { {       5, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       8, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       9, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*   5 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       6, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*   6 */   { {  T_DATA, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*   7 */   { {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {       7, MOV}, {   T_REM, ANC}, {       7, MOV}, {       7, MOV}, {       7, MOV} },
/*   8 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_DEF, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*   9 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_DIM, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  10 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      11, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      74, MOV}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  11 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_END, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  12 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      13, MOV}, {      14, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  13 */   { {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {  T_FNID, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  14 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_FOR, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  15 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      16, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  16 */   { {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {      17, MOV}, {      19, MOV}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC}, {    T_GO, ANC} },
/*  17 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      18, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  18 */   { {       0, REJ}, { T_GOSUB, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  19 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_GOTO, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  20 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {    T_IF, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      21, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  21 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      22, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_INT, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  22 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      23, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  23 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, { T_INPUT, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  24 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      25, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      26, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  25 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_LET, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  26 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_LOG, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  27 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      28, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  28 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      29, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  29 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_NEXT, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  30 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {    T_ON, ACC}, {       0, REJ}, {      31, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  31 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      32, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  32 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      33, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  33 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      34, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  34 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {T_OPTION, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  35 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      36, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  36 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      37, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  37 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      38, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  38 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, { T_PRINT, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  39 */   { {      50, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      40, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      49, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  40 */   { {      41, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       7, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      42, MOV}, {      46, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  41 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_READ, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  42 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      43, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  43 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      44, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  44 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      45, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  45 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ},{T_RESTORE, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  46 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      47, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  47 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      48, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  48 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {T_RETURN, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  49 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_RND, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  50 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      51, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  51 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {      52, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  52 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      53, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  53 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      54, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  54 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      55, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  55 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      56, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  56 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {    0, REJ}, {T_RANDOMIZE, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  57 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      61, MOV}, {       0, REJ}, {      63, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      64, MOV}, {       0, REJ}, {       0, REJ}, {      58, MOV}, {      62, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  58 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      60, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      59, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  59 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_STOP, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  60 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_STEP, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  61 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_SGN, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  62 */   { {       0, REJ}, {   T_SUB, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  63 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_SIN, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  64 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_SQR, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  65 */   { {      68, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      66, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {    T_TO, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  66 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      67, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  67 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_THEN, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  68 */   { {       0, REJ}, {   T_TAB, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_TAN, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  69 */   { {       0, REJ}, {      70, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      71, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  70 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_ABS, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  71 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_ATN, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  72 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      73, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
/*  73 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_COS, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  74 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {   T_EXP, ACC}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  75 */   { {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {      75, MOV}, {       0, REJ}, {      75, MOV}, {T_QSTRING,ACC}, {      75, MOV} },
/*  76 */   { {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {      77, MOV}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {   T_ADD, ANC}, {      78, MOV}, {   T_ADD, ANC}, {   T_ADD, ANC} },
/*  77 */   { {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {      79, MOV}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {      77, MOV}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {T_INTEGER,ANC}, {      78, MOV}, {T_INTEGER,ANC}, {T_INTEGER,ANC} },
/*  78 */   { {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {      79, MOV}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {      78, MOV}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC} },
/*  79 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      81, MOV}, {      80, MOV}, {      80, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  80 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {      81, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },
/*  81 */   { {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {      81, MOV}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC}, {  T_REAL, ANC} },
/*  82 */   { {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {       77, MOV}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC}, {        78,MOV}, {T_SUBTRACT,ANC}, {T_SUBTRACT,ANC} },
/*  83 */   { {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_SVAR, ACC}, {  T_NVAR, ACC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, { T_NAVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC}, {  T_NVAR, ANC} },
//
// These 11 rows are the special FSM for DATA statement data elements, with start state 84 and final states
// 85, 86, 87, 88, 91, 92, and 93 for UQSTRING, INTEGER, REAL, REAL, REAL, UQSTRING and QSTRING(only on final ")
//
//state                 A                B                C                D                E                F                G                H                I                J                K                L                M                N                O                P                Q                R                S                T                U                V                W                X                Y                Z                $            digit               +                 -                *                /                ^                (                )                <                >                =                ,                ;             space           newline              .                "        anything else
/*  84 */   { {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {      86, MOV}, {      85, MOV}, {      85, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, { T_COMMA, ACC}, {       0, REJ}, {      84, MOV}, {   T_EOL, ACC}, {      87, MOV}, {      93, MOV}, {       0, REJ} },
/*  85 */   { {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {      86, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {T_UQSTRING,ANC},{       0, REJ}, {      92, MOV}, {T_UQSTRING,ANC},{      87, MOV}, {       0, REJ}, {       0, REJ} },
/*  86 */   { {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      89, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {      86, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {T_INTEGER,ANC}, {       0, REJ}, {      97, MOV}, {T_INTEGER,ANC}, {      87, MOV}, {       0, REJ}, {       0, REJ} },
/*  87 */   { {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      89, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {      88, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_REAL, ANC}, {       0, REJ}, {      98, MOV}, {  T_REAL, ANC}, {      92, MOV}, {       0, REJ}, {       0, REJ} },
/*  88 */   { {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      89, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {      88, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_REAL, ANC}, {       0, REJ}, {      92, MOV}, {  T_REAL, ANC}, {      92, MOV}, {       0, REJ}, {       0, REJ} },
/*  89 */   { {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {      91, MOV}, {      90, MOV}, {      90, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {T_UQSTRING,ANC},{       0, REJ}, {      98, MOV}, {T_UQSTRING,ANC},{      92, MOV}, {       0, REJ}, {       0, REJ} },
/*  90 */   { {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {      91, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {T_UQSTRING,ANC},{       0, REJ}, {      92, MOV}, {T_UQSTRING,ANC},{      92, MOV}, {       0, REJ}, {       0, REJ} },
/*  91 */   { {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {      91, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_REAL, ANC}, {       0, REJ}, {      98, MOV}, {  T_REAL, ANC}, {      92, MOV}, {       0, REJ}, {       0, REJ} },
/*  92 */   { {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {T_UQSTRING,ANC},{       0, REJ}, {      94, MOV}, {T_UQSTRING,ANC},{      92, MOV}, {       0, REJ}, {       0, REJ} },
/*  93 */   { {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {      93, MOV}, {       0, REJ}, {      93, MOV}, {T_QSTRING,ACC}, {      93, MOV} },
/*  94 */   { {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {T_UQSTRING,ANC},{       0, REJ}, {      94, MOV}, {T_UQSTRING,ANC},{      92, MOV}, {       0, REJ}, {       0, REJ} },
/*  95 */   { {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_NE, ACC}, {    T_LE, ACC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC}, {    T_LT, ANC} },
/*  96 */   { {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GE, ACC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC}, {    T_GT, ANC} },
/*  97 */   { {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {T_INTEGER,ANC}, {       0, REJ}, {      97, MOV}, {T_INTEGER,ANC}, {      87, MOV}, {       0, REJ}, {       0, REJ} },
/*  98 */   { {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {      92, MOV}, {      92, MOV}, {      92, MOV}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {  T_REAL, ANC}, {       0, REJ}, {      98, MOV}, {  T_REAL, ANC}, {      92, MOV}, {       0, REJ}, {       0, REJ} },
};
//state                 A                B                C                D                E                F                G                H                I                J                K                L                M                N                O                P                Q                R                S                T                U                V                W                X                Y                Z                $            digit               +                 -                *                /                ^                (                )                <                >                =                ,                ;             space           newline              .                "        anything else
//*  0 */   { {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ}, {       0, REJ} },

static bool in_data_statement=false; // if true, we are in a data statement and should use the secondary scanner with start state 84
                                     // if false, use the normal scanner with start state 0
static uint32_t current_line_number=UINT32_C(0); // current line number of the input Minimal BASIC program

// This is the definition for one node of the linked list that contains the token stream
// generated by this scanner and consumed by the parser.
struct token_element {
  struct token t;                // token
  struct token_element *next;    // pointer to next node
};
static struct token_element *token_list_head=NULL,  // pointer to head of token stream
                            *token_list_tail=NULL;  // pointer to tail of token stream

static void append_new_token(enum token_ids tidno,const char *txt,const uint32_t lno,const uint32_t cno);
static uint32_t binary_search(const struct tndata thelist[], const uint32_t element_count, const char *s);
static const struct token * peek_last_token(void);

//
// This function will duplcate the information in the next token in the input stream
// and create a new node with that information and return a pointer to it.  If an error
// occurs, abort() is called.  Do not call this with an empty token stream.  The caller
// is responsible for freeing the node when it is no longer needed.
//
struct token *clone_next_token(void) {
  struct token *temp=NULL;
  if (NULL==token_list_head) {
    fprintf(stderr,"clone_next_token() token list is empty\n");
    abort();
  }
  if ((temp=(struct token *)malloc(sizeof(struct token)))==NULL) {
    fprintf(stderr,"Error: Unable to allocate %" PRIu64 " bytes in clone_next_token\n",
      sizeof(struct token));
    abort();
  }
  temp->tid=token_list_head->t.tid;
  if ((temp->toketext=(char *)malloc(sizeof(char)*strlen(token_list_head->t.toketext)+UINT32_C(1)))==NULL) {
    fprintf(stderr,"Error: Unable to allocate %" PRIu64 " bytes #2 in clone_next_token\n",
      sizeof(char)*strlen(token_list_head->t.toketext)+UINT32_C(1));
    abort();
  }
  strncpy(temp->toketext,token_list_head->t.toketext,strlen(token_list_head->t.toketext)+UINT32_C(1));
  temp->toketext[strlen(token_list_head->t.toketext)]=0;
  temp->lno=token_list_head->t.lno;
  temp->cno=token_list_head->t.cno;
  return temp;
}

//
// This function will return a reference to the head token in the input stream.
// This will change each time a token is consumed.  If no token stream exists,
// abort() is called.
//
struct token peek_next_token(void) {
  if (token_list_head!=NULL)
    return token_list_head->t;
  fprintf(stderr,"peek_next_token() token list is empty\n");
  abort();
}

//
// This function will return a reference to the tail token in the input stream,
// or NULL if the token stream is empty.
//
static const struct token * peek_last_token(void) {
  if (token_list_tail!=NULL)
    return &token_list_tail->t;
  debug_printf("peek_last_token() token list is empty\n");
  return NULL;
}

//
// This is a binary search routine used to lookup the parameter s in the array
// specified by the thelist with the array element count specified by element_count.
// If s is not found, UINT32_MAX is returned to indicate error.
//
static uint32_t binary_search(const struct tndata thelist[], const uint32_t element_count, const char *s) {
  uint32_t low=UINT32_C(0),high=element_count-UINT32_C(1),middle=UINT32_C(0);
  while (low<=high) {
    int32_t comparison_result;
    middle=low+(high-low)/UINT32_C(2);
    comparison_result=strncmp(thelist[middle].namestring,s,30);
    if (INT32_C(0)==comparison_result)
       return middle;
    if (comparison_result<INT32_C(0)) {
       low=middle+UINT32_C(1);
    } else {
       if (UINT32_C(0)==middle) break;
       high=middle-UINT32_C(1);
    }
  }
  return UINT32_MAX;
}

//
// This procedure will add a token to the token stream.
// tidno is the token id number
// txt is the ASCIIZ token text
// lno is the source line number
// cno is the source column number
// If an error occurs, abort() is called.
//
static void append_new_token(enum token_ids tidno,const char *txt,const uint32_t lno, const uint32_t cno) {
  uint32_t tlen; // length of token text
  struct token_element *temp=NULL; // pointer to a temporary token node

  if (NULL==txt) { // ensure token text is not empty
    fprintf(stderr,"Error: NULL second argument passed to append_new_token on line %" PRIu32 "\n",lno);
    abort();
  }
  // allocate a new token node
  if ((temp=(struct token_element *)malloc(sizeof(struct token_element)))==NULL) {
    fprintf(stderr,"Error: unable to allocate %" PRIu64 " bytes in append_new_token\n",sizeof(struct token_element));
    abort();
  }
  debug_printf("append_new_token(%s,'%s',%" PRIu32 ",%" PRIu32 ")\n",token_names[tidno].namestring,txt,lno,cno);
  // fill the new token node with data that was passed into this procedure
  tlen=(uint32_t)strlen(txt)+UINT32_C(1);
  temp->t.tid=tidno;
  if ((temp->t.toketext=(char *)malloc(sizeof(char)*tlen))==NULL) {
    fprintf(stderr,"Error: unable to allocate %" PRIu32 " bytes in append_new_token\n",tlen);
    abort();
  }
  strcpy(temp->t.toketext,txt);
  temp->t.lno=lno;
  temp->t.cno=cno;
  temp->next=NULL;
  // append the node to the list
  if (NULL==token_list_head) { // if this is the very first node
    token_list_head=token_list_tail=temp; // head and tail point to new node
  } else { // add to the end of the list
    token_list_tail->next=temp; // tail node points to new node
    token_list_tail=temp;  // tail updated to point to new node
  }
  return;
}

//
// take token from front of list and return it to the caller
//
struct token get_next_token(void) {
  if (token_list_head!=NULL) {
    struct token_element *temp=token_list_head; // pointer to existing head of list
    struct token t=temp->t; // pointer to first node in the list
    token_list_head=token_list_head->next; // move list head
    free(temp);temp=NULL; // free head node
    if (NULL==token_list_head) // if the list is now empty
       token_list_tail=NULL;   // set tail to NULL too
    return t; // return copy of node to caller
  }
  fprintf(stderr,"get_next_token() token list is empty\n");
  abort();
}

//
// delete all remaining tokens
//
void clear_token_stream(void) {
  while (token_list_head!=NULL) { // while there are more tokens
    if (verbose) printf("freeing '%s'\n",token_list_head->t.toketext);
    free(token_list_head->t.toketext); // free the text
    token_list_head->t.toketext=NULL; // NULL it for safety
    get_next_token(); // move to next element
  }
  token_list_tail=token_list_head=NULL; // set head and TAIL of now-empty list to NULL
  return;
}

//
// This proecuedure is used to print a diagnostic display 
// showing the token stream contents.  It is read-only and
// does not change the token stream in any way.
//
void dump_token_stream(void) {
  struct token_element *current=token_list_head; // current token pointer, initialized to the head of the list

  while (current!=NULL) { // while there are more tokens
    // print the current token data
    printf("token id is %2" PRIu32 ", token name is '%s', text is '%s', line is %" PRIu32 ", column is %" PRIu32 "\n",
           current->t.tid,
           token_names[current->t.tid].namestring,
           current->t.toketext,
           current->t.lno,
           current->t.cno);
    current=current->next; // move to the next token (if one exists)
  }
  return;
}

//
// This is the external entry point to the scanner.  The buffer parameter is an ASCIIZ buffer
// containing the entire program to scan, including newlines (UNIX 0x0A style).
//
bool scan_buffer(const char *buffer) {
  uint64_t position=UINT64_C(0); // current position for the next byte of the input buffer to be processed
  uint32_t curline=UINT32_C(1); // this is for error messages; this counts from 1 for each line of input
                                // and is used for 'error on nth line' messages like missing line number
  uint32_t column=UINT32_C(1);  // this is the current column we are processing; if it exceeds MAXCOLUMN, it is an error
                                // if it is MAXCOLUMN, we must be at T_EOL
  uint32_t tl;                  // length of curtext before any T_QSTRING adjustment
  bool at_start_of_line=true;   // true if current position is the first column, false otherwise
  uint64_t buflen;              // length of input buffer
  char curtext[TBUFLEN]="";     // ASCIIZ string buffer where current token text will be built byte-by-byte
  int32_t curtextpos=INT32_C(-1); // current index in curtext buffer for input bytes
  int32_t curstate=INT32_C(0);  // current state in deterministic finite machine
  int32_t oldstate=INT32_C(0);  // previous state (used in diagnostic message)
  bool token_is_a_keyword=false; // if true, the token is a Minimal BASIC keyword, false otherwise
  bool skipped_a_trailing_space=false; // if true, a trailing space was skipped, false otherwise
  static char why[TBUFLEN];     // error message buffer

  if ((NULL==buffer)||(0==buffer[0])) { // if the input buffer is empty, error
    return false;
  }
  buflen=strlen(buffer);   // save the input buffer length
  in_data_statement=false; // initially we are not in a DATA statement
  while (position<buflen) { // while more input exists in the input buffer
    int cc=character_class[(uint32_t)buffer[position]]; // character class id
    struct fsm_element temp2=FSM_TABLE[curstate][cc]; // next element of FSM_TABLE to process

    if (verbose) {
      debug_printf("at top of loop position is %" PRIu64 ", state is %d, char is '%c'/%d, cc is %d\n",position,curstate,buffer[position],buffer[position],cc);
      debug_printf("current column is %" PRIu32 "\n",column);
      debug_printf("temp2 action is %s\n",action_names[temp2.what2do]);
      debug_printf("temp2 next_state is %d\n",temp2.next_state);
    }
    if (column>MAXCOLUMN) { // if the input line is too long
      fprintf(stderr,"ERROR on line %" PRIu32 ": input line too long; maximum permitted line length is %" PRIu32 " including newline\n",
         current_line_number,MAXCOLUMN);
      return false;
    }
    switch (temp2.what2do) { // consider the action to take
      case REJ: // reject
        fprintf(stderr,"reject on %" PRIu32 "%s column of %" PRIu32 "%s line\n",
          column,suffix(column),curline,suffix(curline));
        if (current_line_number>UINT32_C(0))
          fprintf(stderr,"This error is on or near the BASIC input program line number %" PRIu32 "\n",current_line_number);
        else
          fprintf(stderr,"This error is on the very first line of the BASIC input program and\na line number is required.\n");
        if ('\n'==buffer[position])
          fprintf(stderr,"character is a newline (%d)\n",buffer[position]);
        else
          fprintf(stderr,"character is '%c' (%d)\n",buffer[position],buffer[position]);
        if (islower(buffer[position]))
          fprintf(stderr,"Minimal BASIC does not permit lower-case letters.\n");
        if (in_data_statement && (' '==buffer[position-1]) && 
            ((','==buffer[position])||('\n'==buffer[position])))
          fprintf(stderr,"ECMA-55 does not permit leading or trailing spaces in an unquoted string\nin a DATA statement.\n");
        return false;
      case ACC: // accept and consume current input byte
        debug_printf("consume and ");
        curtext[++curtextpos]=buffer[position++]; // consume input byte
        column++; // increment column position
        // special check if this is a T_LPAREN and the previous token was T_NVAR
        // If it was then change that previous token to a T_NAVAR
        if ((token_list_tail!=NULL) && // if we are not the first token
            (T_NVAR==token_list_tail->t.tid) && // and previous token was T_NVAR
            (T_LPAREN==temp2.next_state) && // and this token is T_LPAREN
            (0==token_list_tail->t.toketext[1])) { // and the T_NVAR name was only 1 byte
           token_list_tail->t.tid=T_NAVAR; // convert previous token from T_NVAR to T_NAVAR
        }
        // intentionally fall through
      case ANC: // accept, but do not consume current input byte
        debug_printf("accept\n");
        curtext[++curtextpos]=0;
        // fix for things like 1+2 counting as T_INTEGER(1) T_INTEGER(+2)
        // when it should be T_INTEGER(1) T_ADD T_INTEGER(2)
        // I do this by looking backwards at the last token before this
        // (if there is one)
        if (('+'==curtext[0])||('-'==curtext[0])) {
          if ((T_REAL==temp2.next_state)||(T_INTEGER==temp2.next_state)) {
            const struct token *t=peek_last_token();
            if (t!=NULL) {
              if ((T_RND==t->tid)||       // special strange case for RND function that has no args
                  (T_FNID==t->tid)||      // special strange case for UDF with no args
                  (T_REAL==t->tid)||
                  (T_INTEGER==t->tid)||
                  (T_RPAREN==t->tid)||
                  (T_NVAR==t->tid)) {
                uint32_t i,oldlen=(uint32_t)strlen(curtext);

                // convert to discrete +/- and separate T_REAL/T_INTGER here
                if ('+'==curtext[0]) {
                  append_new_token(T_ADD,"+",current_line_number,column-oldlen);
                } else {
                  append_new_token(T_SUBTRACT,"-",current_line_number,column-oldlen);
                }
                for (i=UINT32_C(0);i<oldlen-UINT32_C(1);i++)
                  curtext[i]=curtext[i+1];
                curtext[oldlen-UINT32_C(1)]=0;
                debug_printf("HAD TO FIXUP\n");
              }
            }
          }
        }
        if (at_start_of_line) {
          if (temp2.next_state!=T_INTEGER) {
            fprintf(stderr,"ERROR: missing line number on %" PRIu32 "%s line of input.\n",curline,suffix(curline));
            return false;
          } else {
            uint32_t lno=UINT32_C(0);
            if (!convert_line_number(curtext,&lno,why)) {
              fputs(why,stderr);
              return false;
            }
            current_line_number=lno;
          }
        }
        // activate inclusion of any optional code that gets used
        switch(temp2.next_state) {
          case T_READ:
            need_READ=true; // must generate READ/DATA/RESTORE statement support in runtime code
            break;
          case T_INPUT:
            need_INPUT=true; // must generate INPUT statement support in runtime code
            break;
          case T_RND:
          case T_RANDOMIZE:
            need_RND=true; // must generate RANDOMIZE statement and RND function support in runtime code
            break;
          case T_SQR:
            need_SQR=true; // must generate SQR() function support in runtime code
            break;
          case T_SIN:
            need_SIN=true; // must generate SIN() function support in runtime code
            break;
          case T_COS:
            need_COS=true; // must generate COS() function support in runtime code
            break;
          case T_TAN:
            need_TAN=true; // must generate TAN() function support in runtime code
            break;
          case T_ATN:
            need_ATN=true; // must generate ATN() function support in runtime code
            break;
          case T_LOG:
            need_LOG=true; // must generate LOG() function support in runtime code
            break;
          case T_EXP:
            need_EXP=true; // must generate EXP() function support in runtime code
            break;
          case T_POW:
            need_POW=true; // must generate support for '^' exponentation operator in runtime code
            break;
          case T_PRINT:
            need_PRINT=true; // must generate support for PRINT statement and TAB() function in runtime code
            break;
          case T_DATA:
            need_strcpy=true; // must include a string copy function in the runtime code
          default:
            ; // do nothing and fall through
        }
        if (T_DATA==temp2.next_state)
          in_data_statement=true;      // set flag for scanning in a DATA statement to switch to alternate scanner
        if (in_data_statement && (T_EOL==temp2.next_state))
          in_data_statement=false;     // clear flag for scanning in a DATA statement to switch back to normal scanner
        // Is the current token a keyword?
        token_is_a_keyword=token_names[binary_search(token_names,numtokens,token_names[temp2.next_state].namestring)].iskeyword;
        // special check for space (or newline) after keywords (except REM)
        skipped_a_trailing_space=false;
        if (token_is_a_keyword) {
          debug_printf("position=%" PRIu32 ",curtextpos=%d\nspot to check is buffer[%d]='%c'\n",
              position,curtextpos,position-curtextpos,buffer[position-curtextpos]);
          if ((UINT32_C(0)==position-curtextpos)||(buffer[position-curtextpos-1]!=' ')) {
              fprintf(stderr,"ERROR on line %" PRIu32 ": space or newline is required before any keyword, including '%s'\n",
                      current_line_number,curtext);
              return false;
          }
          if (temp2.next_state!=T_REM) {
            if ((buffer[position]!=0x20)&&(buffer[position]!=0x0a)) {
              fprintf(stderr,"ERROR on line %" PRIu32 ": space or newline is required after any keyword, including '%s'\n",
                      current_line_number,curtext);
              return false;
            }
            if (' '==buffer[position]) {
              position++;       // skip over the required space
              column++;
              skipped_a_trailing_space=true;
            }
          }
        }
        // remember length before any tweaks
        tl=(uint32_t)strlen(curtext);
        // tweakage to remove the surrounding double quotes from string literals
        if (T_QSTRING==temp2.next_state) {
          memmove(curtext,curtext+1,tl);
          curtext[tl-2]=0;
        }
        // ok, do the accept
        if (token_is_a_keyword && skipped_a_trailing_space) {
          append_new_token(temp2.next_state,
                           curtext,
                           current_line_number,
                           column-tl-1); // subtract one extra for that trailing space
        } else {
          if (curstate==94) { // unquoted string
            debug_printf("curstate is %d FIXUP IN PROGRESS: curtext before is '%s'\n",curstate,curtext);
            uint32_t tlen=(uint32_t)strlen(curtext);
            while (' '==curtext[tlen-1])
              curtext[--tlen]=(char)0;
            debug_printf("curstate is %d FIXUP IN PROGRESS: curtext after is '%s'\n",curstate,curtext);
          } else {
            debug_printf("curstate is %d NO FIXUP IN PROGRESS: curtext is '%s'\n",curstate,curtext);
          }
          append_new_token(temp2.next_state,
                           curtext,
                           current_line_number,
                           column-tl);
        }
        if (T_EOL==temp2.next_state) { // if we are at the end of an input line
          at_start_of_line=true; // we are at start of line
          curline++; // increment current line number
          column=UINT32_C(1); // set column position to first column
        } else {
          at_start_of_line=false; // we are NOT at start of line
        }
        curtextpos=-1; // reset pointer in token text buffer to signal empty buffer
        if (in_data_statement) // if inside a DATA statement
          curstate=84;    // switch to special DATA statement DFSM
        else
          curstate=0;     // switch to normal DFSM
        break;
      case MOV: // move to next state
        debug_printf("move\n");
        // skip leading spaces on new tokens per standard
        if (((0==curstate)||(84==curstate)) && (' '==buffer[position])) {
          debug_printf("skip a leading space in state %" PRIu32 "\n",curstate);
          position++;
        } else {
          debug_printf("append a byte to curtext\n");
          curtext[++curtextpos]=buffer[position++];
        }
        column++; // move to next column position
        oldstate=curstate; // save current state
        curstate=temp2.next_state; // transition to next state
        debug_printf("moved from state %d to state %d\n",oldstate,curstate);
        break;
      default:
        abort();
        break;
    }
  }
  curtext[++curtextpos]=0; // terminate token text
  debug_printf("curstate='%d',curtext='%s'\n",curstate,curtext);
  if (curtext[0]!=0) { // if there is a current token that has not been accepted
    // should be impossible since every line must have a newline at the end, but....
    debug_printf("dribblings are '%s'\n",curtext);
    if (is_final[curstate].can_accept) {
      if (94==curstate) { // unquoted string
        debug_printf("curstate is %d FIXUP IN PROGRESS: curtext before is '%s'\n",curstate,curtext);
        int tlen=(uint32_t)strlen(curtext);
        while (' '==curtext[tlen-1])
          curtext[--tlen]=(char)0;
        debug_printf("curstate is %d FIXUP IN PROGRESS: curtext after is '%s'\n",curstate,curtext);
      } else {
        debug_printf("curstate is %d NO FIXUP IN PROGRESS: curtext is '%s'\n",curstate,curtext);
      }
      append_new_token(is_final[curstate].accept_state,
                       curtext,
                       current_line_number,
                       column-(uint32_t)strlen(curtext));
    }
  }
  // append final synthetic EOF token to list
  append_new_token(T_EOF,"EOF",current_line_number+UINT32_C(1),UINT64_C(0));
  return true;
}

#ifdef UNIT_TEST
static char *cookit(const char *s) {
  static char cookit_buffer[120];
  uint32_t i,j;

  for (i=j=UINT32_C(0);s[i]!=0;i++) {
    if ('\n'==s[i]) {
      cookit_buffer[j++]='\\';
      cookit_buffer[j++]='n';
    } else {
      cookit_buffer[j++]=s[i];
    }
  }
  cookit_buffer[j]=0; // terminate it
  return cookit_buffer;
}

static const struct tt {
    const char *s;
    const uint32_t howmany;
    const struct token t[60];       // yeah, it wastes space
} tests[]={
   {"10 DATA \"\"\n",
     UINT32_C(4),
     {{T_INTEGER,"10",10,1},{T_DATA,"DATA",10,4},{T_QSTRING,"",10,9},{T_EOL,"\n",10,11}}
   },
   {"10 DATA 123\n",
     UINT32_C(4),
     {{T_INTEGER,"10",10,1},{T_DATA,"DATA",10,4},{T_INTEGER,"123",10,9},{T_EOL,"\n",10,12}}
   },
   {"10 DATA 123, 456,  789\n",
     UINT32_C(8),
     {{T_INTEGER,"10",10,1}, {T_DATA,"DATA",10,4}, {T_INTEGER,"123",10,9}, {T_COMMA,",",10,12}, 
      {T_INTEGER,"456",10,14}, {T_COMMA,",",10,17}, {T_INTEGER,"789",10,20}, {T_EOL,"\n",10,23}}
   },
   {"10 PRINT -0.31415926535E-1+2\n",
    UINT32_C(6),
    {{T_INTEGER,"10",10,1}, {T_PRINT,"PRINT",10,4}, {T_REAL,"-0.31415926535E-1",10,10}, 
     {T_ADD,"+",10,27},{T_INTEGER,"2",10,28}, {T_EOL,"\n",10,29}}
   },
   {"10 DATA 1,0.1,.1,-1,-.1,-0.001E+3\n",
    UINT32_C(14),
    {{T_INTEGER,"10",10,1}, {T_DATA,"DATA",10,4}, {T_INTEGER,"1",10,9}, {T_COMMA,",",10,10},
     {T_REAL,"0.1",10,11}, {T_COMMA,",",10,14}, {T_REAL,".1",10,15}, {T_COMMA,",",10,17},
     {T_INTEGER,"-1",10,18}, {T_COMMA,",",10,20}, {T_REAL,"-.1",10,21}, {T_COMMA,",",10,24},
     {T_REAL,"-0.001E+3",10,25}, {T_EOL,"\n",10,34}}
   },
   {"10 PRINT A(1),A(1,2)\n",
    UINT32_C(14),
    {{T_INTEGER,"10",10,1}, {T_PRINT,"PRINT",10,4}, {T_NAVAR,"A",10,10}, {T_LPAREN,"(",10,11},
     {T_INTEGER,"1",10,12}, {T_RPAREN,")",10,13}, {T_COMMA,",",10,14}, {T_NAVAR,"A",10,15},
     {T_LPAREN,"(",10,16}, {T_INTEGER,"1",10,17}, {T_COMMA,",",10,18}, {T_INTEGER,"2",10,19},
     {T_RPAREN,")",10,20}, {T_EOL,"\n",10,21}}
   },
   {"10 LET A(1)=1\n",
    UINT32_C(9),
    {{T_INTEGER,"10",10,1}, {T_LET,"LET",10,4}, {T_NAVAR,"A",10,8}, {T_LPAREN,"(",10,9},
     {T_INTEGER,"1",10,10}, {T_RPAREN,")",10,11}, {T_EQ,"=",10,12}, {T_INTEGER,"1",10,13},
     {T_EOL,"\n",10,14}}
   },
   {"10 PRINT A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z\n",
     UINT32_C(54),
     {{T_INTEGER,"10",10,1}, {T_PRINT,"PRINT",10,4}, {T_NVAR,"A",10,10}, {T_COMMA,",",10,11},
      {T_NVAR,"B",10,12}, {T_COMMA,",",10,13}, {T_NVAR,"C",10,14}, {T_COMMA,",",10,15}, {T_NVAR,"D",10,16},
      {T_COMMA,",",10,17}, {T_NVAR,"E",10,18}, {T_COMMA,",",10,19}, {T_NVAR,"F",10,20}, {T_COMMA,",",10,21},
      {T_NVAR,"G",10,22}, {T_COMMA,",",10,23}, {T_NVAR,"H",10,24}, {T_COMMA,",",10,25}, {T_NVAR,"I",10,26},
      {T_COMMA,",",10,27}, {T_NVAR,"J",10,28}, {T_COMMA,",",10,29}, {T_NVAR,"K",10,30}, {T_COMMA,",",10,31},
      {T_NVAR,"L",10,32}, {T_COMMA,",",10,33}, {T_NVAR,"M",10,34}, {T_COMMA,",",10,35}, {T_NVAR,"N",10,36},
      {T_COMMA,",",10,37}, {T_NVAR,"O",10,38}, {T_COMMA,",",10,39}, {T_NVAR,"P",10,40}, {T_COMMA,",",10,41},
      {T_NVAR,"Q",10,42}, {T_COMMA,",",10,43}, {T_NVAR,"R",10,44}, {T_COMMA,",",10,45}, {T_NVAR,"S",10,46},
      {T_COMMA,",",10,47}, {T_NVAR,"T",10,48}, {T_COMMA,",",10,49}, {T_NVAR,"U",10,50}, {T_COMMA,",",10,51},
      {T_NVAR,"V",10,52}, {T_COMMA,",",10,53}, {T_NVAR,"W",10,54}, {T_COMMA,",",10,55}, {T_NVAR,"X",10,56},
      {T_COMMA,",",10,57}, {T_NVAR,"Y",10,58}, {T_COMMA,",",10,59}, {T_NVAR,"Z",10,60}, {T_EOL,"\n",10,61}}
    },
   {"10 PRINT A9,B9,C9,D9,E9,F9,G9,H9,I9,J9,K9,L9,M9,N9,O9,P9,Q9,R9,S9,T9,U9\n",
     UINT32_C(44),
     {{T_INTEGER,"10",10,1}, {T_PRINT,"PRINT",10,4}, {T_NVAR,"A9",10,10}, {T_COMMA,",",10,12},
      {T_NVAR,"B9",10,13}, {T_COMMA,",",10,15}, {T_NVAR,"C9",10,16}, {T_COMMA,",",10,18}, {T_NVAR,"D9",10,19},
      {T_COMMA,",",10,21}, {T_NVAR,"E9",10,22}, {T_COMMA,",",10,24}, {T_NVAR,"F9",10,25}, {T_COMMA,",",10,27},
      {T_NVAR,"G9",10,28}, {T_COMMA,",",10,30}, {T_NVAR,"H9",10,31}, {T_COMMA,",",10,33}, {T_NVAR,"I9",10,34},
      {T_COMMA,",",10,36}, {T_NVAR,"J9",10,37}, {T_COMMA,",",10,39}, {T_NVAR,"K9",10,40}, {T_COMMA,",",10,42},
      {T_NVAR,"L9",10,43}, {T_COMMA,",",10,45}, {T_NVAR,"M9",10,46}, {T_COMMA,",",10,48}, {T_NVAR,"N9",10,49},
      {T_COMMA,",",10,51}, {T_NVAR,"O9",10,52}, {T_COMMA,",",10,54}, {T_NVAR,"P9",10,55}, {T_COMMA,",",10,57},
      {T_NVAR,"Q9",10,58}, {T_COMMA,",",10,60}, {T_NVAR,"R9",10,61}, {T_COMMA,",",10,63}, {T_NVAR,"S9",10,64},
      {T_COMMA,",",10,66}, {T_NVAR,"T9",10,67}, {T_COMMA,",",10,69}, {T_NVAR,"U9",10,70}, {T_EOL,"\n",10,72}}
    },
   {"10 PRINT A$,B$,C$,D$,E$,F$,G$,H$,I$,J$,K$,L$,M$,N$,O$,P$,Q$,R$,S$,T$,U$\n",
     UINT32_C(44),
     {{T_INTEGER,"10",10,1}, {T_PRINT,"PRINT",10,4}, {T_SVAR,"A$",10,10}, {T_COMMA,",",10,12},
      {T_SVAR,"B$",10,13}, {T_COMMA,",",10,15}, {T_SVAR,"C$",10,16}, {T_COMMA,",",10,18}, {T_SVAR,"D$",10,19},
      {T_COMMA,",",10,21}, {T_SVAR,"E$",10,22}, {T_COMMA,",",10,24}, {T_SVAR,"F$",10,25}, {T_COMMA,",",10,27},
      {T_SVAR,"G$",10,28}, {T_COMMA,",",10,30}, {T_SVAR,"H$",10,31}, {T_COMMA,",",10,33}, {T_SVAR,"I$",10,34},
      {T_COMMA,",",10,36}, {T_SVAR,"J$",10,37}, {T_COMMA,",",10,39}, {T_SVAR,"K$",10,40}, {T_COMMA,",",10,42},
      {T_SVAR,"L$",10,43}, {T_COMMA,",",10,45}, {T_SVAR,"M$",10,46}, {T_COMMA,",",10,48}, {T_SVAR,"N$",10,49},
      {T_COMMA,",",10,51}, {T_SVAR,"O$",10,52}, {T_COMMA,",",10,54}, {T_SVAR,"P$",10,55}, {T_COMMA,",",10,57},
      {T_SVAR,"Q$",10,58}, {T_COMMA,",",10,60}, {T_SVAR,"R$",10,61}, {T_COMMA,",",10,63}, {T_SVAR,"S$",10,64},
      {T_COMMA,",",10,66}, {T_SVAR,"T$",10,67}, {T_COMMA,",",10,69}, {T_SVAR,"U$",10,70}, {T_EOL,"\n",10,72}}
    },
   {"10 OPTION BASE 1\n",
     UINT32_C(5),
     {{T_INTEGER,"10",10,1},{T_OPTION,"OPTION",10,4},{T_BASE,"BASE",10,11},{T_INTEGER,"1",10,16},
      {T_EOL,"\n",10,17}}
    },
   {"10 PRINT (*/<>)+=12.34+12.34-12-12.34-+*-1.,\n",
     UINT32_C(22),
     {{T_INTEGER,"10",10,1}, {T_PRINT,"PRINT",10,4}, {T_LPAREN,"(",10,10}, {T_MUL,"*",10,11}, {T_DIV,"/",10,12},
      {T_NE,"<>",10,13}, {T_RPAREN,")",10,15}, {T_ADD,"+",10,16}, {T_EQ,"=",10,17}, {T_REAL,"12.34",10,18},
      {T_ADD,"+",10,23}, {T_REAL,"12.34",10,24}, {T_SUBTRACT,"-",10,29}, {T_INTEGER,"12",10,30},
      {T_SUBTRACT,"-",10,32}, {T_REAL,"12.34",10,33}, {T_SUBTRACT,"-",10,38}, {T_ADD,"+",10,39}, {T_MUL,"*",10,40},
      {T_REAL,"-1.",10,41}, {T_COMMA,",",10,44}, {T_EOL,"\n",10,45}}
    },
   {"10 DATA -.123,1E5,1.E5,.1E5,-.1E-5,+1.E+1,-+,-314.15926535E-2\n",
    UINT32_C(18),
    {{T_INTEGER,"10",10,1}, {T_DATA,"DATA",10,4}, {T_REAL,"-.123",10,9}, {T_COMMA,",",10,14}, {T_REAL,"1E5",10,15},
     {T_COMMA,",",10,18}, {T_REAL,"1.E5",10,19}, {T_COMMA,",",10,23}, {T_REAL,".1E5",10,24}, {T_COMMA,",",10,28},
     {T_REAL,"-.1E-5",10,29}, {T_COMMA,",",10,35}, {T_REAL,"+1.E+1",10,36}, {T_COMMA,",",10,42},
     {T_UQSTRING,"-+",10,43}, {T_COMMA,",",10,45}, {T_REAL,"-314.15926535E-2",10,46}, {T_EOL,"\n",10,62}}
    },
   {"999 FOR I=1 TO 10 STEP 3\n",
     UINT32_C(10),
     {{T_INTEGER,"999",999,1}, {T_FOR,"FOR",999,5}, {T_NVAR,"I",999,9}, {T_EQ,"=",999,10}, {T_INTEGER,"1",999,11},
      {T_TO,"TO",999,13}, {T_INTEGER,"10",999,16}, {T_STEP,"STEP",999,19}, {T_INTEGER,"3",999,24}, {T_EOL,"\n",999,25}}
    },
   {"10 LET I1 = +3.4E-3\n",
    UINT32_C(6),
    {{T_INTEGER,"10",10,1}, {T_LET,"LET",10,4}, {T_NVAR,"I1",10,8}, {T_EQ,"=",10,11}, {T_REAL,"+3.4E-3",10,13},
     {T_EOL,"\n",10,20}}
    },
   {"10 REM Comment with some ugly stuff\n",
    UINT32_C(3),
    {{T_INTEGER,"10",10,1},{T_REM,"REM Comment with some ugly stuff",10,4},{T_EOL,"\n",10,36}}
    },
   {"10 REM Comment with some ugly stuff\n20 STOP\n30 END\n",
    UINT32_C(9),
    {{T_INTEGER,"10",10,1}, {T_REM,"REM Comment with some ugly stuff",10,4}, {T_EOL,"\n",10,36},
     {T_INTEGER,"20",20,1}, {T_STOP,"STOP",20,4}, {T_EOL,"\n",20,8},
     {T_INTEGER,"30",30,1}, {T_END,"END",30,4}, {T_EOL,"\n",30,7}}
    },
   {"10 PRINT \"!;\";\n",
    UINT32_C(5),
    {{T_INTEGER,"10",10,1},{T_PRINT,"PRINT",10,4},{T_QSTRING,"!;",10,10},{T_SEMI,";",10,14},{T_EOL,"\n",10,15}}
    },
   {"20 PRINT \"this is a !@#$%^&*()_+-=[]{}|\"\n",
    UINT32_C(4),
    {{T_INTEGER,"20",20,1}, {T_PRINT,"PRINT",20,4}, {T_QSTRING,"this is a !@#$%^&*()_+-=[]{}|",20,10}, 
     {T_EOL,"\n",20,41}}
    },
   {"40 GOSUB 123\n",
    UINT32_C(4),
    {{T_INTEGER,"40",40,1}, {T_GOSUB,"GOSUB",40,4}, {T_INTEGER,"123",40,10}, {T_EOL,"\n",40,13}}
    },
   {"41 GO SUB 123\n",
    UINT32_C(5),
    {{T_INTEGER,"41",41,1}, {T_GO,"GO",41,4}, {T_SUB,"SUB",41,7}, {T_INTEGER,"123",41,11}, {T_EOL,"\n",41,14}}
    },
   {"50 GOTO 10\n",
    UINT32_C(4),
    {{T_INTEGER,"50",50,1}, {T_GOTO,"GOTO",50,4}, {T_INTEGER,"10",50,9}, {T_EOL,"\n",50,11}}
    },
   {"50 GO TO 10\n",
    UINT32_C(5),
    {{T_INTEGER,"50",50,1}, {T_GO,"GO",50,4}, {T_TO,"TO",50,7}, {T_INTEGER,"10",50,10}, {T_EOL,"\n",50,12}}
    },
   {"40 RETURN\n",
    UINT32_C(3),
    {{T_INTEGER,"40",40,1}, {T_RETURN,"RETURN",40,4}, {T_EOL,"\n",40,10}}
    },
   {"40 RESTORE\n",
    UINT32_C(3),
    {{T_INTEGER,"40",40,1}, {T_RESTORE,"RESTORE",40,4}, {T_EOL,"\n",40,11}}
    },
   {"40 RANDOMIZE\n",
    UINT32_C(3),
    {{T_INTEGER,"40",40,1}, {T_RANDOMIZE,"RANDOMIZE",40,4}, {T_EOL,"\n",40,13}}
    },
   {"40 END\n",
    UINT32_C(3),
    {{T_INTEGER,"40",40,1}, {T_END,"END",40,4}, {T_EOL,"\n",40,7}}
    },
   {"40 STOP\n",
    UINT32_C(3),
    {{T_INTEGER,"40",40,1}, {T_STOP,"STOP",40,4}, {T_EOL,"\n",40,8}}
    },
   {"40 PRINT\n",
    UINT32_C(3),
    {{T_INTEGER,"40",40,1}, {T_PRINT,"PRINT",40,4}, {T_EOL,"\n",40,9}}
    },
   {"40 PRINT ,,;\n",
    UINT32_C(6),
    {{T_INTEGER,"40",40,1}, {T_PRINT,"PRINT",40,4}, {T_COMMA, ",",40,10}, {T_COMMA, ",",40,11}, 
     {T_SEMI,";",40,12}, {T_EOL,"\n",40,13}}
    },
   {"341 DATA 1,500,PI,-21.,.255,1E10,5E-1,.4E+1,\"XYZ\",\"X - 3B2\",\"1E10\"\n",
    UINT32_C(24),
    {{T_INTEGER,"341",341,1}, {T_DATA,"DATA",341,5}, {T_INTEGER,"1",341,10}, {T_COMMA,",",341,11},
     {T_INTEGER,"500",341,12}, {T_COMMA,",",341,15}, {T_UQSTRING,"PI",341,16}, {T_COMMA,",",341,18},
     {T_REAL,"-21.",341,19}, {T_COMMA,",",341,23}, {T_REAL,".255",341,24}, {T_COMMA,",",341,28},
     {T_REAL,"1E10",341,29}, {T_COMMA,",",341,33}, {T_REAL,"5E-1",341,34}, {T_COMMA,",",341,38},
     {T_REAL,".4E+1",341,39}, {T_COMMA,",",341,44}, {T_QSTRING,"XYZ",341,45}, {T_COMMA,",",341,50},
     {T_QSTRING,"X - 3B2",341,51}, {T_COMMA,",",341,60}, {T_QSTRING,"1E10",341,61}, {T_EOL,"\n",341,67}}
    },
   {"50 LET A$ = \"hello\"\n",
    UINT32_C(6),
    {{T_INTEGER,"50",50,1},{T_LET,"LET",50,4},{T_SVAR,"A$",50,8},{T_EQ,"=",50,11},{T_QSTRING,"hello",50,13},
     {T_EOL,"\n",50,20}}
    },
   {"341 DATA 500,\"dog\",PI\n",
    UINT32_C(8),
    {{T_INTEGER,"341",341,1}, {T_DATA,"DATA",341,5}, {T_INTEGER,"500",341,10}, {T_COMMA,",",341,13},
     {T_QSTRING,"dog",341,14}, {T_COMMA,",",341,19}, {T_UQSTRING,"PI",341,20}, {T_EOL,"\n",341,22}}
    },
   {"99        IF      A$    <>    \"QUIT\"       THEN          200\n",
     UINT32_C(8),
    {{T_INTEGER,"99",99,1}, {T_IF,"IF",99,11}, {T_SVAR,"A$",99,19}, {T_NE,"<>",99,25},
     {T_QSTRING,"QUIT",99,31}, {T_THEN,"THEN",99,44}, {T_INTEGER,"200",99,58}, {T_EOL,"\n",99,61}}
    },
   {"10 LET I1 = -(-2+1)-3\n",
    UINT32_C(13),
    {{T_INTEGER,"10",10,1}, {T_LET,"LET",10,4}, {T_NVAR,"I1",10,8}, {T_EQ,"=",10,11}, {T_SUBTRACT,"-",10,13},
     {T_LPAREN,"(",10,14}, {T_INTEGER,"-2",10,15}, {T_ADD,"+",10,17}, {T_INTEGER,"1",10,18}, {T_RPAREN,")",10,19},
     {T_SUBTRACT,"-",10,20}, {T_INTEGER,"3",10,21}, {T_EOL,"\n",10,22}}
    },
   {"10 LET A   (   3   ) = 4\n",
    UINT32_C(9),
    {{T_INTEGER,"10",10,1}, {T_LET,"LET",10,4}, {T_NAVAR,"A",10,8}, {T_LPAREN,"(",10,12}, {T_INTEGER,"3",10,16},
     {T_RPAREN,")",10,20}, {T_EQ,"=",10,22}, {T_INTEGER,"4",10,24}, {T_EOL,"\n",10,25}}
   },
};
static const uint32_t numtests=sizeof(tests)/sizeof(struct tt);

int main(int argc, char **argv) {
  struct token curtoken;
  uint32_t ndex;
  bool quit_on_fail,outer_ok=true;
  const char bob[]="BOB'S YOUR UNCLE";

  if ((2==argc)&&(strcmp(argv[1],"-V")==0)) {
     verbose=true;
     quit_on_fail=true;
  } else {
     quit_on_fail=false;
     if ((2==argc)&&(strcmp(argv[1],"-v")==0))
       verbose=true;
     else
       verbose=false;
  }
  init_debug_printf(stderr);
  // positive tests
  for (ndex=UINT32_C(0);ndex<numtokens;ndex++) {
    printf("search for %*s'%s': ",(int)(20-strlen(token_names[ndex].namestring))," ",token_names[ndex].namestring);
    if (binary_search(token_names,numtokens,token_names[ndex].namestring)!=ndex) {
      printf("FAIL\n");
      fprintf(stderr,"Error in binary search routine while searching for '%s'\n", token_names[ndex].namestring);
      return EXIT_FAILURE;
    }
    printf("OK, found in slot %" PRIu32 " as expected.\n",ndex);
  }
  // negative test
  printf("search for %*s'%s': ",(int32_t)(20-strlen(bob))," ",bob);
  if (binary_search(token_names,numtokens,bob)!=UINT_MAX) {
    printf("FAIL\n");
    fprintf(stderr,"Error in binary search routine while searching for '%s'\n",bob);
    return EXIT_FAILURE;
  }
  printf("OK, not found as expected.\n");
  // test read-only token stream dump
  dump_token_stream();
  // remove any leftover tokens
  while (token_list_head!=NULL)
    get_next_token();
  for (ndex=UINT32_C(0);ndex<numtests;ndex++) {
    uint32_t whichone;
    char *c;
    char why[120];
    bool ok;

    ok=true;
    whichone=UINT32_C(0);
    why[0]=0;
    c=cookit(tests[ndex].s);
    printf("scanning %*s'%s'  :  ",(int)(80-strlen(c))," ",c);
    if (!scan_buffer(tests[ndex].s)) {
      ok=false;
      sprintf(why,"scan_buffer() failed\n");
    } else {
      if (verbose) {
        uint32_t i;
        fflush(stdout);
        fprintf(stderr,"Scan completed.\nBEGIN TOKEN DUMP\n");
        dump_token_stream();
        fprintf(stderr,"END TOKEN DUMP\n\n");
        fprintf(stderr,"BEGIN EXPECTED TOKEN STREAM\n");
        for (i=UINT32_C(0);i<tests[ndex].howmany;i++)
          fprintf(stderr,"token #%2" PRIu32 ":  (%" PRIu32 ")%s with text '%s' on line %" PRIu32 " in column %" PRIu32 "\n",
             i,
             tests[ndex].t[i].tid,
             token_names[tests[ndex].t[i].tid].namestring,
             tests[ndex].t[i].toketext,
             tests[ndex].t[i].lno,
             tests[ndex].t[i].cno);
        fprintf(stderr,"END EXPECTED TOKEN STREAM\n");
      }
      curtoken=get_next_token();
      do {
        if (verbose)
          printf("token id is %2" PRIu32 ", token name is '%s', text is '%s' on line %" PRIu32 " in column %" PRIu32 "\n",
             curtoken.tid,
             token_names[curtoken.tid].namestring,
             curtoken.toketext,
             curtoken.lno,
             curtoken.cno);
        if (curtoken.tid!=tests[ndex].t[whichone].tid) {
          ok=false;
          if (0==why[0])
             sprintf(why,"1. Token %" PRIu32 " : Expected (%" PRIu32 ")%s but got (%" PRIu32 ")%s",
                  whichone,
                  tests[ndex].t[whichone].tid,
                  token_names[tests[ndex].t[whichone].tid].namestring,
                  curtoken.tid,
                  token_names[curtoken.tid].namestring);
        }
        if (strcmp(curtoken.toketext,tests[ndex].t[whichone].toketext)!=0) {
          ok=false;
          if (0==why[0])
             sprintf(why,"2. Token %" PRIu32 ": Expected token text '%s', but got '%s'",
               whichone,tests[ndex].t[whichone].toketext,curtoken.toketext);
        }
        if (curtoken.lno!=tests[ndex].t[whichone].lno) {
          ok=false;
          if (0==why[0])
             sprintf(why,"3. Token %" PRIu32 ": Expected token lno %" PRIu32 " but got %" PRIu32,
               whichone,tests[ndex].t[whichone].lno,curtoken.lno);
        }
        if (curtoken.cno!=tests[ndex].t[whichone].cno) {
          ok=false;
          if (0==why[0])
             sprintf(why,"4. Token %" PRIu32 ": Expected token cno %" PRIu32 " but got %" PRIu32,
               whichone,tests[ndex].t[whichone].cno,curtoken.cno);
        }
        whichone++;
        if (verbose) fprintf(stderr,"freeing '%s'\n",token_list_head->t.toketext);
        free(curtoken.toketext);
        curtoken.toketext=NULL;
        curtoken=get_next_token();
      } while (ok && (whichone<tests[ndex].howmany) && (curtoken.tid!=T_EOF));
      if (curtoken.tid!=T_EOF) {
        ok=false;
        if (0==why[0])
           sprintf(why,"3. Expected T_EOF, got %s",token_names[curtoken.tid].namestring);
      }
      free(curtoken.toketext);
      curtoken.toketext=NULL;
      if (whichone!=tests[ndex].howmany) {
        ok=false;
        if (0==why[0])
           sprintf(why,"4. Expected %d tokens, but got %d tokens",tests[ndex].howmany,whichone);
      }
    }
    if (ok) {
      printf("OK\n");
    } else {
      printf("FAIL\n     %s\n",why);
      if (quit_on_fail)
         abort();
      outer_ok=false;
    }
    clear_token_stream();
  }
  return (outer_ok?EXIT_SUCCESS:EXIT_FAILURE);
}
#endif
