//
// symbol_table.c - This is the module implementing the symbol table
//                  for Mr. Ham's ECMA-55 Minimal BASIC compiler.
//
// Copyright (C) 2013,2014  John Gatewood Ham
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// version 2 of the Free Software Foundation License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
#define _POSIX_C_SOURCE 200809L
#include <features.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include "globals.h"
#include "symbol_table.h"
#include "g_fmt_BASIC_normal.h"

#define SMALL_BUFFER_MAX 81	// specify size for small ASCIIZ buffers (80 bytes + terminator)
#define SYMNAME_MAX_BYTES 21    // specify maximum length of an ASCIIZ symbol name (20 bytes + terminator)
#define LITVAL_MAX_BYTES 71     // specify maximum length of an ASCIIZ literal value (70 bytes + terminator)

// union used for containing a floating point number
union realnumber {
  float f;	// 32 bit float
  double d;     // 64 bit float
};

// union used to convert between a 32bit integer bit pattern and a 32bit floating point bit pattern
static union fconvert {
  uint32_t i;
  float f;
} nanny={.i=0x7fa00000};        // SNaN (signalling Not-a-Number)

// union used to convert between a 64bit integer bit pattern and a 64bit floating point bit pattern
static union fconvert64 {
  uint64_t i;
  double d;
} nanny64={.i=0x7ff0000000000001L};     // SNaN (signalling Not-a-Number)

// array used to convert type numbers to type names
static const char *data_item_type_names[]={"QSTR","NUM","UQSTR"};

//
// Table of records for all possible user-defined-functions.
// Initially all 26 rows are marked as having no argument and as
// not having been used by the program.
// WARNING: this array must be sorted by fname
//
static struct udf_item {
  const char * const fname;    // name of fuction is FN? where ? is one of 'A'..'Z'
  char argname;         // arguemnt name if there is an argument, one of 'A'..'Z'
                        // 0 means this UDF has no argument
  bool exists;          // true if UDF has been defined in the program
} udf_items[] = {
{"FNA",0,false}, {"FNB",0,false}, {"FNC",0,false}, {"FND",0,false},
{"FNE",0,false}, {"FNF",0,false}, {"FNG",0,false}, {"FNH",0,false},
{"FNI",0,false}, {"FNJ",0,false}, {"FNK",0,false}, {"FNL",0,false},
{"FNM",0,false}, {"FNN",0,false}, {"FNO",0,false}, {"FNP",0,false},
{"FNQ",0,false}, {"FNR",0,false}, {"FNS",0,false}, {"FNT",0,false},
{"FNU",0,false}, {"FNV",0,false}, {"FNW",0,false}, {"FNX",0,false},
{"FNY",0,false}, {"FNZ",0,false}
};

//
// Table of records for all possible floating point arrays.
// Initially all 26 rows are marked as having 0 dimensions, which means they
// have not been used by the BASIC program.
// WARNING: must be sorted by namestring
//
static struct arraydata thealist[] = {
{"A",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"B",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"C",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"D",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"E",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"F",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"G",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"H",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"I",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"J",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"K",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"L",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"M",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"N",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"O",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"P",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"Q",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"R",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"S",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"T",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"U",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"V",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"W",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"X",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"Y",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
{"Z",UINT32_C(0),UINT32_C(0),UINT32_C(0)},
};
static const uint32_t thealist_count=sizeof(thealist)/sizeof(struct arraydata); // number of rows in thealist

//
// DATA statements contain lists of values used by the READ statement.  The
// READ and RESTORE maintain a pointer to the next data item to read.  Each
// READ will advance the pointer, but a RESTORE can reset the pointer to the
// first value.  Since DATA statements often have duplicates, an array of
// pointers to data blocks is used.  If a value is used more than once in
// DATA statements in a program, there will be only one data block with
// the informatoion, but each time it is used a pointer to it will exist.
// The head of the list of the pointer names is data_list_head, and data_list
// is the record type used to hold the list of pointer names.  The actual
// data blocks are in the list of data_item records whose head is specified
// by data_items.
//
static struct data_list {
  char symbol_name[SYMNAME_MAX_BYTES]; // ASCIIZ string with name of symbol
  struct data_list *next; // pointer to next node
} *data_list_head=NULL, // head of singly-linked list of ?
  *data_list_tail=NULL; // tail of singly-linked list of ?
static uint32_t data_list_count=UINT32_C(0);

//
// DATA statements are lists of values that can be quoted strings, unquoted strings, or
// floating point numbers.  The values of all DATA statements are stored in a singly
// linked list, in order of appearance in the source program.  That list is composed
// of nodes specified by the data_item record type.  Each unique value in the list
// gets a data block, and this structure is used to hold the information necessary
// to create thaat data block.
//
static struct data_item {
  enum data_item_type t; // data type number
  union realnumber numba; // floating point value (if it exists)
  char *s;                // pointer to ASCIIZ string containing the value
  char symname[SYMNAME_MAX_BYTES]; // ASCIIZ name of the value
  struct data_item *next; // pointer to next node
} *data_items=NULL,     // head of singly-linked list of DATA statement items
  *last_data_item=NULL; // tail of singly-linked list of DATA statement items
static uint32_t data_item_count=UINT32_C(0); // number of nodes in list pointed to by data_items

//
// string literal values have names generated by the symbol table module.
// The names are stored in a linked list of nodes defined by the literal_item
// record type.
//
static struct literal_item {
  char symname[SYMNAME_MAX_BYTES]; // ASCIIZ string with name of literal value
  char litval[LITVAL_MAX_BYTES];   // ASCIIZ string with the actual literal value
  struct literal_item *next; // pointer to next node
} *literals=NULL,       // head of singly-linked list of string literal values
  *last_literal=NULL;   // tail of singly-linked list of string literal values

//
// Floating point literal values have names generated by the symbol table module.
// The names are stored in a linked list of nodes defined by the fliteral_item
// record type.
//
static struct fliteral_item {
  char symname[SYMNAME_MAX_BYTES]; // ASCIIZ string with name of symbol
  char litval[LITVAL_MAX_BYTES];   // ASCIIZ string with the actual literal value
  struct fliteral_item *next; // pointer to next node
} *fliterals=NULL,      // head of singly-linked list of floating point literal values
  *last_fliteral=NULL;  // tail of singly-linked list of floating point literal values

//
// FOR statements need hidden variables to contain the terminating bound and the increment value
// and a singly-linked list of those generated hidden variable names is created using nodes
// defined by the magicvar_item record type
//
struct magicvar_item {
  char symname[SYMNAME_MAX_BYTES]; // ASCIIZ string with name of symbol
  struct magicvar_item *next; // pointer to next node
};
static struct magicvar_item *magicvars=NULL; // head of singly-linked list of magic variables used by FOR statements

//
// This is a list of type codes for Minimal BASIC variable types
//   NVAR  - floating point scalar variable
//   NAVAR - floating point array variable
//   SVAR  - string scalar variable
//
enum vartype {NVAR=0,NAVAR=1,SVAR=2,UNDEFINED=3};
//
// This is a map to convert the vartype enum values to string names for display
//
static const char *vartype_names[]={ "NVAR","NAVAR","SVAR","UNDEFINED" };

//
// floating point scalar values can use variables names A..Z, A1..Z1, ... A9..Z9
// so an array of records is created with all of those possible names.  The
// records are of type stdata.  Initially all variables are marked as unused.
// Note that the names A..Z can be either scalar or numeric.  Their first use
// in the program determines the type, which cannot be changed after the variable
// is used.  The type can be set with a LET, an INPUT, a READ, with a DIM, etc.
//
// WARNING: must be sorted by namestring
static struct stdata {
  char * const namestring; // pointer to ASCIIZ string specifying the scalar variable name
  enum vartype tipe; // type of the variable (normally NVAR, but A..Z can be NAVAR)
  bool exists; // true if the variable has been used, false otherwise
} thelist[] = {
{"A",UNDEFINED,false}, {"A$",SVAR,false}, {"A0",NVAR,false}, {"A1",NVAR,false},
{"A2",NVAR,false}, {"A3",NVAR,false}, {"A4",NVAR,false}, {"A5",NVAR,false},
{"A6",NVAR,false}, {"A7",NVAR,false}, {"A8",NVAR,false}, {"A9",NVAR,false},
{"B",UNDEFINED,false}, {"B$",SVAR,false}, {"B0",NVAR,false}, {"B1",NVAR,false},
{"B2",NVAR,false}, {"B3",NVAR,false}, {"B4",NVAR,false}, {"B5",NVAR,false},
{"B6",NVAR,false}, {"B7",NVAR,false}, {"B8",NVAR,false}, {"B9",NVAR,false},
{"C",UNDEFINED,false}, {"C$",SVAR,false}, {"C0",NVAR,false}, {"C1",NVAR,false},
{"C2",NVAR,false}, {"C3",NVAR,false}, {"C4",NVAR,false}, {"C5",NVAR,false},
{"C6",NVAR,false}, {"C7",NVAR,false}, {"C8",NVAR,false}, {"C9",NVAR,false},
{"D",UNDEFINED,false}, {"D$",SVAR,false}, {"D0",NVAR,false}, {"D1",NVAR,false},
{"D2",NVAR,false}, {"D3",NVAR,false}, {"D4",NVAR,false}, {"D5",NVAR,false},
{"D6",NVAR,false}, {"D7",NVAR,false}, {"D8",NVAR,false}, {"D9",NVAR,false},
{"E",UNDEFINED,false}, {"E$",SVAR,false}, {"E0",NVAR,false}, {"E1",NVAR,false},
{"E2",NVAR,false}, {"E3",NVAR,false}, {"E4",NVAR,false}, {"E5",NVAR,false},
{"E6",NVAR,false}, {"E7",NVAR,false}, {"E8",NVAR,false}, {"E9",NVAR,false},
{"F",UNDEFINED,false}, {"F$",SVAR,false}, {"F0",NVAR,false}, {"F1",NVAR,false},
{"F2",NVAR,false}, {"F3",NVAR,false}, {"F4",NVAR,false}, {"F5",NVAR,false},
{"F6",NVAR,false}, {"F7",NVAR,false}, {"F8",NVAR,false}, {"F9",NVAR,false},
{"G",UNDEFINED,false}, {"G$",SVAR,false}, {"G0",NVAR,false}, {"G1",NVAR,false},
{"G2",NVAR,false}, {"G3",NVAR,false}, {"G4",NVAR,false}, {"G5",NVAR,false},
{"G6",NVAR,false}, {"G7",NVAR,false}, {"G8",NVAR,false}, {"G9",NVAR,false},
{"H",UNDEFINED,false}, {"H$",SVAR,false}, {"H0",NVAR,false}, {"H1",NVAR,false},
{"H2",NVAR,false}, {"H3",NVAR,false}, {"H4",NVAR,false}, {"H5",NVAR,false},
{"H6",NVAR,false}, {"H7",NVAR,false}, {"H8",NVAR,false}, {"H9",NVAR,false},
{"I",UNDEFINED,false}, {"I$",SVAR,false}, {"I0",NVAR,false}, {"I1",NVAR,false},
{"I2",NVAR,false}, {"I3",NVAR,false}, {"I4",NVAR,false}, {"I5",NVAR,false},
{"I6",NVAR,false}, {"I7",NVAR,false}, {"I8",NVAR,false}, {"I9",NVAR,false},
{"J",UNDEFINED,false}, {"J$",SVAR,false}, {"J0",NVAR,false}, {"J1",NVAR,false},
{"J2",NVAR,false}, {"J3",NVAR,false}, {"J4",NVAR,false}, {"J5",NVAR,false},
{"J6",NVAR,false}, {"J7",NVAR,false}, {"J8",NVAR,false}, {"J9",NVAR,false},
{"K",UNDEFINED,false}, {"K$",SVAR,false}, {"K0",NVAR,false}, {"K1",NVAR,false},
{"K2",NVAR,false}, {"K3",NVAR,false}, {"K4",NVAR,false}, {"K5",NVAR,false},
{"K6",NVAR,false}, {"K7",NVAR,false}, {"K8",NVAR,false}, {"K9",NVAR,false},
{"L",UNDEFINED,false}, {"L$",SVAR,false}, {"L0",NVAR,false}, {"L1",NVAR,false},
{"L2",NVAR,false}, {"L3",NVAR,false}, {"L4",NVAR,false}, {"L5",NVAR,false},
{"L6",NVAR,false}, {"L7",NVAR,false}, {"L8",NVAR,false}, {"L9",NVAR,false},
{"M",UNDEFINED,false}, {"M$",SVAR,false}, {"M0",NVAR,false}, {"M1",NVAR,false},
{"M2",NVAR,false}, {"M3",NVAR,false}, {"M4",NVAR,false}, {"M5",NVAR,false},
{"M6",NVAR,false}, {"M7",NVAR,false}, {"M8",NVAR,false}, {"M9",NVAR,false},
{"N",UNDEFINED,false}, {"N$",SVAR,false}, {"N0",NVAR,false}, {"N1",NVAR,false},
{"N2",NVAR,false}, {"N3",NVAR,false}, {"N4",NVAR,false}, {"N5",NVAR,false},
{"N6",NVAR,false}, {"N7",NVAR,false}, {"N8",NVAR,false}, {"N9",NVAR,false},
{"O",UNDEFINED,false}, {"O$",SVAR,false}, {"O0",NVAR,false}, {"O1",NVAR,false},
{"O2",NVAR,false}, {"O3",NVAR,false}, {"O4",NVAR,false}, {"O5",NVAR,false},
{"O6",NVAR,false}, {"O7",NVAR,false}, {"O8",NVAR,false}, {"O9",NVAR,false},
{"P",UNDEFINED,false}, {"P$",SVAR,false}, {"P0",NVAR,false}, {"P1",NVAR,false},
{"P2",NVAR,false}, {"P3",NVAR,false}, {"P4",NVAR,false}, {"P5",NVAR,false},
{"P6",NVAR,false}, {"P7",NVAR,false}, {"P8",NVAR,false}, {"P9",NVAR,false},
{"Q",UNDEFINED,false}, {"Q$",SVAR,false}, {"Q0",NVAR,false}, {"Q1",NVAR,false},
{"Q2",NVAR,false}, {"Q3",NVAR,false}, {"Q4",NVAR,false}, {"Q5",NVAR,false},
{"Q6",NVAR,false}, {"Q7",NVAR,false}, {"Q8",NVAR,false}, {"Q9",NVAR,false},
{"R",UNDEFINED,false}, {"R$",SVAR,false}, {"R0",NVAR,false}, {"R1",NVAR,false},
{"R2",NVAR,false}, {"R3",NVAR,false}, {"R4",NVAR,false}, {"R5",NVAR,false},
{"R6",NVAR,false}, {"R7",NVAR,false}, {"R8",NVAR,false}, {"R9",NVAR,false},
{"S",UNDEFINED,false}, {"S$",SVAR,false}, {"S0",NVAR,false}, {"S1",NVAR,false},
{"S2",NVAR,false}, {"S3",NVAR,false}, {"S4",NVAR,false}, {"S5",NVAR,false},
{"S6",NVAR,false}, {"S7",NVAR,false}, {"S8",NVAR,false}, {"S9",NVAR,false},
{"T",UNDEFINED,false}, {"T$",SVAR,false}, {"T0",NVAR,false}, {"T1",NVAR,false},
{"T2",NVAR,false}, {"T3",NVAR,false}, {"T4",NVAR,false}, {"T5",NVAR,false},
{"T6",NVAR,false}, {"T7",NVAR,false}, {"T8",NVAR,false}, {"T9",NVAR,false},
{"U",UNDEFINED,false}, {"U$",SVAR,false}, {"U0",NVAR,false}, {"U1",NVAR,false},
{"U2",NVAR,false}, {"U3",NVAR,false}, {"U4",NVAR,false}, {"U5",NVAR,false},
{"U6",NVAR,false}, {"U7",NVAR,false}, {"U8",NVAR,false}, {"U9",NVAR,false},
{"V",UNDEFINED,false}, {"V$",SVAR,false}, {"V0",NVAR,false}, {"V1",NVAR,false},
{"V2",NVAR,false}, {"V3",NVAR,false}, {"V4",NVAR,false}, {"V5",NVAR,false},
{"V6",NVAR,false}, {"V7",NVAR,false}, {"V8",NVAR,false}, {"V9",NVAR,false},
{"W",UNDEFINED,false}, {"W$",SVAR,false}, {"W0",NVAR,false}, {"W1",NVAR,false},
{"W2",NVAR,false}, {"W3",NVAR,false}, {"W4",NVAR,false}, {"W5",NVAR,false},
{"W6",NVAR,false}, {"W7",NVAR,false}, {"W8",NVAR,false}, {"W9",NVAR,false},
{"X",UNDEFINED,false}, {"X$",SVAR,false}, {"X0",NVAR,false}, {"X1",NVAR,false},
{"X2",NVAR,false}, {"X3",NVAR,false}, {"X4",NVAR,false}, {"X5",NVAR,false},
{"X6",NVAR,false}, {"X7",NVAR,false}, {"X8",NVAR,false}, {"X9",NVAR,false},
{"Y",UNDEFINED,false}, {"Y$",SVAR,false}, {"Y0",NVAR,false}, {"Y1",NVAR,false},
{"Y2",NVAR,false}, {"Y3",NVAR,false}, {"Y4",NVAR,false}, {"Y5",NVAR,false},
{"Y6",NVAR,false}, {"Y7",NVAR,false}, {"Y8",NVAR,false}, {"Y9",NVAR,false},
{"Z",UNDEFINED,false}, {"Z$",SVAR,false}, {"Z0",NVAR,false}, {"Z1",NVAR,false},
{"Z2",NVAR,false}, {"Z3",NVAR,false}, {"Z4",NVAR,false}, {"Z5",NVAR,false},
{"Z6",NVAR,false}, {"Z7",NVAR,false}, {"Z8",NVAR,false}, {"Z9",NVAR,false},
};
static const uint32_t thelist_count=sizeof(thelist)/sizeof(struct stdata); // number of rows in the thelist table

static uint32_t binary_search(const struct stdata alist[], const uint32_t element_count, const char *s);

//
// This function will use a binary search to look for the varaible
// specifed by the s parameter in the list specified by the alist
// parameter.  The number of elements in the alist is specified by
// the element_count parameter.  The alist must be sorted in ascending
// order.  Since the variable tables are not dynamic, they are sorted.
// If the variable name is found, the function will return the index of
// the row containing that variable.  If the variable is not found,
// UINT_MAX is returned to indicate failure.  The alist must not
// contain duplicate variable names (the key).
//
static uint32_t binary_search(const struct stdata alist[], // array to search
      const uint32_t element_count,                        // number of elements in alist
      const char *s) {                                     // value we are searching for
  uint32_t low=UINT32_C(0);                  // index of start of partition
  uint32_t high=element_count-UINT32_C(1);   // index of end of partition
  uint32_t middle=UINT32_C(0);               // index of midway point in partition

  // while the partition still contains elements keep looking
  while (low<=high) {
    int32_t comparison_result; // variable containing the result of
                               // comparing ASCIIZ variable names

    middle=low+(high-low)/UINT32_C(2); // compute the midway point
    comparison_result=strncmp(alist[middle].namestring,s,30); // compare ASCIIZ variable names
    if (INT32_C(0)==comparison_result) { // if we found s
       return middle;                    // return the index of the record
    }
    if (comparison_result<INT32_C(0)) {  // if the value in the midway point is too low
       low=middle+UINT32_C(1);           // look in the partition to the right of the midpoint
    } else {                             // otherwise the value was too high so
       if (UINT32_C(0)==middle) break;   // if the midpoint was the first element, the value is not in the list
       high=middle-UINT32_C(1);          // look in the partition to the left of the midway point
    }
  }
  return UINT_MAX; // s was not found, so return the error indicator value
}

//
// This function will emit the required variable and constant definitions
// for all variables in the symbol tables that are needed.  It will return
// true if it is successful, and false if an error occurs.  The f parameter
// is the handle for the output assembly language file to which the
// output must be appended.
// It could be argued that logically this belongs in codegen.c, but that
// would require sharing too much state, so it is done directly here.
// 
bool generate_symbol_table(FILE *f) {
  uint32_t i,j; // index variables for loops, etc.
  struct literal_item *si=literals; // pointer to current item in string literals list
  struct fliteral_item *fi=fliterals; // pointer to current item in floating point literals list
  struct magicvar_item *mvi=magicvars; // pointer to current item in magicvars list

  if (NULL==f) return false; // if the file handle is invalid, return failure
  if (UINT32_C(0)==thelist_count) return true; // if no variables were used, return success

  if (need_BSS_numeric) { // if we need to emit floating point variable definitions
    // emit the .bss assembly section header
    fprintf(f,"        .section .bss,\"aw\",@nobits\n\n");
    fprintf(f,"        .balign %c\n",(use_double?'8':'4'));
    fprintf(f,".LSTART_BSS:\n");
  
    // emit the definitions for scalar numeric variables
    for (i=UINT32_C(0);i<thelist_count;i++) { // for each possible floating point scalar variable
      if (thelist[i].exists) { // if it was used
        switch(thelist[i].tipe) {
          case NVAR: // and it was a scalar numeric variable
            // emit the definition, taking into account whether 32 or 64 bit floats are used
            fprintf(f,"        .balign %c\n",(use_double?'8':'4'));
            fprintf(f,"        .size %s, %c\n",thelist[i].namestring,(use_double?'8':'4'));
            fprintf(f,"%s:\n        .skip %c      # actually %s\n\n",thelist[i].namestring,
                      (use_double?'8':'4'),(use_double?"double":"float"));
            break;
          case NAVAR: // and it was a numeric array variable
          case SVAR:  // or it was a string variable
            // skip these
            break;
          default: // if it was of an unknown type (only possible with an ICE)
            fprintf(stderr,"ERROR: Unknown type when generating data segment\n");
            return false;
        }
      }
    }
    // emit the definitions for array numeric variables
    for (i=UINT32_C(0);i<thelist_count;i++) { // for each possible floating point array variable
      if (thelist[i].exists) { // if it was used
        if (NAVAR==thelist[i].tipe) { // and it was a numeric array variable
            uint32_t thecount1; // number of elements in first dimension
            uint32_t thecount2; // number of elements in second dimension
            uint32_t thesize;   // number of bytes require for entire array
  
            j=thelist[i].namestring[0]-'A'; // compute index in thealist table from variable name
            if (UINT32_C(1)==thealist[j].dims) { // if it has one dimension (vector)
              // emit definition for one-dimensional array, taking into account whether 32 or 64 bit floats are used
              thecount1=thealist[j].dim1+(base_is_one?UINT32_C(0):UINT32_C(1));
              thesize=(use_double?UINT32_C(8):UINT32_C(4))*thecount1;
              fprintf(f,"        # one-dimensional array\n");
              fprintf(f,"        .balign %c\n",(use_double?'8':'4'));
              fprintf(f,"        .size %s, %" PRIu32 "\n",
                        thelist[i].namestring, thesize);
              fprintf(f,"%s:\n", thelist[i].namestring);
              fprintf(f,"        .rept %" PRIu32 " # rows\n",thecount1);
              fprintf(f,"        .skip %c      # actually %s\n        .endr\n\n",
                        (use_double?'8':'4'), (use_double?"double":"float"));
            } else { // it has two dimensions (matrix)
              // emit definition for two-dimensional array, taking into account whether 32 or 64 bit floats are used
              thecount1=thealist[j].dim1+(base_is_one?UINT32_C(0):UINT32_C(1));
              thecount2=thealist[j].dim2+(base_is_one?UINT32_C(0):UINT32_C(1));
              thesize=(use_double?UINT32_C(8):UINT32_C(4))*thecount1*thecount2;
              fprintf(f,"        # two-dimensional array\n");
              fprintf(f,"        .balign %c\n",(use_double?'8':'4'));
              fprintf(f,"        .size %s, %" PRIu32 "\n",
                        thelist[i].namestring, thesize);
              fprintf(f,"%s:\n",thelist[i].namestring);
              fprintf(f,"        .rept %" PRIu32 " # rows\n",thecount1);
              fprintf(f,"          .rept %" PRIu32 " # columns\n",thecount2);
              fprintf(f,"          .skip %c         # actually %s\n", 
                        (use_double?'8':'4'), (use_double?"double":"float"));
              fprintf(f,"          .endr\n");
              fprintf(f,"        .endr\n\n");
            }
        }
      }
    }
    // emit the definitions for hidden internal magic variables
    while (mvi!=NULL) { // while there are any magic variables left in the list
      // emit the definition, taking into acount whether 32 or 64 bit floats are used
      mvi->symname[SYMNAME_MAX_BYTES-1]=0; // terminate ASCIIZ symbol name
      fprintf(f,"        .balign %c\n",(use_double?'8':'4'));
      fprintf(f,"        .size %s, %c\n",mvi->symname,(use_double?'8':'4'));
      fprintf(f,"%s:\n        .skip %c      # actually %s\n\n",mvi->symname,
                (use_double?'8':'4'),(use_double?"double":"float"));
      mvi=mvi->next;
    }
    fprintf(f,"        .set BSS_SIZE, .-.LSTART_BSS\n\n");
  }

  if (need_BSS_string) { // if we need to emit string variable definitions
    fprintf(f,"        .section .data,\"aw\",@progbits\n\n");
    fprintf(f,"        .balign 1\n");
    fprintf(f,".LSTART2_BSS:\n");
    for (i=UINT32_C(0);i<thelist_count;i++) { // for each possible string variable
      if (thelist[i].exists) { // if it was used
        if (SVAR==thelist[i].tipe) { // and was a string
            // emit the definition
            fprintf(f,"        .size %s,19\n",thelist[i].namestring);
            fprintf(f,"%s:\n",thelist[i].namestring);
            fprintf(f,"        .skip 19\n");
        }
      }
    }
    fprintf(f,"        .set BSS_SIZE2, .-.LSTART2_BSS\n\n");
  }

  if ((si!=NULL)||(fi!=NULL)) { // if we have any literal values
    // emit header for .rodata section
    fprintf(f,"        .section .rodata,\"a\",@progbits\n\n");
    // string literals
    while (si!=NULL) { // while we still have some string literals in the list
      // emit definition for string literal
      si->symname[SYMNAME_MAX_BYTES-1]=0;
      fprintf(f,"%s:\n        .asciz \"%s\"\n",si->symname,si->litval);
      fprintf(f,"        .size %s, .-%s\n\n",si->symname,si->symname);
      si=si->next;
    }
    // numeric literals
    while (fi!=NULL) { // while we still have some numeric literals in the list
     // emit definition for numeric (floating point) literal, taking into account whether 32 or 64 bit floats are used
      fi->symname[SYMNAME_MAX_BYTES-1]=0;
      if (use_double) // if we are doing 64bit math
        fprintf(f,"%s:\n        .double %s\n",fi->symname,fi->litval);
      else            // else we are doing 32bit math
        fprintf(f,"%s:\n        .float %s\n",fi->symname,fi->litval);
      fprintf(f,"        .size %s, .-%s\n\n",fi->symname,fi->symname);
      fi=fi->next;
    }
  }
  if (need_READ) { // if a READ statement was used in the Minimal BASIC program
    struct data_list *dli=data_list_head; // pointer to current ?
    struct data_item *dti=data_items;     // pointer to current ?

    // runtime data item pointer
    fprintf(f,"        .section .data,\"aw\",@progbits\n\n");
    fprintf(f,"        .type .Ldata_item_ptr, @object\n");
    fprintf(f,".Ldata_item_ptr:\n");
    fprintf(f,"        .quad 0\n");
    fprintf(f,"        .size .Ldata_item_ptr, .-.Ldata_item_ptr\n\n");
    // the number of data items in the list
    fprintf(f,"        .section .rodata,\"a\",@progbits\n\n");
    fprintf(f,"        .type .Ldata_item_count, @object\n");
    fprintf(f,".Ldata_item_count:\n");
    fprintf(f,"        .quad %" PRIu32 "\n",data_list_count);
    fprintf(f,"        .size .Ldata_item_count, .-.Ldata_item_count\n");
    if (data_item_count>UINT32_C(0)) { // if there were any items in DATA statements
      // first we emit the unique data blocks
      fprintf(f,"\n# DATA statement unique data blocks\n\n");
      while (dti!=NULL) { // while more data blocks need to be written
        char tbuf[SMALL_BUFFER_MAX]; // temporary buffer used by g_fmt() ASCIIZ to float converter

        // emit the data block
        fprintf(f,"%sS:       # string represenatation\n",dti->symname);
        fprintf(f,"        .asciz \"%s\"\n",dti->s);
        fprintf(f,"%s:\n",dti->symname);
        switch (dti->t) { // choose definition depending on the type of the data item
          case QSTR: // quoted string
            fprintf(f,"        .byte %d    # Type is: %s\n",QSTR,data_item_type_names[QSTR]);
            break;
          case NUM:  // floating point value
          case UQSTR: // unquoted string
            fprintf(f,"        .byte %d    # Type is: %s\n",UQSTR,data_item_type_names[UQSTR]);
            if (use_double) { // if we are doing 64bit math
              fprintf(f,"        .double ");
              g_fmt(tbuf,dti->numba.d,SIGNIFICANCE_WIDTH_WIDE_OUTPUT-UINT32_C(1),EXRAD_WIDE_OUTPUT);
            } else {          // else we are doing 32bit math
              fprintf(f,"        .float ");
              g_fmt(tbuf,dti->numba.f,SIGNIFICANCE_WIDTH_NARROW_OUTPUT-UINT32_C(1),EXRAD_NARROW_OUTPUT);
            }
            fprintf(f,"%s # numeric representation\n",tbuf);
            break;
          default: // unknown type (only possible with an ICE)
            abort();
        }
        fprintf(f,"        .byte %d    # length of ASCIIZ string pointed to by %sS including NULL byte\n",(int)(strlen(dti->s)+1),dti->symname);
        fprintf(f,"        .quad %sS   # pointer to string representation\n",dti->symname);
        fprintf(f,"        .size %s, .-%s\n",dti->symname,dti->symname);
        fprintf(f,"        .type %s, @object\n",dti->symname);
        dti=dti->next;
      }
  
      // then we emit the array of pointers to those blocks
      fprintf(f,"\n# List of pointers to all DATA statement items' data blocks\n\n");
      fprintf(f,"        .align 16\n");
      fprintf(f,"        .type .Ldata_items, @object\n");
      fprintf(f,".Ldata_items:\n");
      while (dli!=NULL) { // while more data item pointers need to be written
        // emit the data item pointer
        fprintf(f,"        .quad %s\n",dli->symbol_name);
        dli=dli->next;
      }
      fprintf(f,"        .size .Ldata_items, .-.Ldata_items\n");
    }
  }
  return true;
}

//
// This procedure deallocates the dynamic memory used by the symbol table module
//
void clear_symbol_table(void) {
  struct literal_item *si=literals;     // pointer to current string literal item
  struct fliteral_item *fi=fliterals;   // pointer to current floating point literal item
  struct data_item *dti=data_items;     // pointer to current unique data block item
  struct magicvar_item *mvi=magicvars;  // pointer to current magic variable item
  struct data_list *dli=data_list_head; // pointer to current data statement item pointer

  // deallocate string literal items
  while (si!=NULL) {
    struct literal_item *temp=si;
    si=si->next;
    free(temp);temp=NULL;
  }
  literals=last_literal=NULL;
  // deallocate floating point literal items
  while (fi!=NULL) {
    struct fliteral_item *temp=fi;
    fi=fi->next;
    free(temp);temp=NULL;
  }
  fliterals=last_fliteral=NULL;
  // deallocate unique data block items
  while (dti!=NULL) {
    struct data_item *temp=dti;
    dti=dti->next;
    free(temp->s);temp->s=NULL;
    free(temp);temp=NULL;
  }
  data_items=last_data_item=NULL;
  // deallocate magic variable items
  while (mvi!=NULL) {
    struct magicvar_item *temp=mvi;
    mvi=mvi->next;
    temp->next=NULL;
    free(temp);temp=NULL;
  }
  magicvars=NULL;
  // delete data statement item pointers
  while (dli!=NULL) {
    struct data_list *temp=dli;
    dli=dli->next;
    temp->next=NULL;
    free(temp);
    temp=NULL;
  }
  return;
}

//
// This procedure will dump the data stored in the symbol table module.
// If onlyused is true, then only variables that have been used in the 
// Minimal BASIC input program will be shown.  Otherwise, everyhing is
// shown.
//
void dump_symbol_table(bool onlyused) {
  uint32_t i,j;
  struct literal_item *si=literals;    // pointer to current string literal item
  struct fliteral_item *fi=fliterals;  // pointer to current floating point literal item
  struct magicvar_item *mvi=magicvars; // pointer to current magic variable item

  if (onlyused) { // if only used symbols should be displayed
    for (i=UINT32_C(0);i<thelist_count;i++) { // for each possible numeric variable name
      if (thelist[i].exists) { // if the variable was used
        // display what we know about it
        printf("name='%s', tipe=%s, exists=%s\n",
          thelist[i].namestring,
          vartype_names[thelist[i].tipe],
          (thelist[i].exists?"true":"false"));
        if (NAVAR==thelist[i].tipe) { // if it was a numeric array variable
          // show the dimension information
          j=thelist[i].namestring[0]-'A';
          printf("name='%s', dims=%" PRIu32 ", dim1=%" PRIu32 ", dim2=%" PRIu32 "\n",
            thealist[j].namestring,
            thealist[j].dims,
            thealist[j].dim1,
            thealist[j].dim2);
        }
      }
    }
  } else { // show all symbols
    for (i=UINT32_C(0);i<thelist_count;i++) { // for each possible numeric variable name
      // display what we know about it
      printf("name='%s', tipe=%s, exists=%s\n",
        thelist[i].namestring,
        vartype_names[thelist[i].tipe],
        (thelist[i].exists?"true":"false"));
    }
    for (i=UINT32_C(0);i<thealist_count;i++) { // for each possible numeric array variable name
      // display what we know about it
      printf("name='%s', dims=%" PRIu32 ", dim1=%" PRIu32 ", dim2=%" PRIu32 "\n",
        thealist[i].namestring,
        thealist[i].dims,
        thealist[i].dim1,
        thealist[i].dim2);
    }
  }
  while (mvi!=NULL) { // while more magic variable names are in the list
    // display the name
    printf("magic variable name='%s'\n",mvi->symname);
    mvi=mvi->next;
  }
  while (si!=NULL) { // while more string literal values are in the list
    // display the name and value
    printf("string literal name='%s', value='%s'\n",si->symname,si->litval);
    si=si->next;
  }
  while (fi!=NULL) { // while more numeric values are in the list
    // display the name and value
    printf("float literal name='%s', value='%s'\n",fi->symname,fi->litval);
    fi=fi->next;
  }

  if (need_READ) { // if any READ statement exists in the Minimal BASIC program
    struct data_item *dti=data_items;            // pointer to current data block
    struct data_list *dli=data_list_head;        // pointer to current pointer name in list of data pointers

    printf("%" PRIu32 " DATA ITEM COUNT\n",data_item_count);
    if (data_item_count>UINT32_C(0)) { // if any data blocks exist
      printf("DATA BLOCKS\n");
      while (dti!=NULL) { // while there are more data blocks
        char tbuf[SMALL_BUFFER_MAX]; // temporary buffer used by g_fmt() ASCIIZ to float function

        // display what we know about the data block
        printf("%sS is \"%s\"\n",dti->symname,dti->s);
        printf("%s is marked as a ",dti->symname);
        switch (dti->t) {
          case QSTR: // quoted string
            printf("quoted string\n");
            break;
          case NUM: // numeric value (floating point)
          case UQSTR: // unquoted string
            if (use_double) { // if we are using 64bit math
              g_fmt(tbuf,dti->numba.d,SIGNIFICANCE_WIDTH_WIDE_OUTPUT-UINT32_C(1),EXRAD_WIDE_OUTPUT);
            } else { // we are using 32bit math
              g_fmt(tbuf,dti->numba.f,SIGNIFICANCE_WIDTH_NARROW_OUTPUT-UINT32_C(1),EXRAD_NARROW_OUTPUT);
            }
            printf("an unquoted string with numeric type %s with stored value %s\n",
              (use_double?"double":"float"),tbuf);
            break;
          default:
            abort();
        }
        dti=dti->next;
      }
  
      printf("DATA POINTER LIST\n");
      while (dli!=NULL) { // while we have more data item pointers
        // display the data item pointer
        printf("%s\n",dli->symbol_name);
        dli=dli->next;
      }
    }
  }

  printf("\nUser-Defined Functions\n");
  for (i=UINT32_C(0);i<UINT32_C(26);i++) { // for each possible UDF
    if (udf_items[i].exists) { // if it was defined
      // display what we know about it
      printf("'%s' has been defined and has ",udf_items[i].fname);
      if (udf_items[i].argname!=0) { // if the UDF has an argument show the name
        printf("one argument named '%c'\n",udf_items[i].argname);
      } else { // otherwise indicate it has no argument
        printf("no arguments\n");
      }
    }
  }   

  return;
}

//
// This function is used to add a numeric literal (floating point) value to the
// symbol table.  The litval parameter has the ASCIIZ string containing the literal
// value.  If the value is added successfully, true is returned and the ASCIIZ symbol
// name generated is returned in the symbol parameter.  Otherwise false is returned.
//
bool add_float_literal(const char *litval, char **symbol) {
  static uint32_t last_symbol=UINT32_C(0); // number of next numeric literal (used to generate literal names)
  struct fliteral_item *fi=NULL;  // pointer to current floating point literal item

  if (NULL==litval) {
    fprintf(stderr,"ERROR: NULL pointer for source of floating point literal\n");
    abort();
  }
  if (NULL==symbol) {
    fprintf(stderr,"ERROR: Invalid pointer for generated name of floating point literal\n");
    abort();
  }
  // first lookup the literal value to see if we already have it.  If we do,
  // the symbol name of the existing literal value is returned.
  if (fliterals!=NULL) { // if there are any floating point literals
    fi=fliterals;
    while (fi!=NULL) { // while there are more floating point literals to consider
      if (strcmp(fi->litval,litval)==0) { // if the current floating point literal is the same is litval
        *symbol=fi->symname; // copy the symbol name into symbol
        return true;         // and return true
      }
      fi=fi->next;
    }
  }

  // The literal value was not found, so add it
  if ((fi=(struct fliteral_item *)malloc(sizeof(struct fliteral_item)))==NULL) { // if we are OOM
    fprintf(stderr,"ERROR: out of memory attempting to add float literal\n"); // report the error and crash
    abort();
  }
  fi->next=NULL;
  snprintf(fi->symname,SYMNAME_MAX_BYTES-1,".LFLIT%04" PRIu32,last_symbol++); // generate the ASCIIZ symbol name
  fi->symname[SYMNAME_MAX_BYTES-1]=0; // ensure termination
  *symbol=fi->symname; // copy the symbol name into symbol
  strncpy(fi->litval,litval,LITVAL_MAX_BYTES-1); // save the literal value
  fi->litval[LITVAL_MAX_BYTES-1]=0; // ensure termination
  if (NULL==fliterals) { // if this is the very first floating point literal
    fliterals=last_fliteral=fi; // point head and tail to this node
  } else { // otherwise
    last_fliteral->next=fi; // append the node to the end of the list of floating point literals
    last_fliteral=fi;
  }
  return true;
}

//
// This function is used to add UDF (user-defined function) information to the
// symbol table.  The fname parameter contains the UDF name, argname contains the
// name of the parameter if the UDF has a parameter, or NULL if the UDF has no
// parameter.  The lno parameter contains the Minimal BASIC program line number
// that is used in error messages if a problem occurs add the UDF information.
// If the UDF has already been defined or has an invalid name, false is returned.
// If everything was OK, true is returned.
//
bool add_udf(const char *fname,const char *argname,uint32_t lno) {
  bool retval=false; // return value for function
  uint32_t recno;   // index of row in udf_items where the record for the fname UDF is stored

  if (NULL==fname) {
    fprintf(stderr,"ERROR: NULL pointer for user-defined function name on line %" PRIu32 "\n",lno);
    abort();
  }
  if (0==fname[0]) {
    fprintf(stderr,"ERROR: user-defined function name cannot be an empty string on line %" PRIu32 "\n",lno);
    abort();
  }
  // ensure length is 3 bytes
  if (strlen(fname)!=3) { // invalid udf name (not 3 bytes) only possible with an ICE
    fprintf(stderr,"Invalid user-defined function name '%s' (not exactly 3 bytes) specified to add_udf()\n",fname);
    goto xit;
  }
  // ensure first two letters are 'FN'
  if ((fname[0]!='F')||(fname[1]!='N')) { // invalid udf name, does not start with FN
    fprintf(stderr,"Invalid user-defined function name '%s' (does not begin with 'FN') specified to add_udf()\n",fname);
    goto xit;
  }
  // ensure incoming ASCIIZ string variable name's third letter is valid
  if ((fname[2]<'A')||(fname[2]>'Z')) { // invalid variable name only possible with an ICE
    fprintf(stderr,"Invalid user-defined function name '%s' (third character is not a capital letter) specified to add_udf()\n",fname);
    goto xit;
  }
  recno=fname[2]-'A'; // compute index of record in udf_items for the UDF specified by fname
  if (recno>UINT32_C(25)) { // if the record is out-of-bounds (on possible if we have ICE)
                            // report error and return indicating error
    fprintf(stderr,"ICE: invalid function name '%s'\n",fname);
  } else { // otherwise
    if (udf_items[recno].exists) { // if the UDF is already defined
      // report the error and return indicating error
      fprintf(stderr,"ERROR: user-defined function '%s' on line %" PRIu32 " has already been defined.\n",fname,lno);
      retval=false;
    } else { // otherwise
      // add UDF information
      udf_items[recno].exists=true; // mark UDF as defined
      udf_items[recno].argname=((NULL==argname)?0:argname[0]); // store number of arguments
      retval=true;
    }
  }
xit:
  debug_printf("add_udf(%s) returns %s\n",fname,(retval?"true":"false"));
  return retval;
}

//
// This function is used to add a string literal value to the
// symbol table.  The litval parameter has the ASCIIZ string containing the literal
// value.  If the value is added successfully, true is returned and the ASCIIZ symbol
// name generated is returned in the symbol parameter.  Otherwise false is returned.
//
bool add_string_literal(const char *litval, char **symbol) {
  static uint32_t last_symbol=UINT32_C(0); // number of next string literal (used to generate literal names)
  struct literal_item *si=NULL; // pointer to current string literal item

  if (NULL==litval) {
    fprintf(stderr,"ERROR: NULL pointer for source of string literal\n");
    abort();
  }
  if (NULL==symbol) {
    fprintf(stderr,"ERROR: Invalid pointer for destination of string literal\n");
    abort();
  }
  // first lookup the literal value to see if we already have it.  If we do,
  // the symbol name of the existing literal value is returned.
  if (literals!=NULL) { // if there are any string literals
    si=literals;
    while (si!=NULL) { // while there are more string literals to consider
      if (strcmp(si->litval,litval)==0) { // if the current string literal is the same as litval
        *symbol=si->symname; // copy the symbol name into symbol
        return true;         // and return true
      }
      si=si->next;
    }
  }

  // The literal value was not found, so add it
  if ((si=(struct literal_item *)malloc(sizeof(struct literal_item)))==NULL) { // if we are OOM
    fprintf(stderr,"ERROR: out of memory attempting to add string literal\n"); // report the error and crash
    abort();
  }
  si->next=NULL;
  snprintf(si->symname,SYMNAME_MAX_BYTES-1,".LSLIT%04" PRIu32,last_symbol++); // generate the ASCIIZ symbol name
  si->symname[SYMNAME_MAX_BYTES-1]=0; // ensure termination
  *symbol=si->symname; // copy the symbol name into symbol
  strncpy(si->litval,litval,LITVAL_MAX_BYTES-1); // save the literal value
  si->litval[LITVAL_MAX_BYTES-1]=0; // ensure termination
  if (NULL==literals) { // if this is the very first string literal
    literals=last_literal=si; // point the head to this node
  } else { // otherwise
    last_literal->next=si; // append the node to the end of the list of string literals
    last_literal=si;
  }
  return true;
}

//
// This function is used to lookup information about a numeric array variable.
// The ASCIIZ string variable named is passed in the s argument.  If the
// information is found, a pointer to the associated record in the thelist
// table is returned.  If the information is not found, NULL is returned.
// The caller of this function must NOT attempt to free the record...
//
struct arraydata *get_array_data(const char *s) {
  struct arraydata *retval=NULL; // variable to store this function's return value
  uint32_t recno; // variable to hold index value returned from binary_search()

  // ensure incoming ASCIIZ string variable name is non-NULL
  if (NULL==s) {
    fprintf(stderr,"NULL pointer passed to get_array_data()\n");
    abort();
  }
  // ensure incoming ASCIIZ string variable name is non-empty
  if (0==s[0]) {
    fprintf(stderr,"Empty string passed to get_array_data()\n");
    abort();
  }
  // ensure incoming ASCIIZ string variable name's first letter is valid
  if ((s[0]<'A')||(s[0]>'Z')) { //invalid variable name only possible with an ICE
    fprintf(stderr,"Invalid numeric array variable name (does not begin with capital letter) specified to get_array_data()\n");
    abort();
  }
  if (s[1]!=0) { // invalid variable name (more than 1 byte) only possible with an ICE
    fprintf(stderr,"Invalid numeric array variable name (more than 1 byte) specified to get_array_data()\n");
    abort();
  }
  recno=binary_search(thelist,thelist_count,s); // search for s in thelist
  if (UINT_MAX==recno) { // s not found
    debug_printf("get_array_data: variable '%s' not found in primary symbol table\n",s);
  } else { // s found
    debug_printf("get_array_data: variable '%s' found in primary symbol table\n",s);
    if (thelist[recno].tipe!=NAVAR) { // if s is not a numeric array variable (type check)
      debug_printf("get_array_data: variable '%s' has invalid type %s instead of %s\n",
        s,vartype_names[thelist[recno].tipe],vartype_names[NAVAR]);
    } else {
      debug_printf("get_array_data: variable '%s' has valid type %s\n",
        s,vartype_names[NAVAR]);
      if (thelist[recno].exists==false) { // if s has not been used
        debug_printf("get_array_data: variables '%s' does not yet exist in primary symbol table\n",s);
      } else { // FOUND IT!
        debug_printf("get_array_data: variables '%s' exists in primary symbol table\n",s);
        retval=&thealist[s[0]-'A']; // get pointer to record for return value
      }
    }
  }
  return retval;
}

//
// This function will lookup a scalar numeric variable name.  If the
// name is found and has been used then true is returned, otherwise
// false is returned.
//
bool nvar_exists(const char *s) {
  bool retval=false; // variable to store this function's return value
  uint32_t recno; // variable to hold index value returned from binary_search()

  // ensure incoming ASCIIZ variable name is non-NULL
  if (NULL==s) {
    fprintf(stderr,"NULL pointer passed to nvar_exists()\n");
    abort();
  }
  // ensure incoming ASCIIZ variable name is non-empty
  if (0==s[0]) {
    fprintf(stderr,"Empty string passed to nvar_exists()\n");
    abort();
  }
  // ensure incoming ASCIIZ string variable name's first letter is valid
  if ((s[0]<'A')||(s[0]>'Z')) { // invalid variable name only possible with an ICE
    fprintf(stderr,"Invalid numeric scalar variable name '%s' (does not begin with capital letter) specified to nvar_exists()\n",s);
    goto xit;
  }
  if (s[1]!=0) { // multi-byte variable name (possible values are [A-Z][0-9])
    if ((s[1]<'0')||(s[1]>'9')) { // invalid variable name (more than one byte and second byte is not a digit) 
                                  // only possible with an ICE
      fprintf(stderr,"Invalid numeric scalar variable name '%s' (second byte is not a digit) specified to nvar_exists()\n",s);
      goto xit;
    }
    if (s[2]!=0) { // invalid variable name (more than 2 bytes) only possible with an ICE
      fprintf(stderr,"Invalid numeric scalar variable name '%s' (more than 2 bytes) specified to nvar_exists()\n",s);
      goto xit;
    }
  }
  if ((recno=binary_search(thelist,thelist_count,s))!=UINT_MAX) // if the name s is found
    if (NVAR==thelist[recno].tipe) // and it is a numeric scalar value (type check)
       retval=thelist[recno].exists; // and it has been used then set retval to true, otherwise false
xit:
  debug_printf("nvar_exists(%s) returns %s\n",s,(retval?"true":"false"));
  return retval;
}

//
// This function will lookup a scalar numeric variable name.  If the
// name is found and has been used or defined (DIM) then true is returned, 
// otherwise false is returned.
//
bool navar_exists(const char *s) {
  bool retval=false; // variable to store this function's return value
  uint32_t recno; // variable to hold index value returned from binary_search()

  // ensure incoming ASCIIZ variable name is non-NULL
  if (NULL==s) {
    fprintf(stderr,"NULL pointer passed to navar_exists()\n");
    abort();
  }
  // ensure incoming ASCIIZ variable name is non-empty
  if (0==s[0]) {
    fprintf(stderr,"Empty string passed to navar_exists()\n");
    abort();
  }
  // ensure incoming ASCIIZ variable name's first letter is valid
  if ((s[0]<'A')||(s[0]>'Z')) { // invalid variable name only possible with an ICE
    fprintf(stderr,"Invalid numeric array variable name '%s' (does not begin with capital letter) specified to navar_exists()\n",s);
    goto xit;
  }
  // ensure incoming ASCIIZ sstring variable name is a lone byte
  if (s[1]!=0) {
    fprintf(stderr,"Invalid numeric array variable name '%s' (more than 1 byte) specified to navar_exists()\n",s);
    goto xit;
  }
  if ((recno=binary_search(thelist,thelist_count,s))!=UINT_MAX) // if the name s is found
    if (NAVAR==thelist[recno].tipe) // and it is a numeric array variable (type check)
       retval=thelist[recno].exists; // and it has been used or defined then set reval to true, otherwise false
xit:
  debug_printf("navar_exists(%s) returns %s\n",s,(retval?"true":"false"));
  return retval;
}

//
// This function will lookup a user-defined function name.  If the
// name is found and has defined (DEF) then true is returned, 
// otherwise false is returned.  If it is true then argname will
// have the name of the argument if there is an argument to the UDF,
// or an empty string if the UDF does not take an argument.
//
bool udf_exists(const char *fname,char *argname) {
  bool retval=false; // variable to store this function's return value
  uint32_t recno; // variable to hold index value returned from binary_search()

  // ensure incoming ASCIIZ variable name is non-NULL
  if (NULL==fname) {
    fprintf(stderr,"NULL pointer passed to udf_exists()\n");
    abort();
  }
  // ensure incoming ASCIIZ variable name is non-empty
  if (0==fname[0]) {
    fprintf(stderr,"Empty string passed to udf_exists()\n");
    abort();
  }
  // ensure length is 3 bytes
  if (strlen(fname)!=3) { // invalid udf name (not 3 bytes) only possible with an ICE
    fprintf(stderr,"Invalid user-defined function name '%s' (not exactly 3 bytes) specified to udf_exists()\n",fname);
    goto xit;
  }
  // ensure first two letters are 'FN'
  if ((fname[0]!='F')||(fname[1]!='N')) { // invalid udf name, does not start with FN
    fprintf(stderr,"Invalid user-defined function name '%s' (does not begin with 'FN') specified to udf_exists()\n",fname);
    goto xit;
  }
  // ensure incoming ASCIIZ string variable name's third letter is valid
  if ((fname[2]<'A')||(fname[2]>'Z')) { // invalid variable name only possible with an ICE
    fprintf(stderr,"Invalid user-defined function name '%s' (third character is not a capital letter) specified to udf_exists()\n",fname);
    goto xit;
  }
  recno=fname[2]-'A'; // compute index in udf_items table from variable name
  if (recno>UINT32_C(25)) { // range check
    fprintf(stderr,"ICE: invalid function name '%s'\n",fname);
  }
  retval=udf_items[recno].exists; // set return value to true if UDF has been defined, otherwise false
  argname[0]=0; // set argname name to empty string
  if (retval) {
    if (0!=udf_items[recno].argname) { // if there is an argument to the functions
       argname[0]=udf_items[recno].argname; // copy argument name to argname
       argname[1]=0;
    }
  }
xit:
  debug_printf("udf_exists(%s) returns %s with %" PRIu8 " arguments\n",
    fname, (retval?"true":"false"),((0==argname[0])?0:1));
  return retval;
}

//
// This function will lookup a scalar string variable name.  If the
// name is found and has been used then true is returned, 
// otherwise false is returned.
//
bool svar_exists(const char *s) {
  bool retval=false; // variable to store this function's return value
  uint32_t recno; // variable to hold index value returned from binary_search()

  // ensure incoming ASCIIZ variable name is non-NULL
  if (NULL==s) {
    fprintf(stderr,"NULL pointer passed to svar_exists()\n");
    abort();
  }
  // ensure incoming ASCIIZ variable name is non-empty
  if (0==s[0]) {
    fprintf(stderr,"Empty string passed to svar_exists()\n");
    abort();
  }
  // ensure incoming ASCIIZ string variable name's first letter is valid
  if ((s[0]<'A')||(s[0]>'Z')) { // invalid variable name only possible with an ICE
    fprintf(stderr,"Invalid string scalar variable name '%s' (does not begin with capital letter) specified to svar_exists()\n",s);
    goto xit;
  }
  if (s[1]!='$') { // invalid variable name (second byte not a dollar) only possible with an ICE
    fprintf(stderr,"Invalid string scalar variable name '%s' (second byte is not a dollar sign) specified to svar_exists()\n",s);
    goto xit;
  }
  if (s[2]!=0) { // invalid variable name (more than 2 bytes) only possible with an ICE
    fprintf(stderr,"Invalid string scalar variable name '%s' (more than 2 bytes) specified to svar_exists()\n",s);
    goto xit;
  }
  if ((recno=binary_search(thelist,thelist_count,s))!=UINT_MAX) // if the name s is found
    if (SVAR==thelist[recno].tipe) // and it is a string scalar value (type check)
       retval=thelist[recno].exists; // and it has been used then set retval to true, otherwise false
xit:
  debug_printf("svar_exists(%s) returns %s\n",s,(retval?"true":"false"));
  return retval;
}

//
// This function will attempt to add the magic variable name specified by the parameter s
// to the list of magic variable names.  It will return true on success, false otherwise.
//
bool assign_magicvar(const char *s) {
  struct magicvar_item *temp=NULL;

  debug_printf("assign_magicvar(%s)\n",s);
  // allocate a new node for a magic variable
  if ((temp=(struct magicvar_item *)malloc(sizeof(struct magicvar_item)))==NULL) {
    fprintf(stderr,"assign_magicvar(%s):Unable to allocate memory\n",s);
    abort();
  }
  // copy magic variable name into new node
  strncpy(temp->symname,s,SYMNAME_MAX_BYTES);
  temp->symname[SYMNAME_MAX_BYTES-1]=0;  // terminate
  // prepend new node to linked list of magic variables
  temp->next=magicvars;
  magicvars=temp;

  need_BSS_numeric=true;	// set flag to true to indicate that a .bss section for numeric variables will be needed
  debug_printf("assign_magicvar(%s) returns true\n",s);
  return true;
}

//
// This function will attempt to add the scalar numeric variable name specified by the parameter s
// to the symbol table.  It will return true on success, false otherwise.  The lno parameter is the
// line number of the Minmal BASIC line number where the variable is accessed and is used in
// error messages.  If the variable is already defined and of the correct type, true is returned.
//
bool assign_nvar(const char *s,const uint32_t lno) {
  bool retval=false; // variable to store this functions's return value
  uint32_t recno; // variable to hold index value returned from binary_search()

  debug_printf("assign_nvar(%s)\n",s);
  // ensure incoming ASCIIZ variable name is non-NULL
  if (NULL==s) {
    fprintf(stderr,"ICE: NULL pointer passed to assign_nvar() on line %" PRIu32 "\n",lno);
    abort();
  }
  // ensure incoming ASCIIZ variable name is non-empty
  if (0==s[0]) {
    fprintf(stderr,"ICE: Empty string passed to assign_nvar() on line %" PRIu32 "\n",lno);
    abort();
  }
  recno=binary_search(thelist,thelist_count,s); // search for s in thelist
  if (UINT_MAX==recno) { // s not found
     fprintf(stderr,"ICE: Variable name '%s' not found in symbol table on line %" PRIu32 "\n",s,lno);
     goto xit;
  } 
  // s found
  if (thelist[recno].exists) { // if s is already defined
     if (NVAR==thelist[recno].tipe) { // already defined correctly as numeric scalar variable (type check)
       retval=true;
       goto xit;
     }
     fprintf(stderr,"Array variable '%s' cannot be assigned a numeric scalar value on line %" PRIu32 "\n",s,lno);
     goto xit;
  }
  // everything ok
  retval=true;
  thelist[recno].tipe=NVAR;   // mark this variable as being a numeric scalar variable
  thelist[recno].exists=true; // mark this variable as used
  need_BSS_numeric=true;      // set flag to true to indicate that a .bss section for numeric variables will be needed
xit:
  debug_printf("assign_nvar(%s) returns %s\n",s,(retval?"true":"false"));
  return retval;
}

//
// This function will attempt to add the array numeric variable name specified by the parameter s
// to the symbol table.  It will return true on success, false otherwise.  The lno parameter is the
// line number of the Minmal BASIC line number where the variable is accessed and is used in
// error messages.  If the variable is already defined and of the correct type, true is returned.
//
bool assign_navar(const char *s,const uint32_t lno) {
  bool retval=false; // variable to store this functions's return value
  uint32_t recno; // variable to hold index value returned from binary_search()

  debug_printf("assign_navar(%s)\n",s);
  // ensure incoming ASCIIZ variable name is non-NULL
  if (NULL==s) {
    fprintf(stderr,"NULL pointer passed to assign_navar()\n");
    abort();
  }
  // ensure incoming ASCIIZ variable name is non-empty
  if (0==s[0]) {
    fprintf(stderr,"Empty string passed to assign_navar()\n");
    abort();
  }
  recno=binary_search(thelist,thelist_count,s); // search for s in thelist
  if (UINT_MAX==recno) { // s not found
     fprintf(stderr,"ICE: Variable name %s not found in symbol table on line %" PRIu32 "\n",s,lno);
     goto xit;
  }
  // s found
  if (thelist[recno].exists) { // if s is already defined
    if (NAVAR==thelist[recno].tipe) { // already defined correctly as a scalar array variable (type check)
      retval=true;
      goto xit;
    }
    fprintf(stderr,"Scalar variable '%s' cannot be assigned a numeric array value on line %" PRIu32 "\n",s,lno);
    goto xit;
  }
  // everything ok
  retval=true;
  thelist[recno].tipe=NAVAR;    // mark this variable as being a numeric array variable
  thelist[recno].exists=true;   // mark this variable as used
  need_BSS_numeric=true;        // set flag to true to indicate that a .bss section for numeric variables will be needed
xit:
  debug_printf("assign_navar(%s) returns %s\n",s,(retval?"true":"false"));
  return retval;
}

//
// This function will attempt to add the scalar string variable name specified by the parameter s
// to the symbol table.  It will return true on success, false otherwise.  If the variable is already
// defined and of the correct type, true is returned.
//
bool assign_svar(const char *s) {
  bool retval=true; // variable to store this functions's return value
  uint32_t recno; // variable to hold index value returned from binary_search()

  debug_printf("assign_svar(%s)\n",s);
  // ensure incoming ASCIIZ variable name is non-NULL
  if (NULL==s) {
    fprintf(stderr,"NULL pointer passed to assign_svar()\n");
    abort();
  }
  // ensure incoming ASCIIZ variable name is non-empty
  if (0==s[0]) {
    fprintf(stderr,"Empty string passed to assign_svar()\n");
    abort();
  }
  // search for scalar string variable name s in thelist
  if ( ((recno=binary_search(thelist,thelist_count,s))==UINT_MAX) || // if not found
       (thelist[recno].tipe!=SVAR)) {                                // or not the correct type
     retval=false;                                                   // return false
  } else {                                                           // otherwise
     thelist[recno].exists=true;                                     // mark this scalar string variable as in-use
     need_BSS_string=true;	// set flag to true to indicate that a .bss section for string variables will be needed
  }

  debug_printf("assign_svar(%s) returns %s\n",s,(retval?"true":"false"));
  return retval;
}

//
// This function updates the thealist table with information about a numeric array variable.
// dims specifies the number of dimensions (1 or 2)
// dim1 is the maximum subscript permitted for the first dimension
// dim2 is the maximum subscript permitted for the second dimensions (0 if dims is 1)
// The function returns true on success, false on failure
//
bool put_array_data(const char *s,uint32_t dims,uint32_t dim1, uint32_t dim2) {
  bool retval=true; // variable to store this functions's return value
  uint32_t recno; // variable to hold index value returned from binary_search()

  debug_printf("put_array_data(%s,%" PRIu32 ",%" PRIu32 ",%" PRIu32 ")\n",s,dims,dim1,dim2);
  // ensure incoming ASCIIZ variable name is non-NULL
  if (NULL==s) {
    fprintf(stderr,"put_array_data: NULL pointer passed to put_array_data()\n");
    abort();
  }
  // ensure incoming ASCIIZ variable name is non-empty
  if (0==s[0]) {
    fprintf(stderr,"put_array_data: Empty string passed to put_array_data()\n");
    abort();
  }
  if ((dims<UINT32_C(1)) || (dims>UINT32_C(2))) { // if dims parameter value is invalid
                                                  // report the error and return false
    fprintf(stderr,"put_array_data: Number of dimensions must be 1 or 2\n");
    goto xit;
  }
  if ((s[0]<'A')||(s[0]>'Z')||(s[1]!=0)) {
    fprintf(stderr,"put_array_data: Invalid array name '%s'\n",s);
    retval=false;
    goto xit;
  }
  if ((dims==UINT32_C(1)) && (dim2!=UINT32_C(0))) { // specified a non-zero maximum for 2nd dimension in vector
                                                    // report error and return false
    fprintf(stderr,"put_array_data: specified a non-zero maximum subscript %" PRIu32 " for 2nd dimension in vector '%s'\n",dim2,s); 
    retval=false;
    goto xit;
  }
  recno=binary_search(thelist,thelist_count,s); // search for s in thelist
  if (recno>=thelist_count) { // if not found report error and return false
    debug_printf("put_array_data: variable '%s' not found in primary symbol table\n",s);
    retval=false;
    goto xit;
  }
  debug_printf("put_array_data: variable '%s' was found in primary symbol table\n",s);
  if ((thelist[recno].tipe!=UNDEFINED) && (thelist[recno].tipe!=NAVAR)) { // if type is not correct
                                                                          // report error and return false
    debug_printf("put_array_data: variable '%s' found, but invalid type '%s' is not UNDEFINED or NAVAR\n",
      s,vartype_names[thelist[recno].tipe]);
    retval=false;
    goto xit;
  }
  debug_printf("put_array_data: variable '%s' found and valid type is '%s'\n",
    s,vartype_names[thelist[recno].tipe]);
  if (thealist[s[0]-'A'].dims!=UINT32_C(0)) { // if number of dimensions has already been set
                                              // report error and return false
    debug_printf("put_array_data: variable '%s' found but already has been dimensioned\n",s);
    retval=false;
    goto xit;
  }
  debug_printf("put_array_data: updating primary symbol table\n");
  thelist[recno].tipe=NAVAR;     // set type to numeric array variable
  thelist[recno].exists=true;    // mark variable as used by this program
  debug_printf("put_array_data: updating secondary array symbol table\n");
  thealist[s[0]-'A'].dims=dims;  // set the number of dimensions
  thealist[s[0]-'A'].dim1=dim1;  // set the maximum permitted subscript for the first dimension
  thealist[s[0]-'A'].dim2=dim2;  // set the maximum permitted subscript for the second dimension if dims is 2
xit:
  if (!retval) debug_printf("not updating either primary symbol table or secondary array symbol table\n");
  debug_printf("put_array_data(%s,%" PRIu32 ",%" PRIu32 ",%" PRIu32 ") returns %s\n",
    s,dims,dim1,dim2,(retval?"true":"false"));
  return retval;
}

//
// Add a literal data file from a DATA statement
// The literal value is passed as an ASCIIZ string in s, and the type to use
// is passed in parameter t.  If successful, true is returned, otherwise
// false is returned.  This can update two lists:
//   1) data block list (if this is not a duplicate of an existing block)
//   2) data pointer list (always, and point to the unique data block for this value
//      that was either found or created in step 1)
//
bool add_data_literal(const char *s, enum data_item_type t) {
  struct data_item *temp=data_items; // pointer to data block information
  struct data_list *temp2=NULL;      // pointer to data pointer information
  char *end=NULL;                    // need by strto[df]()
  uint32_t tlen;                     // variable to hold length of s

  debug_printf("add_data_literal(\"%s\",%" PRIu32 ")\n",s,t);
  // do we have a data block for this literal value already?
  while (temp!=NULL) {
    if ((temp->t==t)&&(strcmp(temp->s,s)==0)) break; // found it!
    temp=temp->next;
  }

  if (NULL==temp) { // no, add data block information
    if ((temp=(struct data_item *)malloc(sizeof(struct data_item)))==NULL) {
      fprintf(stderr,"Out of RAM in add_data_literal()\n");
      abort();
    }
    temp->t=t; // save type
    temp->next=NULL; // initialize next pointer
    tlen=strnlen(s,LITVAL_MAX_BYTES);
    if ((temp->s=(char *)malloc(sizeof(char)*(tlen+UINT32_C(1))))==NULL) {
      fprintf(stderr,"Out of RAM in add_data_literal()\n");
      abort();
    }
    strncpy(temp->s,s,tlen); // copy in the actual literal value
    temp->s[tlen]=0; // terminate
    switch (temp->t) { // depending on type...
      case QSTR:  // quoted string
        break;
      case UQSTR: // unquoted string
        if (use_double)
          temp->numba.d=strtod(s,&end); // store 64bit numeric value
        else
          temp->numba.f=strtof(s,&end); // store 32bit numeric value
        if ((end==s)||(*end!=0)) {
          debug_printf("Hey you, '%s' cannot be a float\n",temp->s);
          if (use_double) // cannot be a float (64bit)
            temp->numba.d=nanny64.d;
          else // 32bit
            temp->numba.f=nanny.f;
          temp->t=QSTR;     // change the type to QSTR
        } else {
          debug_printf("end points to %" PRIu32 "\n",(unsigned char)*end);
          if (use_double)
            debug_printf("Hey you, '%s' is a float like %1.8E\n",temp->s,temp->numba.d);
          else
            debug_printf("Hey you, '%s' is a float like %1.8E\n",temp->s,temp->numba.f);
          // do not change the type of UQSTR
        }
        break;
      case NUM: // numeric scalar
        if (use_double) // if using 64 bit math
          temp->numba.d=strtod(s,&end); // store 64bit numeric value
        else            // else 32 bit math
          temp->numba.f=strtof(s,&end); // store 32bit numeric value
        break;
      default: // unknown type (ICE)
        fprintf(stderr,"Unknown type in add_data_literal()\n");
        abort();
    }
    snprintf(temp->symname,SYMNAME_MAX_BYTES-1,".LDATA%04" PRIu32,data_item_count++); // generate symbol name
    temp->symname[SYMNAME_MAX_BYTES-1]=0; // terminate
    debug_printf("Ok, finally type of '%s' is %s\n",temp->s,
       data_item_type_names[temp->t]);
    if (NULL==data_items) { // if this is the very first data item
      last_data_item=data_items=temp; // point head and tail to this node
    } else { // otherwise
      last_data_item->next=temp; // append the node to the end of the list of data items
      last_data_item=temp;
    }
  }
  if ((temp2=(struct data_list *)malloc(sizeof(struct data_list)))==NULL) {
    fprintf(stderr,"Out of RAM in add_data_literal()\n");
    abort();
  }
  strncpy(temp2->symbol_name,temp->symname,SYMNAME_MAX_BYTES); // store generated symbol name
  temp2->symbol_name[SYMNAME_MAX_BYTES-1]=0; // terminate
  temp2->next=NULL;
  if (NULL==data_list_tail) { // if this is the very first data pointer
    data_list_head=data_list_tail=temp2; // point head and tail to this node
  } else { // otherwise
    data_list_tail->next=temp2; // append the node to the end of the list of data pointers
    data_list_tail=temp2;
  }
  data_list_count++;
  return true;
}

//
// This function is used to add an input ASCIIZ format string specified by the
// fmt parameter.  Such a string is required to implement the INPUT statement.
// It contains up to 30 bytes of data followed by a NULL.  Each byte of data
// is either an 'S' or a 'N'.  An 'S' signals that a string value should be
// input, and an 'N' signals that a numeric value should be input.
// The lno paramater is the used to pass the Minimal BASIC source
// file line number for use in generating the literal data item symbol name.
// The symbol parameter is used to pass back the generated identifier for that 
// format string.
//
bool add_input_format(const char *fmt, char **symbol, uint32_t lno) {
  struct literal_item *si=NULL; // pointer used to create and insert new literal item

  // ensure incoming ASCIIZ format string is non-NULL
  if (NULL==fmt) {
    fprintf(stderr,"ERROR: NULL pointer for source of input format\n");
    abort();
  }
  // ensure incoming ASCIIZ format string is non-empty
  if ('0'==fmt[0]) {
    fprintf(stderr,"ERROR: Empty input format\n");
    abort();
  }
  // allocate a new literal item node
  if ((si=(struct literal_item *)malloc(sizeof(struct literal_item)))==NULL) {
    fprintf(stderr,"ERROR: out of memory attempting to add input format\n");
    abort();
  }
  // populate the new literal item node
  si->next=NULL; // this node will be the last node in the list
  snprintf(si->symname,SYMNAME_MAX_BYTES-1,".LNFMT%04" PRIu32,lno); // generate symbol name
  si->symname[SYMNAME_MAX_BYTES-1]=0; // terminate
  *symbol=si->symname; // fill in symbol parameter to pass the generated symbol name back to the caller
  strncpy(si->litval,fmt,LITVAL_MAX_BYTES-1); // copy the actual format data into the node
  si->litval[LITVAL_MAX_BYTES-1]=0; // terminate
  if (NULL==literals) { // if this is the very first data item
    literals=last_literal=si; // point head and tail to this node
  } else { // otherwise
    last_literal->next=si; // append the node to the end of the list of literal data items
    last_literal=si;
  }
  return true;
}

//
// This is code to test the module and is not normally used.
// If you compile with UNIT_TEST defined, then this module will
// be a stand-alone program with its own main() that runs some
// self-tests.
//
#ifdef UNIT_TEST
#include <stdio.h>

int main(int argc, char **argv) {
  bool quit_on_fail=false,passed=true;
  uint32_t i,j;
  struct arraydata *ad=NULL;
  char outfile_name[]="junk.s";
  FILE *outfile=NULL;

  if ((2==argc)&&(strncmp(argv[1],"-V",3)==0)) {
     verbose=true;
     quit_on_fail=true;
  } else {
     quit_on_fail=false;
     if ((2==argc)&&(strncmp(argv[1],"-v",3)==0))
       verbose=true;
     else
       verbose=false;
  }
  init_debug_printf(stderr);
  debug_printf("verbose=%s,quit_on_fail=%s\n",
      (verbose?"true":"false"), (quit_on_fail?"true":"false"));
  debug_printf("thealist_count=%" PRIu32 ",thelist_count=%" PRIu32 ",UINT_MAX=%" PRIu32 "\n",
      thealist_count,thelist_count,UINT_MAX);
  debug_printf("Begin self-test\n");
  // first test with empty symbol table
  for (i=UINT32_C(0);i<UINT32_C(26);i++) {
    char s[3];
    s[0]=(char)(UINT32_C(65)+i);
    s[1]=0;
    printf("nvar_exists(%2s) when undefined: ",s);
    if (nvar_exists(s)) {
      printf("FAIL\n");
      passed=false;
    } else {
      printf("PASS\n");
    }
    if (!passed && quit_on_fail) break;
    fflush(stdout);
    s[0]=(char)(UINT32_C(65)+i);
    s[1]='$';
    s[2]=0;
    printf("nvar_exists(%2s) when undefined: ",s);
    if (nvar_exists(s)) {
      printf("FAIL\n");
      passed=false;
    } else {
      printf("PASS\n");
    }
    if (!passed && quit_on_fail) break;
    fflush(stdout);
    for (j=UINT32_C(0);j<UINT32_C(10);j++) {
      s[1]='0'+j;
      printf("nvar_exists(%2s) when undefined: ",s);
      if (nvar_exists(s)) {
        printf("FAIL\n");
        passed=false;
      } else {
        printf("PASS\n");
      }
      if (!passed && quit_on_fail) break;
      fflush(stdout);
    }
    if (!passed && quit_on_fail) break;
    fflush(stdout);
  }
  // test user-defined functions
  while (true) {
    char s[4]="FN",a[2]={0};

    s[3]=0; // terminate
    for (i=UINT32_C(0);i<UINT32_C(26);i++) {
      s[2]=i+'A';
      printf("udf_exists(%2s) returns false when not defined: ",s);
      if (udf_exists(s,a)) {
        printf("FAIL\n");
        passed=false;
        if (quit_on_fail) break;
      } else {
        printf("PASS\n");
      }
    }
    if (!passed && quit_on_fail) break;
    printf("add_udf(\"FNX\",NULL) returns true when not defined: ");
    if (add_udf("FNX",NULL,10)) {
        printf("PASS\n");
    } else {
        printf("FAIL\n");
        passed=false;
    }
    printf("add_udf(\"FNY\",\"X\") returns true when not defined: ");
    if (add_udf("FNY","X",20)) {
        printf("PASS\n");
    } else {
        printf("FAIL\n");
        passed=false;
    }
    printf("add_udf(\"FNX\",NULL) returns false when already defined: ");
    if (!add_udf("FNX",NULL,10)) {
        printf("PASS\n");
    } else {
        printf("FAIL\n");
        passed=false;
    }
    printf("add_udf(\"BOB\",NULL) returns false when not defined: ");
    if (!add_udf("BOB",NULL,10)) {
        printf("PASS\n");
    } else {
        printf("FAIL\n");
        passed=false;
    }
    printf("udf_exists(\"FNX\") with no arguments returns true with empty argument when defined: ");
    if (udf_exists("FNX",a) && (a[0]==0)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
    }
    printf("udf_exists(\"FNY\") with one argument \"X\" returns true with correct argument when defined: ");
    if (udf_exists("FNY",a) && (a[0]=='X')) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
    }
    break;
  }
  // test scalars
  while (true) {
    char s[3];
    strncpy(s,"A",2);
    printf("assign_nvar(%2s) when undefined: ",s);
    if (assign_nvar(s,1)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"K9",3);
    printf("assign_nvar(%2s) when undefined: ",s);
    if (assign_nvar(s,1)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"G$",3);
    printf("assign_svar(%2s) when undefined: ",s);
    if (assign_svar(s)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"A0",3);
    printf("assign_nvar(%2s) when undefined: ",s);
    if (assign_nvar(s,1)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"A$",3);
    printf("assign_svar(%2s) when undefined: ",s);
    if (assign_svar(s)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"A",2);
    printf("nvar_exists(%2s) when defined: ",s);
    if (nvar_exists(s)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"K9",3);
    printf("nvar_exists(%2s) when defined: ",s);
    if (nvar_exists(s)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"G$",3);
    printf("svar_exists(%2s) when defined: ",s);
    if (svar_exists(s)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"A0",3);
    printf("nvar_exists(%2s) when defined: ",s);
    if (nvar_exists(s)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"A$",3);
    printf("svar_exists(%2s) when defined: ",s);
    if (svar_exists(s)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    // redefine?
    strncpy(s,"A",2);
    printf("assign_nvar(%2s) when defined: ",s);
    if (assign_nvar(s,1)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"K9",3);
    printf("assign_nvar(%2s) when defined: ",s);
    if (assign_nvar(s,1)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"G$",3);
    printf("assign_svar(%2s) when defined: ",s);
    if (assign_svar(s)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"A0",3);
    printf("assign_nvar(%2s) when defined: ",s);
    if (assign_nvar(s,1)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"A$",3);
    printf("assign_svar(%2s) when defined: ",s);
    if (assign_svar(s)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    break;
  }
  // test vector arrays
  while (true) {
    char s[3];
    strncpy(s,"B",2);
    printf("assign_navar(%2s) when undefined: ",s);
    if (assign_navar(s,1)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    printf("put_array_data(%2s,1,10,0) when defined as NAVAR: ",s);
    if (put_array_data(s,UINT32_C(1),UINT32_C(10),UINT32_C(0))) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"A",2);
    printf("assign_navar(%2s) when defined as NVAR: ",s);
    if (assign_navar(s,1)) {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    } else {
      printf("PASS\n");
    }
    printf("put_array_data(%2s,1,10,0) when defined as NVAR: ",s);
    if (put_array_data(s,UINT32_C(1),UINT32_C(10),UINT32_C(0))) {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    } else {
      printf("PASS\n");
    }
    strncpy(s,"B",2);
    printf("get_array_data(%s) when defined as NAVAR: ",s);
    if ((ad=get_array_data(s))==NULL) {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    } else {
      if ((ad->dims!=UINT32_C(1)) || (ad->dim1!=UINT32_C(10)) || (ad->dim2!=UINT32_C(0))) {
        printf("FAIL\n");
        passed=false;
        if (quit_on_fail) break;
      } else {
        printf("PASS\n");
      }
    }
    break;
  }
  // test matrix arrays (default)
  while (true) {
    char s[3];
    strncpy(s,"Z",2);
    printf("assign_navar(%2s) when undefined: ",s);
    if (assign_navar(s,1)) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    printf("put_array_data(%2s,2,10,20) when defined as NAVAR: ",s);
    if (put_array_data(s,UINT32_C(2),UINT32_C(10),UINT32_C(20))) {
      printf("PASS\n");
    } else {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    }
    strncpy(s,"A",2);
    printf("assign_navar(%2s) when defined as NVAR: ",s);
    if (assign_navar(s,1)) {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    } else {
      printf("PASS\n");
    }
    printf("put_array_data(%2s,2,10,20) when defined as NVAR: ",s);
    if (put_array_data(s,UINT32_C(2),UINT32_C(10),UINT32_C(20))) {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    } else {
      printf("PASS\n");
    }
    strncpy(s,"Z",2);
    printf("get_array_data(%s) when defined as NAVAR: ",s);
    if ((ad=get_array_data(s))==NULL) {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    } else {
      if ((ad->dims!=UINT32_C(2)) || (ad->dim1!=UINT32_C(10)) || (ad->dim2!=UINT32_C(20))) {
        printf("FAIL\n");
        passed=false;
        if (quit_on_fail) break;
      } else {
        printf("PASS\n");
      }
    }
    break;
  }
  // test string literals
  while (true) {
    char *s;

    printf("add string literal 'Hello, World!': ");
    if (!add_string_literal("Hello, World!",&s)) {
      printf("FAIL\n");
      passed=false;
      if (quit_on_fail) break;
    } else {
      if (strncmp(".LSLIT0000",s,11)!=0) {
        printf("FAIL\n");
        passed=false;
        if (quit_on_fail) break;
      } else {
        printf("PASS\n");
      }
      printf("add string literal 'craptastic': ");
      if (!add_string_literal("craptastic",&s)) {
        printf("FAIL\n");
        passed=false;
        if (quit_on_fail) break;
      } else {
        if (strncmp(".LSLIT0001",s,11)!=0) {
          printf("FAIL\n");
          passed=false;
          if (quit_on_fail) break;
        } else {
          printf("PASS\n");
        }
        printf("add string literal '1234567890123456789012345678901234567890123456789012345678901234567890': ");
        if (!add_string_literal("1234567890123456789012345678901234567890123456789012345678901234567890",&s)) {
          printf("FAIL\n");
          passed=false;
          if (quit_on_fail) break;
        } else {
          if (strncmp(".LSLIT0002",s,11)!=0) {
            printf("FAIL\n");
            passed=false;
            if (quit_on_fail) break;
          } else {
            printf("PASS\n");
          }
        }
      }
    }
    break;
  }
  if ((outfile=fopen(outfile_name,"w"))==NULL) {
    fprintf(stderr,"Unable to open file '%s for output.\n",outfile_name);
    passed=false;
  } else  {
    if (!generate_symbol_table(outfile)) {
      fprintf(stderr,"Unable to write data segment to file '%s'.\n",outfile_name);
      passed=false;
    }
    fclose(outfile);
  }
  if (verbose) dump_symbol_table(false);
  debug_printf("End self-test\n");
  return (passed?EXIT_SUCCESS:EXIT_FAILURE);
}
#endif
