//
// codegen.c - This file contains the code generation routines
//             for the code segment
//             for Mr. Ham's ECMA-55 Minimal BASIC compiler.
//
// Copyright (C) 2013,2014  John Gatewood Ham
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// version 2 of the Free Software Foundation License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
#include <stdlib.h>
#include <stdarg.h>
#include "globals.h"
#include "codegen.h"
#include "jenkins.h"
#include "shibata.h"
#include "gay.h"

#define COMMENT_BUFFER_MAX 256

static void write_init_RND(FILE *f);
static bool write_strcpy(FILE *f);
static bool write_strcmp(FILE *f);
static bool write_INPUT(FILE *f);
static bool write_printzone(FILE *f);

// map used to convert relop values to strings for display
const char * const relop_names[]={"<","<=","=",">=",">","<>"};
static uint32_t next_label_number=UINT32_C(0); // number used to generate next label name
static uint32_t next_var_number=UINT32_C(0);   // number used to generate next variable name
static uint32_t cur_jumptable_no=UINT32_C(0);  // number used to generate next jump table variable name for ON .. GOTO statement

//
// This function is used by the main driver to emit the assembly prologue.
// The f parameter is the file handle for the output file, and the file must
// already be open for output.  The ASCIIZ file name for the file f is passed
// in the inname parameter.  The code emitted is adjusted depending on many
// of the flags in globals.c, the most significant of which is the use_double
// variable that when true specifies normal 64 bit math will be used, but when
// false specifies 32 bit math will be used instead.  Other flags used are
// need_ATN, need_BSS_numeric, need_BSS_string, need_COS, need_EXP, need_INPUT,
// need_LOG, need_POW, need_READ, need_RND, need_SIN, need_SQR, need_strcmp,
// need_strcpy, and need_TAN.  This function returns true on success, false otherwise.
//
bool write_prologue(FILE *f,const char *inname) {
  if (NULL==f) {
    fprintf(stderr,"write_prologue(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        .file 1 \"%s\"\n",inname);
  fprintf(f,"        .att_syntax\n");
  fprintf(f,"        .att_mnemonic\n");
  fprintf(f,"        .arch core2,nojumps\n");
  if (use_SSE4_1)
    fprintf(f,"        .arch .sse4.1\n");
  if (use_AVX) {
    fprintf(f,"        .arch .avx\n");
    fprintf(f,"        .arch .xsave\n");
  }
  fprintf(f,"        .code64\n\n");

  // compile-time constants

  fprintf(f,"# constant definitions\n\n");
  fprintf(f,"        .equ MAX_STRING_BYTES, %" PRIu32 "\n",MAX_STRING_BYTES);
  if (use_double) { // we are using 64 bit math
    fprintf(f,"        .equ FLOATONE, 0x3ff0000000000000\n");
    fprintf(f,"        .equ FLOATNEGONE, 0xbff0000000000000\n");
    fprintf(f,"        .equ NaN,  0x7ff8000000000000\n");
    fprintf(f,"        .equ SNaN, 0x7ff0000000000001\n");
    fprintf(f,"        .equ PInf, 0x7ff0000000000000\n");
    fprintf(f,"        .equ NInf, 0xfff0000000000000\n");
  } else { // we are using 32 bit math
    fprintf(f,"        .equ FLOATONE, 0x3f800000\n");
    fprintf(f,"        .equ FLOATNEGONE, 0xbf800000\n");
    fprintf(f,"        .equ NaN,  0x7fc00000\n");
    fprintf(f,"        .equ SNaN, 0x7fa00000\n");
    fprintf(f,"        .equ PInf, 0x7f800000\n");
    fprintf(f,"        .equ NInf, 0xff800000\n");
  }
  fprintf(f,"        .equ NAK, 21\n");
  fprintf(f,"        .equ EDOM, 33\n");
  fprintf(f,"        .equ ERANGE, 34\n");
  fprintf(f,"        .equ FE_INVALID, 0x01\n");
  fprintf(f,"        .equ FE_DENORM, 0x02\n");
  fprintf(f,"        .equ FE_DIVBYZERO, 0x04\n");
  fprintf(f,"        .equ FE_OVERFLOW, 0x08\n");
  fprintf(f,"        .equ FE_UNDERFLOW, 0x10\n");
  fprintf(f,"        .equ FE_INEXACT, 0x20\n");
  fprintf(f,"        .equ FE_ALL_EXCEPT, (FE_INEXACT | FE_DIVBYZERO | FE_UNDERFLOW | FE_OVERFLOW | FE_INVALID)\n");
  fprintf(f,"        .equ FE_TONEAREST, 0x000\n");
  fprintf(f,"        .equ FE_DOWNWARD, 0x400\n");
  fprintf(f,"        .equ FE_UPWARD, 0x800\n");
  fprintf(f,"        .equ FE_TOWARDZERO, 0xc00\n");
  fprintf(f,"        .equ STDIN, 0\n");
  fprintf(f,"        .equ STDOUT, 1\n");
  fprintf(f,"        .equ STDERR, 2\n");
  fprintf(f,"        .equ F_GETFL, 3\n");
  fprintf(f,"        .equ F_SETFL, 4\n");
  fprintf(f,"        .equ O_NONBLOCK, 2048\n"); 
  fprintf(f,"        .equ S_IFMT, 00170000   # OCTAL\n");
  fprintf(f,"        .equ S_IFCHR, 0020000   # OCTAL\n");
  fprintf(f,"        .equ __NR_read, 0\n");
  fprintf(f,"        .equ __NR_write, 1\n");
  fprintf(f,"        .equ __NR_fstat, 5\n");
  fprintf(f,"        .equ __NR_exit, 60\n");
  fprintf(f,"        .equ __NR_fcntl, 72\n");
  fprintf(f,"        .equ __NR_gettimeofday, 96\n\n");

  fprintf(f,"        # macros\n\n");

  // macros to move floating point values between memory and SIMD registers
  fprintf(f,"        # store a floating value in a SIMD register to memory (by label)\n");
  fprintf(f,"        #\n");
  fprintf(f,"        # memlabel - destination memory location (label)\n");
  fprintf(f,"        # fromfloatreg - source xmm register (with %%)\n");
  fprintf(f,"\n");
  fprintf(f,"        .macro floatreg_to_floatmem fromfloatreg:req,memlabel:req,gpscratchreg=%%r15\n\n");
  fprintf(f,"        movabsq $\\memlabel\\(), \\gpscratchreg\\()\n");
  fprintf(f,"        movs%c   \\fromfloatreg\\(), (\\gpscratchreg\\())\n\n",(use_double?'d':'s'));
  fprintf(f,"        .endm\n\n");

  fprintf(f,"        # load a floating point value (by label) into a SIMD register\n");
  fprintf(f,"        #\n");
  fprintf(f,"        # memlabel        - label of a floating point value\n");
  fprintf(f,"        # destfloatreg    - name of xmm register to receive the value (with the %%)\n");
  fprintf(f,"        # floatval        - text string containing value for comment\n");
  fprintf(f,"        # gpscratchreg    - GP register used as a scratch register (caller must save if required)\n");
  fprintf(f,"        # floatscratchreg - xmm register used as a scratch register (caller must save if required)\n");
  fprintf(f,"        #\n");
  fprintf(f,"        # trashes xmm0 (required since myisinf needs that for parameter passing)\n");
  fprintf(f,"\n");
  fprintf(f,"        .macro floatlit_to_floatreg memlabel:req,destfloatreg:req,floatval:req,gpscratchreg=%%r15,floatscratchreg=%%xmm1\n\n");
  fprintf(f,"        movabsq $\\memlabel\\(), \\gpscratchreg\\()  #  floating point value \\floatval\\()\n");
  fprintf(f,"        movs%c   (\\gpscratchreg\\()), %%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        movs%c   %%xmm0,\\floatscratchreg\\()\n",(use_double?'d':'s'));
  if (!use_double) // if we are doing 32 bit math
    fprintf(f,"        cvtss2sd %%xmm0,%%xmm0\n"); // convert single to double for library routine
  fprintf(f,"        movabsq $myisinf, \\gpscratchreg\\()\n");
  fprintf(f,"        callq   *\\gpscratchreg\\()\n");
  fprintf(f,"        testq   %%rax, %%rax\n");
  fprintf(f,"        jz      0f\n");
  fprintf(f,"        # Infinity (overflow)\n");
  fprintf(f,"        movabsq $.Loverflow_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,\\gpscratchreg\\()\n");
  fprintf(f,"        callq   *\\gpscratchreg\\()\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movs%c   \\floatscratchreg\\(),\\destfloatreg\\()     # restore from backup copy\n\n",(use_double?'d':'s'));
  fprintf(f,"        .endm\n\n");

  fprintf(f,"        # load a scalar numeric floating point value (by label) into an xmm register\n");
  fprintf(f,"        # sourcelabel     - label of a floating point scalar variable\n");
  fprintf(f,"        # destfloatreg    - name of xmm register to receive the value (with the %%)\n");
  fprintf(f,"        # gpscratchreg    - GP register used as a scratch register (caller must save if required)\n");
  fprintf(f,"\n");
  fprintf(f,"        .macro floatmem_to_floatreg sourcelabel:req,destfloatreg:req,byte0:req,byte1,gpscratchreg=%%r15\n");
  fprintf(f,"\n");
  fprintf(f,"        movabsq $\\sourcelabel\\(),\\gpscratchreg\\()\n");
  fprintf(f,"        movs%c   (\\gpscratchreg\\()), \\destfloatreg\\()\n",(use_double?'d':'s'));
  fprintf(f,"        comis%c  \\destfloatreg\\(),\\destfloatreg\\()\n",(use_double?'d':'s'));
  fprintf(f,"        jnp     1f\n");
  fprintf(f,"        # referenced scalar \\sourcelabel\\() was NaN, uninitialized\n");
  fprintf(f,"        movabsq $.Luninitialized_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, \\gpscratchreg\\()\n");
  fprintf(f,"        callq   *\\gpscratchreg\\()\n");
  fprintf(f,"        # must create string in reverse order for little-endian\n");
  fprintf(f,"        xorq    %%rdi,%%rdi\n");
  fprintf(f,"        .ifnb   \\byte1\\()\n"); // if this is a 2-byte numeric variable like A0 or Z9...
  fprintf(f,"        movb    $\\byte1\\(),%%dil\n");
  fprintf(f,"        shl     $8,%%rdi\n");
  fprintf(f,"        .endif\n");
  fprintf(f,"        movb    $\\byte0\\(),%%dil\n");
  fprintf(f,"        shl     $8,%%rdi\n");
  fprintf(f,"        movb    $32,%%dil\n");
  fprintf(f,"        movabsq $printvarname,\\gpscratchreg\\()\n");
  fprintf(f,"        callq   *\\gpscratchreg\\()\n");
  fprintf(f,"        movabsq $badxit,\\gpscratchreg\\()\n");
  fprintf(f,"        jmpq    *\\gpscratchreg\\()\n");
  fprintf(f,"1:\n");
  fprintf(f,"        .endm\n\n");

  fprintf(f,"        .macro array_1D_to_floatreg arrayname:req,floatreg:req,byte0:req,minidx:req,maxidx:req,gpreg1=%%r8,gpreg2=%%r15\n");
  fprintf(f,"        cvts%c2si \\floatreg\\(),\\gpreg1\\()\n",(use_double?'d':'s'));
  fprintf(f,"        # bounds check for vector array read\n");
  fprintf(f,"        # \\gpreg1\\() must be <= \\maxidx\\() and >= \\minidx\\()\n");
  fprintf(f,"        # technique using just one compare is from \"Hacker's Delight, 2nd Ed.\", Henry S. Warren, Jr. p.67\n");
  fprintf(f,"        .ifeqs \"\\minidx\\()\",\"1\"\n");
  fprintf(f,"        decq    \\gpreg1\\()\n");
  fprintf(f,"        .endif\n");
  fprintf(f,"        cmp     $\\maxidx\\()-\\minidx\\(),\\gpreg1\\()\n");
  fprintf(f,"        jbe     0f\n");
  fprintf(f,"        movabsq $.Larray_out_of_bounds_read, %%rax\n");
  fprintf(f,"        jmpq   *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $\\arrayname\\(),\\gpreg2\\()\n");
  fprintf(f,"        movs%c   0(\\gpreg2\\(),\\gpreg1\\(),%" PRIu32 "),\\floatreg\\()\n",(use_double?'d':'s'),(use_double?UINT32_C(8):UINT32_C(4)));
  fprintf(f,"        comisd  \\floatreg\\(),\\floatreg\\()\n");
  fprintf(f,"        jnp     1f\n");
  fprintf(f,"        # referenced element of array A was NaN, uninitialized\n");
  fprintf(f,"        movabsq $.Luninitialized_element_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        # must create string in reverse order for little endian\n");
  fprintf(f,"        xorq    %%rdi,%%rdi\n");
  fprintf(f,"        movb    $\\byte0\\(),%%dil\n");
  fprintf(f,"        shl     $8,%%rdi\n");
  fprintf(f,"        movb    $32,%%dil\n");
  fprintf(f,"        movabsq $printvarname,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"1:\n");
  fprintf(f,"        .endm\n\n");

  fprintf(f,"        .macro floatreg_to_array_1D arrayname:req,floatvalreg:req,frowreg:req,minidx:req,maxrows:req,gprowreg=%%r8,gpscratchreg=%%r15\n");
  fprintf(f,"        cvts%c2si \\frowreg\\(),\\gprowreg\\()\n",(use_double?'d':'s'));
  fprintf(f,"        # bounds check for vector array write\n");
  fprintf(f,"        # technique using just one compare is from \"Hacker's Delight, 2nd Ed.\", Henry S. Warren, Jr. p.67\n");
  fprintf(f,"        .ifeqs \"\\minidx\\()\",\"1\"\n");
  fprintf(f,"        decq    \\gprowreg\\()\n");
  fprintf(f,"        .endif\n");
  fprintf(f,"        cmp     $\\maxrows\\()-\\minidx\\(),\\gprowreg\\()\n");
  fprintf(f,"        jbe     0f\n");
  fprintf(f,"        movabsq $.Larray_out_of_bounds_write, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $\\arrayname\\(),\\gpscratchreg\\()\n");
  fprintf(f,"        movs%c   \\floatvalreg\\(),0(\\gpscratchreg\\(),\\gprowreg\\(),%" PRIu32 ")\n",(use_double?'d':'s'),(use_double?UINT32_C(8):UINT32_C(4)));
  fprintf(f,"        .endm\n\n");

  fprintf(f,"        .macro array_2D_to_floatreg arrayname:req,frowreg:req,fcolreg:req,byte0:req,minidx:req,maxrow:req,maxcol:req,gpcolreg=%%r8,gprowreg=%%r9,gpscratchreg=%%r15\n");
  fprintf(f,"        cvts%c2si \\fcolreg\\(),\\gpcolreg\\()    # column index\n",(use_double?'d':'s'));
  fprintf(f,"        cvts%c2si \\frowreg\\(),\\gprowreg\\()    # row index\n",(use_double?'d':'s'));
  fprintf(f,"        # bounds check for matrix array read\n");
  fprintf(f,"        # row \\gprowreg\\() must be <= \\maxrow\\() and >= \\minidx\\()\n");
  fprintf(f,"        # column \\gpcolreg\\() must be <= \\maxcol\\() and >= \\minidx\\()\n");
  fprintf(f,"        # technique using just one compare is from \"Hacker's Delight, 2nd Ed.\", Henry S. Warren, Jr. p.67\n");
  fprintf(f,"        .ifeqs \"\\minidx\\()\",\"1\"\n");
  fprintf(f,"        decq    \\gprowreg\\()\n");
  fprintf(f,"        .endif\n");
  fprintf(f,"        cmp     $\\maxrow\\()-\\minidx\\(),\\gprowreg\\() # if row index is out of bounds then goto 0\n");
  fprintf(f,"        ja      0f\n");
  fprintf(f,"        .ifeqs \"\\minidx\\()\",\"1\"\n");
  fprintf(f,"        decq    \\gpcolreg\\()\n");
  fprintf(f,"        .endif\n");
  fprintf(f,"        cmp     $\\maxcol\\()-\\minidx\\(),\\gpcolreg\\() # if column index is OK then goto 1\n");
  fprintf(f,"        jbe     1f\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $.Larray_out_of_bounds_read, %%rax\n");
  fprintf(f,"        jmpq   *%%rax\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movq    \\gprowreg\\(),%%rax\n");
  fprintf(f,"        movabsq $\\maxcol\\()+1-\\minidx\\(),\\gpscratchreg\\() # number of elements in one row\n");
  fprintf(f,"        mul     \\gpscratchreg\\()    # rdx:rax=rax*\\gpscratchreg\\()\n");
  fprintf(f,"        add     %%rax,\\gpcolreg\\()\n");
  fprintf(f,"        movabsq $\\arrayname\\(),\\gpscratchreg\\()\n");
  fprintf(f,"        movsd   0(\\gpscratchreg\\(),\\gpcolreg\\(),%" PRIu32 "),\\frowreg\\()\n",(use_double?UINT32_C(8):UINT32_C(4)));
  fprintf(f,"        comis%c  \\frowreg\\(),\\frowreg\\()\n",(use_double?'d':'s'));
  fprintf(f,"        jnp     2f\n");
  fprintf(f,"        # referenced element of array A was NaN, uninitialized\n");
  fprintf(f,"        movabsq $.Luninitialized_element_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        # must create string in reverse order for little endian\n");
  fprintf(f,"        xorq    %%rdi,%%rdi\n");
  fprintf(f,"        movb    $\\byte0\\(),%%dil\n");
  fprintf(f,"        shl     $8,%%rdi\n");
  fprintf(f,"        movb    $32,%%dil\n");
  fprintf(f,"        movabsq $printvarname,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"2:\n");
  fprintf(f,"        .endm\n\n");

  fprintf(f,"        .macro floatreg_to_array_2D arrayname:req,floatvalreg:req,frowreg:req,fcolreg:req,minidx:req,maxrow:req,maxcol:req,gprowreg=%%r9,gpcolreg=%%r8,gpscratchreg=%%r15\n");
  fprintf(f,"        cvts%c2si \\fcolreg\\(),\\gpcolreg\\()      # column index\n",(use_double?'d':'s'));
  fprintf(f,"        cvts%c2si \\frowreg\\(),\\gprowreg\\()      # row index\n",(use_double?'d':'s'));
  fprintf(f,"        # bounds check for matrix array write\n");
  fprintf(f,"        # \\gprowreg\\() must be <= \\maxrow\\() and >= \\minidx\\()\n");
  fprintf(f,"        # \\gpcolreg\\() must be <= \\maxcol\\() and >= \\minidx\\()\n");
  fprintf(f,"        # bounds check for vector array write\n");
  fprintf(f,"        # technique using just one compare is from \"Hacker's Delight, 2nd Ed.\", Henry S. Warren, Jr. p.67\n");
  fprintf(f,"        .ifeqs \"\\minidx\\()\",\"1\"\n");
  fprintf(f,"        decq    \\gprowreg\\()\n");
  fprintf(f,"        .endif\n");
  fprintf(f,"        cmp     $\\maxrow\\()-\\minidx\\(),\\gprowreg\\()\n");
  fprintf(f,"        ja      0f\n");
  fprintf(f,"        .ifeqs \"\\minidx\\()\",\"1\"\n");
  fprintf(f,"        decq    \\gpcolreg\\()\n");
  fprintf(f,"        .endif\n");
  fprintf(f,"        cmp     $\\maxcol\\()-\\minidx\\(),\\gpcolreg\\()\n");
  fprintf(f,"        jbe     1f\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $.Larray_out_of_bounds_write, %%rax\n");
  fprintf(f,"        jmpq   *%%rax\n");
  fprintf(f,"1:\n");
  fprintf(f,"        # we must use %%rax since mul is hard-wired to use %%rdx:%%rax\n");
  fprintf(f,"        movq    \\gprowreg\\(),%%rax\n");
  fprintf(f,"        movabsq $\\maxcol\\()+1-\\minidx\\(),\\gpscratchreg\\()\n");
  fprintf(f,"        mul     \\gpscratchreg\\()   # rdx:rax=rax*\\gpscratchreg\\()\n");
  fprintf(f,"        add     %%rax,\\gpcolreg\\()\n");
  fprintf(f,"        movabsq $\\arrayname\\(),\\gpscratchreg\\()\n");
  fprintf(f,"        movs%c   \\floatvalreg\\(),0(\\gpscratchreg\\(),\\gpcolreg\\(),%" PRIu32 ")\n",(use_double?'d':'s'),(use_double?UINT32_C(8):UINT32_C(4)));
  fprintf(f,"        .endm\n\n");

  // macro used as prologue when emitting code for the FOR statement
  fprintf(f,"        .macro beginfor ndex_var:req,inc_var:req,limit_var:req,tst_lbl:req,done_lbl:req\n\n");
  fprintf(f,"        .section .rodata,\"a\",@progbits\n\n");
  fprintf(f,"        .type .Lbeginfor\\@_msg, %%object\n");
  fprintf(f,".Lbeginfor\\@_msg:\n");
  fprintf(f,"        .asciz \"\\ndex_var\\()\"\n");
  fprintf(f,"        .size .Lbeginfor\\@_msg, .-.Lbeginfor\\@_msg\n\n");
  fprintf(f,"        .section .text,\"ax\",@progbits\n\n");
  fprintf(f,"        popxmm 0\n");
  fprintf(f,"        comisd  %%xmm0, %%xmm0\n");
  fprintf(f,"        jnp     0f\n");
  fprintf(f,"        # referenced scalar \\inc_var\\() was NaN, uninitialized\n");
  fprintf(f,"        movabsq $.Luninitialized_forinc_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $\\inc_var\\(), %%rax\n");
  fprintf(f,"        movs%c   %%xmm0, (%%rax) # store FOR loop with index \\ndex_var\\() increment value\n",(use_double?'d':'s'));
  fprintf(f,"                              # \\inc_var\\()\n");
  fprintf(f,"        popxmm 1\n");
  fprintf(f,"        comisd  %%xmm1, %%xmm1\n");
  fprintf(f,"        jnp     1f\n");
  fprintf(f,"        # referenced scalar \\limit_var\\() was NaN, uninitialized\n");
  fprintf(f,"        movabsq $.Luninitialized_forlimit_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movabsq $\\limit_var\\(), %%rax\n");
  fprintf(f,"        movs%c   %%xmm1, (%%rax) # store FOR loop with index \\ndex_var\\() limit value in\n",(use_double?'d':'s'));
  fprintf(f,"                              # \\limit_var\\()\n");
  fprintf(f,"        popxmm 2\n");
  fprintf(f,"        comisd  %%xmm2, %%xmm2\n");
  fprintf(f,"        jnp     2f\n");
  fprintf(f,"        # referenced scalar \\ndex_var\\() was NaN, uninitialized\n");
  fprintf(f,"        movabsq $.Luninitialized_forindex_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $.Lbeginfor\\@_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"2:\n");
  fprintf(f,"        movabsq $\\ndex_var\\(), %%rax\n");
  fprintf(f,"        movs%c   %%xmm2, (%%rax) # store FOR loop index var \\ndex_var\\() initial value\n",(use_double?'d':'s'));
  fprintf(f,"\\tst_lbl\\():                  # FOR loop with index var \\ndex_var\\() test code starts here\n");
  fprintf(f,"        pxor    %%xmm3,%%xmm3   # xmm3=0\n");
  fprintf(f,"        movabsq $\\ndex_var\\(), %%rax\n");
  fprintf(f,"        movs%c   (%%rax),%%xmm0  # xmm0=\\ndex_var\\() (FOR loop index variable)\n",(use_double?'d':'s'));
  fprintf(f,"        comisd  %%xmm0, %%xmm0\n");
  fprintf(f,"        jnp     3f\n");
  fprintf(f,"        # referenced scalar \\ndex_var\\() was NaN, uninitialized\n");
  fprintf(f,"        movabsq $.Luninitialized_forindex_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $.Lbeginfor\\@_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"3:\n");
  fprintf(f,"        movabsq $\\limit_var\\(), %%rax\n");
  fprintf(f,"        movs%c   (%%rax),%%xmm1  # xmm1=\\limit_var\\() (FOR loop with index \\ndex_var\\()\n",(use_double?'d':'s'));
  fprintf(f,"                              # limit value)\n");
  fprintf(f,"        comisd  %%xmm1, %%xmm1\n");
  fprintf(f,"        jnp     4f\n");
  fprintf(f,"        # referenced scalar \\limit_var\\() was NaN, uninitialized\n");
  fprintf(f,"        movabsq $.Luninitialized_forlimit_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"4:\n");
  fprintf(f,"        subs%c   %%xmm1,%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        # xmm0=\\ndex_var\\()-\\limit_var\\() (FOR loop index var \\ndex_var\\()\n");
  fprintf(f,"        # FOR loop with index \\ndex_var\\() limit val)\n");
  fprintf(f,"        movabsq $\\inc_var\\(), %%rax\n");
  fprintf(f,"        movs%c   (%%rax),%%xmm2  # xmm2=\\inc_var\\() (FOR loop with index \\ndex_var\\()\n",(use_double?'d':'s'));
  fprintf(f,"                              # increment value)\n");
  fprintf(f,"        ucomis%c %%xmm3,%%xmm2\n",(use_double?'d':'s'));
  fprintf(f,"        jae     5f            # if FOR loop with index \\ndex_var\\() increment \\inc_var\\()\n");
  fprintf(f,"                              #    is non-negative then goto 0 (positive increment)\n");
  fprintf(f,"                              # else (negative increment)\n");
  fprintf(f,"        movabsq $.LFLT_NEG_MASK, %%r15\n");
  fprintf(f,"        movs%c   (%%r15), %%xmm1\n",(use_double?'d':'s'));
  fprintf(f,"        pxor    %%xmm1, %%xmm0  # xmm0=(-xmm0)\n");
  fprintf(f,"5:\n");
  fprintf(f,"        ucomis%c %%xmm3,%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"                              # IF (FORINDEX-FORLIMIT)*SGN(FORINCREMENT)<=0 THEN\n");
  fprintf(f,"                              #   GOTO FORLOOPBODY\n");
  fprintf(f,"        jbe     6f            # if ABS(\\ndex_var\\()-\\limit_var\\()) <=0 then goto 1\n");
  fprintf(f,"                              # (FOR loop with index \\ndex_var\\() body)\n");
  fprintf(f,"        movabsq $\\done_lbl\\(),%%rax\n");
  fprintf(f,"        jmpq    *%%rax         # else goto \\done_lbl\\() (exit FOR loop with\n");
  fprintf(f,"                              # index var \\ndex_var\\())\n");
  fprintf(f,"                              # FOR loop with index var \\ndex_var\\() test code ends here\n");
  fprintf(f,"6:                            # FOR loop with index var \\ndex_var\\() body begins here\n");
  fprintf(f,"        .endm\n\n");

  // macro used as epilogue when emitting code for the FOR statement
  fprintf(f,"        .macro endfor ndex_var:req,inc_var:req,tst_lbl:req,done_lbl:req\n\n");
  fprintf(f,"        .section .rodata,\"a\",@progbits\n\n");
  fprintf(f,"        .type .Lendfor\\@_msg, %%object\n");
  fprintf(f,".Lendfor\\@_msg:\n");
  fprintf(f,"        .asciz \"\\ndex_var\\()\"\n");
  fprintf(f,"        .size .Lendfor\\@_msg, .-.Lendfor\\@_msg\n\n");
  fprintf(f,"        .section .text,\"ax\",@progbits\n\n");
  fprintf(f,"                              # FOR loop with index var \\ndex_var\\() body ends here\n");
  fprintf(f,"                              # FOR loop with index var \\ndex_var\\() increment starts here\n");
  fprintf(f,"        movabsq $\\ndex_var\\(), %%rax\n");
  fprintf(f,"        movs%c   (%%rax),%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        comis%c  %%xmm0, %%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        jnp     1f\n");
  fprintf(f,"        # referenced scalar \\ndex_var\\() was NaN, uninitialized\n");
  fprintf(f,"        movabsq $.Luninitialized_forindex_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $.Lendfor\\@_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"1:\n");
  fprintf(f,"        pushxmm 0             # push FOR loop index variable \\ndex_var\\()\n");
  fprintf(f,"        movabsq $\\inc_var\\(), %%rax\n");
  fprintf(f,"        movs%c   (%%rax),%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        comis%c  %%xmm0, %%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        jnp     2f\n");
  fprintf(f,"        # referenced scalar \\inc_var\\() was NaN, uninitialized\n");
  fprintf(f,"        movabsq $.Luninitialized_forinc_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"2:\n");
  fprintf(f,"        pushxmm 0             # push FOR loop with index \\inc_var\\()\n");
  fprintf(f,"                              # increment variable \\ndex_var\\()\n");
  fprintf(f,"        binary_add\n");
  fprintf(f,"        popxmm  0\n");
  fprintf(f,"                              # increment FOR loop index variable \\ndex_var\\()\n");
  fprintf(f,"        movabsq $\\ndex_var\\(), %%rax      # \\ndex_var\\() = \\ndex_var\\() + \\inc_var\\()\n");
  fprintf(f,"        movs%c   %%xmm0,(%%rax)\n",(use_double?'d':'s'));
  fprintf(f,"                              # FOR loop with index var \\ndex_var\\() increment ends here\n");
  fprintf(f,"        movabsq $\\tst_lbl\\(), %%rax\n");
  fprintf(f,"        jmpq    *%%rax         # goto FOR loop with index var \\ndex_var\\()\n");
  fprintf(f,"                              # test label \\tst_lbl\\()\n");
  fprintf(f,"\\done_lbl\\():\n");
  fprintf(f,"        movabsq $\\inc_var\\(), %%rax\n");
  fprintf(f,"        mov%s $SNaN,%%r15%c\n",(use_double?"absq":"l   "),(use_double?' ':'d'));
  fprintf(f,"        mov%c    %%r15%s,(%%rax)  %s# \\inc_var\\() is now uninitialized again so\n",(use_double?'q':'l'),(use_double?"":"d"),(use_double?" ":""));
  fprintf(f,"                              # illegal jumps into the FOR loop with index var\n");
  fprintf(f,"                              # \\ndex_var\\() can be detected.\n");
  fprintf(f,"        .endm\n\n");

  // macro used when generating code for the GOSUB statement
  fprintf(f,"        .macro gosub from:req to:req\n");
  fprintf(f,"        movabsq $.Lretaddr_stack_ptr, %%r13\n");
  fprintf(f,"        movq    (%%r13),%%r15\n");
  fprintf(f,"        movabsq $retaddr_stack_size, %%r14\n");
  fprintf(f,"        cmp     %%r14,%%r15\n");
  fprintf(f,"        jb      1f\n");
  fprintf(f,"        movabsq $.Lretaddr_stack_full,%%r13\n");
  fprintf(f,"        jmpq    *%%r13\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movabsq $.Lretaddr_stack, %%r14\n");
  fprintf(f,"        movq    $\\from\\(), 0(%%r14,%%r15,8)\n");
  fprintf(f,"        incq    %%r15\n");
  fprintf(f,"        movq    %%r15,(%%r13)\n");
  fprintf(f,"        movabsq $\\to\\(), %%r15\n");
  fprintf(f,"        jmpq    *%%r15\n");
  fprintf(f,"        .endm\n\n");

  // macro used when generating code for the RETURN statement
  fprintf(f,"        .macro return\n");
  fprintf(f,"        movabsq $.Lretaddr_stack_ptr, %%r13\n");
  fprintf(f,"        movq    (%%r13),%%r15\n");
  fprintf(f,"        xorq    %%r14,%%r14\n");
  fprintf(f,"        cmp     %%r14,%%r15\n");
  fprintf(f,"        jne     1f\n");
  fprintf(f,"        movabsq $.Lretaddr_stack_empty,%%r13\n");
  fprintf(f,"        jmpq    *%%r13\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movabsq $.Lretaddr_stack, %%r14\n");
  fprintf(f,"        decq    %%r15\n");
  fprintf(f,"        movq    0(%%r14,%%r15,8), %%r14\n");
  fprintf(f,"        movq    %%r15,(%%r13)\n");
  fprintf(f,"        jmpq    *%%r14\n");
  fprintf(f,"        .endm\n\n");

  // macro used to push a string address on the dedicated string address stack
  fprintf(f,"        # on input address must be in rdi\n");
  fprintf(f,"        .macro pushsaddr\n");
  fprintf(f,"        movabsq $.Lsoperand_stack_ptr, %%rdx\n");
  fprintf(f,"        movq    (%%rdx), %%rbx\n");
  fprintf(f,"        cmpq    $soperand_stack_size, %%rbx\n");
  fprintf(f,"        jb      0f\n");
  fprintf(f,"        movabsq $.Lsoperand_stack_full,%%r13\n");
  fprintf(f,"        jmpq    *%%r13\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $.Lsoperand_stack, %%rax\n");
  fprintf(f,"        movq    %%rdi, (%%rax,%%rbx,8)\n");
  fprintf(f,"        incq    %%rbx\n");
  fprintf(f,"        movq    %%rbx, (%%rdx)\n");
  fprintf(f,"        .endm\n");
  fprintf(f,"\n");

  // macro used to pop a string address from the dedicated string address stack
  fprintf(f,"        # on output address is put into to\n");
  fprintf(f,"        .macro popsaddr to:req\n");
  fprintf(f,"        movabsq $.Lsoperand_stack_ptr, %%rdx\n");
  fprintf(f,"        movq    (%%rdx), %%rbx\n");
  fprintf(f,"        testq   %%rbx, %%rbx\n");
  fprintf(f,"        jnz     0f\n");
  fprintf(f,"        movabsq $.Lsoperand_stack_empty,%%r13\n");
  fprintf(f,"        jmpq    *%%r13\n");
  fprintf(f,"0:\n");
  fprintf(f,"        decq    %%rbx\n");
  fprintf(f,"        movq    %%rbx, (%%rdx)\n");
  fprintf(f,"        movabsq $.Lsoperand_stack, %%rax\n");
  fprintf(f,"        movq    (%%rax,%%rbx,8), \\to\\()\n");
  fprintf(f,"        .endm\n\n");

  // macro used to push a numeric floating point value on the dedicated floating point stack
  fprintf(f,"        # trashes r13, r14, r15\n");
  fprintf(f,"        # 'from' is some xmm register number 0..12\n");
  fprintf(f,"        .macro pushxmm from:req\n");
  fprintf(f,"        movabsq $.Loperand_stack_ptr, %%r13\n");
  fprintf(f,"        movq    (%%r13),%%r15\n");
  fprintf(f,"        movabsq $operand_stack_size, %%r14\n");
  fprintf(f,"        cmp     %%r14,%%r15\n");
  fprintf(f,"        jb      1f\n");
  fprintf(f,"        movabsq $.Loperand_stack_full,%%r13\n");
  fprintf(f,"        jmpq    *%%r13\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movabsq $.Loperand_stack, %%r14\n");
  fprintf(f,"        movs%c   %%xmm\\from\\(), 0(%%r14,%%r15,%" PRIu32 ")\n",(use_double?'d':'s'),(use_double?UINT32_C(8):UINT32_C(4)));
  fprintf(f,"        incq    %%r15\n");
  fprintf(f,"        movq    %%r15,(%%r13)\n");
  fprintf(f,"        .endm\n\n");

  // macro used to pop a numeric floating point value from the dedicated floating point stack
  fprintf(f,"        # trashes r13, r14, r15\n");
  fprintf(f,"        # 'to' is some xmm register number 0..12\n");
  fprintf(f,"        .macro popxmm to:req\n");
  fprintf(f,"        movabsq $.Loperand_stack_ptr, %%r13\n");
  fprintf(f,"        movq    (%%r13),%%r15\n");
  fprintf(f,"        xorq    %%r14,%%r14\n");
  fprintf(f,"        cmp     %%r14,%%r15\n");
  fprintf(f,"        jne     1f\n");
  fprintf(f,"        movabsq $.Loperand_stack_empty,%%r13\n");
  fprintf(f,"        jmpq    *%%r13\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movabsq $.Loperand_stack, %%r14\n");
  fprintf(f,"        decq    %%r15\n");
  fprintf(f,"        movs%c   0(%%r14,%%r15,%" PRIu32 "), %%xmm\\to\\()\n",(use_double?'d':'s'),(use_double?UINT32_C(8):UINT32_C(4)));
  fprintf(f,"        movq    %%r15,(%%r13)\n");
  fprintf(f,"        .endm\n\n");

  // macro used for emitting code for unary minus
  fprintf(f,"        # trashes r13, r14, r15\n");
  fprintf(f,"        .macro unary_minus\n");
  fprintf(f,"        popxmm 0\n");
  fprintf(f,"        movabsq $.LFLT_NEG_MASK, %%r15\n");
  fprintf(f,"        movs%c   (%%r15), %%xmm1\n",(use_double?'d':'s'));
  fprintf(f,"        pxor     %%xmm1, %%xmm0    # xmm0=-xmm0\n");
  fprintf(f,"        pushxmm 0\n");
  fprintf(f,"        .endm\n\n");

  // macro used for emitting code for addition
  fprintf(f,"        # trashes r13, r14, r15, xmm0, xmm1, xmm2\n");
  fprintf(f,"        .macro binary_add\n");
  fprintf(f,"        popxmm  1\n");
  fprintf(f,"        popxmm  0\n");
  fprintf(f,"        movs%c   %%xmm0,%%xmm2     # backup copy of augend in xmm2\n",(use_double?'d':'s'));
  fprintf(f,"                                # since a failed subtract will destroy xmm0\n");
  fprintf(f,"        adds%c   %%xmm1,%%xmm0     # xmm0=xmm0+xmm1\n",(use_double?'d':'s'));
  fprintf(f,"        stmxcsr .Lnewmxcsr\n");
  fprintf(f,"        testl   $FE_INVALID,.Lnewmxcsr\n");
  fprintf(f,"        jz      98f\n");
  fprintf(f,"        movabsq $.Luninitialized, %%r13\n");
  fprintf(f,"        jmpq    *%%r13\n");
  fprintf(f,"98:\n");
  fprintf(f,"        testl   $FE_OVERFLOW,.Lnewmxcsr\n");
  fprintf(f,"        jz      99f\n");
  fprintf(f,"        andl     $0xFFC0,.Lnewmxcsr  # clear exception in memory copy of MXCSR\n");
  fprintf(f,"        ldmxcsr .Lnewmxcsr        # and reload that into real MXCSR\n");
  fprintf(f,"        movabsq $.Loverflow_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r15\n");
  fprintf(f,"        callq   *%%r15\n");
  fprintf(f,"        movabsq $.LFLT_NEG_MASK, %%r15\n");
  fprintf(f,"        movs%c   (%%r15), %%xmm1   # xmm1=0x80000000%s (mask for sign)\n",(use_double?'d':'s'),(use_double?"00000000":""));
  fprintf(f,"        pand    %%xmm2,%%xmm1     # xmm1=0x80000000%s & backup copy of augend\n",(use_double?"00000000":""));
  fprintf(f,"                                # if xmm1 is zero, we got + else we got -\n");
  fprintf(f,"        mov%s    $PInf,%%r13%c     # r13d=+INF\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  fprintf(f,"        mov%s    $NInf,%%r14%c     # r14d=-INF, and is the default\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pextr%c  $0,%%xmm1,%%r15%c   # mov xmm1 into r15\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        subq    $%" PRIu32 ",%%rsp         # allocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
    fprintf(f,"                                # it is a union location for \n");
    fprintf(f,"                                # converting between %" PRIu32 " bit %s and \n",(use_double?UINT32_C(64):UINT32_C(32)),(use_double?"double":"float"));
    fprintf(f,"                                # unsigned %" PRIu32 " int\n",(use_double?UINT32_C(64):UINT32_C(32)));
    fprintf(f,"        movs%c   %%xmm1,(%%rsp)    # move xmm1\n",(use_double?'d':'s'));
    fprintf(f,"        mov%c    (%%rsp),%%r15%c    # into r15\n",(use_double?'q':'l'),(use_double?' ':'d'));
  }
  fprintf(f,"        test%c   %%r15%c,%%r15%c     # if all zeros\n",(use_double?'q':'l'),(use_double?' ':'d'),(use_double?' ':'d'));
  fprintf(f,"        cmovz%c  %%r13%c,%%r14%c     # then use +INF\n",(use_double?'q':'l'),(use_double?' ':'d'),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsr%c  $0,%%r14%c,%%xmm0   # move the + or - INF back into xmm0\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        mov%c    %%r14%c,(%%rsp)    # now move the + or - INF back\n",(use_double?'q':'l'),(use_double?' ':'d'));
    fprintf(f,"        movs%c   (%%rsp), %%xmm0   # into xmm0\n",(use_double?'d':'s'));
    fprintf(f,"        addq    $%" PRIu32 ",%%rsp         # deallocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
  }
  fprintf(f,"        # fall through to exit\n");
  fprintf(f,"99:\n");
  fprintf(f,"        pushxmm 0\n");
  fprintf(f,"        .endm\n\n");

  // macro used for emitting code for subtraction
  fprintf(f,"        # trashes r13, r14, r15, xmm0, xmm1, xmm2\n");
  fprintf(f,"        .macro binary_subtract\n");
  fprintf(f,"        popxmm  1\n");
  fprintf(f,"        popxmm  0\n");
  fprintf(f,"        movs%c   %%xmm0,%%xmm2     # backup copy of subtrahend in xmm2\n",(use_double?'d':'s'));
  fprintf(f,"                                # since a failed subtract will destroy xmm0\n");
  fprintf(f,"        subs%c   %%xmm1,%%xmm0     # xmm0=xmm0-xmm1\n",(use_double?'d':'s'));
  fprintf(f,"        stmxcsr .Lnewmxcsr\n");
  fprintf(f,"        testl   $FE_INVALID,.Lnewmxcsr\n");
  fprintf(f,"        jz      98f\n");
  fprintf(f,"        movabsq $.Luninitialized, %%r13\n");
  fprintf(f,"        jmpq    *%%r13\n");
  fprintf(f,"98:\n");
  fprintf(f,"        testl   $FE_OVERFLOW,.Lnewmxcsr\n");
  fprintf(f,"        jz      99f\n");
  fprintf(f,"        andl     $0xFFC0,.Lnewmxcsr  # clear exception in memory copy of MXCSR\n");
  fprintf(f,"        ldmxcsr .Lnewmxcsr        # and reload that into real MXCSR\n");
  fprintf(f,"        movabsq $.Loverflow_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r15\n");
  fprintf(f,"        callq   *%%r15\n");
  fprintf(f,"        movabsq $.LFLT_NEG_MASK, %%r15\n");
  fprintf(f,"        movs%c   (%%r15), %%xmm1   # xmm1=0x80000000%s (mask for sign)\n",(use_double?'d':'s'),(use_double?"00000000":""));
  fprintf(f,"        pand    %%xmm2,%%xmm1     # xmm1=0x80000000%s & backup copy of subtrahend\n",(use_double?"00000000":""));
  fprintf(f,"                                # if xmm1 is zero, we got + else we got -\n");
  fprintf(f,"        mov%s    $PInf,%%r13%c     # r13d=+INF\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  fprintf(f,"        mov%s    $NInf,%%r14%c     # r14d=-INF, and is the default\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pextr%c  $0,%%xmm1,%%r15%c   # mov xmm1 into r15\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        subq    $%" PRIu32 ",%%rsp         # allocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
    fprintf(f,"                                # it is a union location for \n");
    fprintf(f,"                                # converting between %" PRIu32 " bit %s and \n",(use_double?UINT32_C(64):UINT32_C(32)),(use_double?"double":"float"));
    fprintf(f,"                                # unsigned %" PRIu32 " int\n",(use_double?UINT32_C(64):UINT32_C(32)));
    fprintf(f,"        movs%c   %%xmm1,(%%rsp)    # move xmm1\n",(use_double?'d':'s'));
    fprintf(f,"        mov%c    (%%rsp),%%r15%c    # into r15\n",(use_double?'q':'l'),(use_double?' ':'d'));
  }
  fprintf(f,"        test%c   %%r15%c,%%r15%c     # if all zeros\n",(use_double?'q':'l'),(use_double?' ':'d'),(use_double?' ':'d'));
  fprintf(f,"        cmovz%c  %%r13%c,%%r14%c     # then use +INF\n",(use_double?'q':'l'),(use_double?' ':'d'),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsr%c  $0,%%r14%c,%%xmm0   # move the + or - INF back into xmm0\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        mov%c    %%r14%c,(%%rsp)    # now move the + or - INF back\n",(use_double?'q':'l'),(use_double?' ':'d'));
    fprintf(f,"        movs%c   (%%rsp), %%xmm0   # into xmm0\n",(use_double?'d':'s'));
    fprintf(f,"        addq    $%" PRIu32 ",%%rsp         # deallocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
  }
  fprintf(f,"        # fall through to exit\n");
  fprintf(f,"99:\n");
  fprintf(f,"        pushxmm 0\n");
  fprintf(f,"        .endm\n\n");

  // macro used for emitting code for multiplication
  fprintf(f,"        # trashes r13, r14, r15, xmm0, xmm1, xmm2, xmm3\n");
  fprintf(f,"        .macro binary_multiply\n");
  fprintf(f,"        popxmm  1\n");
  fprintf(f,"        popxmm  0\n");
  fprintf(f,"        movs%c   %%xmm0,%%xmm2     # backup copy of multiplicand in xmm2\n",(use_double?'d':'s'));
  fprintf(f,"                                # since a failed multiply will destroy xmm0\n");
  fprintf(f,"        movs%c   %%xmm1,%%xmm3     # backup copy of multiplier in xmm3\n",(use_double?'d':'s'));
  fprintf(f,"        muls%c   %%xmm1,%%xmm0     # xmm0=xmm0*xmm1\n",(use_double?'d':'s'));
  fprintf(f,"        stmxcsr .Lnewmxcsr\n");
  fprintf(f,"        testl   $FE_INVALID,.Lnewmxcsr\n");
  fprintf(f,"        jz      98f\n");
  fprintf(f,"        movabsq $.Luninitialized, %%r13\n");
  fprintf(f,"        jmpq    *%%r13\n");
  fprintf(f,"98:\n");
  fprintf(f,"        testl   $FE_OVERFLOW,.Lnewmxcsr\n");
  fprintf(f,"        jnz     99f\n");
  fprintf(f,"        jmp     100f\n");
  fprintf(f,"99:\n");
  fprintf(f,"        andl     $0xFFC0,.Lnewmxcsr  # clear exception in memory copy of MXCSR\n");
  fprintf(f,"        ldmxcsr .Lnewmxcsr        # and reload that into real MXCSR\n");
  fprintf(f,"        movabsq $.Loverflow_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r15\n");
  fprintf(f,"        callq   *%%r15\n");
  fprintf(f,"        movabsq $.LFLT_NEG_MASK, %%r15\n");
  fprintf(f,"        movs%c   (%%r15), %%xmm0   # xmm0=0x80000000%s (mask for sign)\n",(use_double?'d':'s'),(use_double?"00000000":""));
  fprintf(f,"        pand    %%xmm3,%%xmm0     # xmm0=0x80000000%s & backup copy of multiplier\n",(use_double?"00000000":""));
  fprintf(f,"        psrld   $31,%%xmm0       # xmm0=xmm0>>31\n");
  fprintf(f,"        movs%c   (%%r15), %%xmm1   # xmm1=0x80000000%s (mask for sign)\n",(use_double?'d':'s'),(use_double?"00000000":""));
  fprintf(f,"        pand    %%xmm2,%%xmm1     # xmm1=0x80000000%s & backup copy of multiplicand\n",(use_double?"00000000":""));
  fprintf(f,"        psrld   $%" PRIu32 ",%%xmm1       # xmm1=xmm1>>%" PRIu32 "\n",(use_double?UINT32_C(63):UINT32_C(31)),(use_double?UINT32_C(63):UINT32_C(31)));
  fprintf(f,"        adds%c   %%xmm0,%%xmm1     # 0,2 (same sign) 1 (different sign)\n",(use_double?'d':'s'));
  fprintf(f,"                                # if xmm1 is one, we got - else we got +\n");
  fprintf(f,"        mov%s    $PInf,%%r13%c     # r13d=+INF\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  fprintf(f,"        mov%s    $NInf,%%r14%c     # r14d=-INF, and is the default\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pextr%c  $0,%%xmm1,%%r15%c   # mov xmm1 into r15\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        subq    $%" PRIu32 ",%%rsp         # allocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
    fprintf(f,"                                # it is a union location for \n");
    fprintf(f,"                                # converting between %" PRIu32 " bit %s and \n",(use_double?UINT32_C(64):UINT32_C(32)),(use_double?"double":"float"));
    fprintf(f,"                                # unsigned %" PRIu32 " int\n",(use_double?UINT32_C(64):UINT32_C(32)));
    fprintf(f,"        movs%c   %%xmm1,(%%rsp)    # move xmm1\n",(use_double?'d':'s'));
    fprintf(f,"        mov%c    (%%rsp),%%r15%c    # into r15\n",(use_double?'q':'l'),(use_double?' ':'d'));
  }
  fprintf(f,"        and%c    $1,%%r15%c        # mask out all but the last bit\n",(use_double?'q':'l'),(use_double?' ':'d'));
  fprintf(f,"        test%c   %%r15%c,%%r15%c     # if all zeros\n",(use_double?'q':'l'),(use_double?' ':'d'),(use_double?' ':'d'));
  fprintf(f,"        cmovz%c  %%r13%c,%%r14%c     # then use +INF\n",(use_double?'q':'l'),(use_double?' ':'d'),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsr%c  $0,%%r14%c,%%xmm0   # move the + or - INF back into xmm0\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        mov%c    %%r14%c,(%%rsp)    # now move the + or - INF back\n",(use_double?'q':'l'),(use_double?' ':'d'));
    fprintf(f,"        movs%c   (%%rsp), %%xmm0   # into xmm0\n",(use_double?'d':'s'));
    fprintf(f,"        addq    $%" PRIu32 ",%%rsp         # deallocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
  }
  fprintf(f,"        jmp     101f\n");
  fprintf(f,"100:\n");
  fprintf(f,"        testl   $FE_UNDERFLOW,.Lnewmxcsr\n");
  fprintf(f,"        jz      101f\n");
  fprintf(f,"        andl     $0xFFC0,.Lnewmxcsr  # clear exception in memory copy of MXCSR\n");
  fprintf(f,"        ldmxcsr .Lnewmxcsr        # and reload that into real MXCSR\n");
  fprintf(f,"        movabsq $.Lunderflow_msg, %%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r15\n");
  fprintf(f,"        callq   *%%r15\n");
  fprintf(f,"        xorp%c   %%xmm0,%%xmm0       # substitute a zero value\n",(use_double?'d':'s'));
  fprintf(f,"        # fall through intentionally\n");
  fprintf(f,"101:\n");
  fprintf(f,"        pushxmm 0\n");
  fprintf(f,"        .endm\n\n");

  // macro used for emitting code for division
  fprintf(f,"        # trashes r13, r14, r15, xmm0, xmm1, xmm2\n");
  fprintf(f,"        .macro binary_divide\n");
  fprintf(f,"        popxmm  1\n");
  fprintf(f,"        popxmm  0\n");
  fprintf(f,"        movs%c   %%xmm0,%%xmm2     # backup copy of numerator in xmm2\n",(use_double?'d':'s'));
  fprintf(f,"                                # since a failed divide will destroy xmm0\n");
  fprintf(f,"        divs%c   %%xmm1,%%xmm0     # xmm0=xmm0/xmm1\n",(use_double?'d':'s'));
  fprintf(f,"        stmxcsr .Lnewmxcsr\n");
  fprintf(f,"        testl   $FE_OVERFLOW,.Lnewmxcsr\n");
  fprintf(f,"        jz      96f\n");
  fprintf(f,"        movabsq $.Loverflow_msg,%%rsi\n");
  fprintf(f,"        jmp     97f\n");
  fprintf(f,"96:\n");
  fprintf(f,"        testl   $FE_DIVBYZERO,.Lnewmxcsr\n");
  fprintf(f,"        jz      98f\n");
  fprintf(f,"        movabsq $.Ldividebyzero_msg,%%rsi\n");
  fprintf(f,"97:\n");
  fprintf(f,"        andl     $0xFFC0,.Lnewmxcsr  # clear exception in memory copy of MXCSR\n");
  fprintf(f,"        ldmxcsr .Lnewmxcsr        # and reload that into real MXCSR\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r15\n");
  fprintf(f,"        callq   *%%r15\n");
  fprintf(f,"        movabsq $.LFLT_NEG_MASK, %%r15\n");
  fprintf(f,"        movs%c   (%%r15), %%xmm1   # xmm1=0x80000000%s (mask for sign)\n",(use_double?'d':'s'),(use_double?"00000000":""));
  fprintf(f,"        pand    %%xmm2,%%xmm1     # xmm1=0x80000000%s & backup copy of numerator\n",(use_double?"00000000":""));
  fprintf(f,"                                # if xmm1 is zero, we got + else we got -\n");
  fprintf(f,"        mov%s    $PInf,%%r13%c     # r13d=+INF\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  fprintf(f,"        mov%s    $NInf,%%r14%c     # r14d=-INF, and is the default\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pextr%c  $0,%%xmm1,%%r15%c   # mov xmm1 into r15\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        subq    $%" PRIu32 ",%%rsp         # allocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
    fprintf(f,"                                # it is a union location for \n");
    fprintf(f,"                                # converting between %" PRIu32 " bit %s and \n",(use_double?UINT32_C(64):UINT32_C(32)),(use_double?"double":"float"));
    fprintf(f,"                                # unsigned %" PRIu32 " int\n",(use_double?UINT32_C(64):UINT32_C(32)));
    fprintf(f,"        movs%c   %%xmm1,(%%rsp)    # move xmm1\n",(use_double?'d':'s'));
    fprintf(f,"        mov%c    (%%rsp),%%r15%c    # into r15\n",(use_double?'q':'l'),(use_double?' ':'d'));
  }
  fprintf(f,"        test%c   %%r15%c,%%r15%c     # if all zeros\n",(use_double?'q':'l'),(use_double?' ':'d'),(use_double?' ':'d'));
  fprintf(f,"        cmovz%c  %%r13%c,%%r14%c     # then use +INF\n",(use_double?'q':'l'),(use_double?' ':'d'),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsr%c  $0,%%r14%c,%%xmm0   # move the + or - INF back into xmm0\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        mov%c    %%r14%c,(%%rsp)    # now move the + or - INF back\n",(use_double?'q':'l'),(use_double?' ':'d'));
    fprintf(f,"        movs%c   (%%rsp), %%xmm0   # into xmm0\n",(use_double?'d':'s'));
    fprintf(f,"        addq    $%" PRIu32 ",%%rsp         # deallocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
  }
  fprintf(f,"        jmp     100f            # continue (CANNOT BE 'jmpq' !!!)\n");
  fprintf(f,"98:\n");
  fprintf(f,"        testl   $FE_INVALID,.Lnewmxcsr\n");
  fprintf(f,"        jz      99f\n");
  fprintf(f,"        # either 0/0 or INF/INF\n");
  if (use_SSE4_1) {
    fprintf(f,"        pextr%c  $0,%%xmm2,%%r15%c   # mov xmm1 into r15\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        subq    $%" PRIu32 ",%%rsp         # allocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
    fprintf(f,"                                # it is a union location for \n");
    fprintf(f,"                                # converting between %" PRIu32 " bit %s and \n",(use_double?UINT32_C(64):UINT32_C(32)),(use_double?"double":"float"));
    fprintf(f,"                                # unsigned %" PRIu32 " int\n",(use_double?UINT32_C(64):UINT32_C(32)));
    fprintf(f,"        movs%c   %%xmm2,(%%rsp)\n",(use_double?'d':'s'));
    fprintf(f,"        mov%c    (%%rsp),%%r15%c\n",(use_double?'q':'l'),(use_double?' ':'d'));
  }
  fprintf(f,"        test%c   %%r15%c,%%r15%c\n",(use_double?'q':'l'),(use_double?' ':'d'),(use_double?' ':'d'));
  fprintf(f,"        jz      101f\n");
  if (!use_SSE4_1) {
    fprintf(f,"        addq    $%" PRIu32 ",%%rsp         # deallocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
  }
  fprintf(f,"        # INF/INF\n");
  fprintf(f,"        movabsq $.Lbaddivide, %%r13\n");
  fprintf(f,"        jmpq    *%%r13\n");
  fprintf(f,"101:\n");
  fprintf(f,"        # 0/0\n");
  fprintf(f,"        movabsq $.Ldividebyzero_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r15\n");
  fprintf(f,"        callq   *%%r15\n");
  fprintf(f,"        andl     $0xFFC0,.Lnewmxcsr  # clear exception in memory copy of MXCSR\n");
  fprintf(f,"        ldmxcsr .Lnewmxcsr        # and reload that into real MXCSR\n");
  fprintf(f,"        mov%s    $PInf,%%r15%c\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsr%c  $0,%%r15%c,%%xmm0   # move +INF into xmm0\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        mov%c    %%r15%c,(%%rsp)    # temp = +INF\n",(use_double?'q':'l'),(use_double?' ':'d'));
    fprintf(f,"        movs%c   (%%rsp),%%xmm0\n",(use_double?'d':'s'));
    fprintf(f,"        addq    $%" PRIu32 ",%%rsp         # deallocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
  }
  fprintf(f,"        jmp     100f\n");
  fprintf(f,"99:\n");
  fprintf(f,"        testl   $FE_UNDERFLOW,.Lnewmxcsr\n");
  fprintf(f,"        jz      100f\n");
  fprintf(f,"        andl     $0xFFC0,.Lnewmxcsr  # clear exception in memory copy of MXCSR\n");
  fprintf(f,"        ldmxcsr .Lnewmxcsr        # and reload that into real MXCSR\n");
  fprintf(f,"        movabsq $.Lunderflow_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r15\n");
  fprintf(f,"        callq   *%%r15\n");
  fprintf(f,"        xorp%c   %%xmm0,%%xmm0       # substitute a zero value\n",(use_double?'d':'s'));
  fprintf(f,"        # fall through intentionally\n");
  fprintf(f,"100:\n");
  fprintf(f,"        pushxmm 0\n");
  fprintf(f,"        .endm\n\n");

  // macro used for emitting code for raising to a power
  fprintf(f,"        # trashes r13, r14, xmm0, xmm1, xmm2, rsi\n");
  fprintf(f,"        .macro binary_power\n");
  fprintf(f,"        movq    %%rax,%%r14\n");
  fprintf(f,"        popxmm  1     # exponent\n");
  if (!use_double) // if we are using 32 bit math
    fprintf(f,"        cvtss2sd %%xmm1,%%xmm1\n");
  fprintf(f,"        popxmm  0     # base\n");
  if (!use_double) // if we are using 32 bit math
    fprintf(f,"        cvtss2sd %%xmm0,%%xmm0\n");
  fprintf(f,"        pxor    %%xmm2, %%xmm2\n");
  fprintf(f,"        comisd  %%xmm2,%%xmm0\n");
  fprintf(f,"        jne     0f\n");
  fprintf(f,"        comisd  %%xmm0,%%xmm1 # we know base is zero at this point\n");
  fprintf(f,"        jae     0f\n");
  fprintf(f,"        # zero to some negative power\n");
  fprintf(f,"        movabsq $.Lzero_to_negative_power_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r13\n");
  fprintf(f,"        callq   *%%r13\n");
  fprintf(f,"        mov%s    $PInf,%%r13%c\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsr%c  $0,%%r13%c,%%xmm0   # move +INF into xmm0\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        subq    $%" PRIu32 ",%%rsp         # allocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
    fprintf(f,"        mov%c    %%r13%c,(%%rsp)    # move +INF into xmm0\n",(use_double?'q':'l'),(use_double?' ':'d'));
    fprintf(f,"        movs%c   (%%rsp),%%xmm0\n",(use_double?'d':'s'));
    fprintf(f,"        addq    $%" PRIu32 ",%%rsp         # deallocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
  }
  fprintf(f,"        jmp     4f\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $xpow, %%r13\n");
  fprintf(f,"        callq   *%%r13\n");
  fprintf(f,"        movabsq $myisnan, %%r13\n");
  fprintf(f,"        callq   *%%r13\n");
  fprintf(f,"        testq   %%rax, %%rax\n");
  fprintf(f,"        jz      1f\n");
  fprintf(f,"        movabsq $.Lbadpower, %%r13\n");
  fprintf(f,"        jmpq    *%%r13\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movabsq $myisinf, %%r13\n");
  fprintf(f,"        callq   *%%r13\n");
  fprintf(f,"        testq   %%rax, %%rax\n");
  fprintf(f,"        jz      3f\n");
  fprintf(f,"        # Infinity (overflow)\n");
  fprintf(f,"        movabsq $.Loverflow_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r13\n");
  fprintf(f,"        callq   *%%r13\n");
  fprintf(f,"3:\n");
  if (!use_double) // if we are using 32 bit math
    fprintf(f,"        cvtsd2ss %%xmm0,%%xmm0\n");
  fprintf(f,"4:\n");
  fprintf(f,"        movq    %%r14,%%rax\n");
  fprintf(f,"        pushxmm 0\n");
  fprintf(f,"        .endm\n\n");

  // read only data for runtime (constants)

  fprintf(f,"        .section .rodata,\"a\",@progbits\n\n");

  //
  // Binary mask used to test if a floating point value is positive
  // or negative, or to force a number to be positive or negative.
  // bitwise AND with this will return this value if a number is negative.
  // bitwise OR with this will force a number to be negative.
  //
  fprintf(f,"        .type .LFLT_NEG_MASK, @object\n");
  fprintf(f,".LFLT_NEG_MASK:\n");
  if (use_double) // using 64 bit math
    fprintf(f,"        .quad 0x8000000000000000\n");
  else // using 32 bit math
    fprintf(f,"        .long 0x80000000\n");
  fprintf(f,"        .size .LFLT_NEG_MASK, .-.LFLT_NEG_MASK\n\n");

  //
  // strings used in error and exception messages
  //
  fprintf(f,"        .type .Lspace, @object\n");
  fprintf(f,".Lspace:\n");
  fprintf(f,"        .asciz \" \"\n");
  fprintf(f,"        .size .Lspace, .-.Lspace\n\n");
  fprintf(f,"        .type .Lnocmov, %%object\n");
  fprintf(f,".Lnocmov:\n");
  fprintf(f,"        .asciz \"CPU lacks required support for CMOV instruction.\\n\"\n");
  fprintf(f,"        .size .Lnocmov, .-.Lnocmov\n");
  fprintf(f,"        .set .Lnocmovlen, .-.Lnocmov\n\n");
  fprintf(f,"        .type .Lnosse2, %%object\n");
  fprintf(f,".Lnosse2:\n");
  fprintf(f,"        .asciz \"CPU lacks required support for SSE2 instructions.\\n\"\n");
  fprintf(f,"        .size .Lnosse2, .-.Lnosse2\n");
  fprintf(f,"        .set .Lnosse2len, .-.Lnosse2\n\n");
  fprintf(f,"        .type .Lnosse3, %%object\n");
  fprintf(f,".Lnosse3:\n");
  fprintf(f,"        .asciz \"CPU lacks required support for SSE3 instructions.\\n\"\n");
  fprintf(f,"        .size .Lnosse3, .-.Lnosse3\n");
  fprintf(f,"        .set .Lnosse3len, .-.Lnosse3\n\n");
  fprintf(f,"        .type .Lnossse3, %%object\n");
  fprintf(f,".Lnossse3:\n");
  fprintf(f,"        .asciz \"CPU lacks required support for SSSE3 instructions.\\n\"\n");
  fprintf(f,"        .size .Lnossse3, .-.Lnossse3\n");
  fprintf(f,"        .set .Lnossse3len, .-.Lnossse3\n\n");
  if (use_SSE4_1) {
    fprintf(f,"        .type .Lnosse4_1, %%object\n");
    fprintf(f,".Lnosse4_1:\n");
    fprintf(f,"        .asciz \"CPU lacks required support for SSE4.1 instructions but you specified -4 compiler option.\\n\"\n");
    fprintf(f,"        .size .Lnosse4_1, .-.Lnosse4_1\n");
    fprintf(f,"        .set .Lnosse4_1len, .-.Lnosse4_1\n\n");
  }
  if (use_AVX) {
    fprintf(f,"        .type .Lnoavx, %%object\n");
    fprintf(f,".Lnoavx:\n");
    fprintf(f,"        .asciz \"CPU lacks required support for AVX instructions but you specified -A compiler option.\\n\"\n");
    fprintf(f,"        .size .Lnoavx, .-.Lnoavx\n");
    fprintf(f,"        .set .Lnoavxlen, .-.Lnoavx\n\n");
  }
  fprintf(f,"        .type .Lon_line_number_msg, %%object\n");
  fprintf(f,".Lon_line_number_msg:\n");
  fprintf(f,"        .asciz \" on line number \"\n");
  fprintf(f,"        .size .Lon_line_number_msg, .-.Lon_line_number_msg\n\n");
  fprintf(f,"        .type .LLOG_ltzero_msg, %%object\n");
  fprintf(f,".LLOG_ltzero_msg:\n");
  fprintf(f,"        .asciz \"Argument value for LOG function not greater than zero\"\n");
  fprintf(f,"        .size .LLOG_ltzero_msg, .-.LLOG_ltzero_msg\n\n");
  fprintf(f,"        .type .Larray_OOB_write_msg, %%object\n");
  fprintf(f,".Larray_OOB_write_msg:\n");
  fprintf(f,"        .asciz \"Array index out of bounds write\"\n");
  fprintf(f,"        .size .Larray_OOB_write_msg, .-.Larray_OOB_write_msg\n\n");
  fprintf(f,"        .type .Larray_OOB_read_msg, %%object\n");
  fprintf(f,".Larray_OOB_read_msg:\n");
  fprintf(f,"        .asciz \"Array index out of bounds read\"\n");
  fprintf(f,"        .size .Larray_OOB_read_msg, .-.Larray_OOB_read_msg\n\n");
  fprintf(f,"        .type .Lstack_full_msg, %%object\n");
  fprintf(f,".Lstack_full_msg:\n");
  fprintf(f,"        .asciz \"Operand stack of %" PRIu32 " entries is full\"\n",MAX_NUMERIC_OPERANDS);
  fprintf(f,"        .size .Lstack_full_msg, .-.Lstack_full_msg\n\n");
  fprintf(f,"        .type .Lsstack_full_msg, %%object\n");
  fprintf(f,".Lsstack_full_msg:\n");
  fprintf(f,"        .asciz \"String stack of %" PRIu32 " entries is full\"\n",MAX_STRING_OPERANDS);
  fprintf(f,"        .size .Lsstack_full_msg, .-.Lsstack_full_msg\n\n");
  fprintf(f,"        .type .Lreturn_stack_full_msg, %%object\n");
  fprintf(f,".Lreturn_stack_full_msg:\n");
  fprintf(f,"        .asciz \"RETURN stack of %" PRIu32 " entries is full\"\n",MAX_GOSUB_DEPTH);
  fprintf(f,"        .size .Lreturn_stack_full_msg, .-.Lreturn_stack_full_msg\n\n");
  fprintf(f,"        .type .Lstack_empty_msg, %%object\n");
  fprintf(f,".Lstack_empty_msg:\n");
  fprintf(f,"        .asciz \"Operand stack empty\"\n");
  fprintf(f,"        .size .Lstack_empty_msg, .-.Lstack_empty_msg\n\n");
  fprintf(f,"        .type .Lsstack_empty_msg, %%object\n");
  fprintf(f,".Lsstack_empty_msg:\n");
  fprintf(f,"        .asciz \"String stack empty\"\n");
  fprintf(f,"        .size .Lsstack_empty_msg, .-.Lsstack_empty_msg\n\n");
  fprintf(f,"        .type .Lreturn_stack_empty_msg, %%object\n");
  fprintf(f,".Lreturn_stack_empty_msg:\n");
  fprintf(f,"        .asciz \"RETURN without GOSUB\"\n");
  fprintf(f,"        .size .Lreturn_stack_empty_msg, .-.Lreturn_stack_empty_msg\n\n");
  if (need_SQR) { // we need to include support for SQR() function
    fprintf(f,"        .type .Lnegative_sqrtf_msg, %%object\n");
    fprintf(f,".Lnegative_sqrtf_msg:\n");
    fprintf(f,"        .asciz \"SQR() is not valid for negative values\"\n");
    fprintf(f,"        .size .Lnegative_sqrtf_msg, .-.Lnegative_sqrtf_msg\n");
  }
  fprintf(f,"        .type .Loverflow_msg, %%object\n");
  fprintf(f,".Loverflow_msg:\n");
  fprintf(f,"        .asciz \"Arithmetic overflow\"\n");
  fprintf(f,"        .size .Loverflow_msg, .-.Loverflow_msg\n\n");
  fprintf(f,"        .type .Loverflow_exp_msg, %%object\n");
  fprintf(f,".Loverflow_exp_msg:\n");
  fprintf(f,"        .asciz \"Overflow in EXP function\"\n");
  fprintf(f,"        .size .Loverflow_exp_msg, .-.Loverflow_exp_msg\n\n");
  fprintf(f,"        .type .Lunderflow_msg, %%object\n");
  fprintf(f,".Lunderflow_msg:\n");
  fprintf(f,"        .asciz \"Arithmetic underflow\"\n");
  fprintf(f,"        .size .Lunderflow_msg, .-.Lunderflow_msg\n\n");
  fprintf(f,"        .type .Ldividebyzero_msg, %%object\n");
  fprintf(f,".Ldividebyzero_msg:\n");
  fprintf(f,"        .asciz \"Divide by zero\"\n");
  fprintf(f,"        .size .Ldividebyzero_msg, .-.Ldividebyzero_msg\n\n");
  fprintf(f,"        .type .Lbaddivide_msg, %%object\n");
  fprintf(f,".Lbaddivide_msg:\n");
  fprintf(f,"        .asciz \"Bad divide\"\n");
  fprintf(f,"        .size .Lbaddivide_msg, .-.Lbaddivide_msg\n\n");
  fprintf(f,"        .type .Lbadpower_msg, %%object\n");
  fprintf(f,".Lbadpower_msg:\n");
  fprintf(f,"        .asciz \"Bad power\"\n");
  fprintf(f,"        .size .Lbadpower_msg, .-.Lbadpower_msg\n\n");
  fprintf(f,"        .type .Lzero_to_negative_power_msg, %%object\n");
  fprintf(f,".Lzero_to_negative_power_msg:\n");
  fprintf(f,"        .asciz \"Zero raised to a negative power\"\n");
  fprintf(f,"        .size .Lzero_to_negative_power_msg, .-.Lzero_to_negative_power_msg\n\n");
  fprintf(f,"        .type .Luninitialized_forindex_msg, %%object\n");
  fprintf(f,".Luninitialized_forindex_msg:\n");
  fprintf(f,"        .asciz \"ICE:Read of uninitialized FOR loop index \"\n");
  fprintf(f,"        .size .Luninitialized_forindex_msg, .-.Luninitialized_forindex_msg\n\n");
  fprintf(f,"        .type .Luninitialized_forinc_msg, %%object\n");
  fprintf(f,".Luninitialized_forinc_msg:\n");
  fprintf(f,"        .asciz \"ICE:Read of uninitialized FOR loop increment\"\n");
  fprintf(f,"        .size .Luninitialized_forinc_msg, .-.Luninitialized_forinc_msg\n\n");
  fprintf(f,"        .type .Luninitialized_forlimit_msg, %%object\n");
  fprintf(f,".Luninitialized_forlimit_msg:\n");
  fprintf(f,"        .asciz \"ICE:Read of uninitialized FOR loop limit\"\n");
  fprintf(f,"        .size .Luninitialized_forlimit_msg, .-.Luninitialized_forlimit_msg\n\n");
  fprintf(f,"        .type .Luninitialized_msg, %%object\n");
  fprintf(f,".Luninitialized_msg:\n");
  fprintf(f,"        .asciz \"Read of uninitialized variable\"\n");
  fprintf(f,"        .size .Luninitialized_msg, .-.Luninitialized_msg\n\n");
  fprintf(f,"        .type .Luninitializedparm_msg, %%object\n");
  fprintf(f,".Luninitializedparm_msg:\n");
  fprintf(f,"        .asciz \"Read of uninitialized function parameter\"\n");
  fprintf(f,"        .size .Luninitializedparm_msg, .-.Luninitializedparm_msg\n\n");
  fprintf(f,"        .type .Luninitializedparm2_msg, %%object\n");
  fprintf(f,".Luninitializedparm2_msg:\n");
  fprintf(f,"        .asciz \" of function FN\"\n");
  fprintf(f,"        .size .Luninitializedparm2_msg, .-.Luninitializedparm2_msg\n\n");
  fprintf(f,"        .type .Luninitialized_element_msg, %%object\n");
  fprintf(f,".Luninitialized_element_msg:\n");
  fprintf(f,"        .asciz \"Read of uninitialized element in array\"\n");
  fprintf(f,"        .size .Luninitialized_element_msg, .-.Luninitialized_element_msg\n\n");
  fprintf(f,"        .type .Longotofail_msg, %%object\n");
  fprintf(f,".Longotofail_msg:\n");
  fprintf(f,"        .asciz \"ON .. GOTO expression value out of range\"\n");
  fprintf(f,"        .size .Longotofail_msg, .-.Longotofail_msg\n\n");
  fprintf(f,"        .type .Lstring_too_long_msg, %%object\n");
  fprintf(f,".Lstring_too_long_msg:\n");
  fprintf(f,"        .asciz \"String literal in assignment too long\"\n");
  fprintf(f,"        .size .Lstring_too_long_msg, .-.Lstring_too_long_msg\n\n");
  if (need_READ) { // need to include support for READ statement
    fprintf(f,"        .type .Lout_of_data_msg, %%object\n");
    fprintf(f,".Lout_of_data_msg:\n");
    fprintf(f,"        .asciz \"Out of data\"\n");
    fprintf(f,"        .size .Lout_of_data_msg, .-.Lout_of_data_msg\n\n");
    fprintf(f,"        .type .Lbad_string_read_msg, %%object\n");
    fprintf(f,".Lbad_string_read_msg:\n");
    fprintf(f,"        .asciz \"Bad string READ\"\n");
    fprintf(f,"        .size .Lbad_string_read_msg, .-.Lbad_string_read_msg\n\n");
    fprintf(f,"        .type .Lbad_number_read_msg, %%object\n");
    fprintf(f,".Lbad_number_read_msg:\n");
    fprintf(f,"        .asciz \"Bad number READ\"\n");
    fprintf(f,"        .size .Lbad_number_read_msg, .-.Lbad_number_read_msg\n\n");
  }
  if (need_INPUT) { // need to include support for INPUT statement
    fprintf(f,"        .type .LEOF_on_input_msg, %%object\n");
    fprintf(f,".LEOF_on_input_msg:\n");
    fprintf(f,"        .asciz \"EOF on input\"\n");
    fprintf(f,"        .size .LEOF_on_input_msg, .-.LEOF_on_input_msg\n\n");
    fprintf(f,"        .type .LERR_on_input_msg, %%object\n");
    fprintf(f,".LERR_on_input_msg:\n");
    fprintf(f,"        .asciz \"file error on input\"\n");
    fprintf(f,"        .size .LERR_on_input_msg, .-.LERR_on_input_msg\n\n");
    fprintf(f,"        .type .Linvalid_number_on_input_msg, %%object\n");
    fprintf(f,".Linvalid_number_on_input_msg:\n");
    fprintf(f,"        .asciz \"file error on input\"\n");
    fprintf(f,"        .size .Linvalid_number_on_input_msg, .-.Linvalid_number_on_input_msg\n\n");
  }

  fprintf(f,"        .section .data,\"aw\",@progbits\n\n");
  //
  // This is the stack pointer for the dedicated floating point stack used in
  // numeric expression evalauation at runtime by the generated program.
  //
  fprintf(f,"        .type .Loperand_stack_ptr, @object\n");
  fprintf(f,".Loperand_stack_ptr:\n");
  fprintf(f,"        .quad 0\n");
  fprintf(f,"        .size .Loperand_stack_ptr, .-.Loperand_stack_ptr\n\n");
  //
  // This is the dedicated floating point stack used in numeric expression 
  // evaluation at runtime by the generated program.
  //
  fprintf(f,"        .type .Loperand_stack, @object\n");
  fprintf(f,".Loperand_stack:\n");
  fprintf(f,"        .rept %" PRIu32 "\n",MAX_NUMERIC_OPERANDS);
  fprintf(f,"        .%s NaN    # (%" PRIu32 " byte units, actually %s)\n",(use_double?"quad":"long"),(use_double?UINT32_C(8):UINT32_C(4)),(use_double?"double":"float"));
  fprintf(f,"        .endr\n");
  fprintf(f,"        .size .Loperand_stack, .-.Loperand_stack\n\n");
  fprintf(f,"        .set operand_stack_size, (.-.Loperand_stack)>>2\n\n");

  //
  // This is the stack pointer for the dedicated string address stack used
  // in string expression evaluation at runtime by the generated program.
  //
  fprintf(f,"        .type .Lsoperand_stack_ptr, @object\n");
  fprintf(f,".Lsoperand_stack_ptr:\n");
  fprintf(f,"        .quad 0\n");
  fprintf(f,"        .size .Lsoperand_stack_ptr, .-.Lsoperand_stack_ptr\n\n");
  //
  // This is the dedicated string address stack used in string expression
  // evaluation at runtime by the generated program.
  //
  fprintf(f,"        .type .Lsoperand_stack, @object\n");
  fprintf(f,".Lsoperand_stack:\n");
  fprintf(f,"        .rept %" PRIu32 "\n",MAX_STRING_OPERANDS);
  fprintf(f,"        .quad 0      # (8 byte units, string addresses)\n");
  fprintf(f,"        .endr\n");
  fprintf(f,"        .size .Lsoperand_stack, .-.Lsoperand_stack\n");
  fprintf(f,"        .set soperand_stack_size, (.-.Lsoperand_stack)>>3\n\n");

  //
  // This is the stack pointer for the dedicated return address stack used
  // by GOSUB/RETURN at runtime by the generated program.
  //
  fprintf(f,"        .type .Lretaddr_stack_ptr, @object\n");
  fprintf(f,".Lretaddr_stack_ptr:\n");
  fprintf(f,"        .quad 0\n");
  fprintf(f,"        .size .Lretaddr_stack_ptr, .-.Lretaddr_stack_ptr\n\n");
  //
  // This is the dedicated return address stack used by GOSUB/RETURN
  // at runtime by the generated program.
  //
  fprintf(f,"        .type .Lretaddr_stack, @object\n");
  fprintf(f,".Lretaddr_stack:\n");
  fprintf(f,"        .rept %" PRIu32 "     # supported levels of nesting....\n",MAX_GOSUB_DEPTH);
  fprintf(f,"        .quad 0      # (8 byte units, return addresses)\n");
  fprintf(f,"        .endr\n");
  fprintf(f,"        .size .Lretaddr_stack, .-.Lretaddr_stack\n");
  fprintf(f,"        .set retaddr_stack_size, (.-.Lretaddr_stack)>>3\n\n");

  fprintf(f,"        .section .bss,\"aw\",@nobits\n\n");
  //
  // This is the line output buffer used by the PRINT statement.
  //
  fprintf(f,"        .align  1,0\n");
  fprintf(f,"        .type .LOUTBUFFER,@object\n");
  fprintf(f,".LOUTBUFFER:\n");
  fprintf(f,"        .skip 32\n");
  fprintf(f,"        .size .LOUTBUFFER,.-.LOUTBUFFER\n\n");
  //
  // This is the line input buffer used by the INPUT statement.
  //
  fprintf(f,"        .align  1,0\n");
  fprintf(f,"        .type .LINBUFFER,@object\n");
  fprintf(f,".LINBUFFER:\n");
  fprintf(f,"        .skip 1024\n");
  fprintf(f,"        .size .LINBUFFER,.-.LINBUFFER\n");
  fprintf(f,"        .set INBUFFER_size, 1024\n\n");

  //
  // This variable is used to hold the current Minimal BASIC 
  // statement line number, which is used for display in some
  // error and exception messages.
  //
  fprintf(f,"        .align  8,0\n");
  fprintf(f,"        .type .LCURLINENO,@object\n");
  fprintf(f,".LCURLINENO:\n");
  fprintf(f,"        .skip 8\n");
  fprintf(f,"        .size .LCURLINENO,.-.LCURLINENO\n\n");

  //
  // This variable is used to hold the original RSP (stack pointer)
  // register contents so they can be restored on program exit.
  //
  fprintf(f,"        .type   .Loldrsp,@object\n");
  fprintf(f,".Loldrsp:\n");
  fprintf(f,"        .skip 8\n");
  fprintf(f,"        .size   .Loldrsp, .-.Loldrsp\n\n");

  //
  // This is used to store the original SSE2 floating point unit
  // MXCSR status register contents so they can be restored on 
  // program exit.
  //
  fprintf(f,"        .type   .Loldmxcsr, @object\n");
  fprintf(f,"        .align 4\n");
  fprintf(f,".Loldmxcsr:\n");
  fprintf(f,"        .zero   4\n");
  fprintf(f,"        .size   .Loldmxcsr, .-.Loldmxcsr\n\n");

  //
  // Variable used by macros implementing arithmetic operators
  // for values on the dedicated floating point stack.  The
  // SIMD floating point unit MXCSR status register contents
  // will be saved here, and then they can be tested to see
  // if any exceptions (like overflow, etc.) have occurred.
  //
  fprintf(f,"        .type   .Lnewmxcsr, @object\n");
  fprintf(f,"        .align 4\n");
  fprintf(f,".Lnewmxcsr:\n");
  fprintf(f,"        .zero   4\n");
  fprintf(f,"        .size   .Lnewmxcsr, .-.Lnewmxcsr\n\n");

  //
  // This is a runtime flag that is used to indicate whether
  // input on STDIN should be echod on STDOUT or not.  When
  // reading on a TTY we should not echo (the tty will do that)
  // but when reading from a file or pipe we need to echo.
  //
  fprintf(f,"        .type   .Lecho_input, @object\n");
  fprintf(f,".Lecho_input:\n");
  fprintf(f,"        .zero   8\n");
  fprintf(f,"        .size   .Lecho_input, .-.Lecho_input\n\n");

  fprintf(f,"        .section .text,\"ax\",@progbits\n");
  fprintf(f,"        .globl  _start\n");
  fprintf(f,"        .type   main, @function\n");
  fprintf(f,"        .align 16, 0x90\n");

  //
  // This label is the actual entry point where program execution
  // will begin.
  //
  fprintf(f,"_start:\n");
  fprintf(f,"main:\n");
  fprintf(f,"        movabsq $.Loldrsp, %%rax\n");
  fprintf(f,"        movq    %%rsp, (%%rax)    # save stack pointer\n");
  fprintf(f,"# checks for required CPU features\n");
  fprintf(f,"        movabsq $havecmov,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        jnz     0f\n");
  fprintf(f,"        movq    $.Lnocmovlen, %%rdx\n");
  fprintf(f,"        movq    $STDERR, %%rdi\n");
  fprintf(f,"        movabsq $.Lnocmov, %%rsi\n");
  fprintf(f,"        movq    $STDERR, %%rdi\n");
  fprintf(f,"        movq    $__NR_write, %%rax\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"checkfail:\n");
  fprintf(f,"        movabsq $1, %%rdi  # return code indicates failure\n");
  fprintf(f,"        movabsq $__NR_exit, %%rax\n");
  fprintf(f,"        syscall           # does not return\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $havesse2,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        jnz     0f\n");
  fprintf(f,"        movq    $.Lnosse2len, %%rdx\n");
  fprintf(f,"        movq    $STDERR, %%rdi\n");
  fprintf(f,"        movabsq $.Lnosse2, %%rsi\n");
  fprintf(f,"        movq    $STDERR, %%rdi\n");
  fprintf(f,"        movq    $__NR_write, %%rax\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"        jmp     checkfail\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $havesse3,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        jnz     0f\n");
  fprintf(f,"        movq    $.Lnosse3len, %%rdx\n");
  fprintf(f,"        movq    $STDERR, %%rdi\n");
  fprintf(f,"        movabsq $.Lnosse3, %%rsi\n");
  fprintf(f,"        movq    $STDERR, %%rdi\n");
  fprintf(f,"        movq    $__NR_write, %%rax\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"        jmp     checkfail\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $havessse3,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        jnz     0f\n");
  fprintf(f,"        movq    $.Lnossse3len, %%rdx\n");
  fprintf(f,"        movq    $STDERR, %%rdi\n");
  fprintf(f,"        movabsq $.Lnossse3, %%rsi\n");
  fprintf(f,"        movq    $STDERR, %%rdi\n");
  fprintf(f,"        movq    $__NR_write, %%rax\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"        jmp     checkfail\n");
  fprintf(f,"0:\n");
  if (use_SSE4_1) {
    fprintf(f,"        movabsq $havesse4_1,%%rax\n");
    fprintf(f,"        callq   *%%rax\n");
    fprintf(f,"        jnz      0f\n");
    fprintf(f,"        movq    $.Lnosse4_1len, %%rdx\n");
    fprintf(f,"        movq    $STDERR, %%rdi\n");
    fprintf(f,"        movabsq $.Lnosse4_1, %%rsi\n");
    fprintf(f,"        movq    $STDERR, %%rdi\n");
    fprintf(f,"        movq    $__NR_write, %%rax\n");
    fprintf(f,"        syscall\n");
    fprintf(f,"        jmp     checkfail\n");
    fprintf(f,"0:\n");
  }
  if (use_AVX) {
    fprintf(f,"        movabsq $haveavx,%%rax\n");
    fprintf(f,"        callq   *%%rax\n");
    fprintf(f,"        jnz      0f\n");
    fprintf(f,"        movq    $.Lnoavxlen, %%rdx\n");
    fprintf(f,"        movq    $STDERR, %%rdi\n");
    fprintf(f,"        movabsq $.Lnoavx, %%rsi\n");
    fprintf(f,"        movq    $STDERR, %%rdi\n");
    fprintf(f,"        movq    $__NR_write, %%rax\n");
    fprintf(f,"        syscall\n");
    fprintf(f,"        jmp     checkfail\n");
    fprintf(f,"0:\n");
  }
  fprintf(f,"        # save original MXCSR\n");
  fprintf(f,"        stmxcsr .Loldmxcsr\n");
  fprintf(f,"        # put MXCSR into eax\n");
  fprintf(f,"        movabsq $.Loldmxcsr, %%rbx\n");
  fprintf(f,"        movl    (%%rbx), %%eax\n");
  fprintf(f,"        # enable masking, set RC to mode 0, disable FTZ and DAZ,\n");
  fprintf(f,"        # and clear any pending exceptions\n");
  fprintf(f,"        movw    $0x1F80, %%ax\n");
  fprintf(f,"        # save new MXCSR\n");
  fprintf(f,"        movabsq $.Lnewmxcsr, %%rbx\n");
  fprintf(f,"        movl    %%eax, (%%rbx)\n");
  fprintf(f,"        # load new MXCSR\n");
  fprintf(f,"        ldmxcsr .Lnewmxcsr\n");
  fprintf(f,"        # is STDIN a tty?\n");
  fprintf(f,"        movabsq $isterminal, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $.Lecho_input, %%rbx\n");
  fprintf(f,"        movq    %%rax, (%%rbx)\n");
  if (need_RND) // need to include support for RND() function
    write_init_RND(f);
  if (need_PRINT) { // need to include support for PRINT statement
    fprintf(f,"        # initialize output buffer\n");
    fprintf(f,"        movabsq $initbuf, %%rax\n");
    fprintf(f,"        callq   *%%rax\n");
  }
  // make call to initialize .bss section at runtime before main program
  // code begins.
  fprintf(f,"        movabsq $.Linit_bss,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fflush(f);
  return true;
}

//
// This function is used by the main driver to emit the assembly epilogue.
// The f parameter is the file handle for the output file, and the file must
// already be open for output.  The code emitted is adjusted depending on many
// of the flags in globals.c, the most significant of which is the use_double
// variable that when true specifies normal 64 bit math will be used, but when
// false specifies 32 bit math will be used instead.  Other flags used are
// need_SQR, need_READ, need_INPUT, need_RND, and need_PRINT.  This function
// returns true on success, false otherwise.
//
bool write_epilogue(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"write_epilogue(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"goodxit:\n"); // code jumps here to exit program successfully
  fprintf(f,"        xorq    %%rdi, %%rdi    # return code\n");
  fprintf(f,"        jmp     finalxit      # short jump is OK here\n");
  fprintf(f,"        .global badxit\n");
  fprintf(f,"badxit:\n");  // code jumps here to exit program with an error
  fprintf(f,"        movabsq $.Lon_line_number_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $.LCURLINENO, %%rax      # line number\n");
  fprintf(f,"        movl    (%%rax), %%eax\n");
  fprintf(f,"        movabsq $.LOUTBUFFER, %%rsi      # buffer\n");
  fprintf(f,"        movl    %%eax, %%edi\n");
  fprintf(f,"        movabsq $myitoa, %%rax\n");
  fprintf(f,"        callq   *%%rax           # call itoa(i,b)\n");
  fprintf(f,"        movabsq $.LOUTBUFFER, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $printnewline, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $1, %%rdi        # return code\n");
  fprintf(f,"finalxit:\n"); // code jumps here from goodxit, or falls through from badxit
                            // since this is common code we need for any kind of exit.
  if (need_INPUT) { // need to include support for INPUT statement
    fprintf(f,"        # drain STDIN\n");
    fprintf(f,"        movabsq $drain_stdin,%%rax\n");
    fprintf(f,"        callq   *%%rax\n");
  }
  fprintf(f,"        # restore original MXCSR\n");
  fprintf(f,"        ldmxcsr .Loldmxcsr\n");
  fprintf(f,"        movabsq $.Loldrsp, %%rax\n");
  fprintf(f,"        movq    (%%rax), %%rsp    # restore stack pointer\n");
  fprintf(f,"        movabsq $__NR_exit, %%rax\n");
  fprintf(f,"        syscall                 # does not return\n");

  // subroutine to initialize .bss values (if required; otherwise it does nothing)
  fprintf(f,".Linit_bss:\n");
  if (need_BSS_numeric) { // we have some numeric variables
    fprintf(f,"        # initialize .bss section numeric variables\n");
    fprintf(f,"        movabsq $.LSTART_BSS, %%rdi\n");
    fprintf(f,"        movabsq $BSS_SIZE/%c, %%rcx\n",(use_double?'8':'4'));
    fprintf(f,"        mov%s $SNaN,%%%cax\n",(use_double?"absq":"l   "),(use_double?'r':'e'));
    fprintf(f,"        cld\n");
    fprintf(f,"        rep stos%c\n",(use_double?'q':'l'));
  } else {
    fprintf(f,"        # .bss section has no numeric variables to initialize\n");
  }
  if (need_BSS_string) { // we have some string scalar variables
    fprintf(f,"        # initialize .bss section string variables\n");
    fprintf(f,"        movabsq $.LSTART2_BSS, %%rdi\n");
    fprintf(f,"        movabsq $BSS_SIZE2, %%rcx\n");
    fprintf(f,"        movb    $NAK,%%al\n");
    fprintf(f,"        cld\n");
    fprintf(f,"        rep stosb\n");
  } else {
    fprintf(f,"        # .bss section has no string variables to initialize\n");
  }
  fprintf(f,"        retq\n");

  // error messages
  fprintf(f,".Larray_out_of_bounds_write:\n");
  fprintf(f,"        movabsq $.Larray_OOB_write_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,".Larray_out_of_bounds_read:\n");
  fprintf(f,"        movabsq $.Larray_OOB_read_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,".Loperand_stack_full:\n");
  fprintf(f,"        movabsq $.Lstack_full_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,".Loperand_stack_empty:\n");
  fprintf(f,"        movabsq $.Lstack_empty_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,".Lsoperand_stack_full:\n");
  fprintf(f,"        movabsq $.Lsstack_full_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,".Lsoperand_stack_empty:\n");
  fprintf(f,"        movabsq $.Lsstack_empty_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,".Lretaddr_stack_full:\n");
  fprintf(f,"        movabsq $.Lreturn_stack_full_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,".Lretaddr_stack_empty:\n");
  fprintf(f,"        movabsq $.Lreturn_stack_empty_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  if (need_SQR) { // need to include support for SQR() function
    fprintf(f,".Lnegative_sqrtf:\n");
    fprintf(f,"        movabsq $.Lnegative_sqrtf_msg, %%rsi\n");
    fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
    fprintf(f,"        jmpq    *%%rax\n");
  }
  fprintf(f,".Loverflow:\n");
  fprintf(f,"        movabsq $.Loverflow_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,".Lbaddivide:\n");
  fprintf(f,"        movabsq $.Lbaddivide_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,".Lbadpower:\n");
  fprintf(f,"        movabsq $.Lbadpower_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,".Luninitialized:\n");
  fprintf(f,"        movabsq $.Luninitialized_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,".Lstring_too_long:\n");
  fprintf(f,"        movabsq $.Lstring_too_long_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  if (need_READ) { // need to include support for READ statement
    fprintf(f,".Lout_of_data:\n");
    fprintf(f,"        movabsq $.Lout_of_data_msg, %%rsi\n");
    fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
    fprintf(f,"        jmpq    *%%rax\n");
    fprintf(f,".Lbad_string_read:\n");
    fprintf(f,"        movabsq $.Lbad_string_read_msg, %%rsi\n");
    fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
    fprintf(f,"        jmpq    *%%rax\n");
    fprintf(f,".Lbad_number_read:\n");
    fprintf(f,"        movabsq $.Lbad_number_read_msg, %%rsi\n");
    fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
    fprintf(f,"        jmpq    *%%rax\n");
  }
  if (need_INPUT) { // need to include support for INPUT statement
    fprintf(f,".LEOF_on_input:\n");
    fprintf(f,"        movabsq $.LEOF_on_input_msg, %%rsi\n");
    fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
    fprintf(f,"        jmpq    *%%rax\n");
    fprintf(f,".LERR_on_input:\n");
    fprintf(f,"        movabsq $.LERR_on_input_msg, %%rsi\n");
    fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
    fprintf(f,"        jmpq    *%%rax\n");
    fprintf(f,".Linvalid_number_on_input:\n");
    fprintf(f,"        movabsq $.Linvalid_number_on_input_msg, %%rsi\n");
    fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
    fprintf(f,"        jmpq    *%%rax\n");
  }

  // code to print an error message, then exit the program indicating error
  fprintf(f,".Lprint_error_message:\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"        .size main, .-main\n\n");

  // procedure to report a non-fatal exception
  fprintf(f,"        # incoming argument is address of buffer in rsi\n");
  fprintf(f,"        # procedure, no return value\n");
  fprintf(f,"        .type exception_non_fatal, @function\n");
  fprintf(f,"exception_non_fatal:\n");
  fprintf(f,"        pushq   %%rbp\n");
  fprintf(f,"        movq    %%rsp, %%rbp\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        subq    $8,%%rsp      # this 8 + following push of rax maintains 16 byte alignment\n");
  fprintf(f,"        pushq   %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $.Lon_line_number_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $.LCURLINENO, %%rax      # line number\n");
  fprintf(f,"        movq    (%%rax), %%rdi\n");
  fprintf(f,"        movabsq $.LOUTBUFFER, %%rsi      # buffer\n");
  fprintf(f,"        movabsq $myitoa, %%rax\n");
  fprintf(f,"        callq   *%%rax           # call itoa(i,b)\n");
  fprintf(f,"        movabsq $.LOUTBUFFER, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $printnewline, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        popq    %%rax\n");
  fprintf(f,"        movq    %%rbp,%%rsp\n");
  fprintf(f,"        popq    %%rbp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size exception_non_fatal, .-exception_non_fatal\n\n");

  // function to determine if SSE2 instructions are supported
  fprintf(f,"        # no args, returns 1 in rax if SSE2 detected, 0 otherwise\n");
  fprintf(f,"        .type   havesse2, @function\n");
  fprintf(f,"havesse2:\n");
  fprintf(f,"        mov    $0x1,%%eax\n");
  fprintf(f,"        cpuid\n");
  fprintf(f,"        xorq   %%rax,%%rax\n");
  fprintf(f,"        andl   $0x4000000,%%edx\n");
  fprintf(f,"        setz   %%ah\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size havesse2, .-havesse2\n\n");

  // function to determine if SSE3 instructions are supported
  fprintf(f,"        # no args, returns 1 in rax if SSE3 detected, 0 otherwise\n");
  fprintf(f,"        .type   havesse3, @function\n");
  fprintf(f,"havesse3:\n");
  fprintf(f,"        mov    $0x1,%%eax\n");
  fprintf(f,"        cpuid\n");
  fprintf(f,"        xorq   %%rax,%%rax\n");
  fprintf(f,"        andl   $0x1,%%ecx\n");
  fprintf(f,"        setz   %%ah\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size havesse3, .-havesse3\n\n");

  // function to determine if SSSE3 instructions are supported
  fprintf(f,"        # no args, returns 1 in rax if SSSE3 detected, 0 otherwise\n");
  fprintf(f,"        .type   havessse3, @function\n");
  fprintf(f,"havessse3:\n");
  fprintf(f,"        mov    $0x1,%%eax\n");
  fprintf(f,"        cpuid\n");
  fprintf(f,"        xorq   %%rax,%%rax\n");
  fprintf(f,"        andl   $0x200,%%ecx\n");
  fprintf(f,"        setz   %%ah\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size havessse3, .-havessse3\n\n");

  if (use_SSE4_1) {
    // function to determine if SSE4.1 instructions are supported
    fprintf(f,"        # no args, returns 1 in rax if SSE4.1 detected, 0 otherwise\n");
    fprintf(f,"        .type   havesse4_1, @function\n");
    fprintf(f,"havesse4_1:\n");
    fprintf(f,"        mov    $0x1,%%eax\n");
    fprintf(f,"        cpuid\n");
    fprintf(f,"        xorq   %%rax,%%rax\n");
    fprintf(f,"        andl   $0x80000,%%ecx\n");
    fprintf(f,"        setz   %%ah\n");
    fprintf(f,"        retq\n");
    fprintf(f,"        .size havesse4_1, .-havesse4_1\n\n");
  }
  if (use_AVX) {
    fprintf(f,"	# no args, returns 1 in rax if AVX detected, 0 otherwise\n");
    fprintf(f,"	.type  haveavx, @function\n");
    fprintf(f,"haveavx:\n");
    fprintf(f,"        # first check hardware support\n");
    fprintf(f,"        mov    $0x1,%%eax\n");
    fprintf(f,"        cpuid\n");
    fprintf(f,"        andl   $0x018000000, %%ecx # AVX and OXSAVE hardware support?\n");
    fprintf(f,"        setz   %%ah\n");
    fprintf(f,"        jnz    0f\n");
    fprintf(f,"        # now check O/S support\n");
    fprintf(f,"        xorl   %%ecx,%%ecx\n");
    fprintf(f,"        xgetbv\n");
    fprintf(f,"        andl   $0x6,%%eax   # O/S has both XMM and YMM state support?\n");
    fprintf(f,"        cmpl   $0x6,%%eax\n");
    fprintf(f,"        setz   %%ah\n");
    fprintf(f,"0:\n");
    fprintf(f,"	retq\n");
    fprintf(f,"	.size haveavx, .-haveavx\n\n");
  }

  // function to determine if CMOV is supported
  fprintf(f,"        # no args, returns 1 in rax if CMOV detected, 0 otherwise\n");
  fprintf(f,"        .type   havecmov, @function\n");
  fprintf(f,"havecmov:\n");
  fprintf(f,"        mov    $0x1,%%eax\n");
  fprintf(f,"        cpuid\n");
  fprintf(f,"        xorq   %%rax,%%rax\n");
  fprintf(f,"        andl   $0x08000000,%%edx\n");
  fprintf(f,"        setz   %%ah\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size havecmov, .-havecmov\n\n");

  // function (with related variable) to determine if STDOUT is a terminal (tty) or not
  fprintf(f,"        .section .bss,\"aw\",@nobits\n\n");
  fprintf(f,"        .balign 4\n");
  fprintf(f,"        .type   buf, @object\n");
  fprintf(f,"        .size   buf, 144\n");
  fprintf(f,"buf:\n");
  fprintf(f,"        .skip   144\n\n");
  fprintf(f,"        .section .text,\"ax\",@progbits\n\n");
  fprintf(f,"        # no arguments, returns value in rax, 0 for not a tty, 1 for a tty\n");
  fprintf(f,"        .type   isterminal, @function\n");
  fprintf(f,"isterminal:\n");
  fprintf(f,"        pushq   %%rbp\n");
  fprintf(f,"        movq    %%rsp, %%rbp\n");
  fprintf(f,"        movq    $STDIN, %%rdi    # rdi = STDIN\n");
  fprintf(f,"        movabsq $buf, %%rsi      # rsi = buf\n");
  fprintf(f,"        movabsq $__NR_fstat, %%rax       # rax = __NR_fstat\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"        movabsq $buf, %%rax      # rax = buf\n");
  fprintf(f,"        movl    24(%%rax), %%eax  # eax = buf.st_mode\n");
  fprintf(f,"        andl    $S_IFMT, %%eax   # eax &= S_IFMT\n");
  fprintf(f,"        cmpl    $S_IFCHR, %%eax  # al = (eax==S_IFCHR)\n");
  fprintf(f,"        sete    %%al\n");
  fprintf(f,"        movzbl  %%al, %%eax       # eax = al (return value)\n");
  fprintf(f,"        movq    %%rbp, %%rsp\n");
  fprintf(f,"        popq    %%rbp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   isterminal, .-isterminal\n\n");

  // procedure (with related constant) to print a newline to STDOUT
  fprintf(f,"        .section .rodata.str1.1,\"aMS\",@progbits,1\n\n");
  fprintf(f,"        .type .LNL, @object\n");
  fprintf(f,".LNL:\n");
  fprintf(f,"        .asciz \"\\n\"\n");
  fprintf(f,"        .size .LNL, .-.LNL\n\n");
  fprintf(f,"        .section .text,\"ax\",@progbits\n");
  fprintf(f,"        .type printnewline, @function\n");
  fprintf(f,"printnewline:\n");
  fprintf(f,"        movq    $1, %%rdx        # rdx = 1 (number of bytes to write, do NOT write NULL)\n");
  fprintf(f,"        movabsq $.LNL, %%rsi     # rsi = address of buffer\n");
  fprintf(f,"        movabsq $STDOUT, %%rdi\n");
  fprintf(f,"        movabsq $__NR_write, %%rax\n");
  fprintf(f,"        syscall                 # call kernel\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size printnewline, .-printnewline\n\n");

  // procedure to print an ASCIIZ string directly to STDOUT, bypassing the PRINT zone system
  fprintf(f,"        .type printstring, @function\n");
  fprintf(f,"printstring:\n");
  fprintf(f,"        # incoming argument is address of buffer in rsi\n");
  fprintf(f,"        pushq   %%rbp\n");
  fprintf(f,"        movq    %%rsp, %%rbp\n");
  fprintf(f,"        movq    %%rsi, %%rdi\n");
  fprintf(f,"        movabsq $mystrlen, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        # rsi still has address of buffer\n");
  fprintf(f,"        movq    %%rax, %%rdx      # rdx = number of bytes in buffer\n");
  fprintf(f,"        movabsq $STDOUT, %%rdi\n");
  fprintf(f,"        movabsq $__NR_write, %%rax\n");
  fprintf(f,"        syscall                 # call kernel\n");
  fprintf(f,"        movq    %%rbp,%%rsp\n");
  fprintf(f,"        popq    %%rbp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size printstring, .-printstring\n\n");

  // alternate procedure to print an ASCIIZ string to STDOUT, bypassing the PRINT zone system
  fprintf(f,"        .type printstring2, @function\n");
  fprintf(f,"        # incoming argument is address of buffer in rdi\n");
  fprintf(f,"        # this procedure returns nothing\n");
  fprintf(f,"printstring2:\n");
  fprintf(f,"        movabsq $mystrlen, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movq    %%rdi, %%rsi      # rsi = address of buffer\n");
  fprintf(f,"        movq    %%rax, %%rdx      # rdx = number of bytes in buffer\n");
  fprintf(f,"        movabsq $STDOUT, %%rdi    # rdi = file number \n");
  fprintf(f,"        movabsq $__NR_write, %%rax\n");
  fprintf(f,"        syscall                 # call kernel\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size printstring2, .-printstring2\n\n");

  // boolean function to determine whether a double value in xmm0 is a number or not
  fprintf(f,"        # input is double value in xmm0\n");
  fprintf(f,"        # return is integer in rax\n");
  fprintf(f,"        # 1 means xmm0 was a NaN value\n");
  fprintf(f,"        # 0 means xmm0 was not a NaN value\n");
  fprintf(f,"        # trashes rax,rcx\n");
  fprintf(f,"        .type   myisnan, @function\n");
  fprintf(f,"myisnan:\n");
  fprintf(f,"        movd    %%xmm0, %%rcx\n");
  fprintf(f,"        movq    %%rcx, %%rax\n");
  fprintf(f,"        shrq    $32, %%rax\n");
  fprintf(f,"        andl    $0x7FFFFFFF, %%eax\n");
  fprintf(f,"        testl   %%ecx, %%ecx\n");
  fprintf(f,"        setne   %%cl\n");
  fprintf(f,"        movzbl  %%cl, %%ecx\n");
  fprintf(f,"        addl    %%eax, %%ecx\n");
  fprintf(f,"        cmpl    $0x7FF00000, %%ecx\n");
  fprintf(f,"        seta    %%al\n");
  fprintf(f,"        movzbl  %%al, %%eax\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size myisnan, .-myisnan\n\n");

  // boolean function to determine whether a double value in xmm0 is a infinite or not
  fprintf(f,"        # input is double value in xmm0\n");
  fprintf(f,"        # return is integer in rax\n");
  fprintf(f,"        # 1 means xmm0 was an Inf value\n");
  fprintf(f,"        # 0 means xmm0 was not an Inf value\n");
  fprintf(f,"        # trashes rax,rcx,rdx,rsi\n");
  fprintf(f,"        .type   myisinf,@function\n");
  fprintf(f,"myisinf:\n");
  fprintf(f,"        movd    %%xmm0, %%rax\n");
  fprintf(f,"        movabsq $0x7FFFFFFF00000000,%%rdx\n");
  fprintf(f,"        andq    %%rax, %%rdx\n");
  fprintf(f,"        xorb    %%cl, %%cl\n");
  fprintf(f,"        movabsq $0x7FF0000000000000,%%rsi\n");
  fprintf(f,"        cmpq    %%rsi, %%rdx\n");
  fprintf(f,"        jne     0f\n");
  fprintf(f,"        testl   %%eax, %%eax\n");
  fprintf(f,"        sete    %%cl\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movzbl  %%cl, %%eax\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   myisinf, .-myisinf\n\n");

  //
  // function to compute floor() of the value at the top of the dedicated floating point stack
  //
  fprintf(f,"        .section .bss,\"aw\",@nobits\n\n");
  fprintf(f,"        .type .Lmxcsr, @object\n");
  fprintf(f,".Lmxcsr:\n");
  fprintf(f,"        .zero 4\n");
  fprintf(f,"        .size .Lmxcsr, .-.Lmxcsr\n\n");
  fprintf(f,"        .type .Lmxcsrtemp, @object\n");
  fprintf(f,".Lmxcsrtemp:\n");
  fprintf(f,"        .zero 4\n");
  fprintf(f,"        .size .Lmxcsrtemp, .-.Lmxcsrtemp\n\n");
  fprintf(f,"        .section .text,\"ax\",@progbits\n");
  fprintf(f,"        # incoming data is float on top of operand stack\n");
  fprintf(f,"        # function will calculate floor() of that value and\n");
  fprintf(f,"        # replace the original value on top of the stack with\n");
  fprintf(f,"        # the new value.\n");
  fprintf(f,"        # trashes r13, r14, r15, xmm0\n");
  fprintf(f,"        .type   intf, @function\n");
  fprintf(f,"intf:\n");
  fprintf(f,"        pushq   %%rbp\n");
  fprintf(f,"        movq    %%rsp, %%rbp\n");
  fprintf(f,"        popxmm  0\n");
  fprintf(f,"        movabsq $.Lmxcsr, %%r13  # save the original mxcsr\n");
  fprintf(f,"        stmxcsr (%%r13)          # into .Lmxcsr for later\n");
  fprintf(f,"        movl    (%%r13), %%r13d   # and put a copy of it into r13d\n");
  fprintf(f,"        movabsq $.Lmxcsrtemp, %%r14      # put a copy of it into .Lmxcsrtemp\n");
  fprintf(f,"        stmxcsr (%%r14)          # put current rounding mode into .Lmxcsrtemp\n");
  fprintf(f,"        movl    (%%r14), %%r13d   # put current rounding mode into ebx\n");
  fprintf(f,"        andl    $0xffff9fff, %%r13d      # clear rouding bits 13/14\n");
  fprintf(f,"        orl     $0x00002000, %%r13d      # set rounding down\n");
  fprintf(f,"        movl    %%r13d, (%%r14)   # save new rounding mode into .Lmxcsrtemp\n");
  fprintf(f,"        ldmxcsr (%%r14)          # load new rounding mode\n");
  fprintf(f,"        cvts%c2si %%xmm0, %%r13\n",(use_double?'d':'s'));
  fprintf(f,"        cvtsi2s%c %%r13, %%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        movabsq $.Lmxcsr, %%r13  # load the original mxcsr\n");
  fprintf(f,"        ldmxcsr (%%r13)          # from .Lmxcsr to restore it\n");
  fprintf(f,"        pushxmm 0\n");
  fprintf(f,"        movq    %%rbp,%%rsp\n");
  fprintf(f,"        popq    %%rbp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size intf, .-intf\n\n");
  //
  // A custom implementation of the well-known C standard library memcpy() function
  // This is needed by David M. Gay's ASCIIZ<-->double conversion code
  //
  fprintf(f,"        .globl  memcpy\n");
  fprintf(f,"        # buffers pointed to by rdi and rsi MUST NOT OVERLAP\n");
  fprintf(f,"        # input:\n");
  fprintf(f,"        # rdi is address of destination\n");
  fprintf(f,"        # rsi is address of source\n");
  fprintf(f,"        # rdx is number of bytes to copy\n");
  fprintf(f,"        # output:\n");
  fprintf(f,"        # rax is address of destination\n");
  fprintf(f,"        .type   memcpy,@function\n");
  fprintf(f,"memcpy:\n");
  fprintf(f,"        movq    %%rdi, %%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        testq   %%rdx, %%rdx\n");
  fprintf(f,"        jz      1f\n");
  fprintf(f,"        movb    (%%rsi), %%cl\n");
  fprintf(f,"        incq    %%rsi\n");
  fprintf(f,"        movb    %%cl, (%%rdi)\n");
  fprintf(f,"        incq    %%rdi\n");
  fprintf(f,"        decq    %%rdx\n");
  fprintf(f,"        jmp     0b\n");
  fprintf(f,"1:\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   memcpy, .-memcpy\n\n");

  // A custom implementation of the well-known C standard library function strlen()
  fprintf(f,"        # on input, address of buffer in rdi\n");
  fprintf(f,"        # output is size in bytes in rax\n");
  fprintf(f,"        .type   mystrlen,@function\n");
  fprintf(f,"mystrlen:\n");
  fprintf(f,"        cld\n");
  fprintf(f,"        xorq    %%rax, %%rax      # length=0\n");
  fprintf(f,"0:\n");
  fprintf(f,"        cmpb    $0, (%%rdi,%%rax) # if %%rax+%%rdi==0?\n");
  fprintf(f,"        je      1f              # goto 1\n");
  fprintf(f,"        incq    %%rax            # rax=rax+1\n");
  fprintf(f,"        jmp     0b              # goto 0, short jump OK here\n");
  fprintf(f,"1:\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   mystrlen, .-mystrlen\n\n");

  // A custom implementation of the well-known C standard library function itoa()
  // with it's required assocated constant
  fprintf(f,"        .section .rodata.cst8,\"aM\",@progbits,8\n\n");
  fprintf(f,"        .align  8\n");
  fprintf(f,".Lmyitoa_minneg:\n"); // minimum 64bit negative number
  fprintf(f,"        .quad -9223372036854775808\n");
  fprintf(f,"        .equ .Lmyitoa_minneg_length, .-.Lmyitoa_minneg\n\n");

  fprintf(f,"        .section .text,\"ax\",@progbits\n\n");

  fprintf(f,"        .type   myitoa, @function\n\n");
  fprintf(f,"#\n");
  fprintf(f,"# Integer To Ascii\n");
  fprintf(f,"# convert signed integer to ASCIIZ\n");
  fprintf(f,"# expects caller to pass the 64bit signed integer value in rdi,\n");
  fprintf(f,"# and pass the 64bit address of the output buffer in rsi.\n");
  fprintf(f,"# Nothing is returned.\n");
  fprintf(f,"# WARING: buffer size must be >= 22 bytes!\n");
  fprintf(f,"#\n");
  fprintf(f,"myitoa:\n");
  fprintf(f,"        pushq   %%rbp\n");
  fprintf(f,"        movq    %%rsp, %%rbp\n");
  fprintf(f,"        subq    $30, %%rsp\n");
  fprintf(f,"        andq    $-16, %%rsp      # force align the stack on 16-byte\n");
  fprintf(f,"        subq    $8, %%rsp        # 8 bytes of padding\n");
  fprintf(f,"        pushfq                  # 8 bytes of flags, maintaining alignment\n");
  fprintf(f,"#\n");
  fprintf(f,"# number is in 1st argument, treated as signed, must be in range\n");
  fprintf(f,"# -9223372036854775808..9223372036854775807\n");
  fprintf(f,"#\n");
  fprintf(f,"        movq    %%rsi, %%rbx      # the buffer address\n");
  fprintf(f,"        movq    %%rdi, %%rdx      # the 64 bit signed integer value\n");
  fprintf(f,"        cmp     $0, %%rdx        # ok, 0 will be sign extended\n");
  fprintf(f,"        jge     0f\n");
  fprintf(f,"        movq $1, -8(%%rbp)       # negative\n");
  fprintf(f,"        neg     %%rdx\n");
  fprintf(f,"        jno     1f\n");
  fprintf(f,"# number is -9223372036854775808, special case it\n");
  fprintf(f,"        movabsq $.Lmyitoa_minneg, %%rsi\n");
  fprintf(f,"        movq    %%rbx, %%rdi\n");
  fprintf(f,"        movabsq $.Lmyitoa_minneg_length, %%rcx\n");
  fprintf(f,"        cld\n");
  fprintf(f,"        rep     movsb\n");
  fprintf(f,"        jmp     4f\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movq $0, -8(%%rbp)       # negative\n");
  fprintf(f,"1:\n");
  fprintf(f,"#\n");
  fprintf(f,"#  put the digits in a buffer (they are calculated backwards)\n");
  fprintf(f,"#\n");
  fprintf(f,"        xorq    %%rcx, %%rcx      # initialize the counter to zero\n");
  fprintf(f,"        lea     -22(%%rbp), %%rdi # point to the buffer\n");
  fprintf(f,"2:\n");
  fprintf(f,"        mov     %%rdx, %%rax      # numerator\n");
  fprintf(f,"        cqto\n");
  fprintf(f,"        movabsq $10, %%r8        # divisor of 10\n");
  fprintf(f,"        idivq   %%r8\n");
  fprintf(f,"        xchg    %%rax, %%rdx      # get quotient\n");
  fprintf(f,"        addb    $48, %%al        # convert to ASCII digit\n");
  fprintf(f,"        movb    %%al, (%%rdi)     # put in buffer\n");
  fprintf(f,"        incq    %%rdi            # move to next byte in output buffer\n");
  fprintf(f,"        incq    %%rcx            # count the digit\n");
  fprintf(f,"        cmp     $0, %%rdx        # done?\n");
  fprintf(f,"        jnz     2b\n");
  fprintf(f,"#\n");
  fprintf(f,"# copy the reversed tbuff buffer to a non-reversed global output buffer\n");
  fprintf(f,"# specified by bx\n");
  fprintf(f,"#\n");
  fprintf(f,"        movq    %%rdi, %%rsi\n");
  fprintf(f,"        movq    %%rbx, %%rdi\n");
  fprintf(f,"        # deal with leading sign if necessary\n");
  fprintf(f,"        cmpq    $0, -8(%%rbp)\n");
  fprintf(f,"        je      3f\n");
  fprintf(f,"        movb    $'-',(%%rdi)\n");
  fprintf(f,"        incq    %%rdi\n");
  fprintf(f,"3:\n");
  fprintf(f,"        decq    %%rsi            # back up through the buffer\n");
  fprintf(f,"        movb    (%%rsi),%%al      # get byte from inbuffer\n");
  fprintf(f,"        movb    %%al,(%%rdi)      # put in outbuffer\n");
  fprintf(f,"        incq    %%rdi            # move to next byte of outbuffer\n");
  fprintf(f,"        decq    %%rcx\n");
  fprintf(f,"        jnz     3b\n");
  fprintf(f,"4:\n");
  fprintf(f,"        movb    $0, (%%rdi)      # append string terminator\n");
  fprintf(f,"        popfq\n");
  fprintf(f,"        movq    %%rbp, %%rsp\n");
  fprintf(f,"        popq    %%rbp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   myitoa, .-myitoa\n\n");

  // include support for runtime BASIC functions if it is needed
  if (need_RND) write_RND(f); // need to include support for RND() function
  if (need_SIN) write_SIN(f); // need to include support for SIN() function
  if (need_COS) write_COS(f); // need to include support for COS() function
  if (need_TAN) write_TAN(f); // need to include support for TAN() function
  if (need_ATN) write_ATN(f); // need to include support for ATN() function
  if (need_LOG) write_LOG(f); // need to include support for LOG() function
  if (need_EXP) write_EXP(f); // need to include support for EXP() function
  if (need_POW) write_POW(f); // need to include support for POW() function
  if (need_strcpy) write_strcpy(f); // need to include support for strcpy() function
  if (need_strcmp) write_strcmp(f); // need to include support for strcmp() function
  if (need_INPUT) write_INPUT(f); // need to include support for INPUT statement
  write_printzone(f);        // always need it, used to support PRINT statment
  dgay_code(f);              // always need it, used to convert float<-->ASCIIZ
  fflush(f);
  return true;
}

//
// Function to emit code to push a float literal value onto the dedicated floating
// point stack.  The f paramter is the file handle for the already-open output
// assembly language file.  The flitval is an ASCIIZ string containing the actual
// literal value for including in the generated comments.  The flitname parameter 
// is the assembly identifier for the floating point literal value's address.
// Returns true on success, false on failure.
//
bool push_float_literal(FILE *f, const char *flitname,const char *flitval) {
  if (NULL==f) {
    fprintf(stderr,"push_float_literal(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==flitname) || (0==flitname[0])) {
    fprintf(stderr,"push_float_literal(): float literal parameter is NULL or empty\n");
    return false;
  }
  fprintf(f,"        floatlit_to_floatreg %s,%%xmm0,%s\n",flitname,(flitval[0]==' '?flitval+1:flitval));
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// Function to emit code to push a string literal value onto the dedicated string
// address stack.  The f paramter is the file handle for the already-open output
// assembly language file.  The txt is an ASCIIZ string containing the actual
// literal value for including in the generated comments.  The slitname parameter 
// is the assembly identifier for the string literal value's address.
// Returns true on success, false on failure.
//
bool push_string_literal(FILE *f, const char *slitname, const char *txt) {
  if (NULL==f) {
    fprintf(stderr,"push_string_literal(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==slitname) || (0==slitname[0])) {
    fprintf(stderr,"push_string_literal(): string literal parameter is NULL or empty\n");
    return false;
  }
  fprintf(f,"        movabsq $%s, %%rdi  # %%rdi=pointer to '%s'\n",slitname,txt);
  fprintf(f,"        pushsaddr\n");
  fflush(f);
  return true;
}

//
// Function to emit code to push a scalar numeric variable's value onto the dedicated floating
// point stack.  The f paramter is the file handle for the already-open output
// assembly language file.  The varname parameter is the assembly identifier for the 
// numeric variable.
// Returns true on success, false on failure.
//
bool push_nvar(FILE *f, const char *varname) {
  if (NULL==f) {
    fprintf(stderr,"push_nvar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"push_nvar(): varname parameter is NULL or empty\n");
    return false;
  }
  if (varname[1]!=0) { // if this is a 2-byte numeric variable like A0 or Z9...
    fprintf(f,"        floatmem_to_floatreg %s,%%xmm0,'%c','%c'\n",varname,varname[0],varname[1]);
  } else {
    fprintf(f,"        floatmem_to_floatreg %s,%%xmm0,'%c'\n",varname,varname[0]);
  }
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// Function to emit code to push a numeric parameter for a UDF onto the 
// dedicated floating point stack.
// The f paramter is the file handle for the already-open output assembly 
// language file.  The varname parameter is the assembly identifier for the 
// scalar numeric variable.  The fname is a variable that contains the
// UDF name for use in error messages.
// Returns true on success, false on failure.
//
bool push_nparm(FILE *f, const char *varname, const char *fname) {
  if (NULL==f) {
    fprintf(stderr,"push_nparm(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"push_nparm(): varname parameter is NULL or empty\n");
    return false;
  }
  fprintf(f,"        floatmem_to_floatreg %s,%%xmm0,'%c'\n",varname,varname[6]);
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// Function to emit code to push a string variable's address onto the dedicated string
// address stack.  The f paramter is the file handle for the already-open output
// assembly language file.  The varname parameter is the assembly identifier for the 
// string variable.
// Returns true on success, false on failure.
//
bool push_svar(FILE *f, const char *varname) {
  if (NULL==f) {
    fprintf(stderr,"push_svar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"push_svar(): varname parameter is NULL or empty\n");
    return false;
  }
  fprintf(f,"        movabsq $%s, %%rdi\n",varname);
  fprintf(f,"        cmpb    $NAK,(%%rdi)\n");
  fprintf(f,"        jne     0f\n");
  fprintf(f,"        # referenced scalar was uninitialized\n");
  fprintf(f,"        movabsq $.Luninitialized_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        # must create string in reverse order\n");
  fprintf(f,"        xorq    %%rdi,%%rdi\n");
  fprintf(f,"        movb    $'%c',%%dil\n",varname[1]);
  fprintf(f,"        shl     $8,%%rdi\n");
  fprintf(f,"        movb    $'%c',%%dil\n",varname[0]);
  fprintf(f,"        shl     $8,%%rdi\n");
  fprintf(f,"        movb    $32,%%dil\n");
  fprintf(f,"        movabsq $printvarname,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        pushsaddr\n");
  fflush(f);
  return true;
}

//
// Function to emit code to push an array numeric variable's value onto the
// dedicated floating point stack.  The f paramter is the file handle for the
// already-open output assembly language file.  The varname parameter is the
// assembly identifier for the numeric variable.  The maxindex parameter is the
// highest permitted index value for the array, and is used to implement array
// bounds checking.  The index for a one-dimensional array will be on the top
// of the dedicated floating point stack already.  This compiler generates code
// that stores arrays in row-major order.  Returns true on success, false on
// failure.  The base_is_one flag is used to adjust the bounds checking.  If
// base_is_one is true, the minimum index is one, otherwise the minimum index
// is zero.
//
bool push_1D_navar(FILE *f, const char *varname, const uint32_t maxindex) {
  if (NULL==f) {
    fprintf(stderr,"push_1D_navar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"push_1D_navar(): varname parameter is NULL or empty\n");
    return false;
  }
  // one dimensional array reference
  // top_of_stack should have index expression as a float
  fprintf(f,"        popxmm 0\n");
  fprintf(f,"        array_1D_to_floatreg %s,%%xmm0,'%c',%" PRIu32 ",%" PRIu32 ",%%r8,%%r15\n",varname,varname[0],(base_is_one?UINT32_C(1):UINT32_C(0)),maxindex);
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// Function to emit code to push a 2 dimensional array numeric variable's value
// onto the dedicated floating point stack.  The f paramter is the file handle
// for the already-open output assembly language file.  The varname parameter
// is the assembly identifier for the numeric variable.  The rows parameter is
// the highest permitted index for the first dimension.  The columns parameter
// is the highest permitted index for the second dimension.  The rows and
// columns are used to implement array bounds checking.  Before calling this,
// the column index must be on the top of the stack, and the row index must be
// just below it in the stack.  This compiler generates code that stores arrays
// in row-major order.  Returns true on success, false on failure.  The
// base_is_one flag is used to adjust the bounds checking.  If base_is_one is
// true, the minimum index is one, otherwise the minimum index is zero.
//
bool push_2D_navar(FILE *f, const char *varname, const uint32_t rows, const uint32_t columns) {
  if (NULL==f) {
    fprintf(stderr,"push_2D_navar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"push_2D_navar(): varname parameter is NULL or empty\n");
    return false;
  }
  // two dimensional array reference
  // top_of_stack should have column index expression as a float
  // top_of_stack-1 should have row index expression as a float
  // we must use rax since mul is hard-wired to use rdx:rax
  fprintf(f,"        popxmm  1 # column index\n");
  fprintf(f,"        popxmm  0 # row index\n");
  fprintf(f,"        array_2D_to_floatreg %s,%%xmm0,%%xmm1,'%c',%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%%r8,%%r9,%%r15\n",
     varname,varname[0],(base_is_one?UINT32_C(1):UINT32_C(0)),rows,columns);
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// This function will emit code to change the sign of the value at the
// top of the dedicated floating point stack.  This effect is
// sematically like this:
// POP TO REGSITER
// REGISTER=0-REGISTER
// PUSH FROM REGISTER
// Returns true on success, false on failure.
//
bool donegate(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"donegate(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        unary_minus\n");
  fflush(f);
  return true;
}

//
// This function will emit code to add the top two values
// on the dedicated floating point stack and replace them
// with their sum.  This effect is semantically like this:
// POP TO REGISTER2
// POP TO REGISTER1
// REGISTER3=REGISTER1+REGISTER2
// PUSH REGISTER3
// Returns true on success, false on failure.
//
bool doadd(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"doadd(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        binary_add\n");
  fflush(f);
  return true;
}

//
// This function will emit code to subtract the value on top
// of the dedicated floating point stack from the value just
// below it and replace those two values with their difference.
// This effect is semantically like this:
// POP TO REGISTER2
// POP TO REGISTER1
// REGISTER3=REGISTER1-REGISTER2
// PUSH REGISTER3
// Returns true on success, false on failure.
//
bool dosubtract(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dosubtract(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        binary_subtract\n");
  fflush(f);
  return true;
}

//
// This function will emit code to multiply the top two values
// on the dedicated floating point stack and replace them with
// their product.  This effect is semantically like this:
// POP TO REGISTER2
// POP TO REGISTER1
// REGISTER3=REGISTER1*REGISTER2
// PUSH REGISTER3
// Returns true on success, false on failure.
//
bool domultiply(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"domultiply(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        binary_multiply\n");
  fflush(f);
  return true;
}

//
// This function will emit code to divide the top value
// on the dedicated floating point stack into the value immediately
// below it and replace those two values with their quotient.  This 
// effect is semantically like this:
// POP TO REGISTER2
// POP TO REGISTER1
// REGISTER3=REGISTER1/REGISTER2
// PUSH REGISTER3
// Returns true on success, false on failure.
//
bool dodivide(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dodivide(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        binary_divide\n");
  fflush(f);
  return true;
}

//
// This function will emit code to raise the value just below the top 
// value on the dedicated floating point stack to the power specified
// by the value immediately on top of that stack replace them with the 
// result.  This effect is semantically like this:
// POP TO REGISTER2
// POP TO REGISTER1
// REGISTER3=REGISTER1^REGISTER2
// PUSH REGISTER3
// where '^' is the power operator.
// Returns true on success, false on failure.
//
bool dopow(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dopow(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        binary_power\n");
  fflush(f);
  return true;
}

//
// This function is used to implement assignment to a scalar numeric variable.
// The f parameter is the name of the assembly language output file and that
// file must be open for output.  The varname parameter contains the name of
// the destination variable.  The value to be stored must be on the top of the
// dedicated floating point stack.  Returns true on success, false otherwise.
//
bool dostore_nvar(FILE *f, const char *varname) {
  if (NULL==f) {
    fprintf(stderr,"dostore_nvar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"dostore_nvar(): varname parameter is NULL or empty\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  fprintf(f,"        floatreg_to_floatmem %%xmm0,%s\n",varname);
  fflush(f);
  return true;
}

//
// This function is used to implement assignment to a scalar string variable.
// The f parameter is the name of the assembly language output file and that
// file must be open for output.  The varname parameter contains the name of
// the destination variable.  The address of the source string value to be 
// stored must be on the top of the dedicated string address stack.  Returns 
// true on success, false otherwise.  The strings are ASCIIZ strings, and
// mystrcpy() is called to actually copy the data.
//
bool dostore_svar(FILE *f, const char *varname) {
  if (NULL==f) {
    fprintf(stderr,"dostore_svar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"dostore_svar(): varname parameter is NULL or empty\n");
    return false;
  }
  need_strcpy=true;
  fprintf(f,"        popsaddr %%rsi\n");
  fprintf(f,"        movq    %%rsi,%%rdi\n");
  fprintf(f,"        movabsq $mystrlen, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        cmpq    $MAX_STRING_BYTES,%%rax\n");
  fprintf(f,"        jbe     0f\n");
  fprintf(f,"        movabsq $.Lstring_too_long,%%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $%s, %%rdi\n",varname);
  fprintf(f,"        movabsq $mystrcpy, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fflush(f);
  return true;
}

//
// This function is used to implement assignment to a numeric one-dimensional
// array variable.  The f parameter is the name of the assembly language output
// file and that file must be open for output.  The varname parameter contains
// the name of the destination variable.  The maxrows parameter contains the
// maximum permitted index value.  The value to be stored must be on the top of
// the dedicated floating point stack.  The index must be the value immediately
// below the top of stack value.  Returns true on success, false otherwise.
// The base_is_one flag is used to adjust the bounds checking.  If base_is_one
// is true, the minimum index is one, otherwise the minimum index is zero.
//
bool dostore_1D_navar(FILE *f, const char *varname, const uint32_t maxrows) {
  if (NULL==f) {
    fprintf(stderr,"dostore_1D_navar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"dostore_1D_navar(): varname parameter is NULL or empty\n");
    return false;
  }
  // top_of_stack should have RHS expression value
  // top_of_stack-1 should have row index expression value as a float
  fprintf(f,"        popxmm 2    #  RHS expression value\n");
  fprintf(f,"        popxmm 0    #  array index expression value\n");
  fprintf(f,"        floatreg_to_array_1D %s,%%xmm2,%%xmm0,%" PRIu32 ",%" PRIu32 ",%%r8,%%r15\n",
    varname,(base_is_one?UINT32_C(1):UINT32_C(0)),maxrows);
  fflush(f);
  return true;
}

//
// This function is used to implement assignment to a numeric two-dimensional
// array variable.  The f parameter is the name of the assembly language output
// file and that file must be open for output.  The varname parameter contains
// the name of the destination variable.  The maxrows parameter contains the
// maximum permitted index for the first dimension.  The maxcolumns parameter
// contains the maximum permitted index for the second dimension.  The value to
// be stored must be on the top of the dedicated floating point stack.  The top
// of stack has the value to store, below that is the column index, and below
// that is the row index.  Returns true on success, false otherwise.  The
// base_is_one flag is used to adjust the bounds checking.  If base_is_one is
// true, the minimum index is one, otherwise the minimum index is zero.
//
bool dostore_2D_navar(FILE *f, const char *varname, const uint32_t maxrows, const uint32_t maxcolumns) {
  if (NULL==f) {
    fprintf(stderr,"dostore_2D_navar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"dostore_2D_navar(): varname parameter is NULL or empty\n");
    return false;
  }
  // top_of_stack should have RHS expression value
  // top_of_stack-1 should have column index expression value as a float
  // top_of_stack-2 should have row index expression value as a float
  fprintf(f,"        popxmm 2    # RHS expression value\n");
  fprintf(f,"        popxmm 1    # column index\n");
  fprintf(f,"        popxmm 0    # row index\n");
  fprintf(f,"        floatreg_to_array_2D %s,%%xmm2,%%xmm0,%%xmm1,%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%%r9,%%r8,%%r15\n",
    varname,(base_is_one?UINT32_C(1):UINT32_C(0)),maxrows,maxcolumns);
  fflush(f);
  return true;
}

//
// This function is called at the start of the sequence for generating code for
// each Minimal BASIC source input line.  The lineno parameter specifies the
// Minimal BASIC line number that code will be emitted for after the call to
// this function.  This function will generate a label from the lineno
// parameter (.LLINE####) and also copy the Minimal BASIC line number to the
// magic internal .LCURLINENO variable which is used for displaying errors and
// exceptions.  The f parameter is the name of the assembly language output
// file and that file must be open for output.  This function returns true on
// success, false otherwise.
//
bool doloc(FILE *f, const uint32_t lineno) {
  if (NULL==f) {
    fprintf(stderr,"doloc(): file parameter is NULL\n");
    return false;
  }
  if (lineno>UINT32_C(9999)) {
    fprintf(stderr,"doloc(): lineno %" PRIu32 " is greater than 9999\n",lineno);  // ECMA-55 MinimalBASIC rule
    return false;
  }
  // Cannot include this since it confuses gdb...
  // fprintf(f,"        .loc 1 %" PRIu32 "\n",lineno);
  fprintf(f,".LLINE%04" PRIu32 ":\n",lineno);
  fprintf(f,"        movabsq $.LCURLINENO, %%rax\n");
  if (lineno>UINT32_C(4294967295)) {
    fprintf(f,"        movabsq $%" PRIu32 ", %%rbx\n",lineno); // x86-64 has only this instruction for 64bit immediates
    fprintf(f,"        movq    %%rbx, (%%rax)\n");
  } else {
    fprintf(f,"        movq    $%" PRIu32 ", (%%rax)\n",lineno);  // zero extend 32 bit val which is <= 9999
  }
  fflush(f);
  return true;
}

//
// This function will emit code to append a floating point value to the output
// buffer.  It is used as part of the implementation of the PRINT statement.
// The f parameter is the name of the assembly language output file and that
// file must be open for output.  This function returns true on success, false
// otherwise.  
//
bool doprint_number(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"doprint_number(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  fprintf(f,"        movabsq $printfloat, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fflush(f);
  return true;
}

//
// This function will emit code to append a string value to the output buffer.
// It is used as part of the implementation of the PRINT statement.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.
//
bool doprint_string(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"doprint_string(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popsaddr %%rdi\n");
  fprintf(f,"        movabsq $appendbuf, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fflush(f);
  return true;
}

//
// This function will emit code to output the current output buffer to STDOUT.
// It is used as part of the implementation of the PRINT statement.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.
//
bool doprint_newline(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"doprint_newline(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        movabsq $outputbuf, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fflush(f);
  return true;
}

//
// This function will emit code to advance the next print zone of the current
// output buffer.  It is used as part of the implementation of the PRINT
// statement to support the comma (,) behavior.  The f parameter is the name of
// the assembly language output file and that file must be open for output.
// This function returns true on success, false otherwise.
//
bool doprint_nextfield(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"doprint_nextfield(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        movabsq $nextzone, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fflush(f);
  return true;
}

//
// This function will emit code to imlement a call to the ABS() function.  The
// argument to the function must be on top of the dedicated floating point
// stack.  The return value of the function will be left on top of the
// dedicated floating point stack.  The f parameter is the name of the assembly
// language output file and that file must be open for output.  This function
// returns true on success, false otherwise.
//
bool doabs(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"doabs(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  fprintf(f,"        # xmm0=abs(xmm0)\n");
  fprintf(f,"        mov%s    $0x7fffffff%s, %%%cax\n",(use_double?"absq":"l"),(use_double?"ffffffff":""),(use_double?'r':'e'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsrq  $0,%%rax,%%xmm1\n");
    fprintf(f,"        pand    %%xmm1,%%xmm0\n");
  } else {
    fprintf(f,"        subq    $%" PRIu32 ",%%rsp         # allocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
    fprintf(f,"                                # it is a union location for \n");
    fprintf(f,"                                # converting between %" PRIu32 " bit %s and \n",(use_double?UINT32_C(64):UINT32_C(32)),(use_double?"double":"float"));
    fprintf(f,"                                # unsigned %" PRIu32 " int\n",(use_double?UINT32_C(64):UINT32_C(32)));
    fprintf(f,"        movs%c   %%xmm0, (%%rsp)\n",(use_double?'d':'s'));
    fprintf(f,"        and%c    %%%cax, (%%rsp)\n",(use_double?'q':'l'),(use_double?'r':'e'));
    fprintf(f,"        movs%c   (%%rsp), %%xmm0\n",(use_double?'d':'s'));
    fprintf(f,"        addq    $%" PRIu32 ",%%rsp         # deallocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
  }
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// This function will emit code to do the SGN() function as inline code.  The
// argument to the function must be on top of the dedicated floating point
// stack.  The return value is left on top of the dedicated floating point
// stack.  The f parameter is the name of the assembly language output file and
// that file must be open for output.  This function returns true on success,
// false otherwise.
//
bool dosgn(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dosgn(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  fprintf(f,"        pxor    %%xmm1,%%xmm1\n");
  fprintf(f,"        comis%c  %%xmm1,%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        ja      0f\n");
  fprintf(f,"        jb      1f\n");
  fprintf(f,"        movs%c   %%xmm1,%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        jmp     2f\n");
  fprintf(f,"0:\n");
  fprintf(f,"        mov%c    $FLOATONE, %%%cax\n",(use_double?'q':'l'),(use_double?'r':'e'));
  fprintf(f,"        subq    $%" PRIu32 ", %%rsp\n",(use_double?UINT32_C(8):UINT32_C(4)));
  fprintf(f,"        mov%c    %%%cax, (%%rsp)\n",(use_double?'q':'l'),(use_double?'r':'e'));
  fprintf(f,"        movs%c   (%%rsp),%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        addq    $%" PRIu32 ", %%rsp\n",(use_double?UINT32_C(8):UINT32_C(4)));
  fprintf(f,"        jmp     2f\n");
  fprintf(f,"1:\n");
  fprintf(f,"        mov%c    $FLOATNEGONE, %%%cax\n",(use_double?'q':'l'),(use_double?'r':'e'));
  fprintf(f,"        subq    $%" PRIu32 ", %%rsp\n",(use_double?UINT32_C(8):UINT32_C(4)));
  fprintf(f,"        mov%c    %%%cax, (%%rsp)\n",(use_double?'q':'l'),(use_double?'r':'e'));
  fprintf(f,"        movs%c   (%%rsp),%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        addq    $%" PRIu32 ", %%rsp\n",(use_double?UINT32_C(8):UINT32_C(4)));
  fprintf(f,"        # intentionally fall through\n");
  fprintf(f,"2:\n");
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// This function will emit code to implement a call of the SQR() function by
// directly calling the square root instruction on the processor.  The argument
// to the function must be on top of the dedicated floating point stack.  The
// return value is left on top of the dedicated floating point stack.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.
//
bool dosqr(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dosqr(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  fprintf(f,"        sqrts%c  %%xmm0,%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        xorp%c   %%xmm15, %%xmm15\n",(use_double?'d':'s'));
  fprintf(f,"        ucomis%c %%xmm15, %%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        jae     1f\n");
  fprintf(f,"        movabsq $.Lnegative_sqrtf, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"1:\n");
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// This function will emit code to implement a call of the  INT() function.
// The argument to the function must be on top of the dedicated floating point
// stack.  The return value is left on top of the dedicated floating point
// stack.  The f parameter is the name of the assembly language output file and
// that file must be open for output.  This function returns true on success,
// false otherwise.
//
bool doint(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"doint(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        movabsq $intf, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fflush(f);
  return true;
}

//
// This function will emit code to implement a call of the RND function.  The
// argument to the function must be on top of the dedicated floating point
// stack.  The return value is left on top of the dedicated floating point
// stack.  The f parameter is the name of the assembly language output file and
// that file must be open for output.  This function returns true on success,
// false otherwise.
//
bool dornd(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dornd(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        movabsq $drand, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  if (!use_double)
    fprintf(f,"        cvtsd2ss %%xmm0,%%xmm0\n");
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// This function will emit code to implement the RANDOMIZE statement.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.
//
bool dorandomize(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dorandomize(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        movabsq    $1, %%rdi\n");
  fprintf(f,"        movabsq $realinit, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fflush(f);
  return true;
}

//
// This function will emit code to call the SIN() function.  The argument to
// that function must be on top of the dedicated floating point stack.  The
// return value is left on top of the dedicated floating point stack.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.
//
bool dosin(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dosin(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  if (!use_double)
    fprintf(f,"        cvtss2sd %%xmm0, %%xmm0\n");
  fprintf(f,"        movabsq $xsin_u1,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  if (!use_double)
    fprintf(f,"        cvtsd2ss %%xmm0, %%xmm0\n");
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// This function will emit code to call the COS() function.  The argument to
// that function must be on top of the dedicated floating point stack.  The
// return value is left on top of the dedicated floating point stack.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.
//
bool docos(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"docos(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  if (!use_double)
    fprintf(f,"        cvtss2sd %%xmm0, %%xmm0\n");
  fprintf(f,"        movabsq $xcos_u1,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  if (!use_double)
    fprintf(f,"        cvtsd2ss %%xmm0, %%xmm0\n");
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// This function will emit code to implement the STOP statement.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.
//
bool dostop(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dostop(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        movabsq $goodxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fflush(f);
  return true;
}

//
// This function will emit code to implement the GOTO statement.  The Minimal
// BASIC statement target line number must be passed in the target_lineno
// parameter.  The f parameter is the name of the assembly language output file
// and that file must be open for output.  This function returns true on
// success, false otherwise.
//
bool dogoto(FILE *f, const uint32_t target_lineno) {
  if (NULL==f) {
    fprintf(stderr,"dogoto(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        movabsq $.LLINE%04" PRIu32 ", %%rax\n",target_lineno);
  fprintf(f,"        jmpq    *%%rax\n");
  fflush(f);
  return true;
}

//
// This function will emit code to call the TAN() function.  The argument to
// that function must be on top of the dedicated floating point stack.  The
// return value is left on top of the dedicated floating point stack.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.  
//
bool dotan(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dotan(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  if (!use_double)
    fprintf(f,"        cvtss2sd %%xmm0,%%xmm0\n");
  fprintf(f,"        movabsq $xtan_u1,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  if (!use_double)
    fprintf(f,"        cvtsd2ss %%xmm0, %%xmm0\n");
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// This function will emit code to call the ATN() function.  The argument to
// that function must be on top of the dedicated floating point stack.  The
// return value is left on top of the dedicated floating point stack.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.
//
bool doatan(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"doatan(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  if (!use_double)
    fprintf(f,"        cvtss2sd %%xmm0,%%xmm0\n");
  fprintf(f,"        movabsq $xatan_u1,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  if (!use_double)
    fprintf(f,"        cvtsd2ss %%xmm0, %%xmm0\n");
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// This function will emit code to call the LOG() function.  The argument to
// that function must be on top of the dedicated floating point stack.  The
// return value is left on top of the dedicated floating point stack.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.
//
bool dolog(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dolog(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  if (!use_double)
    fprintf(f,"        cvtss2sd %%xmm0,%%xmm0\n");
  fprintf(f,"        pxor    %%xmm1,%%xmm1\n");
  fprintf(f,"        comisd  %%xmm1,%%xmm0\n");
  fprintf(f,"        ja      0f\n");
  fprintf(f,"# LOG argument <=0 - fatal exception\n");
  fprintf(f,"        movabsq $.LLOG_ltzero_msg, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $xlog_u1,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  if (!use_double)
    fprintf(f,"        cvtsd2ss %%xmm0, %%xmm0\n");
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// This function will emit code to call the EXP() function.  The argument to
// that function must be on top of the dedicated floating point stack.  The
// return value is left on top of the dedicated floating point stack.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.
//
// WARNING: This FAILS for negative arguments.
//
bool doexp(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"doexp(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  if (!use_double) // 32 bit math
    fprintf(f,"        cvtss2sd %%xmm0,%%xmm0\n");
  fprintf(f,"        movabsq $xexp,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $myisinf, %%r13\n");
  fprintf(f,"        callq   *%%r13\n");
  fprintf(f,"        testq   %%rax, %%rax\n");
  fprintf(f,"        jz      3f\n");
  fprintf(f,"        # Infinity (overflow)\n");
  fprintf(f,"        movabsq $.Loverflow_exp_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r13\n");
  fprintf(f,"        callq   *%%r13\n");
  fprintf(f,"3:\n");
  if (!use_double) { // 32 bit math
    fprintf(f,"        cvtsd2ss %%xmm0, %%xmm0\n");
    fprintf(f,"        stmxcsr .Lnewmxcsr\n");
    fprintf(f,"        testl   $FE_OVERFLOW,.Lnewmxcsr\n");
    fprintf(f,"        jz      4f\n");
    fprintf(f,"        # Infinity (overflow)\n");
    fprintf(f,"        andl     $0xFFC0,.Lnewmxcsr  # clear exception in memory copy of MXCSR\n");
    fprintf(f,"        ldmxcsr .Lnewmxcsr        # and reload that into real MXCSR\n");
    fprintf(f,"        subq    $4,%%rsp         # allocate scratch 4 byte area\n");
    fprintf(f,"        movl    $PInf,%%r13d     # r13d=+INF\n");
    fprintf(f,"        movl    %%r13d,(%%rsp)\n");
    fprintf(f,"        movss   (%%rsp),%%xmm0\n");
    fprintf(f,"        addq    $4,%%rsp         # allocate scratch 4 byte area\n");
    fprintf(f,"        movabsq $.Loverflow_exp_msg,%%rsi\n");
    fprintf(f,"        movabsq $exception_non_fatal,%%r13\n");
    fprintf(f,"        callq   *%%r13\n");
    fprintf(f,"4:\n");
  }
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// This function will emit code for a string comparison in an IF statement.
// The right-hand-side string address should be on top of the dedicated
// floating point stack, and the left-hand-side string address should just
// below that.  The type of comparison is specified by the op parameter.  
// At runtime, if the condition is true, the conditional branch to the target
// Minimal BASIC line number specified by target_lineno is taken, otherwise it
// is not.  The f parameter is the name of the assembly language output file
// and that file must be open for output.  Only comparison for equality and
// inequality are permitted in Minimal BASIC.  This function returns true on
// success, false otherwise.  
//
bool dostring_comparison(FILE *f, const enum relop op, const uint32_t target_lineno) {
  if (NULL==f) {
    fprintf(stderr,"dostring_comparison(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popsaddr %%rsi\n");
  fprintf(f,"        popsaddr %%rdi\n");
  fprintf(f,"        movabsq $mystrcmp, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        testq   %%rax,%%rax\n");
  switch(op) {
    case EQ:
        fprintf(f,"        jnz     0f\n");
      break;
    case NE:
        fprintf(f,"        jz      0f\n");
      break;
    default:
      fprintf(stderr,"INTERNAL COMPILER ERROR: dostring_comparison() encountered unknown operator\n");
      abort();
  }
  fprintf(f,"        movabsq $.LLINE%04" PRIu32 ",%%rax\n",target_lineno);
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fflush(f);
  return true;
}

//
// This function will emit code for a numeric comparison in an IF statement.
// The right-hand-side expression value should be on top of the dedicated
// floating point stack, and the left-hand-side expression value should just
// below that.  The type of comparison is specified by the op parameter.  
// At runtime, if the condition is true, the conditional branch to the target
// Minimal BASIC line number specified by target_lineno is taken, otherwise it
// is not.  The f parameter is the name of the assembly language output file
// and that file must be open for output.  This function returns true on
// success, false otherwise.
//
bool donumeric_comparison(FILE *f, const enum relop op, const uint32_t target_lineno) {
  if (NULL==f) {
    fprintf(stderr,"donumeric_comparison(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm  1    # rhs expression value\n");
  fprintf(f,"        popxmm  0    # lhs expression value\n");
  fprintf(f,"        comis%c  %%xmm1, %%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        jp      0f   # LHS and/or RHS was NaN\n");
  switch(op) {
    case EQ:
      fprintf(f,"        jne     1f\n");
      break;
    case NE:
      fprintf(f,"        je      1f\n");
      break;
    case LT:
      fprintf(f,"        jae     1f\n");
      break;
    case LE:
      fprintf(f,"        ja      1f\n");
      break;
    case GT:
      fprintf(f,"        jbe     1f\n");
      break;
    case GE:
      fprintf(f,"        jb      1f\n");
      break;
    default:
      fprintf(stderr,"INTERNAL COMPILER ERROR: donumeric_comparison() encountered unknown operator\n");
      abort();
  }
  fprintf(f,"        movabsq $.LLINE%04" PRIu32 ",%%rax\n",target_lineno);
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $.Luninitialized,%%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"1:\n");
  fflush(f);
  return true;
}

//
// This function will emit code for the GOSUB statement.  The actual
// implementation is in the gosub assembly language macro.  The target_lineno
// is the Minimal BASIC statement line number of the subroutine to run.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.
//
bool dogosub(FILE *f, const uint32_t target_lineno) {
  if (NULL==f) {
    fprintf(stderr,"dogosub(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        gosub 0f .LLINE%04" PRIu32 "\n", target_lineno);
  fprintf(f,"0:\n");
  fflush(f);
  return true;
}

//
// This function will emit code for the RETURN statement.  The actual
// implementation is in the return assembly language macro.  The f parameter is
// the name of the assembly language output file and that file must be open for
// output.  This function returns true on success, false otherwise.
//
bool doreturn(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"doreturn(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        return\n");
  fflush(f);
  return true;
}

// The f parameter is the name of the assembly language output file and
// that file must be open for output.
// This function returns true on success, false otherwise.
bool doongotostart(FILE *f,const uint32_t maxchoice) {
  if (NULL==f) {
    fprintf(stderr,"doongotostart(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  fprintf(f,"        cvts%c2sil %%xmm0, %%eax\n",(use_double?'d':'s'));
  fprintf(f,"        stmxcsr .Lnewmxcsr\n");
  fprintf(f,"        testl   $FE_INVALID,.Lnewmxcsr\n");
  fprintf(f,"        jz      0f\n");
  fprintf(f,"        jmp     1f\n");
  fprintf(f,"0:\n");
  fprintf(f,"        decl    %%eax\n");
  fprintf(f,"        cmpl    $%" PRIu32 ",%%eax\n",maxchoice-UINT32_C(1));
  fprintf(f,"        ja      1f\n");
  fprintf(f,"        jmp     *.LJT%04" PRIu32 "(,%%rax,8)\n",cur_jumptable_no);
  fprintf(f,"1:\n");
  fprintf(f,"        movabsq $.Longotofail_msg, %%rsi\n");
  fprintf(f,"        movabsq $.Lprint_error_message, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n\n");
  fprintf(f,"        .section .rodata,\"a\",@progbits\n\n");
  fprintf(f,".LJT%04" PRIu32 ":\n",cur_jumptable_no++);
  fflush(f);
  return true;
}


// The f parameter is the name of the assembly language output file and
// that file must be open for output.
// This function returns true on success, false otherwise.
bool doongotochoice(FILE *f, const uint32_t target_lineno) {
  if (NULL==f) {
    fprintf(stderr,"doongotochoice(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        .quad   .LLINE%04" PRIu32 "\n",target_lineno);
  fflush(f);
  return true;
}

// The f parameter is the name of the assembly language output file and
// that file must be open for output.
// This function returns true on success, false otherwise.
bool doongotoend(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"doongotoend(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"\n        .section .text,\"ax\",@progbits\n\n");
  fflush(f);
  return true;
}

//
// This function will emit the code for the mystrcpy assembly language function.
// To support string assignment, a function must exist to copy one ASCIIZ
// string to another location.  The mystrcpy is a simple, unoptimized
// implementation of a C standard library strcpy() function.  The f parameter
// is the name of the assembly language output file and that file must be open
// for output.  This function returns true on success, false otherwise.
//
static bool write_strcpy(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"emit_strcpy(): file parameter is NULL\n");
    return false;
  }

  fprintf(f,"        # input ptr rsi, output ptr rdi\n");
  fprintf(f,"        .type   mystrcpy, @function\n");
  fprintf(f,"mystrcpy:\n");
  fprintf(f,"        cld\n");
  fprintf(f,"        movq    %%rdi, %%rdx\n");
  fprintf(f,"0:\n");
  fprintf(f,"        lodsb\n");
  fprintf(f,"        stosb\n");
  fprintf(f,"        testb   %%al, %%al\n");
  fprintf(f,"        jnz     0b\n");
  fprintf(f,"        movq    %%rdx, %%rax\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   mystrcpy, .-mystrcpy\n\n");
  fflush(f);
  return true;
}

//
// This function will emit the code for the mystrcmp assembly language
// function.  To support string comparison, a function must exist to compare
// one ASCIIZ string to another ASCIIZ string.  The mystrcmp is a simple,
// unoptimized implementation of a C standard library strcmp() function.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  This function returns true on success, false
// otherwise.
//
static bool write_strcmp(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"emit_strcmp(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        # lhs in rdi, rhs in rsi, result in rax (-1 for <, 0 for ==, 1 for >)\n");
  fprintf(f,"        .globl  mystrcmp\n");
  fprintf(f,"        .type   mystrcmp, @function\n");
  fprintf(f,"mystrcmp:\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movsbl  (%%rdi), %%eax\n");
  fprintf(f,"        testb   %%al, %%al\n");
  fprintf(f,"        je      1f\n");
  fprintf(f,"        movb    (%%rsi), %%dl\n");
  fprintf(f,"        testb   %%dl, %%dl\n");
  fprintf(f,"        je      1f\n");
  fprintf(f,"        cmpb    %%dl, %%al\n");
  fprintf(f,"        jne     1f\n");
  fprintf(f,"        incq    %%rdi\n");
  fprintf(f,"        incq    %%rsi\n");
  fprintf(f,"        jmp     0b\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movsbl  (%%rsi), %%edx\n");
  fprintf(f,"        subl    %%edx, %%eax\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   mystrcmp, .-mystrcmp\n\n");
  fflush(f);
  return true;
}

// The f parameter is the name of the assembly language output file and that file must be open for output.  This function returns true on success, false otherwise.
static bool write_INPUT(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"emit_INPUT(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        # BEGIN support for runtime INPUT subsystem\n\n");
  fprintf(f,"        .equ MAX_INPUT_VARS, 30\n\n");
  fprintf(f,"        #\n");
  fprintf(f,"        # This procedure clears the STDIN input buffer, discarding its contents (if any)\n");
  fprintf(f,"        #\n");
  fprintf(f,"drain_stdin:\n");
  fprintf(f,"        pushq   %%rbp\n");
  fprintf(f,"        movq    %%rsp,%%rbp\n");
  fprintf(f,"        subq    $16,%%rsp\n");
  fprintf(f,"        # flags   %%rbp-8\n");
  fprintf(f,"        # tbuffer %%rbp-16\n");
  fprintf(f,"        # get current flags\n");
  fprintf(f,"        movq    $__NR_fcntl, %%rax\n");
  fprintf(f,"        movq    $STDIN, %%rdi\n");
  fprintf(f,"        movq    $F_GETFL, %%rsi\n");
  fprintf(f,"        xorq    %%rdx,%%rdx\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"        movq    %%rax,-8(%%rbp)\n");
  fprintf(f,"        # set nonblocking\n");
  fprintf(f,"        movq    $__NR_fcntl,%%rax\n");
  fprintf(f,"        movq    $STDIN,%%rdi\n");
  fprintf(f,"        movq    $F_SETFL,%%rsi\n");
  fprintf(f,"        movq    -8(%%rbp),%%rdx\n");
  fprintf(f,"        orq     $O_NONBLOCK,%%rdx\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"0:\n");
  fprintf(f,"        # read a byte\n");
  fprintf(f,"        movq    $__NR_read,%%rax\n");
  fprintf(f,"        movq    $STDIN,%%rdi\n");
  fprintf(f,"        lea     -16(%%rsp),%%rsi\n");
  fprintf(f,"        movq    $1, %%rdx\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"        cmpq    $0,%%rax\n");
  fprintf(f,"        jg      0b\n");
  fprintf(f,"        # return to normal\n");
  fprintf(f,"        movq    $__NR_fcntl,%%rax\n");
  fprintf(f,"        movq    $STDIN,%%rdi\n");
  fprintf(f,"        movq    $F_SETFL,%%rsi\n");
  fprintf(f,"        movq    -8(%%rbp),%%rdx\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"        movq    %%rax,-8(%%rbp)\n");
  fprintf(f,"        movq    %%rbp, %%rsp\n");
  fprintf(f,"        popq    %%rbp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   drain_stdin, .-drain_stdin\n\n");
  fprintf(f,"        #\n");
  fprintf(f,"        # This function will ASCII bytes of input from STDIN until a newline\n");
  fprintf(f,"        # is read or the buffer is full (which we know from the length).\n");
  fprintf(f,"        # It expects the caller to pass the address of the buffer to hold the\n");
  fprintf(f,"        # ASCIIZ string in %%rdi, and it expects the caller to pass the\n");
  fprintf(f,"        # length of that buffer in %%rsi.  The function returns its\n");
  fprintf(f,"        # result in %%rax.  On success, the result is the number of\n");
  fprintf(f,"        # bytes read; on failure it is -ERRORNO.\n");
  fprintf(f,"        #\n");
  fprintf(f,"        .type   INPUT, @function\n");
  fprintf(f,"INPUT:\n");
  fprintf(f,"        pushq   %%rbp\n");
  fprintf(f,"        movq    %%rsp, %%rbp\n");
  fprintf(f,"        pushq   %%r13\n");
  fprintf(f,"        pushq   %%r12\n");
  fprintf(f,"        pushq   %%rbx\n");
  fprintf(f,"        leaq    (%%rdi,%%rsi), %%r13       # r13 is bufmax\n");
  fprintf(f,"        movq    %%rdi, %%r12      # r12 is buffer\n");
  fprintf(f,"        movq    %%rdi, %%rbx      # rbx is tbuffer\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movq    $STDIN, %%rdi    # read from STDIN\n");
  fprintf(f,"        movq    $1, %%rdx        # read 1 byte\n");
  fprintf(f,"        movq    %%rbx, %%rsi      # into tbuffer\n");
  fprintf(f,"        movq    $__NR_read,%%rax # read system call\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"        testq   %%rax, %%rax      # if read some bytes then\n");
  fprintf(f,"        jg      1f              # goto 1\n");
  fprintf(f,"        movabsq $.LEOF_on_input,%%rax\n");
  fprintf(f,"        jmp     *%%rax\n");
  fprintf(f,"1:\n");
  fprintf(f,"        leaq    1(%%rbx), %%rax   # temp=tbuffer+1\n");
  fprintf(f,"        cmpb    $10, (%%rbx)     # if tbuffer!='\\n'\n");
  fprintf(f,"        jne     3f              # goto 3\n");
  fprintf(f,"2:\n");
  fprintf(f,"        movb    $0, (%%rax)      # *tbuffer=0 (terminate)\n");
  fprintf(f,"        subq    %%r12, %%rax      # compute tbuffer-buffer as return value\n");
  fprintf(f,"        jmp     4f              # goto 4\n");
  fprintf(f,"3:\n");
  fprintf(f,"        cmpq    %%r13, %%rax      # if tbuffer+1 >= bufmax\n");
  fprintf(f,"        jae     2b              # goto 2\n");
  fprintf(f,"        movq    %%rax, %%rbx      # save temp into tbuffer\n");
  fprintf(f,"        jmp     0b              # goto 0\n");
  fprintf(f,"4:\n");
  fprintf(f,"        # if we are not a tty, we must echo stuff\n");
  fprintf(f,"        movabsq $.Lecho_input, %%rbx\n");
  fprintf(f,"        movq    (%%rbx), %%rbx\n");
  fprintf(f,"        testq   %%rbx, %%rbx\n");
  fprintf(f,"        jnz     5f\n");
  fprintf(f,"        # ok, we are NOT a tty\n");
  fprintf(f,"        pushq   %%rax\n");
  fprintf(f,"        movq    %%r12, %%rsi\n");
  fprintf(f,"        movabsq $printstring, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        popq    %%rax\n");
  fprintf(f,"5:\n");
  fprintf(f,"        popq    %%rbx\n");
  fprintf(f,"        popq    %%r12\n");
  fprintf(f,"        popq    %%r13\n");
  fprintf(f,"        popq    %%rbp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   INPUT, .-INPUT\n\n");

  fprintf(f,"        .section .bss,\"aw\",@nobits\n\n");

  //
  // The INPUT statement for Minimal BASIC has the requirement that all data is input,
  // then validated, and then, only if all validation for every variable in the INPUT
  // statement passed, is the data actually copied to the destination variables.  To
  // make that possible, data must be input into temporary storage.  The .Lrt_nput_stack
  // is used for this, but is actually sometimes used more like a vector than a stack,
  // accessing by a slot number instead of stricly from the top of the stack.  When
  // used as a stack, the .Lrt_nput_stack_top is the stack pointer.  To record the types
  // required for the variables in the INPUT statement, a format statement is created.
  // That is stored in the .Linformat variable.
  //
  fprintf(f,"        .align 32\n");
  fprintf(f,"        .type   .Lrt_nput_stack, @object\n");
  fprintf(f,".Lrt_nput_stack:\n");
  fprintf(f,"        .rept MAX_INPUT_VARS\n");
  fprintf(f,"        .zero 1\n");
  fprintf(f,"        .zero 2*MAX_STRING_BYTES+1  # larger than required to allow LONG numeric constants\n");
  fprintf(f,"        .endr\n");
  fprintf(f,"        .size   .Lrt_nput_stack, .-.Lrt_nput_stack\n\n");
  fprintf(f,"        .align 8\n");
  fprintf(f,"        .type   .Lrt_nput_stack_top, @object\n");
  fprintf(f,"        .size   .Lrt_nput_stack_top, 8\n");
  fprintf(f,".Lrt_nput_stack_top:\n");
  fprintf(f,"        .zero   8\n\n");
  fprintf(f,"        # The .Linformat holds the format specification ASCIIZ string for an INPUT statement\n");
  fprintf(f,"        # each byte of of the string will hold either 'N' or 'S'.  An 'N' signals that a\n");
  fprintf(f,"        # numeric (floating point) value is needed.  An 'S' signals that a 7-bit ASCIIZ string\n");
  fprintf(f,"        # value is needed.  For example, given a Minimal BASIC statement like:\n");
  fprintf(f,"        # INPUT X,Y$,Z[3]\n");
  fprintf(f,"        # the value in .Linformat would be \"NSN\"\n");
  fprintf(f,"        .align 16\n");
  fprintf(f,"        .type   .Linformat, @object\n");
  fprintf(f,".Linformat:\n");
  fprintf(f,"        .zero   MAX_STRING_BYTES+1\n");
  fprintf(f,"        .size   .Linformat, .-.Linformat\n\n");

  //
  // Minimal BASIC rules for values entered in response to an INPUT statement
  // state that for data outside of a quoted string, all whitespace should be
  // ignored.  To remove such space from the raw ASCIIZ data in the input
  // buffer, the mytrim() function is used.
  //
  fprintf(f,"        .section .text,\"ax\",@progbits\n\n");

  fprintf(f,"        # mytrim() procedure, takes one argument, ptr to ASCIIZ string, in rdi\n");
  fprintf(f,"        # nothing is returned - removes all leading/trailing spaces and tabs in place\n");
  fprintf(f,"        .type   mytrim, @function\n");
  fprintf(f,"mytrim:\n");
  fprintf(f,"        xorl    %%eax, %%eax\n");
  fprintf(f,"        cmpb    $0, (%%rdi)\n");
  fprintf(f,"        pushq   %%rbp\n");
  fprintf(f,"        movq    %%rsp, %%rbp\n");
  fprintf(f,"        je      11f\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movb    (%%rdi,%%rax), %%dl\n");
  fprintf(f,"        testb   %%dl, %%dl\n");
  fprintf(f,"        je      3f\n");
  fprintf(f,"        cmpb    $9, %%dl\n");
  fprintf(f,"        jne     2f\n");
  fprintf(f,"1:\n");
  fprintf(f,"        incq    %%rax\n");
  fprintf(f,"        jmp     0b\n");
  fprintf(f,"2:\n");
  fprintf(f,"        cmpb    $32, %%dl\n");
  fprintf(f,"        je      1b\n");
  fprintf(f,"3:\n");
  fprintf(f,"        movq    %%rax, %%rdx\n");
  fprintf(f,"4:\n");
  fprintf(f,"        cmpb    $0, (%%rdi,%%rax)\n");
  fprintf(f,"        je      6f\n");
  fprintf(f,"        incq    %%rax\n");
  fprintf(f,"        jmp     4b\n");
  fprintf(f,"5:\n");
  fprintf(f,"        movb    (%%rdi,%%rax), %%cl\n");
  fprintf(f,"        cmpb    $9, %%cl\n");
  fprintf(f,"        jne     8f\n");
  fprintf(f,"6:\n");
  fprintf(f,"        decq    %%rax\n");
  fprintf(f,"        cmpq    %%rdx, %%rax\n");
  fprintf(f,"        ja      5b\n");
  fprintf(f,"7:\n");
  fprintf(f,"        subq    %%rdx, %%rax\n");
  fprintf(f,"        addq    %%rdi, %%rdx\n");
  fprintf(f,"        leaq    1(%%rax), %%rcx\n");
  fprintf(f,"        xorl    %%eax, %%eax\n");
  fprintf(f,"        jmp     9f\n");
  fprintf(f,"8:\n");
  fprintf(f,"        cmpb    $32, %%cl\n");
  fprintf(f,"        je      6b\n");
  fprintf(f,"        jmp     7b\n");
  fprintf(f,"9:\n");
  fprintf(f,"        cmpq    %%rcx, %%rax\n");
  fprintf(f,"        je      10f\n");
  fprintf(f,"        movb    (%%rdx,%%rax), %%sil\n");
  fprintf(f,"        movb    %%sil, (%%rdi,%%rax)\n");
  fprintf(f,"        incq    %%rax\n");
  fprintf(f,"        jmp     9b\n");
  fprintf(f,"10:\n");
  fprintf(f,"        movb    $0, (%%rdi,%%rax)\n");
  fprintf(f,"11:\n");
  fprintf(f,"        popq    %%rbp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   mytrim, .-mytrim\n\n");

  //
  // At runtime, scanning values is required for the INPUT statment.
  // This is done using a DFSM (deterministic finite state machine).
  // This machine uses character classes and states.
  //
  fprintf(f,"        .section .rodata,\"a\",@progbits\n\n");
  fprintf(f,"        # The .Lcharacter_class is a conversion map to convert between\n");
  fprintf(f,"        # an input 7 bit ASCII value and a character class number for use\n");
  fprintf(f,"        # with the DFSM used by the INPUT statement scanner.  .Lcharacter_class\n");
  fprintf(f,"        # is a vector in which there are 127 elements.\n");
  fprintf(f,"        # The ASCII code of the byte in question is used as an\n");
  fprintf(f,"        # index into the .Lcharacter_class vector to get the\n");
  fprintf(f,"        # character class number which is stored in that element.\n");
  fprintf(f,"        #\n");
  fprintf(f,"        #  0  is A..D,F..Z upper case\n");
  fprintf(f,"        #  1  is E upper case\n");
  fprintf(f,"        #  2  is DIGIT\n");
  fprintf(f,"        #  3  is ADD\n");
  fprintf(f,"        #  4  is MINUS\n");
  fprintf(f,"        #  5  is COMMA\n");
  fprintf(f,"        #  6  is SPACE\n");
  fprintf(f,"        #  7  is NEWLINE\n");
  fprintf(f,"        #  8  is DOT\n");
  fprintf(f,"        #  9  is QUOTE\n");
  fprintf(f,"        # 10  is anything else\n");
  fprintf(f,"        .align 32\n");
  fprintf(f,"        .type   .Lcharacter_class, @object\n");
  fprintf(f,".Lcharacter_class:\n");
  fprintf(f,"        .byte  10, 10, 10, 10, 10, 10, 10, 10, 10, 10\n");
  fprintf(f,"        .byte   7, 10, 10, 10, 10, 10, 10, 10, 10, 10\n");
  fprintf(f,"        .byte  10, 10, 10, 10, 10, 10, 10, 10, 10, 10\n");
  fprintf(f,"        .byte  10, 10,  6, 10,  9, 10, 10, 10, 10, 10\n");
  fprintf(f,"        .byte  10, 10, 10,  3,  5,  4,  8, 10,  2,  2\n");
  fprintf(f,"        .byte   2,  2,  2,  2,  2,  2,  2,  2, 10, 10\n");
  fprintf(f,"        .byte  10, 10, 10, 10, 10,  0,  0,  0,  0,  1\n");
  fprintf(f,"        .byte   0,  0,  0,  0,  0,  0,  0,  0,  0,  0\n");
  fprintf(f,"        .byte   0,  0,  0,  0,  0,  0,  0,  0,  0,  0\n");
  fprintf(f,"        .byte   0, 10, 10, 10, 10, 10, 10, 10, 10, 10\n");
  fprintf(f,"        .byte  10, 10, 10, 10, 10, 10, 10, 10, 10, 10\n");
  fprintf(f,"        .byte  10, 10, 10, 10, 10, 10, 10, 10, 10, 10\n");
  fprintf(f,"        .byte  10, 10, 10, 10, 10, 10, 10, 10\n");
  fprintf(f,"        .size   .Lcharacter_class, .-.Lcharacter_class\n\n");
  fprintf(f,"         # .Lstates is a matrix containing the DFSM used to scan\n");
  fprintf(f,"         # data from an INPUT statement.  The rows correspond to the\n");
  fprintf(f,"         # states, and the columns correspond to character classes.\n");
  fprintf(f,"         # To find the next action, the character class value for the\n");
  fprintf(f,"         # current byte is used as the column index, and the current\n");
  fprintf(f,"         # state is used as the row index in the .Lstates matrix.\n");
  fprintf(f,"         # Actions can be accept (see list below), move to a new\n");
  fprintf(f,"         # state, or reject.  Values 0 through 12 move to a new state.\n");
  fprintf(f,"         # Values 13 through 18 accept tokens.  Value 19 is a reject.\n");
  fprintf(f,"        .align 32\n");
  fprintf(f,"        .type   .Lstates, @object\n");
  fprintf(f,".Lstates:\n");
  fprintf(f,"         # 13 accept T_COMMA\n");
  fprintf(f,"         # 14 accept T_EOL\n");
  fprintf(f,"         # 15 accept T_INTEGER\n");
  fprintf(f,"         # 16 accept T_REAL\n");
  fprintf(f,"         # 17 accept T_UQSTRING\n");
  fprintf(f,"         # 18 accept T_QSTRING\n");
  fprintf(f,"         # 19 reject\n");
  fprintf(f,"         #    A..D,                       new\n");
  fprintf(f,"         #    F..Z E  0..9 +   -   ,  ' ' line .   \" other\n");
  fprintf(f,"        .byte  8,  8,  2,  1,  1, 13,  0, 14,  3,  9, 19  # state 0\n");
  fprintf(f,"        .byte  8,  8,  2,  8,  8, 17,  8, 17,  3, 19, 19  # state 1\n");
  fprintf(f,"        .byte  8,  5,  2,  8,  8, 15, 11, 15,  3, 19, 19  # state 2\n");
  fprintf(f,"        .byte  8,  5,  4,  8,  8, 16, 12, 16,  8, 19, 19  # state 3\n");
  fprintf(f,"        .byte  8,  5,  4,  8,  8, 16, 12, 16,  8, 19, 19  # state 4\n");
  fprintf(f,"        .byte  8,  8,  7,  6,  6, 17,  8, 17,  8, 19, 19  # state 5\n");
  fprintf(f,"        .byte  8,  8,  7,  8,  8, 17,  8, 17,  8, 19, 19  # state 6\n");
  fprintf(f,"        .byte  8,  8,  7,  8,  8, 16, 12, 16,  8, 19, 19  # state 7\n");
  fprintf(f,"        .byte  8,  8,  8,  8,  8, 17, 10, 17,  8, 19, 19  # state 8\n");
  fprintf(f,"        .byte  9,  9,  9,  9,  9,  9,  9, 19,  9, 18,  9  # state 9\n");
  fprintf(f,"        .byte  8,  8,  8,  8,  8, 17, 10, 17,  8, 19, 19  # state 10\n");
  fprintf(f,"        .byte  8,  8,  8,  8,  8, 15, 11, 15,  8, 19, 19  # state 11\n");
  fprintf(f,"        .byte  8,  8,  8,  8,  8, 16, 12, 16,  8, 19, 19  # state 12\n");
  fprintf(f,"        .size    .Lstates, .-.Lstates\n\n");
  // error messages
  fprintf(f,"        .section .rodata.str1.1,\"aMS\",@progbits,1\n\n");
  fprintf(f,".LparseinputC1:\n");
  fprintf(f,"        .string \"Error: comma encountered when data expected\\n\"\n");
  fprintf(f,".LparseinputC2:\n");
  fprintf(f,"        .string \"Error: newline encountered when data expected\\n\"\n");
  fprintf(f,".LparseinputC3:\n");
  fprintf(f,"        .string \"Error: number encountered when comma or newline expected\\n\"\n");
  fprintf(f,".LparseinputC4:\n");
  fprintf(f,"        .string \"Error: numeric constant string maximum size is 36 bytes\\n\"\n");
  fprintf(f,".LparseinputC5:\n");
  fprintf(f,"        .string \"Error: Runtime failure: too many input variables\\n\"\n");
  fprintf(f,".LparseinputC6:\n");
  fprintf(f,"        .string \"Error: unquoted string encountered when comma or newline expected\\n\"\n");
  fprintf(f,".LparseinputC7:\n");
  fprintf(f,"        .string \"Error: unquoted string maximum size is 18 bytes\\n\"\n");
  fprintf(f,".LparseinputC8:\n");
  fprintf(f,"        .string \"Error: quoted string encountered when comma or newline expected\\n\"\n");
  fprintf(f,".LparseinputC9:\n");
  fprintf(f,"        .string \"Error: quoted string maximum size is 18 bytes\\n\"\n\n");
  //
  // This boolean function will parse the input line buffer passed in %rdi with the length passed in %rsi.
  // It will push the values obtained onto the .Lrt_nput_stack and generate a new format string in .Linformat.
  // The return value is passed in the rax register, and will 0 or 1 corresponding to false (error) or true (success).
  //
  fprintf(f,"        .section .text,\"ax\",@progbits\n\n");
  fprintf(f,"        .type   parseinput, @function\n");
  fprintf(f,"parseinput:\n");
  fprintf(f,"        pushq   %%rbp\n");
  fprintf(f,"        movq    %%rsp, %%rbp\n");
  fprintf(f,"        movabsq $.Lrt_nput_stack_top, %%rax\n");
  fprintf(f,"        xorl    %%edx, %%edx\n");
  fprintf(f,"        pushq   %%r14\n");
  fprintf(f,"        movq    %%rsi, %%r14\n");
  fprintf(f,"        pushq   %%r13\n");
  fprintf(f,"        movq    %%rdi, %%r13\n");
  fprintf(f,"        pushq   %%r12\n");
  fprintf(f,"        xorl    %%r12d, %%r12d\n");
  fprintf(f,"        pushq   %%rbx\n");
  fprintf(f,"        xorl    %%ebx, %%ebx\n");
  fprintf(f,"        subq    $96, %%rsp\n");
  fprintf(f,"        movq    $0, (%%rax)\n");
  fprintf(f,"        movabsq $.Linformat, %%rax\n");
  fprintf(f,"        movb    $0, (%%rax)\n");
  fprintf(f,"        xorl    %%eax, %%eax\n");
  fprintf(f,"        movb    $0, -113(%%rbp)\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movb    0(%%r13,%%r12), %%sil\n");
  fprintf(f,"        imulq   $11, %%rax, %%rax\n");
  fprintf(f,"        movabsq $.Lcharacter_class, %%rdi\n");
  fprintf(f,"        movsbl  %%sil, %%ecx\n");
  fprintf(f,"        movzbl  (%%rdi,%%rcx), %%ecx\n");
  fprintf(f,"        addq    %%rcx, %%rax\n");
  fprintf(f,"        movabsq $.Lstates, %%rcx\n");
  fprintf(f,"        movb    (%%rcx,%%rax), %%cl\n");
  fprintf(f,"        cmpb    $16, %%cl\n");
  fprintf(f,"        movzbl  %%cl, %%eax\n");
  fprintf(f,"        ja      1f\n");
  fprintf(f,"        cmpb    $15, %%cl\n");
  fprintf(f,"        jae.d32 5f\n");
  fprintf(f,"        cmpb    $13, %%cl\n");
  fprintf(f,"        je      2f\n");
  fprintf(f,"        cmpb    $14, %%cl\n");
  fprintf(f,"        je      3f\n");
  fprintf(f,"        jmp     21f\n");
  fprintf(f,"1:\n");
  fprintf(f,"        cmpb    $18, %%cl\n");
  fprintf(f,"        je.d32  16f\n");
  fprintf(f,"        jb.d32  13f\n");
  fprintf(f,"        cmpb    $19, %%cl\n");
  fprintf(f,"        je.d32  24f\n");
  fprintf(f,"        jmp     21f\n");
  fprintf(f,"2:\n");
  fprintf(f,"        incq    %%rbx\n");
  fprintf(f,"        incq    %%r12\n");
  fprintf(f,"        testb   $1, %%bl\n");
  fprintf(f,"        je.d32  12f\n");
  fprintf(f,"        movabsq $.LparseinputC1, %%rdi\n");
  fprintf(f,"        jmp     18f\n");
  fprintf(f,"3:\n");
  fprintf(f,"        andb    $1, %%bl\n");
  fprintf(f,"        jne     4f\n");
  fprintf(f,"        movabsq $printstring2, %%rax\n");
  fprintf(f,"        movabsq $.LparseinputC2, %%rdi\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movl    $19, %%eax\n");
  fprintf(f,"4:\n");
  fprintf(f,"        movabsq $.Lrt_nput_stack_top, %%rsi\n");
  fprintf(f,"        movabsq $.Linformat, %%rcx\n");
  fprintf(f,"        movq    (%%rsi), %%rdx\n");
  fprintf(f,"        movb    $0, (%%rcx,%%rdx)\n");
  fprintf(f,"        jmp     24f\n");
  fprintf(f,"5:\n");
  fprintf(f,"        incq    %%rbx\n");
  fprintf(f,"        testb   $1, %%bl\n");
  fprintf(f,"        jne     6f\n");
  fprintf(f,"        movabsq $.LparseinputC3, %%rdi\n");
  fprintf(f,"        jmp     18f\n");
  fprintf(f,"6:\n");
  fprintf(f,"        leaq    -113(%%rbp), %%rdi\n");
  fprintf(f,"        movb    $0, -113(%%rbp,%%rdx)\n");
  fprintf(f,"        movabsq $mytrim, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        leaq    -113(%%rbp), %%rdi\n");
  fprintf(f,"        movabsq $mystrlen, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        cmpq    $36, %%rax\n");
  fprintf(f,"        jbe     7f\n");
  fprintf(f,"        movabsq $.LparseinputC4, %%rdi\n");
  fprintf(f,"        jmp     18f\n");
  fprintf(f,"7:\n");
  fprintf(f,"        movabsq .Lrt_nput_stack_top, %%rax\n");
  fprintf(f,"        cmpq    $29, %%rax\n");
  fprintf(f,"        jbe     9f\n");
  fprintf(f,"8:\n");
  fprintf(f,"        movabsq $.LparseinputC5, %%rdi\n");
  fprintf(f,"        jmp     18f\n");
  fprintf(f,"9:\n");
  fprintf(f,"        movabsq $.Linformat, %%rdx\n");
  fprintf(f,"        movabsq $.Lrt_nput_stack, %%rdi\n");
  fprintf(f,"        movb    $'N', (%%rdx,%%rax)\n");
  fprintf(f,"        imulq   $38, %%rax, %%rdx\n");
  fprintf(f,"        addq    %%rdx, %%rdi\n");
  fprintf(f,"        movb    $43, (%%rdi)\n");
  fprintf(f,"10:\n");
  fprintf(f,"        leaq    -113(%%rbp), %%rsi\n");
  fprintf(f,"        incq    %%rax\n");
  fprintf(f,"        incq    %%rdi\n");
  fprintf(f,"        movabsq %%rax, .Lrt_nput_stack_top\n");
  fprintf(f,"11:\n");
  fprintf(f,"        movabsq $mystrcpy, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"12:\n");
  fprintf(f,"        movb    $0, -113(%%rbp)\n");
  fprintf(f,"        xorl    %%ecx, %%ecx\n");
  fprintf(f,"        xorl    %%eax, %%eax\n");
  fprintf(f,"        jmp     23f\n");
  fprintf(f,"13:\n");
  fprintf(f,"        incq    %%rbx\n");
  fprintf(f,"        testb   $1, %%bl\n");
  fprintf(f,"        jne     14f\n");
  fprintf(f,"        movabsq $.LparseinputC6, %%rdi\n");
  fprintf(f,"        jmp     18f\n");
  fprintf(f,"14:\n");
  fprintf(f,"        leaq    -113(%%rbp), %%rdi\n");
  fprintf(f,"        movb    $0, -113(%%rbp,%%rdx)\n");
  fprintf(f,"        movabsq $mytrim, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        leaq    -113(%%rbp), %%rdi\n");
  fprintf(f,"        movabsq $mystrlen, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        cmpq    $18, %%rax\n");
  fprintf(f,"        jbe     15f\n");
  fprintf(f,"        movabsq $.LparseinputC7, %%rdi\n");
  fprintf(f,"        jmp     18f\n");
  fprintf(f,"15:\n");
  fprintf(f,"        movabsq .Lrt_nput_stack_top, %%rax\n");
  fprintf(f,"        cmpq    $29, %%rax\n");
  fprintf(f,"        ja.d32  8b\n");
  fprintf(f,"        movabsq $.Linformat, %%rdx\n");
  fprintf(f,"        movabsq $.Lrt_nput_stack, %%rdi\n");
  fprintf(f,"        movb    $'S', (%%rdx,%%rax)\n");
  fprintf(f,"        imulq   $38, %%rax, %%rdx\n");
  fprintf(f,"        addq    %%rdx, %%rdi\n");
  fprintf(f,"        movb    $40, (%%rdi)\n");
  fprintf(f,"        jmp     10b\n");
  fprintf(f,"16:\n");
  fprintf(f,"        leaq    1(%%rdx), %%rax\n");
  fprintf(f,"        incq    %%rbx\n");
  fprintf(f,"        movb    %%sil, -113(%%rbp,%%rdx)\n");
  fprintf(f,"        testb   $1, %%bl\n");
  fprintf(f,"        jne     17f\n");
  fprintf(f,"        movabsq $.LparseinputC8, %%rdi\n");
  fprintf(f,"        jmp     18f\n");
  fprintf(f,"17:\n");
  fprintf(f,"        movb    $0, -113(%%rbp,%%rax)\n");
  fprintf(f,"        leaq    -113(%%rbp), %%rdi\n");
  fprintf(f,"        movabsq $mytrim, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        leaq    -113(%%rbp), %%rdi\n");
  fprintf(f,"        movabsq $mystrlen, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        cmpq    $20, %%rax\n");
  fprintf(f,"        jbe     19f\n");
  fprintf(f,"        movabsq $.LparseinputC9, %%rdi\n");
  fprintf(f,"18:\n");
  fprintf(f,"        movabsq $printstring2, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        jmp     22f\n");
  fprintf(f,"19:\n");
  fprintf(f,"        cmpb    $0, -113(%%rbp)\n");
  fprintf(f,"        je      20f\n");
  fprintf(f,"        movb    $0, -114(%%rbp,%%rax)\n");
  fprintf(f,"20:\n");
  fprintf(f,"        movabsq .Lrt_nput_stack_top, %%rax\n");
  fprintf(f,"        incq    %%r12\n");
  fprintf(f,"        cmpq    $29, %%rax\n");
  fprintf(f,"        ja.d32  8b\n");
  fprintf(f,"        movabsq $.Linformat, %%rdx\n");
  fprintf(f,"        movabsq $.Lrt_nput_stack, %%rdi\n");
  fprintf(f,"        movb    $'S', (%%rdx,%%rax)\n");
  fprintf(f,"        imulq   $38, %%rax, %%rdx\n");
  fprintf(f,"        incq    %%rax\n");
  fprintf(f,"        movabsq %%rax, .Lrt_nput_stack_top\n");
  fprintf(f,"        leaq    -113(%%rbp), %%rax\n");
  fprintf(f,"        leaq    1(%%rax), %%rsi\n");
  fprintf(f,"        addq    %%rdx, %%rdi\n");
  fprintf(f,"        movb    $12, (%%rdi)\n");
  fprintf(f,"        incq    %%rdi\n");
  fprintf(f,"        jmp     11b\n");
  fprintf(f,"21:\n");
  fprintf(f,"        leaq    1(%%rdx), %%rcx\n");
  fprintf(f,"        incq    %%r12\n");
  fprintf(f,"        movb    %%sil, -113(%%rbp,%%rdx)\n");
  fprintf(f,"        cmpq    %%r14, %%r12\n");
  fprintf(f,"        ja      22f\n");
  fprintf(f,"        cmpq    $80, %%rcx\n");
  fprintf(f,"        jbe     23f\n");
  fprintf(f,"22:\n");
  fprintf(f,"        movl    $19, %%eax\n");
  fprintf(f,"        jmp     24f\n");
  fprintf(f,"23:\n");
  fprintf(f,"        movq    %%rcx, %%rdx\n");
  fprintf(f,"        jmp     0b\n");
  fprintf(f,"24:\n");
  fprintf(f,"        cmpq    $14, %%rax\n");
  fprintf(f,"        sete    %%al\n");
  fprintf(f,"        addq    $96, %%rsp\n");
  fprintf(f,"        popq    %%rbx\n");
  fprintf(f,"        popq    %%r12\n");
  fprintf(f,"        popq    %%r13\n");
  fprintf(f,"        popq    %%r14\n");
  fprintf(f,"        popq    %%rbp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   parseinput, .-parseinput\n\n");

  fprintf(f,"        .section .rodata.str1.1,\"aMS\",@progbits,1\n\n");

  fprintf(f,".LdoinputC12:\n");
  fprintf(f,"        .string \"STRING\"\n");
  fprintf(f,".LdoinputC13:\n");
  fprintf(f,"        .string \"NUMBER\"\n");
  fprintf(f,".LdoinputC14:\n");
  fprintf(f,"        .string \"? \"\n");
  fprintf(f,".LdoinputC15:\n");
  fprintf(f,"        .string \"INVALID INPUT: expected \"\n");
  fprintf(f,".LdoinputC16:\n");
  fprintf(f,"        .string \"INVALID INPUT: string exceeds 18 bytes\"\n");
  fprintf(f,".LdoinputC17:\n");
  fprintf(f,"        .string \",\"\n\n");

  fprintf(f,"        .section .text,\"ax\",@progbits\n\n");

  fprintf(f,"        .type   doinput, @function\n");
  fprintf(f,"doinput:\n");
  fprintf(f,"        pushq   %%rbp\n");
  fprintf(f,"        movq    %%rsp, %%rbp\n");
  fprintf(f,"        pushq   %%r15\n");
  fprintf(f,"        pushq   %%r14\n");
  fprintf(f,"        pushq   %%r13\n");
  fprintf(f,"        pushq   %%r12\n");
  fprintf(f,"        pushq   %%rbx\n");
  fprintf(f,"        movq    %%rdi, %%rbx\n");
  fprintf(f,"        subq    $136, %%rsp\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $.LdoinputC14, %%rdi\n");
  fprintf(f,"        movb    $0, -129(%%rbp)\n");
  fprintf(f,"        movabsq $printstring2, %%r12\n");
  fprintf(f,"        movb    $0, -128(%%rbp)\n");
  fprintf(f,"        callq   *%%r12\n");
  fprintf(f,"        leaq    -129(%%rbp), %%rdi\n");
  fprintf(f,"        movl    $80, %%esi\n");
  fprintf(f,"        movabsq $INPUT, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        testq   %%rax,%%rax\n");
  fprintf(f,"        jnz     1f\n");
  fprintf(f,"        movabsq $.LEOF_on_input,%%rax\n");
  fprintf(f,"        jmp     *%%rax\n");
  fprintf(f,"1:\n");
  fprintf(f,"        leaq    -129(%%rbp), %%rdi\n");
  fprintf(f,"        movq    %%rax, %%rsi\n");
  fprintf(f,"        movabsq $parseinput, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        testb   %%al, %%al\n");
  fprintf(f,"        jne     3f\n");
  fprintf(f,"2:\n");
  fprintf(f,"        movabsq $.LdoinputC15, %%rdi\n");
  fprintf(f,"        xorl    %%r13d, %%r13d\n");
  fprintf(f,"        movabsq $.LdoinputC13, %%r15\n");
  fprintf(f,"        callq   *%%r12\n");
  fprintf(f,"        movq    %%rbx, %%rdi\n");
  fprintf(f,"        movabsq $mystrlen, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movl    %%eax, %%r14d\n");
  fprintf(f,"        jmp     10f\n");
  fprintf(f,"3:\n");
  fprintf(f,"        movabsq $mystrcmp, %%r13\n");
  fprintf(f,"        movq    %%rbx, %%rdi\n");
  fprintf(f,"        movabsq $.Linformat, %%rsi\n");
  fprintf(f,"        callq   *%%r13\n");
  fprintf(f,"        movq    %%r13, -160(%%rbp)\n");
  fprintf(f,"        testl   %%eax, %%eax\n");
  fprintf(f,"        je.d32  11f\n");
  fprintf(f,"        movabsq $mystrlen, %%r14\n");
  fprintf(f,"        movabsq $.Linformat, %%rdi\n");
  fprintf(f,"        callq   *%%r14\n");
  fprintf(f,"        movq    %%r14, -168(%%rbp)\n");
  fprintf(f,"        movabsq $.Lrt_nput_stack_top, %%rdx\n");
  fprintf(f,"        movl    %%eax, -148(%%rbp)\n");
  fprintf(f,"        movl    %%eax, %%eax\n");
  fprintf(f,"        cmpq    %%rax, (%%rdx)\n");
  fprintf(f,"        jne.d32 2b\n");
  fprintf(f,"        movabsq $.Lrt_nput_stack+1, %%r13\n");
  fprintf(f,"        xorl    %%r14d, %%r14d\n");
  fprintf(f,"        movabsq $.Linformat, %%r15\n");
  fprintf(f,"        jmp     7f\n");
  fprintf(f,"4:\n");
  fprintf(f,"        leaq    (%%r14,%%r15), %%rcx\n");
  fprintf(f,"        cmpb    $'N', (%%rcx)\n");
  fprintf(f,"        jne     6f\n");
  fprintf(f,"        cmpb    $'S', (%%rbx,%%r14)\n");
  fprintf(f,"        jne     6f\n");
  fprintf(f,"        movq    -168(%%rbp), %%rax\n");
  fprintf(f,"        movq    %%r13, %%rdi\n");
  fprintf(f,"        movq    %%rcx, -176(%%rbp)\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movq    -176(%%rbp), %%rcx\n");
  fprintf(f,"        cmpq    $18, %%rax\n");
  fprintf(f,"        jbe     5f\n");
  fprintf(f,"        movabsq $.LdoinputC16, %%rdi\n");
  fprintf(f,"        callq   *%%r12\n");
  fprintf(f,"        movabsq $printnewline, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        jmp     6f\n");
  fprintf(f,"5:\n");
  fprintf(f,"        movb    $'S', (%%rcx)\n");
  fprintf(f,"        movb    $40, -1(%%r13)\n");
  fprintf(f,"6:\n");
  fprintf(f,"        incq    %%r14\n");
  fprintf(f,"        addq    $38, %%r13\n");
  fprintf(f,"7:\n");
  fprintf(f,"        cmpl    %%r14d, -148(%%rbp)\n");
  fprintf(f,"        ja      4b\n");
  fprintf(f,"        movq    -160(%%rbp), %%rax\n");
  fprintf(f,"        movq    %%rbx, %%rdi\n");
  fprintf(f,"        movabsq $.Linformat, %%rsi\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        testl   %%eax, %%eax\n");
  fprintf(f,"        je      11f\n");
  fprintf(f,"        jmp     2b\n");
  fprintf(f,"8:\n");
  fprintf(f,"        testl   %%r13d, %%r13d\n");
  fprintf(f,"        je      9f\n");
  fprintf(f,"        movabsq $.LdoinputC17, %%rdi\n");
  fprintf(f,"        callq   *%%r12\n");
  fprintf(f,"9:\n");
  fprintf(f,"        movabsq $.LdoinputC12, %%rdi\n");
  fprintf(f,"        cmpb    $'S', (%%rbx,%%r13)\n");
  fprintf(f,"        cmovne  %%r15, %%rdi\n");
  fprintf(f,"        incq    %%r13\n");
  fprintf(f,"        callq   *%%r12\n");
  fprintf(f,"10:\n");
  fprintf(f,"        cmpl    %%r14d, %%r13d\n");
  fprintf(f,"        jb      8b\n");
  fprintf(f,"        movabsq $printnewline, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        jmp     0b\n");
  fprintf(f,"11:\n");
  fprintf(f,"        addq    $136, %%rsp\n");
  fprintf(f,"        popq    %%rbx\n");
  fprintf(f,"        popq    %%r12\n");
  fprintf(f,"        popq    %%r13\n");
  fprintf(f,"        popq    %%r14\n");
  fprintf(f,"        popq    %%r15\n");
  fprintf(f,"        popq    %%rbp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   doinput, .-doinput\n\n");
  fprintf(f,"        # END support for runtime INPUT subsystem\n\n");
  fflush(f);
  return true;
}

//
// This function is used to push the constant value one onto the dedicated
// floating point stack when a FOR statement does not specify a STEP value and
// the default value of one need to be used.  The f parameter is the name of
// the assembly language output file and that file must be open for output.
//
bool doforpush1(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"doforpush1(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        movabsq $1, %%rax\n");
  fprintf(f,"        cvtsi2s%c %%rax,%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return true;
}

//
// This function emits the the code for the FOR statement.  It will create the
// synthetic variables that contain the value of the termination exrpession
// (.LFORLIMIT####) and increment (.LFORINCREMENT####) using the
// next_var_number, and pass those back to the caller in the inc_var and
// limit_var parameters.  It will also create the label for the jump from next
// (.FORTEST####) and the label for the loop exit (.LFORDONE####) from the
// next_label_number.    The loop index variable must be passed in the
// ndex_var parameter.  The actual implementation is contained in the beginfor
// assembly macro.  The f parameter is the name of the assembly language output
// file and that file must be open for output.  tol_label, bol_label, inc_var,
// limit_var must be allocated by caller and exist and be appropriately sized.
// This function returns true on succes, false otherwise.
//
bool doforloopstart(FILE *f, const char *ndex_var, char *tol_label, char *bol_label, char *inc_var, char *limit_var) {
  if (NULL==f) {
    fprintf(stderr,"doforloopstart(): file parameter is NULL\n");
    return false;
  }
  snprintf(tol_label,MAX_FOR_LABEL_BYTES,".LFORTEST%04" PRIu32,next_label_number);
  snprintf(bol_label,MAX_FOR_LABEL_BYTES,".LFORDONE%04" PRIu32,next_label_number++);
  snprintf(limit_var,MAX_FOR_VAR_BYTES,".LFORLIMIT%04" PRIu32,next_var_number);
  snprintf(inc_var,MAX_FOR_VAR_BYTES,".LFORINCREMENT%04" PRIu32,next_var_number++);
  fprintf(f,"        beginfor %s,%s,%s,%s,%s\n",ndex_var,inc_var,limit_var,tol_label,bol_label);
  fflush(f);
  return true;
}

//
// This function emits the code for the NEXT statement.  The synthetic variable
// for the increment created in doforloopstart() must be passed in the ndex_var
// parameter (.LFORINCREMENT####).  The label for the jump from next
// (.FORTEST####) to the loop termination test generated by the
// doforloopstart() must be passed in the tol_label parameter.  The label for
// the for loop termination (.LFORDONE####) generated by the doforloopstart()
// must be passed in the bol_label parameter.  Finally, the loop index variable
// must be passed in the ndex_var parameter.  The actual implementation is
// contained in the endfor assembly language macro.  The f parameter is the
// name of the assembly language output file and that file must be open for
// output.  This function returns true on succes, false otherwise.
//
bool doforloopend(FILE *f,const char *tol_label, const char *bol_label, const char *ndex_var, const char *inc_var) {
  if (NULL==f) {
    fprintf(stderr,"doforloopend(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        endfor %s,%s,%s,%s\n",ndex_var,inc_var,tol_label,bol_label);
  fflush(f);
  return true;
}

//
// This function emits the code for the TAB() function call, adjust the current
// column output buffer position.  The column number value must be stored on
// top of the dedicated floating point stack, but only the integer part will be
// used.  The f parameter is the name of the assembly language output file and
// that file must be open for output.  This function returns true on success,
// false otherwise.
//
bool dotab(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dotab(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        popxmm 0\n");
  fprintf(f,"        cvts%c2si %%xmm0,%%rdi\n",(use_double?'d':'s'));
  fprintf(f,"        movabsq $tab, %%rax\n");
  fprintf(f,"        callq  *%%rax\n");
  fflush(f);
  return true;
}

//
// This function will emit code for an INPUT statement.  The nfmt parameter
// will contain the format string that specifies the type and number of
// fields.  Each byte of the ASCIIZ string will contain either an 'N' or
// an 'S'.  An 'N' indicates that a numeric (floating point) value is required.
// An 'S' indicates that a string value is required.  For example, given
// the Minimal BASIC statement:
// INPUT A,B$,C[1]
// the corresponding format string would be "NSN".  The f parameter is the name
// of the assembly language output file and that file must be open for output.
// This function returns true on success, false otherwise.
//
bool do_INPUT(FILE *f,const char *nfmt) {
  if (NULL==f) {
    fprintf(stderr,"do_INPUT(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==nfmt) || (0==nfmt[0])) {
    fprintf(stderr,"do_INPUT(): nfmt parameter is NULL or empty\n");
    return false;
  }
  fprintf(f,"        movabsq $%s, %%rdi\n",nfmt);
  fprintf(f,"        movabsq $doinput, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  return true;
}

//
// This function will emit code for input from STDIN of a scalar numeric
// variable whose ASCIIZ name is passed in the parameter varname.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  The lno parameter is the BASIC source line number
// of the INPUT statement that required this input, and is used in generating
// error messages.  The slotno parameter is the index of the variable in the
// INPUT statement, zero-based.  So for INPUT I,J, the slotno for I is zero,
// and the slnot for J is one.  This function returns true on success, false
// otherwise.
//
bool doinput_nvar(FILE *f,const char *varname,const uint32_t slotno,const uint32_t lno) {
  if (NULL==f) {
    fprintf(stderr,"doinput_nvar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"doinput_nvar(): varname parameter is NULL or empty\n");
    return false;
  }
  fprintf(f,"# scalar numeric INPUT %s (convert, check for overflow, save in temp slot)\n",varname);
  fprintf(f,"        movabsq $.Lrt_nput_stack, %%r15\n");
  fprintf(f,"        movabsq $%" PRIu32 ",%%rax           # slot number %" PRIu32 " on .Lrt_nput_stack\n",slotno,slotno);
  fprintf(f,"        movabsq $%" PRIu32 ",%%r9           # width of a record on .Lrt_nput_stack\n",UINT32_C(2)*MAX_STRING_BYTES+UINT32_C(2));
  fprintf(f,"        mulq    %%r9\n");
  fprintf(f,"        leaq    1(%%r15,%%rax),%%rdi # 1st arg is ASCIZ buffer\n");
  fprintf(f,"        xorl    %%esi, %%esi        # 2nd arg is NULL\n");
  fprintf(f,"        movabsq $strtod, %%rax     # call code to do actual conversion from ASCIIZ to floating point\n");
  fprintf(f,"        callq   *%%rax\n");
  if (!use_double) // 32 bit math
    fprintf(f,"        cvtsd2ss %%xmm0, %%xmm1\n");
  else // 64 bit math
    fprintf(f,"        movsd    %%xmm0, %%xmm1\n");
  fprintf(f,"        # if xmm1 is +Infinity then overflow\n");
  fprintf(f,"        mov%s    $PInf,%%r13%c  # move constant +INF into r13\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsr%c  $0,%%r13%c,%%xmm0  # move +INF into xmm0\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        subq    $%" PRIu32 ",%%rsp         # allocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
    fprintf(f,"                                # it is a union location for \n");
    fprintf(f,"                                # converting between %" PRIu32 " bit %s and \n",(use_double?UINT32_C(64):UINT32_C(32)),(use_double?"double":"float"));
    fprintf(f,"                                # unsigned %" PRIu32 " int\n",(use_double?UINT32_C(64):UINT32_C(32)));
    fprintf(f,"        mov%c    %%r13%c,(%%rsp)    # temp = +INF\n",(use_double?'q':'l'),(use_double?' ':'d'));
    fprintf(f,"        movs%c   (%%rsp),%%xmm0    # move +INF into xmm0\n",(use_double?'d':'s'));
    fprintf(f,"        addq    $%" PRIu32 ",%%rsp         # deallocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
  }
  fprintf(f,"        comis%c  %%xmm1,%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        jne     0f\n");
  fprintf(f,"        # got overflow\n");
  fprintf(f,"        movabsq $.Loverflow_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r13\n");
  fprintf(f,"        callq   *%%r13\n");
  fprintf(f,"        movabsq $.LLINE%04" PRIu32 ",%%rax\n",lno);
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $.Lrt_nput_stack, %%r15\n");
  fprintf(f,"        movabsq $%" PRIu32 ", %%rax\n",slotno);
  fprintf(f,"        movabsq $%" PRIu32 ", %%r9\n",UINT32_C(2)*MAX_STRING_BYTES+UINT32_C(2));
  fprintf(f,"        mulq    %%r9\n");
  fprintf(f,"        movs%c   %%xmm1,1(%%r15,%%rax)\n",(use_double?'d':'s'));
  fflush(f);
  return true;
}

//
// This function will emit code for input from STDIN of a scalar string
// variable whose ASCIIZ name is passed in the parameter varname.  The slotno
// parameter is the index of the variable in the INPUT statement, zero-based.
// So for INPUT I$,J$, the slotno for I$ is zero, and the slnot for J$ is one.
// The f parameter is the name of the assembly language output file and that
// file must be open for output.  This function returns true on success, false
// otherwise.
//
bool doinput_svar(FILE *f,const char *varname,const uint32_t slotno) {
  if (NULL==f) {
    fprintf(stderr,"doinput_svar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"doinput_nvar(): varname parameter is NULL or empty\n");
    return false;
  }
  need_strcpy=true;
  fprintf(f,"# scalar string INPUT %s (do nothing right now, already in temp slot)\n",varname);
  fflush(f);
  return true;
}

//
// This function will emit code for input from STDIN of an array numeric
// variable whose ASCIIZ name is passed in the parameter varname.  The f
// parameter is the name of the assembly language output file and that file
// must be open for output.  The lno parameter is the BASIC source line number
// of the INPUT statement that required this input, and is used in generating
// error messages.  The slotno parameter is the index of the variable in the
// INPUT statement, zero-based.  So for INPUT I,J, the slotno for I is zero,
// and the slnot for J is one.  This function returns true on success, false
// otherwise.
//
bool doinput_navar(FILE *f, const char *varname, const uint32_t slotno,const uint32_t lno) {
  if (NULL==f) {
    fprintf(stderr,"doinput_navar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"doinput_navar(): varname parameter is NULL or empty\n");
    return false;
  }
  fprintf(f,"# array numeric INPUT %s (convert, check for overflow, save in temp slot)\n",varname);
  fprintf(f,"        movabsq $.Lrt_nput_stack, %%r15\n");
  fprintf(f,"        movabsq $%" PRIu32 ",%%rax           # slot number %" PRIu32 " on .Lrt_nput_stack\n",slotno,slotno);
  fprintf(f,"        movabsq $%" PRIu32 ",%%r9           # width of a record on .Lrt_nput_stack\n",UINT32_C(2)*MAX_STRING_BYTES+UINT32_C(2));
  fprintf(f,"        mulq    %%r9\n");
  fprintf(f,"        leaq    1(%%r15,%%rax),%%rdi # 1st arg is ASCIZ buffer\n");
  fprintf(f,"        xorl    %%esi, %%esi        # 2nd arg is NULL\n");
  fprintf(f,"        movabsq $strtod, %%rax     # call code to do actual conversion from ASCIIZ to floating point\n");
  fprintf(f,"        callq   *%%rax\n");
  if (!use_double) // 32 bit math
    fprintf(f,"        cvtsd2ss %%xmm0, %%xmm1\n");
  else // 64 bit math
    fprintf(f,"        movsd    %%xmm0, %%xmm1\n");
  fprintf(f,"        # if xmm1 is +Infinity then overflow\n");
  fprintf(f,"        mov%s    $PInf,%%r13%c  # move constant +INF into r13\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsr%c  $0,%%r13%c,%%xmm0  # move +INF into xmm0\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        subq    $%" PRIu32 ",%%rsp         # allocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
    fprintf(f,"                                # it is a union location for \n");
    fprintf(f,"                                # converting between %" PRIu32 " bit %s and \n",(use_double?UINT32_C(64):UINT32_C(32)),(use_double?"double":"float"));
    fprintf(f,"                                # unsigned %" PRIu32 " int\n",(use_double?UINT32_C(64):UINT32_C(32)));
    fprintf(f,"        mov%c    %%r13%c,(%%rsp)    # temp = +INF\n",(use_double?'q':'l'),(use_double?' ':'d'));
    fprintf(f,"        movs%c   (%%rsp),%%xmm0    # move +INF into xmm0\n",(use_double?'d':'s'));
    fprintf(f,"        addq    $%" PRIu32 ",%%rsp         # deallocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
  }
  fprintf(f,"        comis%c  %%xmm1,%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        jne     0f\n");
  fprintf(f,"        # got overflow\n");
  fprintf(f,"        movabsq $.Loverflow_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r13\n");
  fprintf(f,"        callq   *%%r13\n");
  fprintf(f,"        movabsq $.LLINE%04" PRIu32 ",%%rax\n",lno);
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $.Lrt_nput_stack, %%r15\n");
  fprintf(f,"        movabsq $%" PRIu32 ", %%rax\n",slotno);
  fprintf(f,"        movabsq $%" PRIu32 ", %%r9\n",UINT32_C(2)*MAX_STRING_BYTES+UINT32_C(2));
  fprintf(f,"        mulq    %%r9\n");
  fprintf(f,"        movs%c   %%xmm1,1(%%r15,%%rax)\n",(use_double?'d':'s'));
  return true;
}

//
// This function is used to copy the temporary value input for a scalar numeric variable
// from the temporary slot specified by slotno to the actual variable specified by varname.
// The lno parameter is the BASIC source code line number of the INPUT statement that is
// used when reporting errors.  The f parameter is the name of the assembly
// language output file and that file must be open for output.  This function
// returns true on success, false otherwise.
//
bool doinput_nvar_phase2(FILE *f,const char *varname,const uint32_t slotno,const uint32_t lno) {
  if (NULL==f) {
    fprintf(stderr,"doinput_nvar_phase2(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"doinput_nvar_phase2(): varname parameter is NULL or empty\n");
    return false;
  }
  if (UINT32_C(0)==slotno)
    fprintf(f,"\n        # INPUT ok, copy from temp locations to real locations\n\n");
  fprintf(f,"# scalar numeric INPUT %s phase 2\n",varname);
  fprintf(f,"        movabsq $.Lrt_nput_stack, %%r15\n");
  switch(slotno) {
    case UINT32_C(0):
      fprintf(f,"        movs%c   1(%%r15),%%xmm1\n",(use_double?'d':'s'));
      break;
    case UINT32_C(1):
      fprintf(f,"        movabsq $%" PRIu32 ", %%rax\n",UINT32_C(2)*MAX_STRING_BYTES+UINT32_C(2));
      fprintf(f,"        movs%c   1(%%r15,%%rax),%%xmm1\n",(use_double?'d':'s'));
      break;
    default:
      fprintf(f,"        movabsq $%" PRIu32 ", %%r9\n",UINT32_C(2)*MAX_STRING_BYTES+UINT32_C(2));
      fprintf(f,"        movabsq $%" PRIu32 ", %%rax\n",slotno);
      fprintf(f,"        mulq    %%r9\n");
      fprintf(f,"        movs%c   1(%%r15,%%rax),%%xmm1\n",(use_double?'d':'s'));
      break;
  }
  fprintf(f,"        movabsq $%s, %%r14\n",varname);
  fprintf(f,"        movs%c   %%xmm1, (%%r14)\n",(use_double?'d':'s'));
  fflush(f);
  return true;
}

//
// This function is used to copy the temporary value input for a scalar string
// variable from the temporary slot specified by slotno to the actual variable
// specified by varname.  The f parameter is the name of the assembly language
// output file and that file must be open for output.  This function returns
// true on success, false otherwise.
//
bool doinput_svar_phase2(FILE *f,const char *varname,const uint32_t slotno) {
  if (NULL==f) {
    fprintf(stderr,"doinput_svar_phase2(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"doinput_nvar_phase2(): varname parameter is NULL or empty\n");
    return false;
  }
  if (UINT32_C(0)==slotno)
    fprintf(f,"\n        # INPUT ok, copy from temp locations to real locations\n\n");
  fprintf(f,"# scalar string INPUT %s phase 2\n",varname);
  fprintf(f,"        movabsq $.Lrt_nput_stack, %%r15\n");
  switch(slotno) {
    case UINT32_C(0):
      fprintf(f,"        leaq    1(%%r15),%%rsi # 2nd arg is source\n");
      break;
    case UINT32_C(1):
      fprintf(f,"        movabsq $%" PRIu32 ", %%rax\n",UINT32_C(2)*MAX_STRING_BYTES+UINT32_C(2));
      fprintf(f,"        leaq    1(%%r15,%%rax),%%rsi # 2nd arg is source\n");
      break;
    default:
      fprintf(f,"        movabsq $%" PRIu32 ", %%r9\n",UINT32_C(2)*MAX_STRING_BYTES+UINT32_C(2));
      fprintf(f,"        movabsq $%" PRIu32 ", %%rax\n",slotno);
      fprintf(f,"        mulq    %%r9\n");
      fprintf(f,"        leaq    1(%%r15,%%rax),%%rsi # 2nd arg is source\n");
      break;
  }
  fprintf(f,"        movabsq $%s, %%rdi  # 1st arg is destination\n",varname);
  fprintf(f,"        movabsq $mystrcpy, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fflush(f);
  return true;
}

//
// This function is used to copy the temporary value input for an array numeric
// variable from the temporary slot specified by slotno to the actual variable
// specified by varname.  The lno parameter is the BASIC source code line
// number of the INPUT statement that is used when reporting errors.  The
// maximum permitted index for the first dimension is passed in the maxrows
// parameter, and the maximum permitted index for the second dimension is
// passed in the maxcolumns parameter.  If the array is one-dimensional (vector)
// then maxcolumns must be zero.  The maxrows and maxcolumns values are used for array
// index bounds checking.  The f parameter is the name of the assembly language
// output file and that file must be open for output.  This function returns
// true on success, false otherwise.
//
bool doinput_navar_phase2(FILE *f, const char *varname, const uint32_t maxrows, const uint32_t maxcolumns,const uint32_t slotno,const uint32_t lno) {
  if (NULL==f) {
    fprintf(stderr,"doinput_navar_phase2(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"doinput_navar_phase2(): varname parameter is NULL or empty\n");
    return false;
  }
  if (UINT32_C(0)==slotno)
    fprintf(f,"\n        # INPUT ok, copy from temp locations to real locations\n\n");
  fprintf(f,"# array numeric INPUT %s phase 2\n",varname);
  fprintf(f,"        movabsq $.Lrt_nput_stack, %%r15\n");
  switch(slotno) {
    case UINT32_C(0):
      fprintf(f,"        movs%c   1(%%r15),%%xmm1\n",(use_double?'d':'s'));
      break;
    case UINT32_C(1):
      fprintf(f,"        movabsq $%" PRIu32 ", %%rax\n",UINT32_C(2)*MAX_STRING_BYTES+UINT32_C(2));
      fprintf(f,"        movs%c   1(%%r15,%%rax),%%xmm1\n",(use_double?'d':'s'));
      break;
    default:
      fprintf(f,"        movabsq $%" PRIu32 ", %%r9\n",UINT32_C(2)*MAX_STRING_BYTES+UINT32_C(2));
      fprintf(f,"        movabsq $%" PRIu32 ", %%rax\n",slotno);
      fprintf(f,"        mulq    %%r9\n");
      fprintf(f,"        movs%c   1(%%r15,%%rax),%%xmm1\n",(use_double?'d':'s'));
      break;
  }
  fprintf(f,"        pushxmm 1\n");
  return ((maxcolumns==UINT32_C(0))?dostore_1D_navar(f,varname,maxrows):dostore_2D_navar(f,varname,maxrows,maxcolumns));
}

//
// This function emits code for the READ statement when the destination
// is a numeric scalar variable.  The f parameter is the name of the
// assembly language output file and that file must be open.  The varname
// parameter is the name of the scalar numeric variable where the
// value read from the DATA list will be stored.  This function returns
// true on success, false otherwise.
//
bool doread_nvar(FILE *f,const char *varname) {
  if (NULL==f) {
    fprintf(stderr,"doread_nvar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"doread_nvar(): varname parameter is NULL or empty\n");
    return false;
  }
  fprintf(f,"# scalar numeric READ %s\n",varname);
  fprintf(f,"        movabsq $.Ldata_item_ptr, %%rax\n");
  fprintf(f,"        movq    (%%rax), %%rax\n");
  fprintf(f,"        movabsq $.Ldata_item_count, %%rbx\n");
  fprintf(f,"        movq    (%%rbx), %%rbx\n");
  fprintf(f,"        cmpq    %%rbx, %%rax\n");
  fprintf(f,"        jb      0f\n");
  fprintf(f,"        movabsq $.Lout_of_data, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $.Ldata_items, %%rbx     # base of data_items array of struct pointers\n");
  fprintf(f,"        movq    (%%rbx,%%rax,8),%%rax      # load pointer value in slot rax\n");
  fprintf(f,"        cmpb    $2,(%%rax)\n");
  fprintf(f,"        je      1f\n");
  fprintf(f,"        movabsq $.Lbad_number_read, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movs%c   1(%%rax),%%xmm0           # ok, load %s 2 bytes into that structure into xmm0\n",(use_double?'d':'s'),(use_double?"double":"float"));
  fprintf(f,"        movabsq $%s, %%rax\n",varname);
  fprintf(f,"        movs%c   %%xmm0,(%%rax)\n",(use_double?'d':'s'));
  fprintf(f,"        movabsq $.Ldata_item_ptr, %%rax\n");
  fprintf(f,"        incq    (%%rax)                  # increment data_item_ptr in memory but not register\n");
  fprintf(f,"        mov%s    $PInf,%%r15%c\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsr%c  $0,%%r15%c,%%xmm1   # move the +INF into xmm1\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        subq    $%" PRIu32 ",%%rsp         # allocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
    fprintf(f,"                                # it is a union location for \n");
    fprintf(f,"                                # converting between %" PRIu32 " bit %s and \n",(use_double?UINT32_C(64):UINT32_C(32)),(use_double?"double":"float"));
    fprintf(f,"                                # unsigned %" PRIu32 " int\n",(use_double?UINT32_C(64):UINT32_C(32)));
    fprintf(f,"        mov%c    %%r15%c,(%%rsp)    # temp = +INF\n",(use_double?'q':'l'),(use_double?' ':'d'));
    fprintf(f,"        movs%c   (%%rsp),%%xmm1    # xmm1 = +INF\n",(use_double?'d':'s'));
  }
  fprintf(f,"        ucomis%c %%xmm1,%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        jne     2f\n");
  fprintf(f,"        jmp     3f\n");
  fprintf(f,"2:\n");
  fprintf(f,"        mov%s    $NInf,%%r15%c\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsr%c  $0,%%r15%c,%%xmm1   # move the -INF into xmm1\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        mov%c    %%r15%c,(%%rsp)    # temp = -INF\n",(use_double?'q':'l'),(use_double?' ':'d'));
    fprintf(f,"        movs%c   (%%rsp),%%xmm1    # xmm1 = -INF\n",(use_double?'d':'s'));
  }
  fprintf(f,"        ucomis%c %%xmm1,%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        jne     4f\n");
  fprintf(f,"3:\n");
  fprintf(f,"        # overflow message\n");
  fprintf(f,"        movabsq $.Loverflow_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r15\n");
  fprintf(f,"        callq   *%%r15\n");
  fprintf(f,"4:\n");
  if (!use_SSE4_1) {
    fprintf(f,"        addq    $%" PRIu32 ",%%rsp         # deallocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
  }
  fflush(f);
  return true;
}

//
// This function emits code for the READ statement when the destination
// is a scalar string variable.  The f parameter is the name of the
// assembly language output file and that file must be open.  The varname
// parameter is the name of the scalar string variable where the
// value read from the DATA list will be stored.  This function returns
// true on success, false otherwise.
//
bool doread_svar(FILE *f,const char *varname) {
  if (NULL==f) {
    fprintf(stderr,"doread_svar(): file parameter is NULL\n");
    return false;
  }
  need_strcpy=true;
  fprintf(f,"# READ %s\n",varname);
  fprintf(f,"        movabsq $.Ldata_item_ptr, %%rax\n");
  fprintf(f,"        movq    (%%rax), %%rax\n");
  fprintf(f,"        movabsq $.Ldata_item_count, %%rbx\n");
  fprintf(f,"        movq    (%%rbx), %%rbx\n");
  fprintf(f,"        cmpq    %%rbx, %%rax\n");
  fprintf(f,"        jb      0f\n");
  fprintf(f,"        movabsq $.Lout_of_data, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $.Ldata_items, %%rbx     # base of data_items array of struct pointers\n");
  fprintf(f,"        movq    (%%rbx,%%rax,8),%%rsi      # load pointer value in slot rax\n");
  fprintf(f,"        cmpb    $2,(%%rsi)               # unquoted string?\n");
  fprintf(f,"        jne     1f\n");
  fprintf(f,"        movq    %d(%%rsi),%%rsi            # ok, load string pointer 6 bytes into that structure into rsi\n",(use_double?UINT32_C(10):UINT32_C(6)));
  fprintf(f,"        jmp     3f\n");
  fprintf(f,"1:\n");
  fprintf(f,"        cmpb    $0,(%%rsi)               # quoted string?\n");
  fprintf(f,"        je      2f\n");
  fprintf(f,"        movabsq $.Lbad_string_read, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"2:\n");
  fprintf(f,"        movq    2(%%rsi),%%rsi            # ok, load string pointer 2 bytes into that structure into rsi\n");
  fprintf(f,"3:\n");
  fprintf(f,"        movabsq $%s, %%rdi\n",varname);
  fprintf(f,"        movabsq $mystrcpy, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movabsq $.Ldata_item_ptr, %%rax\n");
  fprintf(f,"        incq    (%%rax)                  # increment data_item_ptr in memory but not register\n");
  fflush(f);
  return true;
}

//
// This function emits code for the READ statement when the destination is a
// numeric array variable.  The f parameter is the name of the assembly
// language output file and that file must be open.  The varname parameter is
// the name of the numeric array variable where the value read from the DATA
// list will be stored.  This function returns true on success, false
// otherwise.  Note that this function calls dostore_navar() to store the value
// read from the xmm0 register, and that dstore_navar() does the required array
// bounds checking.
//
bool doread_navar(FILE *f, const char *varname, const uint32_t maxrows, const uint32_t maxcolumns) {
  if (NULL==f) {
    fprintf(stderr,"dosread_navar(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==varname) || (0==varname[0])) {
    fprintf(stderr,"doread_navar(): varname parameter is NULL or empty\n");
    return false;
  }
  fprintf(f,"# array numeric READ %s\n",varname);
  fprintf(f,"        movabsq $.Ldata_item_ptr, %%rax\n");
  fprintf(f,"        movq    (%%rax), %%rax\n");
  fprintf(f,"        movabsq $.Ldata_item_count, %%rbx\n");
  fprintf(f,"        movq    (%%rbx), %%rbx\n");
  fprintf(f,"        cmpq    %%rbx, %%rax\n");
  fprintf(f,"        jb      0f\n");
  fprintf(f,"        movabsq $.Lout_of_data, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq $.Ldata_items, %%rbx     # base of data_items array of struct pointers\n");
  fprintf(f,"        movq    (%%rbx,%%rax,8),%%rax      # load pointer value in slot rax\n");
  fprintf(f,"        cmpb    $2,(%%rax)               # number?\n");
  fprintf(f,"        je      1f\n");
  fprintf(f,"        movabsq $.Lbad_number_read, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movs%c   1(%%rax),%%xmm0           # ok, load %s 2 bytes into that structure into xmm0\n",(use_double?'d':'s'),(use_double?"double":"float"));
  fprintf(f,"        movabsq $.Ldata_item_ptr, %%rax\n");
  fprintf(f,"        incq    (%%rax)                  # increment data_item_ptr in memory but not register\n");
  fprintf(f,"        mov%s    $PInf,%%r15%c\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsr%c  $0,%%r15%c,%%xmm1   # move the +INF into xmm1\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        subq    $%" PRIu32 ",%%rsp         # allocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
    fprintf(f,"                                # it is a union location for \n");
    fprintf(f,"                                # converting between %" PRIu32 " bit %s and \n",(use_double?UINT32_C(64):UINT32_C(32)),(use_double?"double":"float"));
    fprintf(f,"                                # unsigned %" PRIu32 " int\n",(use_double?UINT32_C(64):UINT32_C(32)));
    fprintf(f,"        mov%c    %%r15%c,(%%rsp)    # temp = +INF\n",(use_double?'q':'l'),(use_double?' ':'d'));
    fprintf(f,"        movs%c   (%%rsp),%%xmm1\n",(use_double?'d':'s'));
    fprintf(f,"        ucomis%c %%xmm1,%%xmm0\n",(use_double?'d':'s'));
  }
  fprintf(f,"        jne     2f\n");
  fprintf(f,"        jmp     3f\n");
  fprintf(f,"2:\n");
  fprintf(f,"        mov%s    $NInf,%%r15%c\n",(use_double?"absq":"l"),(use_double?' ':'d'));
  if (use_SSE4_1) {
    fprintf(f,"        pinsr%c  $0,%%r15%c,%%xmm1   # move the -INF into xmm1\n",(use_double?'q':'d'),(use_double?' ':'d'));
  } else {
    fprintf(f,"        mov%c    %%r15%c,(%%rsp)    # temp = -INF\n",(use_double?'q':'l'),(use_double?' ':'d'));
    fprintf(f,"        movs%c   (%%rsp),%%xmm1\n",(use_double?'d':'s'));
  }
  fprintf(f,"        ucomis%c %%xmm1,%%xmm0\n",(use_double?'d':'s'));
  fprintf(f,"        jne     4f\n");
  fprintf(f,"3:\n");
  fprintf(f,"        # overflow message\n");
  fprintf(f,"        movabsq $.Loverflow_msg,%%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r15\n");
  fprintf(f,"        callq   *%%r15\n");
  fprintf(f,"4:\n");
  if (!use_SSE4_1) {
    fprintf(f,"        addq    $%" PRIu32 ",%%rsp         # deallocate scratch %" PRIu32 " byte area\n",(use_double?UINT32_C(8):UINT32_C(4)),(use_double?UINT32_C(8):UINT32_C(4)));
  }
  fprintf(f,"        pushxmm 0\n");
  fflush(f);
  return ((maxcolumns==UINT32_C(0))?dostore_1D_navar(f,varname,maxrows):dostore_2D_navar(f,varname,maxrows,maxcolumns));
}

//
// This function emits code to implement the Minimal BASIC RESTORE statement.
// The f parameter is the file handle for the output file, and the file must
// already be open for output.  This will reset the DATA statement pointer to
// the very first data item.  It returns true on success, false otherwise.
//
bool dorestore(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"dorestore(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        # RESTORE data pointer\n");
  fprintf(f,"        movabsq $.Ldata_item_ptr, %%rax\n");
  fprintf(f,"        movq    $0, (%%rax)\n");
  fflush(f);
  return true;
}

//
// This function emits code to implement the Minimal BASIC PRINT statement
// support.  The f parameter is the file handle for the output file, and the
// file must already be open for output.  It returns true on success, false
// otherwise.
//
static bool write_printzone(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"write_printzone(): file parameter is NULL\n");
    return false;
  }
  fprintf(f,"        # begin PRINT ZONE code\n\n");
  fprintf(f,"        .set YES, 1\n");
  fprintf(f,"        .set NO, 0\n");
  fprintf(f,"        .set OUTPUT_WIDTH, %" PRIu32 "\n",(use_double_output?UINT32_C(132):UINT32_C(80)));
  fprintf(f,"        .set SIGNIFICANCE_WIDTH, %" PRIu32 "\n",(use_double_output?SIGNIFICANCE_WIDTH_WIDE_OUTPUT:SIGNIFICANCE_WIDTH_NARROW_OUTPUT));
  fprintf(f,"        .set EXRAD_WIDTH, %" PRIu32 "\n",(use_double_output?EXRAD_WIDE_OUTPUT:EXRAD_NARROW_OUTPUT));
  fprintf(f,"        .set ZONE_WIDTH, SIGNIFICANCE_WIDTH+EXRAD_WIDTH+%" PRIu32 "\n",(use_double_output?UINT32_C(5):UINT32_C(6)));
  fprintf(f,"        .set MAX_ZONE, OUTPUT_WIDTH/ZONE_WIDTH\n\n");

  fprintf(f,"        .section .data,\"aw\",@progbits\n\n");

  fprintf(f,"        .align 8\n\n");
  fprintf(f,"        .type   .Lobuf_next, @object\n");
  fprintf(f,"        .size   .Lobuf_next, 8\n");
  fprintf(f,".Lobuf_next:\n");
  fprintf(f,"        .quad   1\n\n");
  fprintf(f,"        .type   .Lcurzone, @object\n");
  fprintf(f,"        .size   .Lcurzone, 8\n");
  fprintf(f,".Lcurzone:\n");
  fprintf(f,"        .quad   1\n\n");
  fprintf(f,"        .size   .LOBUF, OUTPUT_WIDTH+1\n");
  fprintf(f,".LOBUF:\n");
  fprintf(f,"        .asciz  \"OUTPUT BUFFER IS UNINITIALIZED\"\n");
  fprintf(f,"        .set    msgcnt, .-.LOBUF\n");
  fprintf(f,"        .zero   OUTPUT_WIDTH-30+1\n\n");

  fprintf(f,"        .section .rodata.str1.1,\"aMS\",@progbits,1\n\n");

  fprintf(f,"        .type   .LTABERRMSG,@object\n");
  fprintf(f,".LTABERRMSG:\n");
  fprintf(f,"        .asciz  \"ERROR: tab column is less than minimum output column 1\"\n");
  fprintf(f,"        .size   .LTABERRMSG, .-.LTABERRMSG\n\n");
  fprintf(f,"        .type   .LAPPENDBUFERRMSG1,@object\n");
  fprintf(f,".LAPPENDBUFERRMSG1:\n");
  fprintf(f,"        .string \"NULL pointer sent to appendbuf\"\n");
  fprintf(f,"        .set    .LAPPENDBUFERRMSG1len, .-.LAPPENDBUFERRMSG1\n\n");

  fprintf(f,"        .section .text,\"ax\",@progbits\n\n");

  fprintf(f,"        # procedure takes 1 arg, column number, in rdi\n");
  fprintf(f,"        .type   tab, @function\n");
  fprintf(f,"tab:\n");
  fprintf(f,"        testq   %%rdi, %%rdi\n");
  fprintf(f,"        pushq   %%rbx\n");
  fprintf(f,"        movq    %%rdi, %%rbx\n");
  fprintf(f,"        pushq   %%rdx\n");
  fprintf(f,"        jne     0f\n");
  fprintf(f,"        movabsq $.LTABERRMSG, %%rsi\n");
  fprintf(f,"        movabsq $exception_non_fatal,%%r15\n");
  fprintf(f,"        callq   *%%r15\n");
  fprintf(f,"        movq    $1, %%rdi\n");
  fprintf(f,"0:\n");
  fprintf(f,"        cmpq    $OUTPUT_WIDTH, %%rdi\n");
  fprintf(f,"        jbe     1f\n");
  fprintf(f,"        movl    $OUTPUT_WIDTH, %%ecx\n");
  fprintf(f,"        movq    %%rdi, %%rax\n");
  fprintf(f,"        xorl    %%edx, %%edx\n");
  fprintf(f,"        divq    %%rcx\n");
  fprintf(f,"        movq    %%rdx, %%rbx\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movabsq $.Lobuf_next, %%rax\n");
  fprintf(f,"        cmpq    (%%rax), %%rbx\n");
  fprintf(f,"        jae     2f\n");
  fprintf(f,"        movabsq $outputbuf, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"2:\n");
  fprintf(f,"        movq    %%rbx, %%rax\n");
  fprintf(f,"        movabsq %%rax, .Lobuf_next\n");
  fprintf(f,"        movl    $ZONE_WIDTH, %%ecx\n");
  fprintf(f,"        xorl    %%edx, %%edx\n");
  fprintf(f,"        decq    %%rax\n");
  fprintf(f,"        divq    %%rcx\n");
  fprintf(f,"        incq    %%rax\n");
  fprintf(f,"        movabsq %%rax, .Lcurzone\n");
  fprintf(f,"        popq    %%rax\n");
  fprintf(f,"        popq    %%rbx\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   tab, .-tab\n\n");
  fprintf(f,"        # procedure takes no arguments\n");
  fprintf(f,"        .type   initbuf, @function\n");
  fprintf(f,"initbuf:\n");
  fprintf(f,"        movabsq $.LOBUF, %%rdx\n");
  fprintf(f,"        xorl    %%eax, %%eax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movb    $' ', (%%rdx,%%rax)\n");
  fprintf(f,"        incq    %%rax\n");
  fprintf(f,"        cmpq    $OUTPUT_WIDTH, %%rax\n");
  fprintf(f,"        jne     0b\n");
  fprintf(f,"        movabsq $.LOBUF+OUTPUT_WIDTH, %%rax\n");
  fprintf(f,"        movb    $0, (%%rax)\n");
  fprintf(f,"        movabsq $.Lobuf_next, %%rax\n");
  fprintf(f,"        movq    $1, (%%rax)\n");
  fprintf(f,"        movabsq $.Lcurzone, %%rax\n");
  fprintf(f,"        movq    $1, (%%rax)\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   initbuf, .-initbuf\n\n");
  fprintf(f,"        # procedure takes no arguments\n");
  fprintf(f,"        .type   outputbuf, @function\n");
  fprintf(f,"outputbuf:\n");
  fprintf(f,"        movabsq $.LOBUF, %%rax\n");
  fprintf(f,"        movb    $0, OUTPUT_WIDTH(%%rax)\n");
  fprintf(f,"        movq    $OUTPUT_WIDTH, %%rdx\n");
  fprintf(f,"        movabsq $.LOBUF, %%rsi\n");
  fprintf(f,"        movq    $STDOUT, %%rdi\n");
  fprintf(f,"        movq    $__NR_write, %%rax\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"        movq    $1, %%rdx\n");
  fprintf(f,"        movabsq $.LNL, %%rsi\n");
  fprintf(f,"        movabsq $STDOUT, %%rdi\n");
  fprintf(f,"        movabsq $__NR_write, %%rax\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"        movabsq $initbuf, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   outputbuf, .-outputbuf\n\n");
  fprintf(f,"        # procedure takes one argument pointer to ASCIIZ string to append in rdi\n");
  fprintf(f,"        .type   appendbuf, @function\n");
  fprintf(f,"appendbuf:\n");
  fprintf(f,"        testq   %%rdi, %%rdi\n");
  fprintf(f,"        pushq   %%r14\n");
  fprintf(f,"        pushq   %%r13\n");
  fprintf(f,"        pushq   %%r12\n");
  fprintf(f,"        movq    %%rdi, %%r12\n");
  fprintf(f,"        pushq   %%rbx\n");
  fprintf(f,"        jne     0f\n");
  fprintf(f,"        movq    $.LAPPENDBUFERRMSG1len, %%rdx\n");
  fprintf(f,"        movq    $STDERR, %%rdi\n");
  fprintf(f,"        movabsq $.LAPPENDBUFERRMSG1, %%rsi\n");
  fprintf(f,"        movq    $STDERR, %%rdi\n");
  fprintf(f,"        movq    $__NR_write, %%rax\n");
  fprintf(f,"        syscall\n");
  fprintf(f,"        movabsq $badxit, %%rax\n");
  fprintf(f,"        jmpq    *%%rax\n");
  fprintf(f,"0:\n");
  fprintf(f,"        movabsq .Lcurzone, %%rax\n");
  fprintf(f,"        movabsq $.Lobuf_next, %%rdx\n");
  fprintf(f,"        imulq   $ZONE_WIDTH, %%rax, %%rax\n");
  fprintf(f,"        subq    $ZONE_WIDTH-1, %%rax\n");
  fprintf(f,"        cmpq    %%rax, (%%rdx)\n");
  fprintf(f,"        jae     1f\n");
  fprintf(f,"        movq    %%rax, (%%rdx)\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movabsq $mystrlen, %%rax\n");
  fprintf(f,"        movq    %%r12, %%rdi\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movq    %%rax, %%r13\n");
  fprintf(f,"        movabsq .Lobuf_next, %%rax\n");
  fprintf(f,"        cmpq    $1, %%rax\n");
  fprintf(f,"        jbe     2f\n");
  fprintf(f,"        leaq    -1(%%r13,%%rax), %%rax\n");
  fprintf(f,"        cmpq    $OUTPUT_WIDTH, %%rax\n");
  fprintf(f,"        jbe     2f\n");
  fprintf(f,"        movabsq $outputbuf, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"2:\n");
  fprintf(f,"        movabsq $.LOBUF, %%r14\n");
  fprintf(f,"        xorl    %%ebx, %%ebx\n");
  fprintf(f,"        jmp     4f\n");
  fprintf(f,"3:\n");
  fprintf(f,"        leaq    1(%%rsi), %%rdx\n");
  fprintf(f,"        incq    %%rbx\n");
  fprintf(f,"        movq    %%rdx, %%rax\n");
  fprintf(f,"        cmpq    %%r13, %%rbx\n");
  fprintf(f,"        movabsq %%rax, .Lobuf_next\n");
  fprintf(f,"        movb    -1(%%r12,%%rbx), %%cl\n");
  fprintf(f,"        movb    %%cl, -1(%%r14,%%rsi)\n");
  fprintf(f,"        jae     4f\n");
  fprintf(f,"        cmpq    $OUTPUT_WIDTH, %%rdx\n");
  fprintf(f,"        jbe     4f\n");
  fprintf(f,"        movabsq $outputbuf, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"4:\n");
  fprintf(f,"        movabsq $.Lobuf_next, %%rax\n");
  fprintf(f,"        cmpq    %%r13, %%rbx\n");
  fprintf(f,"        movq    (%%rax), %%rsi\n");
  fprintf(f,"        jne     3b\n");
  fprintf(f,"        popq    %%rbx\n");
  fprintf(f,"        leaq    -1(%%rsi), %%rax\n");
  fprintf(f,"        movl    $ZONE_WIDTH, %%ecx\n");
  fprintf(f,"        xorl    %%edx, %%edx\n");
  fprintf(f,"        divq    %%rcx\n");
  fprintf(f,"        incq    %%rax\n");
  fprintf(f,"        movabsq %%rax, .Lcurzone\n");
  fprintf(f,"        popq    %%r12\n");
  fprintf(f,"        popq    %%r13\n");
  fprintf(f,"        popq    %%r14\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   appendbuf, .-appendbuf\n\n");

  fprintf(f,"        # This procedure will update curzone to the next avaiable zone, wrapping\n");
  fprintf(f,"        # back to zone 1 after MAZ_ZONE.  If the forceit parameter is one then move to the next \n");
  fprintf(f,"        # output zone even if we are on the first character of an output zone already.  If\n");
  fprintf(f,"        # forceit is zero and we are on the first character of an output zone already, do\n");
  fprintf(f,"        # nothing.\n");
  fprintf(f,"        # argument forecit is passed in rdi\n");
  fprintf(f,"        # trashes rax, rcx, rdx, rsi\n\n");
  fprintf(f,"        .type   nextzone, @function\n");
  fprintf(f,"nextzone:\n");
  fprintf(f,"        movabsq .Lcurzone, %%rax\n");
  fprintf(f,"        leaq    1(%%rax), %%rcx        # next=curzone+1\n");
  fprintf(f,"        testq   %%rdi, %%rdi           # if we do not want to force a move then\n");
  fprintf(f,"        jne     0f                   # goto 0\n");
  fprintf(f,"        movabsq .Lobuf_next, %%rax\n");
  fprintf(f,"        cmpq    $1, %%rax             # if obuf_next >= 1 then\n");
  fprintf(f,"        jbe     0f                   # goto 0\n");
  fprintf(f,"        decq    %%rax\n");
  fprintf(f,"        xorl    %%edx, %%edx\n");
  fprintf(f,"        movl    $ZONE_WIDTH, %%esi\n");
  fprintf(f,"        divq    %%rsi\n");
  fprintf(f,"        testq   %%rdx, %%rdx           # if ((buf_next-1)%%ZONE_WIDTH)==0 then\n");
  fprintf(f,"        je      2f                   # goto 2\n");
  fprintf(f,"0:\n");
  fprintf(f,"        cmpq    $MAX_ZONE, %%rcx      # if next <= MAX_ZONE then\n");
  fprintf(f,"        jbe     1f                   # goto 1\n");
  fprintf(f,"        movabsq $outputbuf, %%rax\n");
  fprintf(f,"        call    *%%rax                # output the current buffered line\n");
  fprintf(f,"        movl    $1, %%ecx             # next=1\n");
  fprintf(f,"1:\n");
  fprintf(f,"        movq    %%rcx, %%rax           # curzone=next\n");
  fprintf(f,"        imulq   $ZONE_WIDTH, %%rcx, %%rcx   # rcx*=ZONE_WIDTH\n");
  fprintf(f,"        movabsq %%rax, .Lcurzone\n");
  fprintf(f,"        leaq    -ZONE_WIDTH(%%rcx), %%rax   # rax=rcx-ZONE_WIDTH\n");
  fprintf(f,"        movabsq %%rax, .Lobuf_next         # obuf_next=(curzone-1)*ZONE_WIDTH\n");
  fprintf(f,"2:\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size   nextzone, .-nextzone\n\n");
  fprintf(f,"        # end PRINT ZONE code\n\n");

  // used to give variable name when uninitialized variable is read
  fprintf(f,"        .type   printvarname, @function\n");
  fprintf(f,"        # procedure, no return value\n");
  fprintf(f,"printvarname:\n");
  fprintf(f,"        # incoming argument is ASCIIZ varname in rdi, left justified\n");
  fprintf(f,"        subq    $16,%%rsp\n");
  fprintf(f,"        movq    %%rdi,(%%rsp)\n");
  fprintf(f,"        movq    %%rsp,%%rsi\n");
  fprintf(f,"        movabsq $printstring,%%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        addq    $16,%%rsp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size printvarname, .-printvarname\n\n");
  // we only need this if we had a PRINT statement
  fprintf(f,"        .type   printfloat, @function\n");
  fprintf(f,"printfloat:\n");
  fprintf(f,"        # incoming argument is real number in xmm0\n");
  fprintf(f,"        pushq   %%rbp\n");
  fprintf(f,"        movq    %%rsp, %%rbp\n");
  fprintf(f,"        subq    $32, %%rsp       # reserve room for a 64 byte buffer at -32(%%rsp)\n");
  if (!use_double)
    fprintf(f,"        cvtss2sd %%xmm0,%%xmm0    # xmm0=(double)xmm0, value to convert\n");
  fprintf(f,"        leaq    -32(%%rbp), %%rdi   # buffer\n");
  fprintf(f,"        movq    $SIGNIFICANCE_WIDTH-1, %%rsi  # number of digits after decimal point\n");
  fprintf(f,"        movl    $EXRAD_WIDTH, %%edx # number of digits in expondent\n");
  fprintf(f,"        movabsq $g_fmt, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movq    %%rax,%%rdi\n");
  fprintf(f,"        movabsq $appendbuf, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        movq    %%rbp,%%rsp\n");
  fprintf(f,"        popq    %%rbp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size printfloat, .-printfloat\n\n");
  fflush(f);
  return true;
}

//
// This function emits code to call a UDF.
// The f parameter is the file handle for the output file, and the 
// file must already be open for output.
// The fname parameter is the name of the UDF, and the hasarg
// parameter indicates whether the function requires an argument.
// If hasarg is true, the function requires an argument, otherwise
// it does not.  If an argument is required, it must be the top
// value on the dedicated floating point stack before calling this
// function.  However, the UDFs use a register-based calling convention
// so if there is an argument, this function will pop the argument value 
// from the decidated floating point stack into xmm0.  After the function
// returns, the return value passed in xmm0 will be pushed onto the
// dedicated floating point stack.
//
bool doudf(FILE *f,const char *fname,bool hasarg) {
  if (NULL==f) {
    fprintf(stderr,"doudf(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==fname) || (0==fname[0])) {
    fprintf(stderr,"doudf(): fname is NULL or empty\n");
    return false;
  }
  if (hasarg)
    fprintf(f,"        popxmm 0     # put argument into register xmm0\n");
  fprintf(f,"        movabsq $%s,%%rax\n",fname);
  fprintf(f,"        callq   *%%rax\n");
  fprintf(f,"        pushxmm 0     # put result on stack\n");
  fflush(f);
  return true;
}

//
// This function emits code for the beginning of a UDF.  The f parameter is the
// file handle for the output file, and the file must already be open for
// output.  The fname parameter is the name of the UDF, and the argname
// parameter is the name of the UDF's argument.  If the UDF does not have an
// argument, argname should be NULL.  The UDFs use a register calling convention
// so if there is an argument, it will be passed in xmm0.
//
bool doudf_header(FILE *f,const char *fname,const char *argname) {
  if (NULL==f) {
    fprintf(stderr,"doudf_header(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==fname) || (0==fname[0])) {
    fprintf(stderr,"doudf_header(): fname is NULL or empty\n");
    return false;
  }
  if ((argname!=NULL) && (argname[0]!=0)) { // if there is a parameter, allocate storage for it initializing it to SNaN
    fprintf(f,"        .section .data,\"aw\",@progbits\n\n");
    fprintf(f,".LFN%c_%c:\n",fname[2],argname[0]);
    fprintf(f,"        .%s SNaN  # %" PRIu32 " bytes, actually %s\n",(use_double?"quad":"long"),(use_double?UINT32_C(8):UINT32_C(4)),(use_double?"double":"float"));
    fprintf(f,"        .size .LFN%c_%c, .-.LFN%c_%c\n\n",
      fname[2],argname[0],fname[2],argname[0]);
    fprintf(f,"        .section .text,\"ax\",@progbits\n\n");
  }
  fprintf(f,"        .type   %s, @function\n",fname);
  fprintf(f,"%s:\n",fname);
  fprintf(f,"        pushq   %%rbp\n");
  fprintf(f,"        movq    %%rsp, %%rbp\n");
  if ((argname!=NULL) && (argname[0]!=0)) { // if there is a parameter, save it in the allocated storage
    fprintf(f,"        movabsq $.LFN%c_%c, %%r15\n", fname[2],argname[0]);
    fprintf(f,"        movs%c   %%xmm0, (%%r15)      # store function argument %c\n",
      (use_double?'d':'s'),argname[0]);
  }
  fflush(f);
  return true;
}

//
// This function emits code for the end of a UDF.  The f parameter is the file
// handle for the output file, and the file must already be open for output.
// The fname parameter is the name of the UDF.  The body of the UDF is an
// arithmetic expression and the computed value will be left on top of the
// dedicated floating point stack.  This function will pop that into the
// xmm0 register before actually returning from the UDF.  The UDFs use a
// register calling convention, and the return value must be in xmm0.
//
bool doudf_footer(FILE *f,const char *fname) {
  if (NULL==f) {
    fprintf(stderr,"doudf_footer(): file parameter is NULL\n");
    return false;
  }
  if ((NULL==fname) || (0==fname[0])) {
    fprintf(stderr,"doudf_footer(): fname is NULL or empty\n");
    return false;
  }
  fprintf(f,"        popxmm 0     # pop stack into xmm0 for return\n");
  fprintf(f,"        movq    %%rbp,%%rsp\n");
  fprintf(f,"        popq    %%rbp\n");
  fprintf(f,"        retq\n");
  fprintf(f,"        .size %s, .-%s\n\n",fname,fname);
  fflush(f);
  return true;
}

//
// This procedure will emit code to setup the main runtime stack.
// This is actually the last code emitted to the assembly output.
// The f parameter is the file handle for the output file, and the 
// file must already be open for output.  This procedure also emits
// the ELF identification string.
//
void gen_stack(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"gen_stack(): file parameter is NULL\n");
    abort();
  }
  fprintf(f,"\n# Configure runtime system stack\n\n");
  fprintf(f,"        .section .note.GNU-stack,\"\",@progbits\n");
  fprintf(f,"\n# Low-level compiler identification information.\n\n");
  fprintf(f,"        .ident \"Mr. Ham's ECMA-55 Minimal BASIC version %s for x86_64 Linux\"\n",version);
  fprintf(f,"        .end\n");
  fflush(f);
  return;
}

//
// This procedure will emit code to call the realinit subroutine to
// initilialize (seed) the random number generator.  That is needed 
// to support the Minimal BASIC RANDOMIZE statement.  The f parameter 
// is the file handle for the output file, and the file must already 
// be open for output.
//
static void write_init_RND(FILE *f) {
  if (NULL==f) {
    fprintf(stderr,"write_init_RND(): file parameter is NULL\n");
    abort();
  }
  fprintf(f,"        # initialize random number generator\n");
  fprintf(f,"        xorq   %%rdi, %%rdi\n");
  fprintf(f,"        movabsq $realinit, %%rax\n");
  fprintf(f,"        callq   *%%rax\n");
  return;
}

//
// This function is used to write a comment line in the generated
// assembly language output.  The f parameter is the file handle 
// for the output file, and the file must already be open for output.
// The fmt parameter is a format string like fprintf() uses.  The
// API for this function is the same as the well-known C standard 
// library function fprintf().  On success, true is returned,
// otherwise false is returned.
//
bool emit_comment(FILE *f, const char *fmt, ...) {
  va_list arg;
  static char buffer[COMMENT_BUFFER_MAX];

  if (NULL==f) {
    fprintf(stderr,"emit_comment(): file parameter is NULL\n");
    abort();
  }
  if (NULL==fmt) {
    fprintf(stderr,"emit_comment(): fmt parameter is NULL\n");
    return false;
  }
  if (0==fmt[0]) {
    fprintf(stderr,"emit_comment(): fmt parameter is empty\n");
    return false;
  }
  snprintf(buffer,COMMENT_BUFFER_MAX,"        # %s\n",fmt);
  buffer[COMMENT_BUFFER_MAX-2]='\n';
  buffer[COMMENT_BUFFER_MAX-1]=0;
  va_start(arg, fmt);
  vfprintf(f,buffer,arg);
  va_end(arg);
  fflush(f);
  return true;
}
